
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
//
//	Revision History
//
//		
//		
//		
//
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////

/*
BOOLEAN StoreAuthenticationEvent(PCUSTOM_AUTHENTICATION_EVENT pAuthenticationEventOut, P_802_11_AUTHENTICATION_REQUEST pAuthenticationEventIn);
BOOLEAN StoreEapolPacket(PEAPOL_BUFFER pEapolBuffer, PUCHAR pbRxBuffer, ULONG dwRxBufferLength);
void 	ParseEAPOLPacket(PUCHAR pbEapolMsg, PUCHAR pbLeapChallenge, PUCHAR pbPacket, BOOLEAN fReceived, BOOLEAN fDisplay);
PUCHAR 	RxProtocolConversion(PUCHAR pbInputBuffer, ULONG dwBufferLength, UCHAR bOperatingMode, PULONG pdwPacketSize);
BOOLEAN GetChannelFromBeaconOrProbe(PUCHAR pBeaconOrProbeFrame, PUCHAR pChannel);
BOOLEAN GetSupportedRatesfromBeaconOrProbe(PUCHAR pSupportedRates, PUCHAR pSupportedRatesLength, PUCHAR pBeaconOrProbeFrame, USHORT ElementOffset);
BOOLEAN GetSSIDfromBeaconOrProbe(PUCHAR pbSSID, PUCHAR pbSSIDlength, PUCHAR pbBeaconOrProbeFrame);
BOOLEAN GetElementFromBeaconOrProbe(PUCHAR pbBeaconOrProbeFrame, USHORT wFrameLen, UCHAR bElementID, PUCHAR pbOUI, PUCHAR pbElementBody, PUSHORT pwElementBodyLen);
BOOLEAN IsTxKeyAvailable(PSECURITY_INFO pSecurityInfo);
VOID 	SelectCiphers(PSECURITY_INFO pSecurityInfo, BOOLEAN fIsBroadcast, BOOLEAN* pfTxEapPending, BOOLEAN fForceUnencrypted, PUCHAR pbCipherType, PUCHAR pbCipherLength);
BOOLEAN MatchSecuritySettings(PSECURITY_INFO pSecurityInfo, PBSS_INFO_EX pBssInfo, UCHAR bOperatingMode, STATISTICS* pStats);
void 	AddIEsInRequest(PUCHAR pbFrameBody, PUSHORT pwBytesAdded, PSECURITY_INFO pSecurityInfo, PBSS_INFO_EX pBssInfo, PUCHAR pbIPAddress, USHORT wCwMin, USHORT wCwMax);
void 	BuildNullFrame(PUCHAR pbFrame, PASSOCIATION_INFO pAssociationInfo, PUSHORT pwFrameLength);
void 	BuildDisAssociationRequestFrame(PUCHAR pbFrame, PASSOCIATION_INFO pAssociationInfo, USHORT wReasonCode, PUSHORT pwFrameLength);
void 	BuildReAssociationRequestFrame(PUCHAR pbFrame, PASSOCIATION_INFO pAssociationInfo, PSECURITY_INFO pSecurityInfo, PUSHORT pwFrameLength);
void 	BuildAssociationRequestFrame(PUCHAR pbFrame, PASSOCIATION_INFO pAssociationInfo, PSECURITY_INFO pSecurityInfo, PUSHORT pwFrameLength);
void 	BuildDeAuthenticationRequestFrame(PUCHAR pbFrame, PASSOCIATION_INFO pAssociationInfo, USHORT wReasonCode, PUSHORT pwFrameLength);
void 	BuildAuthenticationRequestFrame(PUCHAR pbFrame, PAUTHENTICATION_INFO pAuthenticationInfo, PASSOCIATION_INFO pAssociationInfo, PSECURITY_INFO pSecurityInfo, PUSHORT pwFrameLength);
BOOLEAN IsEncryptionEnabled(PSECURITY_INFO pSecurityInfo);
void 	BuildMgmtFrameHeader(PUCHAR pbHeader, UCHAR bFrameSubType, PAUTHENTICATION_INFO pAuthenticationInfo, PASSOCIATION_INFO pAssociationInfo, PSECURITY_INFO pSecurityInfo);

#ifdef WPA2
VOID 	WPA2_FlashPMKIDCandidateList(PPMKID_CANDIDATELIST pPMKIDCandidateList);
BOOLEAN	WPA2_ComparePMKIDCandidateLists(PSECURITY_INFO pSecurityInfo);
BOOLEAN WPA2_BuildPMKIDCandidateList(PPMKID_CANDIDATELIST pPMKIDCandidateList, BSS_INFO_EX* pBSSinfoEx, UCHAR bBSSinfoExEntries, BSS_INFO_EX* pCurrentBSSinfo);
ULONG 	WPA2_BuildAuthEncryptionCapability(P_802_11_AUTHENTICATION_ENCRYPTION pAuthEncryptionList);
USHORT 	WPA2_AddPMKsInElement(PSECURITY_INFO pSecurityInfo, PBSS_INFO_EX pBssInfo, PUCHAR pbIEBuffer);
BOOLEAN WPA2_AddPMKIDInList(PSECURITY_INFO pSecurityInfo, P__802_11_PMKID pPMKID);
#endif
*/


////////////////////////////////////////////////////////////////////////////
//
//	Here, you can include any files you need
//
////////////////////////////////////////////////////////////////////////////
#include "vnet.h"
#include "frame.h"
#include "appint.h"



////////////////////////////////////////////////////////////////////////////
//
//	Here, you can add any custom definitions
//
////////////////////////////////////////////////////////////////////////////

// size of MAC ADDRESS.
#define VNet_LENGTH_OF_ADDRESS 	6

// definition of the mask that indicates that an AP has WPA2 pre-authentication capabilities enabled.
// we compare it with the WpaCapabilities of the WPA_INFO in the BSS_INFO_EX structure.
#define RSN_PREAUTHENTICATION_CAPABILITY_ON		1

// definition of the offset in the beacon/probe frame that contains the length of the SSID.
#define SSID_ELEMENT_LENGTH_OFFSET              13		//12+1

// definitions for security capabilities of BSS
// The configuration of the InWpaMode parameter in the SecurityInfo, 
// must match with the Capabilities parameter in the BSS_INFO_EX of an AP,
// before attempting to associate with this AP.
#define BSS_CAPABILITY_WPA    1		// WPA
#define BSS_CAPABILITY_RSN    2		// WPA2
#define BSS_CAPABILITY_CCX    4		// CCX 1.0

// definitions for the cipher suites in the WPA information elements
#define OUI_CIPHER_SUITE_WEP_64   	1	// WEP 40bits
#define OUI_CIPHER_SUITE_TKIP   	2	// TKIP
#define OUI_CIPHER_SUITE_CCMP   	4	// AES
#define OUI_CIPHER_SUITE_WEP_128  	5	// WEP 104bits

// definition of the OUI size in the WPA information elements
#define OUI_SIZE          			4

#if(CCX2)
#define ELEMENT_ID_CELL_POWER_INFO 0x96
#define ELEMENT_ID_VERSION_NUM	   0xDD	
#endif



#if defined(PCMCIA_DEBUG)||defined(PCI_DEBUG)
#define DbgCommon(L,A) {printk A;}
#else
#define DbgCommon(L,A) { }
#endif


////////////////////////////////////////////////////////////////////////////
//
//	Here, you can add any custom variables
//
////////////////////////////////////////////////////////////////////////////

// the rates we use to suild the association request
UCHAR 	BasicRates[2][4] = {{0x82,0x84,0x0b,0x16},{0x82,0x84,0x8b,0x96}};

// ethernet frame types. we use them to convert a received wireless packet to an ethernet packet.
UCHAR   SNAP_ETHERNETII[6] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0xF8};
UCHAR   STATIC_ENTRIES_ETHERTYPE_8137[] = {0x81, 0x37};
UCHAR   STATIC_ENTRIES_ETHERTYPE_80F3[] = {0x80, 0xF3};

//
// definitions used in the WPA Information Element
//

// the OUI that is contained in the WPA information element
UCHAR	WPA_OUI[3] = {0x00, 0x50, 0xF2};
// the version of the WPA information element
USHORT	WPA_VERSION	= 0x0001;

// the OUIs that we use in the cipher type field of the WPA information element.
UCHAR	WPA_OUI_GROUP_WEP_64[4]  = {0x00, 0x50, 0xF2, 0x01};
UCHAR	WPA_OUI_GROUP_WEP_128[4] = {0x00, 0x50, 0xF2, 0x05};
UCHAR	WPA_OUI_GROUP_TKIP[4]    = {0x00, 0x50, 0xF2, 0x02};
UCHAR	WPA_OUI_GROUP_AES[4]     = {0x00, 0x50, 0xF2, 0x04};
UCHAR	WPA_OUI_PW_TKIP[4]		 = {0x00, 0x50, 0xF2, 0x02};
UCHAR	WPA_OUI_PW_AES[4]		 = {0x00, 0x50, 0xF2, 0x04};

// the OUIs that we use in the authentication type field of the WPA information element.
UCHAR   WPA_OUI_AUTH_1X[4]  = {0x00, 0x50, 0xF2, 0x01}; 	//802.1x
UCHAR   WPA_OUI_AUTH_PSK[4] = {0x00, 0x50, 0xF2, 0x02}; 	//PSK

// the number of the pairwise cipher suites we include in the WPA information element.
USHORT 	WPA_PW_SUITE_CNT = 0x0001;

// the Capabilities we add in the capabilities field in the WPA information element.
UCHAR 	WPA_CAP[2] = {0x00, 0x00};


//
// definitions used in the WPA2 Information Element
//

// the OUI that is contained in the WPA2 information element
UCHAR	RSN_OUI[3] = {0x00, 0x0F, 0xAC};
// the version of the WPA2 information element
USHORT	RSN_VERSION	= 0x0001;
	
// the OUIs that we use in the cipher type field of the WPA2 information element.
UCHAR	RSN_OUI_GROUP_WEP_64[4]  = {0x00, 0x0F, 0xAC, 0x01};
UCHAR	RSN_OUI_GROUP_WEP_128[4] = {0x00, 0x0F, 0xAC, 0x05};
UCHAR	RSN_OUI_GROUP_TKIP[4]    = {0x00, 0x0F, 0xAC, 0x02};
UCHAR	RSN_OUI_GROUP_AES[4]     = {0x00, 0x0F, 0xAC, 0x04};
UCHAR	RSN_OUI_PW_TKIP[4]		 = {0x00, 0x0F, 0xAC, 0x02};
UCHAR	RSN_OUI_PW_AES[4]		 = {0x00, 0x0F, 0xAC, 0x04};

// the OUIs that we use in the authentication type field of the WPA2 information element.
UCHAR   RSN_OUI_AUTH_1X[4]  = {0x00, 0x0F, 0xAC, 0x01}; //802.1x
UCHAR   RSN_OUI_AUTH_PSK[4] = {0x00, 0x0F, 0xAC, 0x02}; //PSK

// the number of the pairwise cipher suites we include in the WPA2 information element.
USHORT 	RSN_PW_SUITE_CNT = 0x0001;

// the Capabilities we add in the capabilities field in the WPA information element.
UCHAR 	RSN_CAP[2] = {0x00, 0x00};

#ifdef CCX
UCHAR   CCX_ELEMENT_BODY[30] = {0x00, 0x01, 0x66, 0x00, 0x00, 
								0x00, 0xff, 0x03, 0x18, 0x00, 
								0x64, 0x73, 0x6d, 0x69, 0x74, 
								0x2d, 0x77, 0x32, 0x6b, 0x00, 
								0x00, 0x00, 0x00, 0x00,	0x00, 
								0x00, 0x00, 0x00, 0x00, 0x21};


UCHAR CCX_IP_BODY[11] = {0x0a, 0x00, 0x40, 0x96, 0x00,
						 0x00, 0x00, 0x00, 0x00,
						 0x00, 0x00};
#endif

#if(CCX2)
UCHAR	CCKM_AUTH[4]			 = {0x00, 0x40, 0x96, 0x00};
UCHAR	CKIP_OUI[4]				 = {0x00, 0x40, 0x96, 0x00};
UCHAR	CMIC_OUI[4]				 = {0x00, 0x40, 0x96, 0x02};
UCHAR	CKIPCMIC_OUI[4]			 = {0x00, 0x40, 0x96, 0x01};
#endif



////////////////////////////////////////////////////////////////////////////
//
//	Here, you can add any custom structures
//
////////////////////////////////////////////////////////////////////////////

typedef UCHAR   _802_11_MAC_ADDRESS[6];

// we use this structure to notify the supplicant for TKIP MIC errors.
// without this notification, there can not be any countermeasures.
typedef struct __802_11_AUTHENTICATION_REQUEST
{
    ULONG 				Length;            	// Length of structure
    _802_11_MAC_ADDRESS Bssid;				// Current BSSID
    ULONG 				Flags;				// Reason of the request
} _802_11_AUTHENTICATION_REQUEST, *P_802_11_AUTHENTICATION_REQUEST;

// a custom structure that we keep the authentication events that 
// we indicate to the supplicant.
#define MAX_CUSTOM_AUTHENTICATION_EVENTS 3

typedef struct __CUSTOM_AUTHENTICATION_EVENT{

	// if the AuthenticationRequest contains a valid request set the InUse to TRUE.
	// invalidate it as soon as we indicate it to the supplicant
	UCHAR							InUse;
		
	_802_11_AUTHENTICATION_REQUEST	AuthenticationRequest;
}CUSTOM_AUTHENTICATION_EVENT, *PCUSTOM_AUTHENTICATION_EVENT;


////////////////////////////////////////////////////////////////////////////
//
//	Here, you can add any custom functions
//
////////////////////////////////////////////////////////////////////////////


#if 0
static inline int VNetMemCompare(PUCHAR px, PUCHAR py, ULONG bytes)
{
	return memcmp(px, py, bytes);
}
#endif




////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	BuildMgmtFrameHeader()
//	Creates a header for any management frame.
//
//	Parameters
//		OUT: PUCHAR					Pointer to a buffer that we will store the header. 
//									Buffer has to be allocated by the caller.
//		IN:  UCHAR					Type of the managment frame we will create.
//		IN:  PAUTHENTICATION_INFO	Pointer to the authentication info structure of the Adapter.
//		IN:  PASSOCIATION_INFO		Pointer to the association info structure of the Adapter.
//		IN:	 PSECURITY_INFO			Pointer to the security info structure of the Adapter.
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////

void BuildMgmtFrameHeader(PUCHAR pbHeader, UCHAR bFrameSubType, PAUTHENTICATION_INFO pAuthenticationInfo, PASSOCIATION_INFO pAssociationInfo, PSECURITY_INFO pSecurityInfo)
{
	// cast the buffer to a managment frame header structure
	s80211_MgmtFrm	*pMgmtFrame = (s80211_MgmtFrm*)pbHeader;

	// build the header according to the type of the requested frame
	switch (bFrameSubType)
	{
		case C80211_SUBTYPE_MGMT_Authentication:

			pMgmtFrame->FrameControl[0]	= C80211_PROTOCOL_VERSION | C80211_TYPE_MGMT | C80211_SUBTYPE_MGMT_Authentication;

			if(pAuthenticationInfo->CurrentAuthentTransactionSeqNum == 1 || !IsEncryptionEnabled(pSecurityInfo)) 	
				pMgmtFrame->FrameControl[1]	= 0x00;
			else if(IsEncryptionEnabled(pSecurityInfo))
				pMgmtFrame->FrameControl[1]	= C80211_FRMCTRL_ENCRYPTION;

			break;

		case C80211_SUBTYPE_MGMT_Deauthentication:
		case C80211_SUBTYPE_MGMT_ASS_REQUEST:
		case C80211_SUBTYPE_MGMT_REASS_REQUEST:
		case C80211_SUBTYPE_MGMT_DISASSOSIATION:

			pMgmtFrame->FrameControl[0] = bFrameSubType | C80211_PROTOCOL_VERSION | C80211_TYPE_MGMT;
			pMgmtFrame->FrameControl[1]	= 0x00;
			break;

	}

	// add the common fields
	
	pMgmtFrame->DurationID			= 0x8000;			
	
	VNetMoveMemory(pMgmtFrame->uDA,		pAssociationInfo->CurrentBSSID, VNet_LENGTH_OF_ADDRESS); 
	VNetMoveMemory(pMgmtFrame->uSA,		pAssociationInfo->CurrentAddress, VNet_LENGTH_OF_ADDRESS);
	VNetMoveMemory(pMgmtFrame->uBSSID,	pAssociationInfo->CurrentBSSID, VNet_LENGTH_OF_ADDRESS); 
	
	pMgmtFrame->SequenceControl		= 0x0000;

}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	IsEncryptionEnabled()
//	Returns the current encryption status of the driver.
//
//	Parameters
//		IN:	 PSECURITY_INFO			Pointer to the security info structure of the Adapter.
//		
//
//	returns:	TRUE  : if the encryption is enabled
//				FALSE : if the encryption is disabled
/////////////////////////////////////////////////////////////////////////////
BOOLEAN IsEncryptionEnabled(PSECURITY_INFO pSecurityInfo)
{
	if(	pSecurityInfo->EncryptionStatus == _802_11Encryption1Enabled ||
		pSecurityInfo->EncryptionStatus == _802_11Encryption2Enabled ||
		pSecurityInfo->EncryptionStatus == _802_11Encryption3Enabled ||
		pSecurityInfo->EncryptionStatus == _802_11Encryption2KeyAbsent ||
		pSecurityInfo->EncryptionStatus == _802_11Encryption3KeyAbsent)
		return TRUE;
	else
		return FALSE;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	BuildAuthenticationRequestFrame()
//	Creates an authentication request frame.
//
//	Parameters
//		OUT: PUCHAR					Pointer to a buffer that we will store the frame. 
//									Buffer has to be allocated by the caller.
//		IN:  PAUTHENTICATION_INFO	Pointer to the authentication info structure of the Adapter.
//		IN:  PASSOCIATION_INFO		Pointer to the association info structure of the Adapter.
//		IN:	 PSECURITY_INFO			Pointer to the security info structure of the Adapter.
//		OUT: PUSHORT				Size of the frame returned, in bytes.
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////
void BuildAuthenticationRequestFrame(PUCHAR pbFrame, PAUTHENTICATION_INFO pAuthenticationInfo, PASSOCIATION_INFO pAssociationInfo, PSECURITY_INFO pSecurityInfo, PUSHORT pwFrameLength)
{
	s80211_MgmtFrm					*pMgmtFrame;
	s80211_MgmtBody_Authentication	*pAuthFrameBody;

	// get a pointer to the header of the frame
	pMgmtFrame = (s80211_MgmtFrm*)pbFrame;
	// get a pointer to the body of the frame
	pAuthFrameBody = (s80211_MgmtBody_Authentication*)pMgmtFrame->Body;
	
	//
	// Format Frame Header 
	//
	BuildMgmtFrameHeader((PUCHAR)pMgmtFrame, (UCHAR)C80211_SUBTYPE_MGMT_Authentication, pAuthenticationInfo, pAssociationInfo, pSecurityInfo);

	
	//
	// Frame Body
	//

	// the authentication algorithm will be decided by the Authentication Mode of the Adapter.
	switch (pSecurityInfo->AuthenticationMode)
	{
		case _802_11AuthModeShared:
			pAuthFrameBody->AuthenticationAlgorithmNumber = C80211_MGMT_AAN_SHAREDKEY;
			break;

		case _802_11AuthModeWPA2PSK:
		case _802_11AuthModeWPA2:
		case _802_11AuthModeWPAPSK:
		case _802_11AuthModeWPA:
		case _802_11AuthModeWPANone:
		case _802_11AuthModeOpen:
#ifdef CCX
			// This would normally have the Open value unless the Funk has set a EAP value.
			
			DbgCommon(1,("AUTH: EAP Request (%x)...\n", pSecurityInfo->CCXinfo.FswNetworkEAPToUse))

			pAuthFrameBody->AuthenticationAlgorithmNumber = pSecurityInfo->CCXinfo.FswNetworkEAPToUse;
#else

			pAuthFrameBody->AuthenticationAlgorithmNumber = C80211_MGMT_AAN_OPENSYSTEM;
#endif
			break;
		
		default:
			break;
	}

	//
	// Set AuthenticationTransactionSequenceNumber
	//
	pAuthFrameBody->AuthenticationTransactionSequenceNumber = pAuthenticationInfo->CurrentAuthentTransactionSeqNum;
	DbgCommon(1,("Sending Authentication Request with Transaction Sequence Number:%d...\n", pAuthenticationInfo->CurrentAuthentTransactionSeqNum))

	pAuthenticationInfo->ExpectedAuthentTransactionSeqNum = pAuthenticationInfo->CurrentAuthentTransactionSeqNum+1; 

	pAuthenticationInfo->CurrentAuthentTransactionSeqNum += 2;

	pAuthFrameBody->StatusCode = 0x0000;

	// add the challenge in the case of shared authentication
	if((pAuthenticationInfo->ChallengeTextLength != 0) &&
		(pAuthenticationInfo->ExpectedAuthentTransactionSeqNum != 2))
	{
		pAuthFrameBody->ChallengeText.ElementID = C80211_MGMT_ElementID_ChallengeText;
		pAuthFrameBody->ChallengeText.Length = pAuthenticationInfo->ChallengeTextLength;
	
		if(pAuthenticationInfo->ChallengeTextLength)
			VNetMoveMemory(&(pMgmtFrame->Body[8]), pAuthenticationInfo->pChallengeText, pAuthenticationInfo->ChallengeTextLength);

		*pwFrameLength = 24 + 6 + 2 + pAuthenticationInfo->ChallengeTextLength; //Header + Body + element Header + element length
	}
	else
		*pwFrameLength = 24 + 6; //Header + Body
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	BuildDeAuthenticationRequestFrame()
//	Creates a deauthentication request frame.
//
//	Parameters
//		OUT: PUCHAR					Pointer to a buffer that we will store the frame. 
//									Buffer has to be allocated by the caller.
//		IN:  PASSOCIATION_INFO		Pointer to the association info structure of the Adapter
//		IN:	 USHORT					The reason that will be reported in the request
//		OUT: PUSHORT				Size of the frame returned in bytes
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////
void BuildDeAuthenticationRequestFrame(PUCHAR pbFrame, PASSOCIATION_INFO pAssociationInfo, USHORT wReasonCode, PUSHORT pwFrameLength)
{
	s80211_MgmtFrm						*pMgmtFrame;
	s80211_MGMTBody_DEAUTHENTICATION	*DeAuReqFrameBody;
	
	// get a pointer to the header of the frame
	pMgmtFrame = (s80211_MgmtFrm*)pbFrame;
	// get a pointer to the data of the frame
	DeAuReqFrameBody = (s80211_MGMTBody_DEAUTHENTICATION*)(pMgmtFrame->Body);
	
	// Format Frame Header 
	BuildMgmtFrameHeader((PUCHAR)pMgmtFrame, (UCHAR)C80211_SUBTYPE_MGMT_Deauthentication, NULL, pAssociationInfo, NULL);
		
	// add the reason of the deauthentication
	DeAuReqFrameBody->ReasonCode = wReasonCode;
		
	// calculate the frame length
	*pwFrameLength = 24 + 2;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	BuildAssociationRequestFrame()
//	Creates an association request frame.
//
//	Parameters
//		OUT: PUCHAR					Pointer to a buffer that we will store the frame. 
//									Buffer has to be allocated by the caller.
//		IN:  PASSOCIATION_INFO		Pointer to the association info structure of the Adapter
//		IN:	 PSECURITY_INFO			Pointer to the security info structure of the Adapter
//		OUT: PUSHORT				Size of the frame returned in bytes
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////
void BuildAssociationRequestFrame(PUCHAR pbFrame, PASSOCIATION_INFO pAssociationInfo, PSECURITY_INFO pSecurityInfo, PUSHORT pwFrameLength)
{
    s80211_MgmtFrm                  *pMgmtFrame;
    s80211_MgmtBody_AssRqst         *pAssReqFrameBody;
    
    USHORT							wOffset = 0, wBytesAdded;
    
	// get a pointer to the header of the frame
    pMgmtFrame 			= (s80211_MgmtFrm*) pbFrame;
	// get a pointer to the data of the frame
	pAssReqFrameBody 	= (s80211_MgmtBody_AssRqst*)(pMgmtFrame->Body);
	
	// Format Frame Header 
	BuildMgmtFrameHeader((PUCHAR)pMgmtFrame, (UCHAR)C80211_SUBTYPE_MGMT_ASS_REQUEST, NULL, pAssociationInfo, NULL);
	
	//
	// Capability Information
	//
	pAssReqFrameBody->CapabilityInformation = (USHORT)C80211_MGMT_CAPABILITY_ESS;

	if(IsEncryptionEnabled(pSecurityInfo) || pSecurityInfo->InWpaMode)
		pAssReqFrameBody->CapabilityInformation |= C80211_MGMT_CAPABILITY_Privacy;

	if(pAssociationInfo->PreambleType == PREAMBLE_TYPE_SHORT)
		pAssReqFrameBody->CapabilityInformation |= C80211_MGMT_CAPABILITY_ShortPreamble;

	pAssociationInfo->AssocCapabilities = pAssReqFrameBody->CapabilityInformation;

	//Listen Interval
	pAssReqFrameBody->ListenInterval = pAssociationInfo->ListenInterval* pAssociationInfo->BeaconPeriod;
	wOffset = 4;

	// SSID Element
	*((PUCHAR)pAssReqFrameBody + wOffset) = C80211_MGMT_ElementID_SSID;
	wOffset += 1;

	*((PUCHAR)pAssReqFrameBody + wOffset) = pAssociationInfo->DesiredSSIDsize;
	wOffset += 1;

	VNetMoveMemory(((PUCHAR)pAssReqFrameBody + wOffset), pAssociationInfo->DesiredSSID, pAssociationInfo->DesiredSSIDsize);
	wOffset += pAssociationInfo->DesiredSSIDsize;

	// Supported Rates Element
	*((PUCHAR)pAssReqFrameBody + wOffset) = C80211_MGMT_ElementID_SupportedRates;
	wOffset += 1;

	*((PUCHAR)pAssReqFrameBody + wOffset) = 4;
	wOffset += 1;

	VNetMoveMemory((PUCHAR)pAssReqFrameBody + wOffset, (PVOID)&BasicRates[0][0], 4);
	wOffset += 4;

	// Other Elements
	AddIEsInRequest(((PUCHAR)pAssReqFrameBody + wOffset), &wBytesAdded, pSecurityInfo, 
					&pAssociationInfo->CurrentBSSinfo, pAssociationInfo->IPAddress,
					 pAssociationInfo->CwMin, pAssociationInfo->CwMax);
	
	// store the info of the variable elements we have added in the request
    VNetZeroMemory(pAssociationInfo->AssReassIEInfoBuffer, MAX_ASSOC_BUFF_LEN); 
    VNetMoveMemory(pAssociationInfo->AssReassIEInfoBuffer, ((PUCHAR)pAssReqFrameBody + wOffset), wBytesAdded); 
    pAssociationInfo->AssReassIEInfoBufferLength = wBytesAdded;

	wOffset += wBytesAdded;

	// calculate the frame length
	*pwFrameLength = 24 + wOffset;

	// store the info of the request
	pAssociationInfo->AssociationRespFrameBodyBytes	= 0;
	pAssociationInfo->AssociationMethod				= AssocMethodAssociationRequest;
	pAssociationInfo->LastAssReasReqBytes			= wOffset;  //FrameLen-24;

	VNetMoveMemory(pAssociationInfo->LastAssReasReqFrameBody, 
					(PUCHAR)pAssReqFrameBody, pAssociationInfo->LastAssReasReqBytes);

}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	BuildReAssociationRequestFrame()
//	Creates an reassociation request frame.
//
//	Parameters
//		OUT: PUCHAR					Pointer to a buffer that we will store the frame. 
//									Buffer has to be allocated by the caller.
//		IN:  PASSOCIATION_INFO		Pointer to the association info structure of the Adapter
//		IN:	 PSECURITY_INFO			Pointer to the security info structure of the Adapter
//		OUT: PUSHORT				Size of the frame returned in bytes
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////
void BuildReAssociationRequestFrame(PUCHAR pbFrame, PASSOCIATION_INFO pAssociationInfo, PSECURITY_INFO pSecurityInfo, PUSHORT pwFrameLength)
{
    s80211_MgmtFrm                  *pMgmtFrame;
    s80211_MGMT_REASS_REQUEST       *pReAssReqFrameBody;
    
    USHORT	wOffset = 0, wBytesAdded;

	// get a pointer to the header of the frame
    pMgmtFrame 			= (s80211_MgmtFrm*) pbFrame;
	// get a pointer to the data of the frame
	pReAssReqFrameBody 	= (s80211_MGMT_REASS_REQUEST*)(pMgmtFrame->Body);
	
	// Format Frame Header 
	BuildMgmtFrameHeader((PUCHAR)pMgmtFrame, (UCHAR)C80211_SUBTYPE_MGMT_REASS_REQUEST, NULL, pAssociationInfo, NULL);
	
	//
	// Capability Information
	//
	pReAssReqFrameBody->CapabilityInformation = (USHORT)C80211_MGMT_CAPABILITY_ESS;

	if(IsEncryptionEnabled(pSecurityInfo) || pSecurityInfo->InWpaMode)
		pReAssReqFrameBody->CapabilityInformation |= C80211_MGMT_CAPABILITY_Privacy;

	if(pAssociationInfo->PreambleType == PREAMBLE_TYPE_SHORT)
		pReAssReqFrameBody->CapabilityInformation |= C80211_MGMT_CAPABILITY_ShortPreamble;

	pAssociationInfo->AssocCapabilities = pReAssReqFrameBody->CapabilityInformation;

	//Listen Interval
	pReAssReqFrameBody->ListenInterval = pAssociationInfo->ListenInterval * pAssociationInfo->BeaconPeriod;
	wOffset = 4;

	VNetMoveMemory(((PUCHAR)pReAssReqFrameBody + wOffset), pAssociationInfo->LastBSSID, VNet_LENGTH_OF_ADDRESS); 
	wOffset += 6;

	// SSID element
	*((PUCHAR)pReAssReqFrameBody + wOffset) = C80211_MGMT_ElementID_SSID;
	wOffset += 1;

	*((PUCHAR)pReAssReqFrameBody + wOffset) = pAssociationInfo->DesiredSSIDsize;
	wOffset += 1;

	VNetMoveMemory(((PUCHAR)pReAssReqFrameBody + wOffset), pAssociationInfo->DesiredSSID, pAssociationInfo->DesiredSSIDsize);
	wOffset += pAssociationInfo->DesiredSSIDsize;

	// Supported Rates element
	*((PUCHAR)pReAssReqFrameBody + wOffset) = C80211_MGMT_ElementID_SupportedRates;
	wOffset += 1;

	*((PUCHAR)pReAssReqFrameBody + wOffset) = 4;
	wOffset += 1;

	VNetMoveMemory(((PUCHAR)pReAssReqFrameBody + wOffset), (PVOID)&BasicRates[0][0], 4);
	wOffset += 4;

#ifdef CCX2
	// CCKM reassociation request element
	if((pAssociationInfo->CurrentBSSinfo.Capabilities & BSS_CAPAPILITY_CCKM)
		&&pSecurityInfo->CCKMinfo.FastHandOffEnabled)
	{
		VNetMoveMemory(((PUCHAR)pReAssReqFrameBody + wOffset), (PUCHAR)pSecurityInfo->CCKMinfo.CCKMAssociationRequestIE, pSecurityInfo->CCKMinfo.CCKMAssociationRequestIElength);
		wOffset +=(USHORT)pSecurityInfo->CCKMinfo.CCKMAssociationRequestIElength;
	}
	//
#endif
	// other elements
	AddIEsInRequest(((PUCHAR)pReAssReqFrameBody + wOffset), &wBytesAdded, pSecurityInfo, 
					&pAssociationInfo->CurrentBSSinfo, pAssociationInfo->IPAddress,
					pAssociationInfo->CwMin, pAssociationInfo->CwMax);
	
	// store the info of the variable elements we have added in the request
    VNetZeroMemory(pAssociationInfo->AssReassIEInfoBuffer, MAX_ASSOC_BUFF_LEN); 
    VNetMoveMemory(pAssociationInfo->AssReassIEInfoBuffer, ((PUCHAR)pReAssReqFrameBody + wOffset), wBytesAdded); 
    pAssociationInfo->AssReassIEInfoBufferLength = wBytesAdded;

	wOffset += wBytesAdded;

	// calculate the frame length
	*pwFrameLength = 24 + wOffset;

	// store the info of the request
	pAssociationInfo->AssociationRespFrameBodyBytes	= 0;
	pAssociationInfo->AssociationMethod				= AssocMethodReAssociationRequest;

	pAssociationInfo->LastAssReasReqBytes			= wOffset; //FrameLen-24;
	VNetMoveMemory(pAssociationInfo->LastAssReasReqFrameBody, 
				(PUCHAR)pReAssReqFrameBody , pAssociationInfo->LastAssReasReqBytes);

}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	BuildDisAssociationRequestFrame()
//	Creates an disassociation request frame.
//
//	Parameters
//		OUT: PUCHAR					Pointer to a buffer that we will store the frame. 
//									Buffer has to be allocated by the caller.
//		IN:  PASSOCIATION_INFO		Pointer to the association info structure of the Adapter
//		IN:	 USHORT					The reason that will be reported in the request
//		OUT: PUSHORT				Size of the frame returned in bytes
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////
void BuildDisAssociationRequestFrame(PUCHAR pbFrame, PASSOCIATION_INFO pAssociationInfo, USHORT wReasonCode, PUSHORT pwFrameLength)
{
	s80211_MgmtFrm					*pMgmtFrame;
	s80211_MgmtBody_DeAss			*DeAssReqFrameBody;
	
	// get a pointer to the header of the frame
	pMgmtFrame = (s80211_MgmtFrm*)pbFrame;
	// get a pointer to the data of the frame
	DeAssReqFrameBody = (s80211_MgmtBody_DeAss*)(pMgmtFrame->Body);
	
	// Format Frame Header 
	BuildMgmtFrameHeader((PUCHAR)pMgmtFrame, (UCHAR)C80211_SUBTYPE_MGMT_DISASSOSIATION, NULL, pAssociationInfo, NULL);
	
	// Add the reason of the request 
	DeAssReqFrameBody->ReasonCode = wReasonCode;
		
	// calculate the frame length
	*pwFrameLength = 24 + 2;

	pAssociationInfo->AssociationRespFrameBodyBytes	= 0;
	pAssociationInfo->AssociationMethod				= AssocMethodAssociationRequest;

}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	BuildNullFrame()
//	Creates an NULL data frame.
//
//	Parameters
//		OUT: PUCHAR					Pointer to a buffer that we will store the frame. 
//									Buffer has to be allocated by the caller.
//		IN:  PASSOCIATION_INFO		Pointer to the association info structure of the Adapter
//		OUT: PUSHORT				Size of the frame returned in bytes
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////
void BuildNullFrame(PUCHAR pbFrame, PASSOCIATION_INFO pAssociationInfo, PUSHORT pwFrameLength)
{
	s80211_MgmtFrm					*pMgmtFrame;
	
	// get a pointer to the header of the frame
	pMgmtFrame = (s80211_MgmtFrm*)pbFrame;
	
	//
	// Create Data Frame Header
	//
	pMgmtFrame->FrameControl[0] = 0x48;		// NULL type
	pMgmtFrame->FrameControl[1]	= C80211_FRMCTRL_TO_DS;
	pMgmtFrame->DurationID = 0x00;

	VNetMoveMemory(pMgmtFrame->uDA, pAssociationInfo->CurrentBSSID, VNet_LENGTH_OF_ADDRESS); 
	VNetMoveMemory(pMgmtFrame->uSA, pAssociationInfo->CurrentAddress, VNet_LENGTH_OF_ADDRESS);
	VNetMoveMemory(pMgmtFrame->uBSSID, pAssociationInfo->CurrentBSSID, VNet_LENGTH_OF_ADDRESS); 

	pMgmtFrame->SequenceControl = 0x0000;
		
	// calculate the frame length (the frame has no data)
	*pwFrameLength = 24;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	AddIEsInRequest()
//	Adds WPA, WPA2, CCX information elements to a buffer that will used in a
//	request management frame, according to the security configuration of the Adapter.
//
//	Parameters
//		OUT: PUCHAR					Pointer to a buffer that we will store the frame. 
//									Buffer has to be allocated by the caller.
//		OUT: PUSHORT				Pointer to return the bytes written in the buffer.
//		IN:	 PSECURITY_INFO			Pointer to the security info structure of the Adapter.
//		IN:  PBSS_INFO_EX			Pointer to the current BSS info.
//		IN:  PUCHAR					Pointer to a 4-bytes buffer that contains the Adapter's IP address.
//		IN:  USHORT					The CWMin of the Adapter.
//		IN:  USHORT					The CWMax of the Adapter.
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////
void AddIEsInRequest(PUCHAR pbFrameBody, PUSHORT pwBytesAdded, PSECURITY_INFO pSecurityInfo, PBSS_INFO_EX pBssInfo, PUCHAR pbIPAddress, USHORT wCwMin, USHORT wCwMax)
{
	
	USHORT 	dOffset = 0;
	BOOLEAN fPwCsSet = 0;
#ifdef AES
	USHORT j;
#endif

#ifdef WPA2
	USHORT	PSN_PMKsLength=0;
#endif

	// first we add RSN IE, in the case that the driver has been configured for WPA2
	// add the selected AP supports WPA2.
#ifdef WPA2
	if((pBssInfo->Capabilities & BSS_CAPABILITY_RSN) && (pSecurityInfo->InWpaMode & WPA_MODE_2))
	{
		//id	
		pbFrameBody[dOffset] = C80211_MGMT_ElementID_RSN;
		dOffset += 1;
		//length (skip it for now, we will add it later)
		dOffset += 1;
		//version
		VNetMoveMemory((pbFrameBody + dOffset), &RSN_VERSION, 2);
		dOffset += 2;

		//group suite tkip
		if(pSecurityInfo->WPA_MixedModeDisabled)
		{
#ifdef AES
			if(pBssInfo->Wpa2Info.GroupSuite == OUI_CIPHER_SUITE_CCMP)
				VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_GROUP_AES, OUI_SIZE);
			else
#endif
				VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_GROUP_TKIP, OUI_SIZE);
		}
		else
		{
			if(pBssInfo->Wpa2Info.GroupSuite == OUI_CIPHER_SUITE_WEP_64)
				VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_GROUP_WEP_64, OUI_SIZE);
			else
			if(pBssInfo->Wpa2Info.GroupSuite == OUI_CIPHER_SUITE_WEP_128)
				VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_GROUP_WEP_128, OUI_SIZE);
#ifdef AES
			else
			if(pBssInfo->Wpa2Info.GroupSuite == OUI_CIPHER_SUITE_CCMP)
				VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_GROUP_AES, OUI_SIZE);
#endif					
			else
			{
				VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_GROUP_TKIP, OUI_SIZE);
			}
		}
		dOffset += OUI_SIZE;

		//pairwise suite count
		VNetMoveMemory((pbFrameBody + dOffset), &RSN_PW_SUITE_CNT, 2);
		dOffset += 2;
		//pw suite tkip
		if(pSecurityInfo->WPA_MixedModeDisabled)
		{
#ifdef AES
			if((pSecurityInfo->EncryptionStatus == _802_11Encryption3KeyAbsent) || 
			   (pSecurityInfo->EncryptionStatus == _802_11Encryption3Enabled))
				VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_PW_AES, OUI_SIZE);
			else
#endif					
				VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_PW_TKIP, OUI_SIZE);
		}
		else
		{
#ifdef AES
			if((pSecurityInfo->EncryptionStatus == _802_11Encryption3KeyAbsent) || 
			   (pSecurityInfo->EncryptionStatus == _802_11Encryption3Enabled))
			{
				for(j=0; j<pBssInfo->Wpa2Info.PairwiseSuiteCnt; j++)
				{
					if(pBssInfo->Wpa2Info.PairwiseSuiteList[j] == OUI_CIPHER_SUITE_CCMP)
					{
						VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_PW_AES, OUI_SIZE);
						fPwCsSet = 1;
						break;
					}
				}
			}
#endif
			if(!fPwCsSet)
			{
				VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_PW_TKIP, OUI_SIZE);
			}
		}		
		dOffset += OUI_SIZE;
		//auth suite count
		VNetMoveMemory((pbFrameBody + dOffset), &RSN_PW_SUITE_CNT, 2);
		dOffset += 2;
		//auth list
		if(pSecurityInfo->AuthenticationMode == _802_11AuthModeWPA2)
			VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_AUTH_1X, OUI_SIZE);
		else
			VNetMoveMemory((pbFrameBody + dOffset), &RSN_OUI_AUTH_PSK, OUI_SIZE);
		dOffset += OUI_SIZE;
		//wpa capabilities
		// CHECK IT
		VNetMoveMemory((pbFrameBody + dOffset), &RSN_CAP, 2);
		dOffset += 2;

		//PMKID
		PSN_PMKsLength = WPA2_AddPMKsInElement(pSecurityInfo, pBssInfo, (pbFrameBody + dOffset));
		dOffset += PSN_PMKsLength;

		// put the size of the produced IE.
		pbFrameBody[1] = 20 + (UCHAR)PSN_PMKsLength;
	}
	else
#endif

	// then we add WPA IE, in the case that the driver has been configured for WPA
	// add the selected AP supports WPA.
	if((pBssInfo->Capabilities & BSS_CAPABILITY_WPA)
#ifdef CCX2
		||(pBssInfo->Capabilities & BSS_CAPAPILITY_CCKM)
#endif
		)
	{
		//id	
		pbFrameBody[dOffset] = C80211_MGMT_ElementID_WPA;
		dOffset += 1;
		//length
		pbFrameBody[dOffset] = 24; //length
		dOffset += 1;
		//oui
		VNetMoveMemory((pbFrameBody + dOffset), WPA_OUI, 3);
		dOffset += 3;
		//oui type
		pbFrameBody[dOffset] = 0x01; 
		dOffset += 1;
		//version
		VNetMoveMemory((pbFrameBody + dOffset), &WPA_VERSION, 2);
		dOffset += 2;

		//group suite tkip
		if(pSecurityInfo->WPA_MixedModeDisabled)
		{
#ifdef AES
			if(pBssInfo->WpaInfo.GroupSuite == OUI_CIPHER_SUITE_CCMP)
				VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_GROUP_AES, OUI_SIZE);
			else
#endif
				VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_GROUP_TKIP, OUI_SIZE);
		}
		else
		{
			if(pBssInfo->WpaInfo.GroupSuite == OUI_CIPHER_SUITE_WEP_64)
				VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_GROUP_WEP_64, OUI_SIZE);
			else
			if(pBssInfo->WpaInfo.GroupSuite == OUI_CIPHER_SUITE_WEP_128)
				VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_GROUP_WEP_128, OUI_SIZE);
#ifdef AES
			else
			if(pBssInfo->WpaInfo.GroupSuite == OUI_CIPHER_SUITE_CCMP)
				VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_GROUP_AES, OUI_SIZE);
#endif
			else			
				VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_GROUP_TKIP, OUI_SIZE);

#ifdef CCX2
			if((pBssInfo->Capabilities & BSS_CAPAPILITY_CCX)&&
			   (pBssInfo->Capabilities & BSS_CAPAPILITY_CCKM))
			{
				if(pBssInfo->WpaInfo.GroupSuite == 0)
					VNetMoveMemory((pbFrameBody + dOffset), &CKIP_OUI, OUI_SIZE);
				else
				if(pBssInfo->WpaInfo.GroupSuite == 1)
					VNetMoveMemory((pbFrameBody + dOffset), &CKIPCMIC_OUI, OUI_SIZE);
				else
				if(pBssInfo->WpaInfo.GroupSuite == 2)
					VNetMoveMemory((pbFrameBody + dOffset), &CMIC_OUI, OUI_SIZE);			
			}
#endif		
		}
		dOffset += OUI_SIZE;

		//pairwise suite count
		VNetMoveMemory((pbFrameBody + dOffset), &WPA_PW_SUITE_CNT, 2);
		dOffset += 2;
		//pw suite tkip
		if(pSecurityInfo->WPA_MixedModeDisabled)
		{
			VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_PW_TKIP, OUI_SIZE);
		}
		else
		{
#ifdef AES
			if((pSecurityInfo->EncryptionStatus == _802_11Encryption3KeyAbsent) || (pSecurityInfo->EncryptionStatus == _802_11Encryption3Enabled))
			{
				for(j=0; j<pBssInfo->WpaInfo.PairwiseSuiteCnt; j++)
				{
					if(pBssInfo->WpaInfo.PairwiseSuiteList[j] == OUI_CIPHER_SUITE_CCMP)
					{
						VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_PW_AES, OUI_SIZE);
						fPwCsSet = 1;
						break;
					}
				}
			}
#endif
			if(!fPwCsSet)
			{
				VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_PW_TKIP, OUI_SIZE);
			}

#ifdef CCX2
			if((pBssInfo->Capabilities & BSS_CAPAPILITY_CCX)&&
			   (pBssInfo->Capabilities & BSS_CAPAPILITY_CCKM))
			{
				for(j=0; j<pBssInfo->WpaInfo.PairwiseSuiteCnt; j++)
				{
					if(pBssInfo->WpaInfo.PairwiseSuiteList[j] == 2)	// CMIC + CCKM
					{// overwrite the previous value
						VNetMoveMemory((pbFrameBody + dOffset), &CMIC_OUI, OUI_SIZE);
						break;
					}
					else if(pBssInfo->WpaInfo.PairwiseSuiteList[j] == 1) // CKIP+CMIC+CCKM
					{// overwrite the previous value
						VNetMoveMemory((pbFrameBody + dOffset), &CKIPCMIC_OUI, OUI_SIZE);
						break;
					}
					else if(pBssInfo->WpaInfo.PairwiseSuiteList[j] == 0) // CKIP + CCKM
					{// overwrite the previous value
						VNetMoveMemory((pbFrameBody + dOffset), &CKIP_OUI, OUI_SIZE);
						break;
					}
				}
			}
			else
			if(pBssInfo->Capabilities & BSS_CAPAPILITY_CCKM)
			{
				for(j=0; j<pBssInfo->WpaInfo.PairwiseSuiteCnt; j++)
				{
					if(pBssInfo->WpaInfo.PairwiseSuiteList[j] == OUI_CIPHER_SUITE_WEP_64) // CCKM + WEP64
					{// overwrite the previous value	
						VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_GROUP_WEP_64, OUI_SIZE);
						break;
					}
					else
					if(pBssInfo->WpaInfo.PairwiseSuiteList[j] == OUI_CIPHER_SUITE_WEP_128) // CCKM + WEP 128
					{// overwrite the previous value
						VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_GROUP_WEP_128, OUI_SIZE);
						break;
					}
				}
			}
#endif
		}

		dOffset += OUI_SIZE;
		//auth suite count
		VNetMoveMemory((pbFrameBody + dOffset), &WPA_PW_SUITE_CNT, 2);
		dOffset += 2;
		//auth list
		if(pSecurityInfo->AuthenticationMode == _802_11AuthModeWPA)
			VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_AUTH_1X, OUI_SIZE);
		else
			VNetMoveMemory((pbFrameBody + dOffset), &WPA_OUI_AUTH_PSK, OUI_SIZE);

#ifdef CCX2
		if(pBssInfo->Capabilities & BSS_CAPAPILITY_CCKM)
		{// overwrite the previous value
			VNetMoveMemory((pbFrameBody + dOffset), &CCKM_AUTH, OUI_SIZE);
		}
#endif

		dOffset += OUI_SIZE;
		//auth suite count
		VNetMoveMemory((pbFrameBody + dOffset), &WPA_CAP, 2);
		dOffset += 2;
	}


	// then we add CCX IE, in the case that the driver has been configured for CCX
#ifdef CCX
	if(pBssInfo->Capabilities & BSS_CAPABILITY_CCX)
	{
		DbgCommon(1,("CCX Element in Association Request\n");)

		pbFrameBody[dOffset] = C80211_MGMT_ElementID_CCX;
		dOffset += 1;

		pbFrameBody[dOffset] = 30;//length
		dOffset += 1;

		*(PUSHORT)(&CCX_ELEMENT_BODY[4]) = wCwMin;
		*(PUSHORT)(&CCX_ELEMENT_BODY[6]) = wCwMax;

		VNetMoveMemory((pbFrameBody + dOffset), CCX_ELEMENT_BODY, 30);
		dOffset += 30;
	
		DbgCommon(1,("CCX IP element in Association Request\n"))

		pbFrameBody[dOffset] = C80211_MGMT_ElementID_IP;
		dOffset += 1;

		VNetMoveMemory(&CCX_IP_BODY[5], pbIPAddress, OUI_SIZE);

		VNetMoveMemory((pbFrameBody + dOffset), CCX_IP_BODY, 11);
		dOffset += 11;
	}
#endif

	// then we add TAP DANCE IE, in the case that the driver has been configured for CCX
#ifdef TAP_DANCE
	DbgCommon(1,("TAP DANCE Element in Association Request\n"))

	VNetMoveMemory(pSecurityInfo->TapDance.MN_IPAddress, pbIPAddress, 4);
	
	pbFrameBody[dOffset] = ElementID_TAP_DANCE;
	dOffset += 1;

	pbFrameBody[dOffset] = 0x0D;	//length
	dOffset += 1;

	pbFrameBody[dOffset] = 0x01;	//ROAM_CTRL
	dOffset += 1;

	DbgCommon(1,("MN IP: %02X %02X %02X %02X\n",pSecurityInfo->TapDance.MN_IPAddress[0],pSecurityInfo->TapDance.MN_IPAddress[1],pSecurityInfo->TapDance.MN_IPAddress[2],pSecurityInfo->TapDance.MN_IPAddress[3]))
	DbgCommon(1,("HAP IP: %02X %02X %02X %02X\n",pSecurityInfo->TapDance.HAP_IPAddress[0],pSecurityInfo->TapDance.HAP_IPAddress[1],pSecurityInfo->TapDance.HAP_IPAddress[2],pSecurityInfo->TapDance.HAP_IPAddress[3]))
	DbgCommon(1,("FAP IP: %02X %02X %02X %02X\n",pSecurityInfo->TapDance.FA_IPAddress[0],pSecurityInfo->TapDance.FA_IPAddress[1],pSecurityInfo->TapDance.FA_IPAddress[2],pSecurityInfo->TapDance.FA_IPAddress[3]))

	// STA's IP address
	VNetMoveMemory((pbFrameBody + dOffset), pSecurityInfo->TapDance.MN_IPAddress, 4); 
	dOffset +=4;
	
	// STA's Home Agent IP address
	VNetMoveMemory((pbFrameBody + dOffset), pSecurityInfo->TapDance.HAP_IPAddress, 4); 
	dOffset +=4;

	// STA's Foreign Agent IP address
	VNetMoveMemory((pbFrameBody + dOffset), pSecurityInfo->TapDance.FA_IPAddress, 4); 
	dOffset +=4;
#endif

	// return the total bytes added.
	*pwBytesAdded = dOffset;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	MatchSecuritySettings()
//	Compares the settings of the Adapter to the settings of the BSS.
//
//	Parameters
//		IN:	 PSECURITY_INFO			Pointer to the security info structure of the Adapter.
//		IN:  PBSS_INFO_EX			Pointer to the current BSS info.
//		IN:  UCHAR					The operating mode that the Adapter has been configured to.
//		OUT: STATISTICS*			Pointer to the statistics structure of the Adapter.
//		
//
//	returns: 	TRUE 	if the settings match
//				FALSE 	if the settings do not match
/////////////////////////////////////////////////////////////////////////////
BOOLEAN MatchSecuritySettings(PSECURITY_INFO pSecurityInfo, PBSS_INFO_EX pBssInfo, UCHAR bOperatingMode, STATISTICS* pStats)
{
	BOOLEAN		fEncryptionIsOn = FALSE;
#ifdef CCX2
	BOOLEAN		fCCKMMatch = FALSE;
#endif
	USHORT		i;
	UCHAR		bEntriesInBlackListBeforeCheck;
	
#ifdef UNDER_CE
	SYSTEMTIME	tCurrentTime;
#else
	ULONG64		tCurrentTime;
#endif
	
	DbgCommon(1,("Match Security...\n"))

	// Get the Encryption Status of the Adapter
	if( pSecurityInfo->EncryptionStatus == _802_11Encryption1Enabled ||
		pSecurityInfo->EncryptionStatus == _802_11Encryption2Enabled ||
		pSecurityInfo->EncryptionStatus == _802_11Encryption3Enabled ||
		pSecurityInfo->EncryptionStatus == _802_11Encryption2KeyAbsent ||
		pSecurityInfo->EncryptionStatus == _802_11Encryption3KeyAbsent)
	{
		fEncryptionIsOn = TRUE;
	}
	else
	{
		fEncryptionIsOn = FALSE;
	}
	
	// if the Encryption Status of the Adapter mismatches with the AP's 
	// Encryption Status, return false.
	if(pBssInfo->EncryptionIsOn ^ fEncryptionIsOn)
	{	
		DbgCommon(1,("Encryption Status mismatch, UsingWep=%d, EncryptionOn=%d \n",pBssInfo->EncryptionIsOn, fEncryptionIsOn))
		return FALSE;
	}

	// In Ad-Hoc, we do not need the next checks.
	if(bOperatingMode == AD_HOC_MODE)
		return TRUE;

	// CCX Rogue AP List
#ifdef CCX
	if (fEncryptionIsOn && pSecurityInfo->CCXinfo.FswNetworkEAP_Set)
	{
		if(CCX_IsAPInRogueAPList(pSecurityInfo->CCXinfo, (PUCHAR)&(pBssInfo->BSSID)))
		{
			DbgCommon(1,("BSS is marked as Rogue AP...\n"))
			return FALSE;
		}
	}
#endif

	//If Encryption is OFF or InWpaMode is NONE, we do not need the next checks.
	if(!(fEncryptionIsOn && pSecurityInfo->InWpaMode))
		return TRUE;

	//WPA capabilities
	if (((!(pBssInfo->Capabilities & pSecurityInfo->InWpaMode)) && 
		 (pBssInfo->Capabilities || pSecurityInfo->InWpaMode)) //mismatch...
#ifdef CCX2
		 &&((pBssInfo->Capabilities & BSS_CAPAPILITY_CCKM) ^ (pSecurityInfo->CCKMinfo.FswCCKMEnable==0))
#endif
		)
	{
		DbgCommon(1,("(2) INFO: Mismatching capabilities: %d - %d\n", (UINT)pBssInfo->Capabilities, pSecurityInfo->InWpaMode))
		return FALSE;
	}

	//WPA Mic Error Black List
	VNetGetSystemTime((PVOID)&tCurrentTime);

	bEntriesInBlackListBeforeCheck = pSecurityInfo->MicBlackListEntries;
	if(!WPA_IsOkToConnect(pSecurityInfo->MicBlackList, 
						  MAX_BSS_ENTRIES, 
						  &pSecurityInfo->MicBlackListEntries, 
						  (PUCHAR)&(pBssInfo->BSSID), 
						  (PVOID)(&tCurrentTime)) )
	{
		DbgCommon(1,("BSS is in MIC Error black List\n"))
		return FALSE;
	}
	else
	{
		if(pSecurityInfo->MicBlackListEntries < bEntriesInBlackListBeforeCheck)
				pStats->WpaBlackListEntries -= (bEntriesInBlackListBeforeCheck - pSecurityInfo->MicBlackListEntries);

		DbgCommon(1,("MatchSecurity: Stats->WpaBlackListEntries=%d EntriesBeforeCheck=%d MicBlackListEntries=%d\n",(UINT)pStats->WpaBlackListEntries,bEntriesInBlackListBeforeCheck,pSecurityInfo->MicBlackListEntries))
	}


#ifdef WPA2
	if(pSecurityInfo->InWpaMode == WPA_MODE_2)
	{
		// if the mixed mode is disabled we need to have the group cipher suite 
		// set to the driver's WPA encryption type.
		if(pSecurityInfo->WPA_MixedModeDisabled)
		{

			// Group suite must be either AES or TKIP
			if( (pBssInfo->Wpa2Info.GroupSuite != OUI_CIPHER_SUITE_TKIP)
#ifdef AES
				&&(pBssInfo->Wpa2Info.GroupSuite != OUI_CIPHER_SUITE_CCMP)

#endif
			  )
			  return FALSE;
		}

		// at least one pairwise cipher suite must be AES or TKIP.
		for(i=0; i<pBssInfo->Wpa2Info.PairwiseSuiteCnt; i++)
		{
			if(((pBssInfo->Wpa2Info.PairwiseSuiteList[i] == OUI_CIPHER_SUITE_TKIP) && 
				 ((pSecurityInfo->EncryptionStatus == _802_11Encryption2KeyAbsent) || (pSecurityInfo->EncryptionStatus == _802_11Encryption2Enabled) ) ) 
#ifdef AES
			   ||
			   ((pBssInfo->Wpa2Info.PairwiseSuiteList[i] == OUI_CIPHER_SUITE_CCMP) &&
				 ((pSecurityInfo->EncryptionStatus == _802_11Encryption3KeyAbsent) || (pSecurityInfo->EncryptionStatus == _802_11Encryption3Enabled) ) )
#endif
			  )
				break;

			// if the next check is true this means that we have
			// exhausted the Pairwise Suites without having found 
			// an RSN cipher.
			if(i == pBssInfo->Wpa2Info.PairwiseSuiteCnt - 1)
			{
				DbgCommon(1,("INFO: RSN Mismatching Capabilities: PairwiseSuite != RSN\n"))
				return FALSE;
			}
		}
	
		// WPA2 Authentication Suite

		if(pSecurityInfo->AuthenticationMode == _802_11AuthModeWPA2)
		{
			fEncryptionIsOn = FALSE;
			for(i=0;i<pBssInfo->Wpa2Info.AuthMgmtSuiteCnt;i++)
			{
				if(pBssInfo->Wpa2Info.AuthMgmtSuiteList[i] == 1)
				{
					fEncryptionIsOn = TRUE;
					break;
				}
			}
		}
		else
		if(pSecurityInfo->AuthenticationMode == _802_11AuthModeWPA2PSK)
		{
			fEncryptionIsOn = FALSE;
			for(i=0;i<pBssInfo->Wpa2Info.AuthMgmtSuiteCnt;i++)
			{
				if(pBssInfo->Wpa2Info.AuthMgmtSuiteList[i] == 2)
				{
					fEncryptionIsOn = TRUE;
					break;
				}
			}
		}
		else
		{
//			DbgCommon(1,("INFO: AuthenticationMode: %d\n", pSecurityInfo->AuthenticationMode))
		}

		return fEncryptionIsOn;
	}
	else
#endif
	if((pSecurityInfo->InWpaMode == WPA_MODE_1)
#ifdef CCX2
		||(pBssInfo->Capabilities & BSS_CAPAPILITY_CCKM)
#endif
		)
	{
		// if the mixed mode is disabled we need to have the group cipher suite 
		// set to the driver's WPA encryption type.
		if(pSecurityInfo->WPA_MixedModeDisabled)
		{
#ifdef AES
			if( !( ((pBssInfo->WpaInfo.GroupSuite == OUI_CIPHER_SUITE_TKIP) && 
				    ((pSecurityInfo->EncryptionStatus == _802_11Encryption2KeyAbsent) || 
					 (pSecurityInfo->EncryptionStatus == _802_11Encryption2Enabled) ||
					 (pSecurityInfo->EncryptionStatus == _802_11Encryption3KeyAbsent) || 
					 (pSecurityInfo->EncryptionStatus == _802_11Encryption3Enabled)) 
				   ) ||
				   ((pBssInfo->WpaInfo.GroupSuite == OUI_CIPHER_SUITE_CCMP) &&
				    ((pSecurityInfo->EncryptionStatus == _802_11Encryption3KeyAbsent) || 
					 (pSecurityInfo->EncryptionStatus == _802_11Encryption3Enabled)) 
				   )
				 )
			  )
#else
			if(pBssInfo->WpaInfo.GroupSuite != OUI_CIPHER_SUITE_TKIP)
#endif
			{
				DbgCommon(1,("INFO: WPA_MixedModeDisabled: GroupSuite = %d\n", pBssInfo->WpaInfo.GroupSuite))
				return FALSE;
			}
			
		}

#ifdef CCX2
		if(!(pBssInfo->Capabilities & BSS_CAPAPILITY_CCKM))
		{// in case of CCKM the pairwise suite is not necessary TKIP
#endif

		// at least one pairwise cipher suite must be WPA.
		for(i=0; i<pBssInfo->WpaInfo.PairwiseSuiteCnt; i++)
		{
#ifdef AES
			if( ((pBssInfo->WpaInfo.PairwiseSuiteList[i] == OUI_CIPHER_SUITE_TKIP) && 
				 ((pSecurityInfo->EncryptionStatus == _802_11Encryption2KeyAbsent) || (pSecurityInfo->EncryptionStatus == _802_11Encryption2Enabled) ) ) ||
				((pBssInfo->WpaInfo.PairwiseSuiteList[i] == OUI_CIPHER_SUITE_CCMP) && !pSecurityInfo->WPA_MixedModeDisabled &&
				 ( (pSecurityInfo->EncryptionStatus == _802_11Encryption3KeyAbsent) || (pSecurityInfo->EncryptionStatus == _802_11Encryption3Enabled) ) )
			  )
#else
			if( (pBssInfo->WpaInfo.PairwiseSuiteList[i] == OUI_CIPHER_SUITE_TKIP) && 
				((pSecurityInfo->EncryptionStatus == _802_11Encryption2KeyAbsent) || (pSecurityInfo->EncryptionStatus == _802_11Encryption2Enabled))
			  )
#endif
				break;

			// if the next check is true this means that we have
			// exhausted the Pairwise Suites without having found 
			// a WPA cipher.
			if(i == pBssInfo->WpaInfo.PairwiseSuiteCnt - 1)
			{
				DbgCommon(1,("INFO: WPA Mismatching Capabilities: PairwiseSuite != WPA\n"))
#ifdef CCX2
				if (!(pBssInfo->Capabilities & BSS_CAPAPILITY_CCKM))
#endif
				return FALSE;
			}
		}
#ifdef CCX2
		}// in case of CCKM the pairwise suite is not necessary TKIP
#endif
	
		// WPA Authentication Suite

		if(pSecurityInfo->AuthenticationMode == _802_11AuthModeWPA)
		{
			fEncryptionIsOn = FALSE;

			for(i=0;i<pBssInfo->WpaInfo.AuthMgmtSuiteCnt;i++)
			{
				if(pBssInfo->WpaInfo.AuthMgmtSuiteList[i] == 1)
				{
					fEncryptionIsOn = TRUE;
#ifndef CCX2
					break;
#endif
				}
#ifdef CCX2
				// CCKM Authentication Suite
				else
				if(pBssInfo->WpaInfo.AuthMgmtSuiteList[i] == 0)
				{
					if((pSecurityInfo->CCKMinfo.FswCCKMEnable)&&
					   (pBssInfo->Capabilities & BSS_CAPAPILITY_CCKM))
					{
						fEncryptionIsOn = TRUE;
						break;
					}
				}
#endif
			}
		}
		else
		if(pSecurityInfo->AuthenticationMode == _802_11AuthModeWPAPSK)
		{
			fEncryptionIsOn = FALSE;
			for(i=0;i<pBssInfo->WpaInfo.AuthMgmtSuiteCnt;i++)
			{
				if(pBssInfo->WpaInfo.AuthMgmtSuiteList[i] == 2)
				{
					fEncryptionIsOn = TRUE;
					break;
				}
			}
		}
		DbgCommon(1,("Match Security: %d\n",fEncryptionIsOn))
		
		return fEncryptionIsOn;
	}
	else //in non-WPA mode
		return TRUE;

}



////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	SelectCiphers()
//	Returns the cipher type that we must use to transmit a packet.
// 	Send the unicast packets with the pairwise key if available 
// 	else send the unicast packets with the group key. 
// 	Send the multicast packets with the group key if available 
// 	else if the encryption is WEP we use the pairwise key.
//
//	Parameters
//		IN:	 PSECURITY_INFO			Pointer to the security info structure of the Adapter.
//		IN:  BOOLEAN				Indicates whether the packet is broadcast or not.
//		IN:  BOOLEAN				Indicates whether the packet is EAPOL or not.
//		IN:  BOOLEAN				Forces the function to return no encryption.
//		OUT: PUCHAR					Pointer to store the returned cipher type.
//		OUT: PUCHAR					Pointer to store the returned cipher size.
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////
VOID SelectCiphers(PSECURITY_INFO pSecurityInfo, BOOLEAN fIsBroadcast, BOOLEAN* pfTxEapPending, BOOLEAN fForceUnencrypted, PUCHAR pbCipherType, PUCHAR pbCipherLength)
{

	if(*pfTxEapPending)
	{
		if(!pSecurityInfo->InWpaMode)
		{
#if DBG
			DbgCommon(1, ("Tx: EAP Transmit unencrypted\n"))
#endif
			*pbCipherType = CIPHER_SUITE_NONE;
			*pbCipherLength = 0;

			*pfTxEapPending = 0;
			
			return;
		}
		
		*pfTxEapPending = 0;
	}
	
	if(fForceUnencrypted)
	{
		*pbCipherType = CIPHER_SUITE_NONE;
		*pbCipherLength = 0;

		return;
	}
	
#ifdef CCX
	if(pSecurityInfo->CCXinfo.CkipEnabled && pSecurityInfo->CCXinfo.CkipKeyPermutation)
	{
		*pbCipherType = CIPHER_SUITE_CCX;
		*pbCipherLength = 8;
		return;
	}
#endif

	if(fIsBroadcast)
	{
		*pbCipherType = pSecurityInfo->EncryptionInfo.GroupCipherSuite;
		
		if ((pSecurityInfo->EncryptionInfo.GroupCipherSuite != CIPHER_SUITE_WEP_64)&&
			(pSecurityInfo->EncryptionInfo.GroupCipherSuite != CIPHER_SUITE_WEP_128)&&
			(pSecurityInfo->EncryptionInfo.GroupCipherSuite != CIPHER_SUITE_TKIP)
#ifdef AES
			&&(pSecurityInfo->EncryptionInfo.GroupCipherSuite != CIPHER_SUITE_AES)
#endif
		   )
		{
			*pbCipherType = CIPHER_SUITE_NONE;
		}

		if ( *pbCipherType == CIPHER_SUITE_NONE &&
			(pSecurityInfo->EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_WEP_64 || 
			 pSecurityInfo->EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_WEP_128))
		{
			*pbCipherType = pSecurityInfo->EncryptionInfo.PairwiseCipherSuite;
		}
	}
	else
	{//Infrastructure
		
		*pbCipherType = pSecurityInfo->EncryptionInfo.PairwiseCipherSuite;
		
		if ((pSecurityInfo->EncryptionInfo.PairwiseCipherSuite != CIPHER_SUITE_WEP_64)&&
			(pSecurityInfo->EncryptionInfo.PairwiseCipherSuite != CIPHER_SUITE_WEP_128)&&
			(pSecurityInfo->EncryptionInfo.PairwiseCipherSuite != CIPHER_SUITE_TKIP)
#ifdef AES
			&&(pSecurityInfo->EncryptionInfo.PairwiseCipherSuite != CIPHER_SUITE_AES)
#endif
		   )
		{
			*pbCipherType = CIPHER_SUITE_NONE;
		}
		
		if (*pbCipherType == CIPHER_SUITE_NONE)
		{
			*pbCipherType = pSecurityInfo->EncryptionInfo.GroupCipherSuite;

			if ((pSecurityInfo->EncryptionInfo.GroupCipherSuite != CIPHER_SUITE_WEP_64)&&
				(pSecurityInfo->EncryptionInfo.GroupCipherSuite != CIPHER_SUITE_WEP_128)&&
				(pSecurityInfo->EncryptionInfo.GroupCipherSuite != CIPHER_SUITE_TKIP)
#ifdef AES
				&&(pSecurityInfo->EncryptionInfo.GroupCipherSuite != CIPHER_SUITE_AES)
#endif
			   )
			{
				*pbCipherType = CIPHER_SUITE_NONE;
			}
		}
	}

	if ((*pbCipherType) == CIPHER_SUITE_NONE)
	{
		*pbCipherLength = 0;
	}
	else
	if ((*pbCipherType) == CIPHER_SUITE_WEP_64||(*pbCipherType) == CIPHER_SUITE_WEP_128)
	{
		*pbCipherLength = 8;
	}
	else
	if ((*pbCipherType) == CIPHER_SUITE_TKIP)
	{
		*pbCipherLength = 12;
	}
#ifdef AES
	else
	if ((*pbCipherType) == CIPHER_SUITE_AES)
	{
		*pbCipherLength = 16;
	}
#endif
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	IsTxKeyAvailable()
//	Indicates whether there is a key available for transmission.
//
//	Parameters
//		IN:	 PSECURITY_INFO			Pointer to the security info structure of the Adapter.
//		
//
//	returns: TRUE	if there is an available key
//			 FALSE if there is not an available key
/////////////////////////////////////////////////////////////////////////////
BOOLEAN IsTxKeyAvailable(PSECURITY_INFO pSecurityInfo)
{
	if(pSecurityInfo->PairwiseKeyAvailable || pSecurityInfo->GroupKeyAvailable)
		return TRUE;
	else
		return FALSE;
}


#ifdef WPA2

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	WPA2_AddPMKIDInList()
//	Stores a PMKID list to the Adapter.
//
//	Parameters
//		OUT: PSECURITY_INFO			Pointer to the security info structure of the Adapter.
//		IN:  P__802_11_PMKID		Pointer to the PMKID list.
//		
//
//	returns: TRUE  always
/////////////////////////////////////////////////////////////////////////////
BOOLEAN WPA2_AddPMKIDInList(PSECURITY_INFO pSecurityInfo, P__802_11_PMKID pPMKID)
{
	ULONG i;

	pSecurityInfo->PMKIDList.BSSIDInfoCount = pPMKID->BSSIDInfoCount;
	
	// Store as many entries as BSSIDInfoCount indicates.
	VNetMoveMemory((PUCHAR)&(pSecurityInfo->PMKIDList.BSSIDInfo), (PUCHAR)&pPMKID->BSSIDInfo, pPMKID->BSSIDInfoCount*sizeof(BSSID_INFO));
	
	return TRUE;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	WPA2_AddPMKsInElement()
//	Retrieves any PMKID from the stored and are availaible for the selected BSS.
//	Then, it adds them to a caller suplied buffer, to be used in a (re)association request.
//
//	Parameters
//		IN:	 PSECURITY_INFO		Pointer to the security info structure of the Adapter.
//		IN:  PBSS_INFO_EX		Pointer to the current BSS info.
//		OUT: PUCHAR				Pointer to a buffer that we will store the PMKIDs. 
//								Buffer has to be allocated by the caller.
//		
//
//	returns: USHORT  the bytes written in the buffer.
/////////////////////////////////////////////////////////////////////////////
USHORT WPA2_AddPMKsInElement(PSECURITY_INFO pSecurityInfo, PBSS_INFO_EX pBssInfo, PUCHAR pbIEBuffer)
{
	UCHAR i, ValidEntries=0;

	// We go throught the PMKID list, to look for entries that have a matching BSSID.
	for(i=0; i<pSecurityInfo->PMKIDList.BSSIDInfoCount; i++)
	{
		if(!VNetMemCompare(pBssInfo->BSSID, pSecurityInfo->PMKIDList.BSSIDInfo[i].BSSID, 6))
		{
			// Copy the valid entry to the buffer
			VNetMoveMemory(pbIEBuffer + (ValidEntries * sizeof(_802_11_PMKID)) + 2, pSecurityInfo->PMKIDList.BSSIDInfo[i].PMKID, sizeof(_802_11_PMKID));
			ValidEntries++;
			DbgCommon(1,("MATCHING PMKID entry FOUND..\n"))
		}
	}

	if(ValidEntries)
	{
		// the first 2 bytes of the buffer must contain the number of the PMKIDs
		*((PUSHORT)pbIEBuffer) = (USHORT)ValidEntries;
			
		DbgCommon(1,("Num of MATCHING PMKID entries = %d\n", ValidEntries))

		// return the number of bytes written in the buffer, so the caller
		// can update the offset, and calculate the whole request size.
		return (ValidEntries * sizeof(_802_11_PMKID_VALUE)) + sizeof(USHORT);
	}

	// return the number of bytes written
	return 0;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	WPA2_BuildAuthEncryptionCapability()
//	Builds a list with the supported pairs of the authentication-encryption settings.
//
//	Parameters
//		OUT: P_802_11_AUTHENTICATION_ENCRYPTION		Pointer to a buffer that we will store the pairs. 
//													Buffer has to be allocated by the caller.
//		
//
//	returns: ULONG  the bytes written in the buffer.
/////////////////////////////////////////////////////////////////////////////
ULONG WPA2_BuildAuthEncryptionCapability(P_802_11_AUTHENTICATION_ENCRYPTION pAuthEncryptionList)
{
	pAuthEncryptionList[0].AuthModeSupported		= _802_11AuthModeOpen;
	pAuthEncryptionList[0].EncryptStatusSupported	= _802_11EncryptionDisabled;
	pAuthEncryptionList[1].AuthModeSupported		= _802_11AuthModeOpen;
	pAuthEncryptionList[1].EncryptStatusSupported	= _802_11Encryption1Enabled;
	pAuthEncryptionList[2].AuthModeSupported		= _802_11AuthModeShared;
	pAuthEncryptionList[2].EncryptStatusSupported	= _802_11EncryptionDisabled;
	pAuthEncryptionList[3].AuthModeSupported		= _802_11AuthModeShared;
	pAuthEncryptionList[3].EncryptStatusSupported	= _802_11Encryption1Enabled;
	pAuthEncryptionList[4].AuthModeSupported		= _802_11AuthModeWPA;
	pAuthEncryptionList[4].EncryptStatusSupported	= _802_11Encryption2Enabled;

#ifdef AES
	pAuthEncryptionList[5].AuthModeSupported		= _802_11AuthModeWPA;
	pAuthEncryptionList[5].EncryptStatusSupported	= _802_11Encryption3Enabled;
	pAuthEncryptionList[6].AuthModeSupported		= _802_11AuthModeWPAPSK;
	pAuthEncryptionList[6].EncryptStatusSupported	= _802_11Encryption2Enabled;
	pAuthEncryptionList[7].AuthModeSupported		= _802_11AuthModeWPAPSK;
	pAuthEncryptionList[7].EncryptStatusSupported	= _802_11Encryption3Enabled;
	pAuthEncryptionList[8].AuthModeSupported		= _802_11AuthModeWPANone;
	pAuthEncryptionList[8].EncryptStatusSupported	= _802_11Encryption2Enabled;
	pAuthEncryptionList[9].AuthModeSupported		= _802_11AuthModeWPANone;
	pAuthEncryptionList[9].EncryptStatusSupported	= _802_11Encryption3Enabled;
	pAuthEncryptionList[10].AuthModeSupported		= _802_11AuthModeWPA2;
	pAuthEncryptionList[10].EncryptStatusSupported	= _802_11Encryption2Enabled;
	pAuthEncryptionList[11].AuthModeSupported		= _802_11AuthModeWPA2;
	pAuthEncryptionList[11].EncryptStatusSupported	= _802_11Encryption3Enabled;
	pAuthEncryptionList[12].AuthModeSupported		= _802_11AuthModeWPA2PSK;
	pAuthEncryptionList[12].EncryptStatusSupported	= _802_11Encryption2Enabled;
	pAuthEncryptionList[13].AuthModeSupported		= _802_11AuthModeWPA2PSK;
	pAuthEncryptionList[13].EncryptStatusSupported	= _802_11Encryption3Enabled;

	return 14;
#else
	pAuthEncryptionList[5].AuthModeSupported		= _802_11AuthModeWPAPSK;
	pAuthEncryptionList[5].EncryptStatusSupported	= _802_11Encryption2Enabled;
	pAuthEncryptionList[6].AuthModeSupported		= _802_11Encryption2Enabled;
	pAuthEncryptionList[7].AuthModeSupported		= _802_11AuthModeWPA2;
	pAuthEncryptionList[7].EncryptStatusSupported	= _802_11Encryption2Enabled;
	pAuthEncryptionList[8].AuthModeSupported		= _802_11AuthModeWPA2PSK;
	pAuthEncryptionList[8].EncryptStatusSupported	= _802_11Encryption2Enabled;

	return 9;
#endif
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	WPA2_BuildPMKIDCandidateList()
//	Retrieves any PMKID from the stored and are availaible for the selected BSS.
//	Then, it adds them to a caller suplied buffer, to be used in a (re)association request.
//
//	Parameters
//		OUT: PPMKID_CANDIDATELIST	Pointer to the BSS candidate list for pre-authentication.
//									Buffer has to be allocated by the caller.
//		IN:  BSS_INFO_EX*			Pointer to the a BSS info list, that we will search for candidates.
//		IN:  UCHAR					Number of entries in the BSS_INFO_EX list.
//		IN:  BSS_INFO_EX*			Pointer to the current BSS info.
//		
//
//	returns: TRUE	always
/////////////////////////////////////////////////////////////////////////////
BOOLEAN WPA2_BuildPMKIDCandidateList(PPMKID_CANDIDATELIST pPMKIDCandidateList, BSS_INFO_EX* pBSSinfoEx, UCHAR bBSSinfoExEntries, BSS_INFO_EX* pCurrentBSSinfo)
{
	UCHAR i;

	// clear the entries count
	pPMKIDCandidateList->NumCandidates = 0;

	// check all the entries in the BSSinfo list
	for(i=0; i<bBSSinfoExEntries; i++)
	{
		// the BSS must have the pre-authentication bit on
//		if(pBSSinfoEx[i].Wpa2Info.WpaCapabilities[0] & RSN_PREAUTHENTICATION_CAPABILITY_ON)
		{
			// we accept APs with the same SSID
			if(pCurrentBSSinfo->SSIDsize == pBSSinfoEx[i].SSIDsize)
			{
				if(!VNetMemCompare(pCurrentBSSinfo->SSID, pBSSinfoEx[i].SSID, pCurrentBSSinfo->SSIDsize))
				{
					// Add the BSSID in the candidates list
					VNetMoveMemory(pPMKIDCandidateList->CandidateList[pPMKIDCandidateList->NumCandidates].BSSID, pBSSinfoEx[i].BSSID, 6);
					pPMKIDCandidateList->CandidateList[pPMKIDCandidateList->NumCandidates].Flags = RSN_PREAUTHENTICATION_CAPABILITY_ON;
					pPMKIDCandidateList->NumCandidates ++;

					// if the list is full, return.
					if(pPMKIDCandidateList->NumCandidates == MAX_PKMID_INFO_ENTRIES)
						break;
				}
			}
		}
	}

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	WPA2_FlashPMKIDCandidateList()
//	Flashes the PMKID candicate list
//
//	Parameters
//		IN: PPMKID_CANDIDATELIST		Pointer to a PMKID candidates list. 
//										Buffer has to be allocated by the caller.
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////
VOID WPA2_FlashPMKIDCandidateList(PPMKID_CANDIDATELIST pPMKIDCandidateList)
{
	VNetZeroMemory((PUCHAR)pPMKIDCandidateList, sizeof(PMKID_CANDIDATELIST));
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	WPA2_ComparePMKIDCandidateLists()
//	Compares the two PMKID candidate lists of the Adapter.
//	We use it to compare the list we have already indicated with any new.
//
//	Parameters
//		IN: PSECURITY_INFO		Pointer to the security info of the Adapter. 
//		
//
//	returns: TRUE	If the lists are identical.
//			 FALSE	If the lists are not identical.
/////////////////////////////////////////////////////////////////////////////
BOOLEAN	WPA2_ComparePMKIDCandidateLists(PSECURITY_INFO pSecurityInfo)
{
	BOOLEAN ReturnValue = TRUE;
	UCHAR	i, j;

	// if the CandidateList is empty we assume that nothing has changed
	// so we keep the CandidateList2App and return ok.
	if(!pSecurityInfo->PMKIDCandidateList.NumCandidates)
		return TRUE;

	// if the CandidateList2App is empty while the CandidateList is not, 
	// we update the CandidateList2App list and indicate the change.
	if(pSecurityInfo->PMKIDCandidateList.NumCandidates && !pSecurityInfo->PMKIDCandidateList2App.NumCandidates)
	{
		VNetMoveMemory((PUCHAR)&pSecurityInfo->PMKIDCandidateList2App, (PUCHAR)&pSecurityInfo->PMKIDCandidateList, sizeof(PMKID_CANDIDATELIST));
		return FALSE;
	}

	// we search for each member of the CandidateList in the CandidateList2App
	for(i=0; i<pSecurityInfo->PMKIDCandidateList.NumCandidates; i++)
	{
		for(j=0; j<pSecurityInfo->PMKIDCandidateList2App.NumCandidates; j++)
		{
			// if we find a match, we procced to the next CandidateList member.
			if(!VNetMemCompare(pSecurityInfo->PMKIDCandidateList.CandidateList[i].BSSID, 
						  pSecurityInfo->PMKIDCandidateList2App.CandidateList[j].BSSID, 6)
			  )
			  break;
		}

		// if the indexs are equal, it means that the CandidateList entry is new, 
		// so we add it in the CandidateList2App list and indicate the change, 
		// as long as the new entry fits in the CandidateList2App.
		if( (j == pSecurityInfo->PMKIDCandidateList2App.NumCandidates) && 
			(pSecurityInfo->PMKIDCandidateList2App.NumCandidates != MAX_PKMID_INFO_ENTRIES) )
		{
			VNetMoveMemory(pSecurityInfo->PMKIDCandidateList2App.CandidateList[pSecurityInfo->PMKIDCandidateList2App.NumCandidates].BSSID, 
						   pSecurityInfo->PMKIDCandidateList.CandidateList[i].BSSID, 6);
			pSecurityInfo->PMKIDCandidateList2App.CandidateList[pSecurityInfo->PMKIDCandidateList2App.NumCandidates].Flags = 1;
			pSecurityInfo->PMKIDCandidateList2App.NumCandidates ++;

			ReturnValue = FALSE;
		}
	}

	return ReturnValue;
}
#endif


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	GetElementFromBeaconOrProbe()
//	Searches a buffer that contains a beacon or probe for a specific IE 
// 	and stores the IE's info to a caller supplied buffer.
//
//	Parameters
//		IN:	 PUCHAR			Pointer to the buffer that contains the beacon or probe.
//		IN:  USHORT			Size of the beacon/probe buffer.
//		IN:  UCHAR			The element's ID we are looking for.
//		IN:  PUCHAR			Pointer to the OUI that corresponds to the Element ID.
//		OUT: PUCHAR			Pointer to store the returned IE.
//							Buffer has to be allocated by the caller.
//		OUT: PUSHORT		Pointer to store the returned IE size, in bytes.
//		
//
//	returns: TRUE	if the element has been found
//			 FALSE  if the element has not been found
/////////////////////////////////////////////////////////////////////////////
BOOLEAN GetElementFromBeaconOrProbe(PUCHAR pbBeaconOrProbeFrame, USHORT wFrameLen, UCHAR bElementID, PUCHAR pbOUI, 
							PUCHAR pbElementBody, PUSHORT pwElementBodyLen)
{
	USHORT			wOffset, wLen;
	UCHAR			bval1,bval2,bval3,bval4;
	
	wOffset = 12;

	if(wOffset > (*pbBeaconOrProbeFrame))
		return FALSE;

	while(wOffset < wFrameLen)
	{
		wLen = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 1));
	
		if(*(pbBeaconOrProbeFrame + wOffset) == bElementID) 
		{
			if(wLen > *pwElementBodyLen)
				return FALSE;

#ifdef CCX2
			if(bElementID == ELEMENT_ID_CELL_POWER_INFO)			/*  ELEMENT 0x96*/
			{
				if (wLen >= 6)
				{
					bval1 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 2));
					bval2 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 3));
					bval3 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 4));
					bval4 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 5));

					// CCX TPC information element
					if((bval1==pbOUI[0]) && (bval2==pbOUI[1]) && (bval3==pbOUI[2]) && (bval4 == pbOUI[3])) 
					{
						VNetMoveMemory(pbElementBody, pbBeaconOrProbeFrame+wOffset+2, wLen);
						*pwElementBodyLen = wLen;
						return TRUE;
					}
				}
			}
			else
#endif
#ifdef CCX
			if(bElementID == C80211_MGMT_ElementID_CCX)
			{
				if( *(pbBeaconOrProbeFrame + wOffset + 10) & 0x18 )
				{
					VNetMoveMemory(pbElementBody, pbBeaconOrProbeFrame+wOffset+2, wLen);
					*pwElementBodyLen = wLen;
					return TRUE;
				}
			}
			else
#endif
			if(bElementID == C80211_MGMT_ElementID_WPA)
			{	
				// validate the minimum WPA element size
				if(wLen >= MIN_WPA_ELEMENT_DATA_SIZE)  /* 6 BYTES*/
				{
					bval1 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 2));
					bval2 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 3));
					bval3 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 4));
					bval4 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 5));

					// WPA information element
					if((bval1==pbOUI[0]) && (bval2==pbOUI[1])&&(bval3==pbOUI[2])&&(bval4 == pbOUI[3])) 
					{
						// if the length is the minimum, we accept the element
						if(wLen == MIN_WPA_ELEMENT_DATA_SIZE)
						{
							VNetMoveMemory(pbElementBody, pbBeaconOrProbeFrame+wOffset+2, wLen);
							*pwElementBodyLen = wLen;
							return TRUE;
						}
						else
						{
#ifdef CCX2
							// Group Suite Selectors MUST not be 0
							bval1 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 8));
							bval2 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 9));
							bval3 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 10));
							bval4 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 11));
							
							if((bval1==CKIP_OUI[0])&&(bval2==CKIP_OUI[1])&&(bval3==CKIP_OUI[2])&&(bval4==CKIP_OUI[3]))
							{// - exception - CCKM+CKIP case where Group Suite = 00:40:96:00 
								VNetMoveMemory(pbElementBody, pbBeaconOrProbeFrame+wOffset+2, wLen);
								*pwElementBodyLen = wLen;
								return TRUE;
							}
							else
#endif

							// Group Suite Selectors MUST not be 0.
							if( *(pbBeaconOrProbeFrame + wOffset + 11) )
							{
								VNetMoveMemory(pbElementBody, pbBeaconOrProbeFrame+wOffset+2, wLen);
								*pwElementBodyLen = wLen;
								return TRUE;
							}
						}
					}
				}
#ifdef CCX2
				if (wLen >= 5)
				{
					bval1 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 2));
					bval2 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 3));
					bval3 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 4));
					bval4 = (*((PUCHAR)(pbBeaconOrProbeFrame) + wOffset + 5));
					
					// CCX version information element
					if((bval1==pbOUI[0]) && (bval2==pbOUI[1])&&(bval3==pbOUI[2])&& (bval4 == pbOUI[3])) 
					{
						VNetMoveMemory(pbElementBody, pbBeaconOrProbeFrame+wOffset+2, wLen);
						*pwElementBodyLen = wLen;
						return TRUE;
					}
				}
#endif
			}
#ifdef WPA2
			else
			if(bElementID == C80211_MGMT_ElementID_RSN)
			{
				// validate the minimum element size
				if(wLen >= MIN_RSN_ELEMENT_DATA_SIZE)	// 2 bytes
				{
					// Version MUST be 1.
					if( (*(pbBeaconOrProbeFrame + wOffset + 2)==(UCHAR)1) && 
						(*(pbBeaconOrProbeFrame + wOffset + 3)==(UCHAR)0) )
					{
						VNetMoveMemory(pbElementBody, pbBeaconOrProbeFrame+wOffset+2, wLen);
						*pwElementBodyLen = wLen;
						return TRUE;
					}
				}
			}
#endif

			if((bElementID != C80211_MGMT_ElementID_WPA)
#ifdef CCX2
			 &&(bElementID != ELEMENT_ID_CELL_POWER_INFO)
#endif
#ifdef CCX
			 &&(bElementID != C80211_MGMT_ElementID_CCX)
#endif
#ifdef WPA2
			 &&(bElementID != C80211_MGMT_ElementID_RSN)
#endif
			)
			{
				VNetMoveMemory(pbElementBody, pbBeaconOrProbeFrame+wOffset+2, wLen);
				*pwElementBodyLen = wLen;
				return TRUE;
			}
		}

		wOffset += (wLen+2);
	}
		
	return FALSE;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	GetSSIDfromBeaconOrProbe()
//	Searches a buffer that contains a beacon or probe for the SSID IE
// 	and stores the SSID to a caller supplied buffer.
//
//	Parameters
//		OUT: PUCHAR			Pointer to store the returned SSID.
//							Buffer has to be allocated by the caller.
//		OUT: PUSHORT		Pointer to store the returned SSID's size, in bytes.
//		IN:	 PUCHAR			Pointer to the buffer that contains the beacon or probe.
//		
//
//	returns: TRUE	if the SSID has been found
//			 FALSE  if the SSID has not been found
/////////////////////////////////////////////////////////////////////////////
BOOLEAN GetSSIDfromBeaconOrProbe(PUCHAR pbSSID, PUCHAR pbSSIDlength, PUCHAR pbBeaconOrProbeFrame)
{
	UCHAR i;

	*pbSSIDlength = *(pbBeaconOrProbeFrame + SSID_ELEMENT_LENGTH_OFFSET);

	if(*pbSSIDlength > MAX_SSID_LENGTH)
		return FALSE;
		
	VNetZeroMemory(pbSSID, MAX_SSID_LENGTH);

	for(i = 0; i < *pbSSIDlength; i++){
		*((PUCHAR)pbSSID+i) = *((PUCHAR)pbBeaconOrProbeFrame + SSID_ELEMENT_OFFSET + i);
	}

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	GetSupportedRatesfromBeaconOrProbe()
//	Searches a buffer that contains a beacon or probe for the supported rates IE
// 	and stores the rates to a caller supplied buffer.
//
//	Parameters
//		OUT: PUCHAR			Pointer to store the returned SSID.
//							Buffer has to be allocated by the caller.
//		OUT: PUSHORT		Pointer to store the returned SSID's size, in bytes.
//		IN:	 PUCHAR			Pointer to the buffer that contains the beacon or probe.
//		IN:  USHORT			Offset in the beacon or probe buffer that we should start.
//		
//
//	returns: TRUE	if the IE has been found
//			 FALSE  if the IE has not been found
/////////////////////////////////////////////////////////////////////////////
BOOLEAN GetSupportedRatesfromBeaconOrProbe(PUCHAR pSupportedRates, PUCHAR pSupportedRatesLength,
									PUCHAR pBeaconOrProbeFrame, USHORT ElementOffset)
{
	UCHAR i;

	if( *(pBeaconOrProbeFrame + ElementOffset) != SUPPORTED_RATES_ELEMENT_ID)
		return FALSE;

	*pSupportedRatesLength = *(pBeaconOrProbeFrame + ElementOffset + 1);

	if(*pSupportedRatesLength > MAX_SUPPORTED_RATES_LENGTH)
		*pSupportedRatesLength = MAX_SUPPORTED_RATES_LENGTH;

	for(i=0; i <*pSupportedRatesLength; i++)
	{
		*(pSupportedRates+i) = *(pBeaconOrProbeFrame + ElementOffset + 2 + i);
	}

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	GetSupportedRatesfromBeaconOrProbe()
//	Searches a buffer that contains a beacon or probe for the BSS channel
// 	and stores the channel to a caller supplied buffer.
//
//	Parameters
//		IN:	 PUCHAR			Pointer to the buffer that contains the beacon or probe.
//		OUT: PUCHAR			Pointer to store the returned channel.
//		
//
//	returns: TRUE	if the IE has been found
//			 FALSE  if the IE has not been found
/////////////////////////////////////////////////////////////////////////////
BOOLEAN GetChannelFromBeaconOrProbe(PUCHAR pBeaconOrProbeFrame, PUCHAR pChannel)
{
	UCHAR	bSSIDLen, bSRatesLen;

	bSSIDLen = *(pBeaconOrProbeFrame + SSID_ELEMENT_LENGTH_OFFSET);
	
	if(bSSIDLen + SUP_RATES_ELEMENT_LENGTH_REL_OFFSET < MIN_BEACON_PROBE_BYTES)
	{
		bSRatesLen	= *((PUCHAR)(pBeaconOrProbeFrame) + bSSIDLen + SUP_RATES_ELEMENT_LENGTH_REL_OFFSET);
	}
	else
		return FALSE;

	if(bSSIDLen + bSRatesLen + DS_PARAM_SET_ELEMENT_REL_OFFSET < MIN_BEACON_PROBE_BYTES)
		*pChannel	= *(pBeaconOrProbeFrame + bSSIDLen + bSRatesLen + DS_PARAM_SET_ELEMENT_REL_OFFSET);
	else
		return FALSE;

	if(*pChannel >= 1 && *pChannel <= 14)  
		return TRUE;
	else
		return FALSE;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	RxProtocolConversion()
//	Converts a wireless packet to an ethernet packet, ready to be indicated to the OS.
//	This function uses the wireless packet's allocated space, to store the ethernet packet.
//
//	Parameters
//		IN:	 PUCHAR			Pointer to the buffer that contains the wireless packet.
//		IN:  ULONG			Size of the wireless packet.
//		IN:  UCHAR			The operating mode of the Adapter.
//		IN:  PULONG			Pointer to store the returned ethernet packet's size.
//		
//
//	returns: PUCHAR		Pointer to the ethernet packet.
/////////////////////////////////////////////////////////////////////////////
PUCHAR RxProtocolConversion(PUCHAR pbInputBuffer, ULONG dwBufferLength, UCHAR bOperatingMode, PULONG pdwPacketSize)
{
	UCHAR	bSourceAddr[6];
	UCHAR	bStartOfEthPacket;

	//
	// Source Address resides at offset 10 or 16 according to OpMode
	// Destination Address is alway at offset 4 and we do not overwrite it
	//
	if(bOperatingMode == AD_HOC_MODE)
		VNetMoveMemory(bSourceAddr, pbInputBuffer + 10, 6);
	else
		VNetMoveMemory(bSourceAddr, pbInputBuffer + 16, 6);

	if (( !VNetMemCompare(pbInputBuffer + 24, SNAP_ETHERNETII, 6) &&
		  !VNetMemCompare(pbInputBuffer + 30, STATIC_ENTRIES_ETHERTYPE_8137, 2)
		) ||
		( *(pbInputBuffer + 24) == 0xaa &&
		   *(pbInputBuffer + 25) == 0xaa &&
		   VNetMemCompare(pbInputBuffer + 30, STATIC_ENTRIES_ETHERTYPE_8137, 2) &&
		   VNetMemCompare(pbInputBuffer + 30, STATIC_ENTRIES_ETHERTYPE_80F3, 2)
		)
	   )
	{
		
		//
		// We will add 12 bytes (DA(6)+SA(6)) and since the Addr4 is NOT used,
		// 802.3 FrameBody starts at an offset of +30 in the 802.11 Frame. 
		// We must start creating the 802.3 header at an offset of +18 (18 + 12 = 30)
		//
		bStartOfEthPacket = 18;

		*pdwPacketSize = dwBufferLength - 18;

		//
		//DA
		//
		VNetMoveMemory((pbInputBuffer + bStartOfEthPacket), (pbInputBuffer + 4), 6);				
		//
		//SA
		//
		VNetMoveMemory((pbInputBuffer + bStartOfEthPacket + 6), bSourceAddr, 6);

	}
	else //Non - IP packet
	{
		//
		// We will add 14 bytes (DA(6)+SA(6)+Length(2)) and since
		// the Addr4 is used, 802.3 FrameBody starts at an offset of +24 in the 802.11
		// Frame. We must start creating the 802.3 header at an offset of +10 (10+14 = 24) 
		//
		bStartOfEthPacket = 10;  

		*pdwPacketSize = dwBufferLength - 10;

		//
		//DA
		//
		VNetMoveMemory((pbInputBuffer + bStartOfEthPacket), (pbInputBuffer + 4), 6);				
		//
		//SA
		//
		VNetMoveMemory((pbInputBuffer + bStartOfEthPacket + 6), bSourceAddr, 6);
		//
		// Length
		//
		*(pbInputBuffer + bStartOfEthPacket + 12) = (UCHAR) ((dwBufferLength-24)>>8);
		*(pbInputBuffer + bStartOfEthPacket + 13) = (UCHAR) (dwBufferLength-24);
	}

	return (pbInputBuffer + bStartOfEthPacket);

}


extern USHORT Bytes2Ushort(PUCHAR pBuf);


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	ParseEAPOLPacket()
//	Checks if a packet has the EAPOL type.
//
//	Parameters
//		OUT: PUCHAR			Pointer to store the EAPOL type of the packet.
//		OUT: PUCHAR 		Pointer to store the LEAP challenge info.
//		IN:  PUCHAR			Pointer to the packet.
//		IN:  BOOLEAN		Indicates whether this is a received packet.
//		IN:	 BOOLEAN		Enables/disables debug information about the packet.
//		
//
//	returns: -
/////////////////////////////////////////////////////////////////////////////
void ParseEAPOLPacket(PUCHAR pbEapolMsg, PUCHAR pbLeapChallenge, PUCHAR pbPacket, BOOLEAN fReceived, BOOLEAN fDisplay)
{
	USHORT wLen, KeyInformation, KeyMaterialLength;
	UCHAR  lbEapolMsg = EAPOL_NONE;
	
	if(((*pbPacket == 0x88) && (*(pbPacket+1) == 0x8E)) || ((*pbPacket == 0x88) && (*(pbPacket+1) == 0xC7)))


	{
		wLen = Bytes2Ushort((pbPacket+4));
		
		if(*(pbPacket+2) != 0x01)
		{
#if DBG
//			DbgCommon(1, ("\nERROR: BAD Version (%d)------------\n", *(pbPacket+2)))
#endif
			return;
		}
	
		if(*(pbPacket+3) == 0x00)
		{
			if(*(pbPacket+6) == 0x01)
				lbEapolMsg = EAP_REQUEST;
			else if(*(pbPacket+6) == 0x02)
				lbEapolMsg = EAP_RESPONSE; 
			else if(*(pbPacket+6) == 0x03)
				lbEapolMsg = EAP_SUCCESS;
			else  if(*(pbPacket+6) == 0x04)
				lbEapolMsg = EAP_REJECT;
			
			if(*(pbPacket+10) == 0x11)
			{
				lbEapolMsg = LEAP_CHALLENGE;
				
				// check for NULL pointer
				if(pbLeapChallenge)
					*pbLeapChallenge = LEAP_CHALLENGE_ON;
			}   
		}
		else if(*(pbPacket+3) == 0x01)
		{
			lbEapolMsg = EAP_START;
		}
		else if(*(pbPacket+3) == 0x02)
		{
			lbEapolMsg = EAPOL_LOG_OFF;
		}
		else if(*(pbPacket+3) == 0x03)
		{
			if( ((*(pbPacket+6) != 0xFE) && (*(pbPacket+6) != 0x02)) )
			{
#if DBG
				DbgCommon(1, ("\nERROR: BAD WPA Descriptor Type (0x%X)------------\n", *(pbPacket+6)))
#endif
				return; 
			}
	
			KeyInformation = Bytes2Ushort(pbPacket+7);
			KeyMaterialLength = Bytes2Ushort(pbPacket+99);
	
			if(KeyInformation & EAPOL_KEY_INFO_ERROR)
			{
#if DBG
				DbgCommon(1, ("\n\n\nDATA INTEGRITY ERROR...\n"))
#endif
				lbEapolMsg = EAPOL_TX_MIC_ERROR_MSG;

				if(pbEapolMsg)
					*pbEapolMsg = lbEapolMsg;

				return;
			}
	
			if(KeyInformation & EAPOL_KEY_INFO_TYPE) //pairwise
			{
				if(!(KeyInformation & EAPOL_KEY_INFO_MIC)) //No - MIC
				{
					if(KeyInformation & EAPOL_KEY_INFO_ACK) //ACK
					{
						lbEapolMsg = EAPOL_RX_HANDSHAKE_MSG1;
					}
#if DBG
					else
					{
						DbgCommon(1, ("\n\n\n%s: EAPOL-KEY Message\n\n\n", fReceived ? "Rx":"Tx"))
					}
#endif
				}
				else // MIC is On
				{
					if(KeyInformation & EAPOL_KEY_INFO_ACK) //ACK
					{
						lbEapolMsg = EAPOL_RX_HANDSHAKE_MSG3;
					}
					else
					{
						if(KeyMaterialLength)
						{
							lbEapolMsg = EAPOL_TX_HANDSHAKE_MSG2;
						}
						else
						{
							lbEapolMsg = EAPOL_TX_HANDSHAKE_MSG4;
						}
					}
				}
			}
			else//group
			{
				if(KeyInformation & EAPOL_KEY_INFO_ACK) //ACK
				{
					lbEapolMsg = EAPOL_RX_GROUP_UPDATE_MSG1;
				}
				else
				{
					lbEapolMsg = EAPOL_TX_GROUP_UPDATE_MSG2;
				}
			}
		}
#if DBG
		if(*(pbPacket+3) == 0x04)
			DbgCommon(1, ("------------ EAPOL ASF ALERT ------------\n"))
	
		if (fDisplay)
		{
			if (lbEapolMsg == EAPOL_RX_HANDSHAKE_MSG1)
			{ DbgCommon(1, ("\n\n\nRx: EAPOL-KEY 4way Handshake Message #1\n\n\n")) }
			else if (lbEapolMsg == EAPOL_RX_HANDSHAKE_MSG3)
			{ DbgCommon(1, ("\n\n\nRx: EAPOL-KEY 4way Handshake Message #3\n\n\n")) }
			else if (lbEapolMsg == EAPOL_TX_HANDSHAKE_MSG2)
			{ DbgCommon(1, ("\n\n\nTx: EAPOL-KEY 4way Handshake Message #2\n\n\n")) }
			else if (lbEapolMsg == EAPOL_TX_HANDSHAKE_MSG4)
			{ DbgCommon(1, ("\n\n\nTx: EAPOL-KEY 4way Handshake Message #4\n\n\n")) }
			else if (lbEapolMsg == EAPOL_RX_GROUP_UPDATE_MSG1)
			{ DbgCommon(1, ("\n\n\nRx: EAPOL-KEY Group Key Update Message #1\n\n\n")) }
			else if (lbEapolMsg == EAPOL_TX_GROUP_UPDATE_MSG2)
			{ DbgCommon(1, ("\n\n\nTx: EAPOL-KEY Group Key Update Message #2\n\n\n")) }
			else if (lbEapolMsg == EAP_REQUEST)
			{ DbgCommon(1, ("------------ EAP Request ------------\n")) }
			else if (lbEapolMsg == EAP_RESPONSE)
			{ DbgCommon(1, ("------------ EAP Response ------------\n")) }
			else if (lbEapolMsg == EAP_SUCCESS)
			{ DbgCommon(1, ("------------ EAP SUCCESS ------------\n")) }
			else if (lbEapolMsg == EAP_REJECT)
			{ DbgCommon(1, ("------------ EAP REJECT  ------------\n")) }
			else if (lbEapolMsg == EAP_START)
			{ DbgCommon(1, ("------------ EAP START ------------\n")) }
			else if (lbEapolMsg == EAPOL_LOG_OFF)
			{ DbgCommon(1, ("------------ EAPOL LOG OFF ------------\n")) }
			else if (lbEapolMsg == LEAP_CHALLENGE)
			{ DbgCommon(1, ("----------- LEAP challenge ----------\n")) }
		}
#endif
	}
	
	// check for NULL pointer
	if(pbEapolMsg)
		*pbEapolMsg = lbEapolMsg;
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	StoreEapolPacket()
//	Stores an EAPOL packet's info from a received ethernet packet.
//
//	Parameters
//		OUT: PUCHAR			Pointer to store the EAPOL info.
//		IN:  PUCHAR 		Pointer to the ethernet packet.
//		IN:  ULONG			Ethernet packet's length.
//		
//
//	returns: TRUE	if the EAPOL info has been stored.
//			 FALSE  if the EAPOL info has not been stored.
/////////////////////////////////////////////////////////////////////////////
BOOLEAN StoreEapolPacket(PEAPOL_BUFFER pEapolBuffer, PUCHAR pbRxBuffer, ULONG dwRxBufferLength)
{
	UCHAR i;

	for(i=0; i<MAX_EAPOL_PACKETS; i++)
	{
		if(!(pEapolBuffer+i)->InUse)
		{
			VNetMoveMemory((pEapolBuffer+i)->Packet, pbRxBuffer, dwRxBufferLength-14);
			(pEapolBuffer+i)->Length = dwRxBufferLength-14;
			(pEapolBuffer+i)->InUse = TRUE;

			break;
		}
	}

	if(i == MAX_EAPOL_PACKETS)
		return FALSE;

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	StoreAuthenticationEvent()
//	Stores an authentication event to custom structure.
//	Later we use the custom structure to indicate the event to the supplicant.
//
//	Parameters
//		OUT: PCUSTOM_AUTHENTICATION_EVENT		Pointer to store authentication event.
//		IN:  P_802_11_AUTHENTICATION_REQUEST	Pointer to the authentication event to be stored.
//		IN:  ULONG								Ethernet packet length.
//		
//
//	returns: TRUE	if the event has been stored.
//			 FALSE  if the event has not been stored.
/////////////////////////////////////////////////////////////////////////////
BOOLEAN StoreAuthenticationEvent(PCUSTOM_AUTHENTICATION_EVENT pAuthenticationEventOut, P_802_11_AUTHENTICATION_REQUEST pAuthenticationEventIn)
{
	UCHAR i;

	for(i=0; i<MAX_CUSTOM_AUTHENTICATION_EVENTS; i++)
	{
		if(!(pAuthenticationEventOut+i)->InUse)
		{
			VNetMoveMemory((PUCHAR)(&(pAuthenticationEventOut+i)->AuthenticationRequest), (PUCHAR)pAuthenticationEventIn, sizeof(_802_11_AUTHENTICATION_REQUEST));
			(pAuthenticationEventOut+i)->InUse = TRUE;

			break;
		}
	}

	if(i == MAX_CUSTOM_AUTHENTICATION_EVENTS)
		return FALSE;

	return TRUE;
}

