/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/

/***************************************************************************/
/***************************************************************************/
/**                                                                        */
/** FastVNET (PCMCIA) Linux Driver		                           	  	   */
/**                                                                        */
/***************************************************************************/


#include "vnet.h"
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/wireless.h>
#include <net/iw_handler.h>
#include "vnetioctl.h"
#include "interrupt.h"

#ifndef __IN_PCMCIA_PACKAGE__

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,1,25))
#define skb_tx_check(dev, skb) \
    do { if (skb == NULL) { dev_tint(dev); return 0; } \
    if {skb->len <= 0) return 0; } while (0)
#define add_tx_bytes(stats, n) do { int x; x = (n); } while (0)
#else
#define skb_tx_check(dev,skb) do { } while (0)
#define add_tx_bytes(stats,n) do { (stats)->tx_bytes += n; } while (0)
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,1,86))
#define DEV_KFREE_SKB(skb)      dev_kfree_skb(skb, FREE_WRITE)
#else
#define DEV_KFREE_SKB(skb)      dev_kfree_skb(skb)
#endif

#endif /* !__IN_PCMCIA_PACKAGE__ */

#ifndef SIOCSIWNAME
#define SIOCSIWNAME SIOCSIWCOMMIT	//Changed in kernel 2.4.19
#endif

/********************************************************************
 * Module parameters, Globals and miscelenious definitions
 ********************************************************************/
#ifndef PCI
#if DBG
static char *version = "2.0.1.0";
#endif
#if defined(R504)
static dev_info_t dev_info = "pcmf504";
#elif defined(R504_2958)
static dev_info_t dev_info = "pcmf504_2958";
#elif defined(R504A_2958)
static dev_info_t dev_info = "pcmf504A_2958";
#elif defined(REV_D)
static dev_info_t dev_info = "pcmf502rd";
#elif defined(REV_E)
static dev_info_t dev_info = "pcmf502re";
#elif defined(RFMD)
static dev_info_t dev_info = "pcmf502r";
#elif defined(RFMD_3COM)
static dev_info_t dev_info = "pcmf502r3";
#endif
static dev_link_t *dev_list = NULL;

static u_int irq_mask = 0xdeb8;	// Interrupt mask
static int irq_list[4] = { -1 };	// Interrupt list (alternative)
#endif //PCI
int mtu = 1500;
int eth = 1;
UCHAR channel = 4;
UCHAR TxRate = 3;
USHORT RTSThreshold = 2347;
USHORT FragThreshold = 2346;
UCHAR OpMode = INFRASTRUCTURE_MODE;
char ESSID[MAX_SSID_LENGTH + 1] = "";
UCHAR WepKeyToUse = 0;
char WepKey1[(LONG_WEP_KEY_SIZE * 2) + 1] = "0000000000\0";
char WepKey2[(LONG_WEP_KEY_SIZE * 2) + 1] = "0000000000\0";
char WepKey3[(LONG_WEP_KEY_SIZE * 2) + 1] = "0000000000\0";
char WepKey4[(LONG_WEP_KEY_SIZE * 2) + 1] = "0000000000\0";
UCHAR WepMode = WEP_MODE_MANDATORY;
UCHAR EncryptionLevel = WEP_DISABLED;
UCHAR AuthenticationType = C80211_MGMT_AAN_OPENSYSTEM;
UCHAR PreambleType = PREAMBLE_TYPE_LONG;
UCHAR PwrMgmtMode = 0;
ULONG TxPowerLevel = 200;
USHORT BeaconPeriod = 100;
UCHAR RadioOn = 1;
int pc_debug = 513;
extern UCHAR TxPowerLevels[96];

MODULE_PARM (pc_debug, "i");
#ifndef PCI
MODULE_PARM (irq_mask, "i");
MODULE_PARM (irq_list, "1-4i");
#endif

MODULE_PARM (mtu, "i");
MODULE_PARM (eth, "i");
MODULE_PARM (channel, "i");
MODULE_PARM (TxRate, "i");
MODULE_PARM (RadioOn, "i");

MODULE_PARM (RTSThreshold, "i");
MODULE_PARM (FragThreshold, "i");

MODULE_PARM (OpMode, "i");
MODULE_PARM (ESSID, "c" __MODULE_STRING (MAX_SSID_LENGTH));
MODULE_PARM (WepKeyToUse, "i");
MODULE_PARM (WepKey1, "c" __MODULE_STRING (11));
MODULE_PARM (WepKey2, "c" __MODULE_STRING (11));
MODULE_PARM (WepKey3, "c" __MODULE_STRING (11));
MODULE_PARM (WepKey4, "c" __MODULE_STRING (11));

MODULE_PARM (WepMode, "i");
MODULE_PARM (EncryptionLevel, "i");
MODULE_PARM (AuthenticationType, "i");
MODULE_PARM (PreambleType, "i");
MODULE_PARM (PwrMgmtMode, "i");
MODULE_PARM (BeaconPeriod, "i");
MODULE_LICENSE ("GPL");
static const struct iw_handler_def vnet_handler_def;

// Ethernet timeout is ((400*HZ)/1000), but we should use a higher
// value, because wireless transmissions are much slower

#define TX_TIMEOUT ((4000*HZ)/1000)

#define MAX_VNET_CARDS 	16

#ifndef PCI
#if (LINUX_VERSION_CODE >= 0x020213)
static struct net_device *vnet_index[MAX_VNET_CARDS];
#else
static struct device *vnet_index[MAX_VNET_CARDS];
#endif
#endif
//
// Local data for netdevice
//
struct net_local
{

#if (LINUX_VERSION_CODE >= 0x020213)
  struct net_device *dev;	// backtrack device
#else
  struct device *dev;		// backtrack device
#endif
#ifndef PCI
  dev_link_t *link;		// backtrack link
  dev_node_t node;
#endif
  spinlock_t slock;		// spinlock
  int interrupt;		// interrupt
  struct net_device_stats stats;	// device stats

#ifdef WIRELESS_EXT
#ifdef WIRELESS_SPY
  int spy_number;
  u_char spy_address[IW_MAX_SPY][VNet_LENGTH_OF_ADDRESS];
  struct iw_quality spy_stat[IW_MAX_SPY];
#endif
  struct iw_statistics wstats;
#endif
  VNet_ADAPTER Adapter;
};
//      Driver Info
/********************************************************************
 * Function Prototypes
 ********************************************************************/
#if (LINUX_VERSION_CODE >= 0x020213)
static int vnet_tx (struct sk_buff *skb, struct net_device *dev);
static int vnet_open (struct net_device *dev);
static int vnet_close (struct net_device *dev);
#else
static int vnet_tx (struct sk_buff *skb, struct device *dev);
static int vnet_open (struct device *dev);
static int vnet_close (struct device *dev);
#endif
#if (LINUX_VERSION_CODE < 0x020500)
static void vnet_interrupt (int irq, void *dev_id, struct pt_regs *regs);
#else
static irqreturn_t vnet_interrupt (int irq, void *dev_id, struct pt_regs *regs);
#endif  // Linux v.< 2.5
#ifndef PCI
static int vnet_config (dev_link_t * link);
static void vnet_release (u_long arg);
static dev_link_t *vnet_attach (void);
static void vnet_detach (dev_link_t * link);
static int vnet_event (event_t event, int priority,
		       event_callback_args_t * args);
void cs_error (client_handle_t handle, int func, int ret);
extern int init_module (void);
extern void cleanup_module (void);
#endif // PCI
#if (LINUX_VERSION_CODE >= 0x020213)
struct net_device_stats *vnet_get_stats (struct net_device *dev);
#else
struct net_device_stats *vnet_get_stats (struct device *dev);
#endif
#ifdef	WIRELESS_EXT
#if (LINUX_VERSION_CODE >= 0x020213)
static int vnet_ioctl (struct net_device *dev, struct ifreq *rq, int cmd);
struct iw_statistics *vnet_get_wireless_stats (struct net_device *dev);
#else
static int vnet_ioctl (struct device *dev, struct ifreq *rq, int cmd);
struct iw_statistics *vnet_get_wireless_stats (struct device *dev);
#endif
#endif

static const struct {
	int reg_domain;
	int min, max;
} channel_table[] = { { REG_DOMAIN_FCC, 1, 11},
		      { REG_DOMAIN_DOC, 1, 11},
		      { REG_DOMAIN_ETSI, 1, 13},
		      { REG_DOMAIN_SPAIN, 10, 11},
		      { REG_DOMAIN_FRANCE, 10, 13},
		      { REG_DOMAIN_MKK, 14, 14},
		      { REG_DOMAIN_MKK1, 1, 14},
		      { REG_DOMAIN_ISRAEL, 9} };
#ifndef PCI

/********************************************************************
 * PCMCIA CONFIG / RELEASE
 ********************************************************************/

#if (LINUX_VERSION_CODE < 0x020602)
#define CS_CHECK(fn, args...) while ((last_ret=CardServices(last_fn=(fn),args))!=0) goto cs_failed
#define CFG_CHECK(fn, args...) if (CardServices(fn, args) != 0) goto next_entry
#else
#define CS_CHECK(fn, ret) \
do { last_fn = (fn); if ((last_ret = (ret)) != 0) goto cs_failed; } while (0)
#endif

static int
vnet_config (dev_link_t * link)
{
  client_handle_t handle = link->handle;
  tuple_t tuple;
  cisparse_t parse;
  struct net_device *dev = (struct net_device *) link->priv;
  struct net_local *local = (struct net_local *) dev->priv;
  int last_fn, last_ret;
  u_char buf[64];
  win_req_t req;
  memreq_t map;
  int i, rc;
  //int rc, i, cs_r;
  config_info_t config;
//      cistpl_cftable_entry_t dflt = { 0 };
  //PVNet_ADAPTER Adapter = (PVNet_ADAPTER)&local->Adapter;

  IF_LOUD (DbgPrint ("-> vnet_config(0x%p)\n", link);)
    // This reads the card's CONFIG tuple to find its configuration registers.
    tuple.DesiredTuple = CISTPL_CONFIG;
  tuple.Attributes = 0;
  tuple.TupleData = buf;
  tuple.TupleDataMax = sizeof (buf);
  tuple.TupleOffset = 0;
#if (LINUX_VERSION_CODE < 0x020602)
  CS_CHECK (GetFirstTuple, handle, &tuple);
  CS_CHECK (GetTupleData, handle, &tuple);
  CS_CHECK (ParseTuple, handle, &tuple, &parse);
#else
  CS_CHECK (GetFirstTuple, pcmcia_get_first_tuple(handle, &tuple));
  CS_CHECK (GetTupleData, pcmcia_get_tuple_data(handle, &tuple));
  CS_CHECK (ParseTuple, pcmcia_parse_tuple(handle, &tuple, &parse));
#endif

  link->conf.ConfigBase = parse.config.base;
  link->conf.Present = parse.config.rmask[0];

  // Configure card

  link->state |= DEV_CONFIG;

  // 2-05-2001 Test : Added GetConfigurationInfo. Check if needed
#if (LINUX_VERSION_CODE < 0x020602)
  CS_CHECK (GetConfigurationInfo, handle, &config);
#else
  CS_CHECK (GetConfigurationInfo, pcmcia_get_configuration_info(handle, &config));
#endif

  link->conf.Vcc = config.Vcc;

  // In this loop, we scan the CIS for configuration table entries,
  // each of which describes a valid card configuration, including
  // voltage, IO window, memory window, and interrupt settings.
  // We make no assumptions about the card to be configured: we use
  // just the information available in the CIS.  In an ideal world,
  // this would work for any PCMCIA card, but it requires a complete
  // and accurate CIS.  In practice, a driver usually "knows" most of
  // these things without consulting the CIS, and most client drivers
  // will only use the CIS to fill in implementation-defined details.
  tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
#if (LINUX_VERSION_CODE < 0x020602)
  CS_CHECK (GetFirstTuple, handle, &tuple);
#else
  CS_CHECK (GetFirstTuple, pcmcia_get_first_tuple(handle, &tuple));
#endif

  while (1)
    {
      cistpl_cftable_entry_t dflt = { 0 };
      cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);
#if (LINUX_VERSION_CODE < 0x020602)
      CFG_CHECK (GetTupleData, handle, &tuple);
      CFG_CHECK (ParseTuple, handle, &tuple, &parse);
#else
     if (pcmcia_get_tuple_data(handle, &tuple) != 0) goto next_entry;
     if (pcmcia_parse_tuple(handle, &tuple, &parse) != 0) goto next_entry;
#endif

      if (cfg->index == 0)
	goto next_entry;
      link->conf.ConfigIndex = cfg->index;

      // Does this card need audio output?
      if (cfg->flags & CISTPL_CFTABLE_AUDIO)
	{
	  link->conf.Attributes |= CONF_ENABLE_SPKR;
	  link->conf.Status = CCSR_AUDIO_ENA;
	}

      // Use power settings for Vcc and Vpp if present
      // Note that the CIS values need to be rescaled
      if (cfg->vcc.present & (1 << CISTPL_POWER_VNOM))
	link->conf.Vcc = cfg->vcc.param[CISTPL_POWER_VNOM] / 10000;
      else if (dflt.vcc.present & (1 << CISTPL_POWER_VNOM))
	link->conf.Vcc = dflt.vcc.param[CISTPL_POWER_VNOM] / 10000;

      if (cfg->vpp1.present & (1 << CISTPL_POWER_VNOM))
	link->conf.Vpp1 = link->conf.Vpp2 =
	  cfg->vpp1.param[CISTPL_POWER_VNOM] / 10000;
      else if (dflt.vpp1.present & (1 << CISTPL_POWER_VNOM))
	link->conf.Vpp1 = link->conf.Vpp2 =
	  dflt.vpp1.param[CISTPL_POWER_VNOM] / 10000;

      // Do we need to allocate an interrupt?
      if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1)
	link->conf.Attributes |= CONF_ENABLE_IRQ;

      // IO window settings
      link->io.NumPorts1 = link->io.NumPorts2 = 0;
      if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0))
	{
	  cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
	  link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
	  if (!(io->flags & CISTPL_IO_8BIT))
	    link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
	  if (!(io->flags & CISTPL_IO_16BIT))
	    link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
	  link->io.BasePort1 = io->win[0].base;
	  link->io.NumPorts1 = io->win[0].len;
	  if (io->nwin > 1)
	    {
	      link->io.Attributes2 = link->io.Attributes1;
	      link->io.BasePort2 = io->win[1].base;
	      link->io.NumPorts2 = io->win[1].len;
	    }
	}

      // This reserves IO space but doesn't actually enable it
#if (LINUX_VERSION_CODE < 0x020602)
      CFG_CHECK (RequestIO, link->handle, &link->io);
#else
      if (pcmcia_request_io(link->handle, &link->io) != 0) goto next_entry;
#endif
      // Now set up a common memory window, if needed.  There is room
      // in the dev_link_t structure for one memory window handle,
      // but if the base addresses need to be saved, or if multiple
      // windows are needed, the info should go in the private data
      // structure for this device.
      // Note that the memory window base is a physical address, and
      // needs to be mapped to virtual space with ioremap() before it
      // is used.
      if ((cfg->mem.nwin > 0) || (dflt.mem.nwin > 0))
	{
	  cistpl_mem_t *mem = (cfg->mem.nwin) ? &cfg->mem : &dflt.mem;
	  req.Attributes = WIN_DATA_WIDTH_16 | WIN_MEMORY_TYPE_CM;
	  req.Base = mem->win[0].host_addr;
	  req.Size = mem->win[0].len;
	  req.AccessSpeed = 0;
	  link->win = (window_handle_t) link->handle;
#if (LINUX_VERSION_CODE < 0x020602)
	  CFG_CHECK (RequestWindow, &link->win, &req);
#else
          if (pcmcia_request_window(&link->handle, &req, &link->win)  != 0) goto next_entry;
#endif
	  map.Page = 0;
	  map.CardOffset = mem->win[0].card_addr;
#if (LINUX_VERSION_CODE < 0x020602)
	  CFG_CHECK (MapMemPage, link->win, &map);
#else
          if (pcmcia_map_mem_page(link->win, &map)  != 0) goto next_entry;
#endif
	}
      break;

    next_entry:
      if (cfg->flags & CISTPL_CFTABLE_DEFAULT)
	dflt = *cfg;
#if (LINUX_VERSION_CODE < 0x020602)
      CS_CHECK (GetNextTuple, handle, &tuple);
#else
      CS_CHECK (GetNextTuple, pcmcia_get_next_tuple(handle, &tuple));
#endif
    }

  // Allocate an interrupt line.  Note that this does not assign a
  // handler to the interrupt, unless the 'Handler' member of the
  // irq structure is initialized.
  if (link->conf.Attributes & CONF_ENABLE_IRQ)
    {
      link->irq.Attributes = IRQ_TYPE_EXCLUSIVE | IRQ_HANDLE_PRESENT;
      link->irq.IRQInfo1 = IRQ_INFO2_VALID | IRQ_LEVEL_ID;
      if (irq_list[0] == -1)
	link->irq.IRQInfo2 = irq_mask;
      else
	for (i = 0; i < 4; i++)
	  link->irq.IRQInfo2 |= 1 << irq_list[i];
      link->irq.Handler = vnet_interrupt;
      link->irq.Instance = dev;
#if (LINUX_VERSION_CODE < 0x020602)
      CS_CHECK (RequestIRQ, link->handle, &link->irq);
#else
      CS_CHECK (RequestIRQ, pcmcia_request_irq(link->handle, &link->irq));
#endif
    }

  // This actually configures the PCMCIA socket -- setting up
  // the I/O windows and the interrupt mapping, and putting the
  // card and host interface into "Memory and IO" mode.

#if (LINUX_VERSION_CODE < 0x020602)
  CS_CHECK (RequestConfiguration, link->handle, &link->conf);
#else
  CS_CHECK (RequestConfiguration, pcmcia_request_configuration(link->handle, &link->conf));
#endif

  // Feed the netdevice with this info
  dev->irq = link->irq.AssignedIRQ;
  dev->base_addr = link->io.BasePort1;
  netif_start_queue (dev);

  // Report what we've done

  IF_LOUD (DbgPrint
	   ("%s: index 0x%02x: Vcc %d.%d", dev_info, link->conf.ConfigIndex,
	    link->conf.Vcc / 10, link->conf.Vcc % 10);) if (link->conf.Vpp1)
    {
    IF_LOUD (DbgPrint
	       (", Vpp %d.%d", link->conf.Vpp1 / 10, link->conf.Vpp1 % 10);)}
  if (link->conf.Attributes & CONF_ENABLE_IRQ)
    {
    IF_LOUD (DbgPrint (", irq %d", link->irq.AssignedIRQ);)}
  if (link->io.NumPorts1)
    {
    IF_LOUD (DbgPrint
	       (", io 0x%04x-0x%04x", link->io.BasePort1,
		  link->io.BasePort1 + link->io.NumPorts1 - 1);)}
  if (link->io.NumPorts2)
    {
    IF_LOUD (DbgPrint
	       (" & 0x%04x-0x%04x", link->io.BasePort2,
		  link->io.BasePort2 + link->io.NumPorts2 - 1);)}
  if (link->win)
    {
    IF_LOUD (DbgPrint
	       (", mem 0x%06lx-0x%06lx", req.Base, req.Base + req.Size - 1);)}
  IF_LOUD (DbgPrint ("\n");) link->state &= ~DEV_CONFIG_PENDING;

  if (!eth)
    {
      for (i = 0; i < MAX_VNET_CARDS; ++i)
	if (!vnet_index[i])
	  {
	    sprintf (dev->name, "vnet%d", i);
	    vnet_index[i] = dev;
	    break;
	  }
    }

  // Register the netdevice
  rc = register_netdev (dev);
  if (rc)
    {
      IF_DEBUG_ERRORS (DbgPrint
		       ("%s: register_netdev() failed!\n",
			dev_info);) 
          vnet_release ((u_long) link);
      return 0;
    }
  IF_LOUD (DbgPrint ("%s: Registered netdevice %s\n", dev_info, dev->name);)
    copy_dev_name (local->node, dev);
  local->Adapter.dev = dev;
  local->Adapter.IoBase = dev->base_addr;

  link->dev = &local->node;
  IF_VERY_LOUD (DbgPrint ("<- vnet_config()\n");) return 1;

cs_failed:
  cs_error (link->handle, last_fn, last_ret);
  vnet_release ((u_long) link);
  IF_VERY_LOUD (DbgPrint ("<- vnet_config()\n");) return 0;
}

static void
vnet_release (u_long arg)
{
  dev_link_t *link = (dev_link_t *) arg;
  struct net_device *dev = (struct net_device *) link->priv;
  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;

  IF_LOUD (DbgPrint ("-> vnet_release(0x%p)\n", link);)
    // If the device is currently in use, we won't release
    // until it's actually closed.
    if (link->open)
    {
      IF_LOUD (DbgPrint
		    ("%s: vnet_release: release postponed, %s still locked\n",
		     dev_info, link->dev->dev_name);) 
          link->state |= DEV_STALE_CONFIG;
      return;
    }
  del_timer (Adapter->MgmtTimer);
  kfree (Adapter->MgmtTimer);
  CardStop (Adapter);
  Adapter->IsUp = FALSE;
  if (link->win)
#if (LINUX_VERSION_CODE < 0x020602)
  CardServices (ReleaseWindow, link->win);
  CardServices (ReleaseConfiguration, link->handle);
#else 
  pcmcia_release_window(link->win);
  pcmcia_release_configuration(link->handle);
#endif 


  if (link->io.NumPorts1)
#if (LINUX_VERSION_CODE < 0x020602)
    CardServices (ReleaseIO, link->handle, &link->io);
#else
  pcmcia_release_io(link->handle, &link->io);
#endif


  if (link->irq.AssignedIRQ)
#if (LINUX_VERSION_CODE < 0x020602)
    CardServices (ReleaseIRQ, link->handle, &link->irq);
#else
  pcmcia_release_irq(link->handle, &link->irq);
#endif

  link->state &= ~DEV_CONFIG;

IF_LOUD (DbgPrint ("<- vnet_release() - success\n");)
}

/********************************************************************
* PCMCIA ATTACH / DETACH
********************************************************************/
static dev_link_t *
vnet_attach (void)
{
  dev_link_t *link;
  struct net_device *dev;
  struct net_local *local;
  int rc;
  client_reg_t client_reg;

  IF_LOUD (DbgPrint ("-> vnet_attach()\n");)
    // Flush stale links
    for (link = dev_list; link; link = link->next)
    if (link->state & DEV_STALE_LINK)
      vnet_detach (link);

  // Initialize the dev_link_t structure
  link = kmalloc (sizeof (struct dev_link_t), GFP_KERNEL);
  memset (link, 0, sizeof (struct dev_link_t));
#if (LINUX_VERSION_CODE < 0x020600)
  link->release.function = &vnet_release;
  link->release.data = (u_long) link;
#endif
  link->conf.Vcc = 33;
  link->conf.IntType = INT_MEMORY_AND_IO;

  // Allocate space for netdevice (private data of link)
  dev = kmalloc (sizeof (struct net_device), GFP_KERNEL);
  memset (dev, 0, sizeof (struct net_device));
  link->priv = dev;
  // Allocate space for netdevice priv (private data of netdevice)
  local = kmalloc (sizeof (struct net_local), GFP_KERNEL);
  memset (local, 0, sizeof (struct net_local));
  dev->priv = local;

  // Initialize specific data
  local->link = link;
  local->dev = dev;
  spin_lock_init (&local->slock);
  spin_lock_init (&local->Adapter.qlock);

  /* Set the mgmt timer */
  local->Adapter.MgmtTimer = kmalloc (sizeof (struct timer_list), GFP_KERNEL);
  init_timer (local->Adapter.MgmtTimer);
  local->Adapter.MgmtTimer->function = MgmtTimer;
  local->Adapter.MgmtTimer->data = (unsigned long) dev;

  // Standard setup for generic data
  ether_setup (dev);
  // kernel callbacks
  dev->open = vnet_open;
  dev->stop = vnet_close;
  dev->hard_start_xmit = vnet_tx;
  dev->get_stats = vnet_get_stats;

#ifdef WIRELESS_EXT
  dev->get_wireless_stats = vnet_get_wireless_stats;
  dev->wireless_handlers = (struct iw_handler_def *)&vnet_handler_def;
  dev->do_ioctl = vnet_ioctl;
#endif
  strcpy(dev->name, "atml%d");

  // Other netdevice data
  init_dev_name (dev, local->node);
  dev->mtu = mtu;

  netif_stop_queue (dev);
  // Register with CardServices
  link->next = dev_list;
  dev_list = link;
  client_reg.dev_info = &dev_info;
  client_reg.Attributes = INFO_IO_CLIENT;
  client_reg.EventMask = CS_EVENT_REGISTRATION_COMPLETE |
    CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
    CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
    CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
  client_reg.event_handler = &vnet_event;
  client_reg.Version = 0x0210;
  client_reg.event_callback_args.client_data = link;

#if (LINUX_VERSION_CODE < 0x020602)
  rc = CardServices (RegisterClient, &link->handle, &client_reg);
#else
  rc = pcmcia_register_client( &link->handle, &client_reg);
#endif
  if (rc)
    {
      cs_error (link->handle, RegisterClient, rc);
      vnet_detach (link);
      return NULL;
    }

  IF_VERY_LOUD (DbgPrint ("<- vnet_attach()\n");) 

    return link;
}

static void
vnet_detach (dev_link_t * link)
{
  dev_link_t **linkp;

  IF_LOUD (DbgPrint ("-> vnet_detach(0x%p)\n", link);)
    // Locate device structure
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
    if (*linkp == link)
      break;
  if (!*linkp)
    {
      IF_DEBUG_ERRORS (DbgPrint
		       ("%s: Attempt to detach non-existing PCMCIA client!\n",
			dev_info);) return;
    }

  // If the device is currently configured and active, we won't
  // actually delete it yet. Instead, it is marked so that when the
  // release() function is called, that will trigger a proper
  // detach()
#if (LINUX_VERSION_CODE < 0x020600)
  del_timer (&link->release);
#else
vnet_release ((u_long) link);
#endif

  if (link->state & DEV_CONFIG)
    {
      IF_LOUD (DbgPrint ("%s: vnet_detach: detach postponed, %s still locked\n", dev_info, link->dev->dev_name);) 
      vnet_release ((u_long) link);
      if (link->state & DEV_STALE_CONFIG) {
          link->state |= DEV_STALE_LINK;
          return;
      }
    }

  // Break the line with CardServices
  if (link->handle)
#if (LINUX_VERSION_CODE < 0x020602)
    CardServices (DeregisterClient, link->handle);
#else
  pcmcia_deregister_client(link->handle);
#endif

  // Unlink device structure, free pieces
  *linkp = link->next;
  if (link->priv)
    {
      struct net_device *dev = (struct net_device *) link->priv;
      if (link->dev)
	{
	  unregister_netdev (dev);
	IF_LOUD (DbgPrint ("%s: Netdevice unregistered\n", dev_info);)}
      if (dev->priv)
	    kfree (dev->priv);
      kfree (link->priv);
    }
  kfree (link);

IF_VERY_LOUD (DbgPrint ("<- vnet_detach()\n");)}

/********************************************************************
 * PCMCIA EVENT HANDLER
********************************************************************/
static int
vnet_event (event_t event, int priority, event_callback_args_t * args)
{
  dev_link_t *link = (dev_link_t *) args->client_data;
  struct net_device *dev = (struct net_device *) link->priv;
  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;

  IF_LOUD (DbgPrint ("-> vnet_event(%s, %d, 0x%p)\n",
		     ((event ==
		       CS_EVENT_REGISTRATION_COMPLETE) ?
		      "registration complete"
		      : ((event ==
			  CS_EVENT_CARD_INSERTION) ? "card insertion"
			 : ((event ==
			     CS_EVENT_CARD_REMOVAL) ? "card removal" : ((event
									 ==
									 CS_EVENT_RESET_PHYSICAL)
									?
									"physical physical"
									: ((event == CS_EVENT_CARD_RESET) ? "card reset" : ((event == CS_EVENT_PM_SUSPEND) ? "pm suspend" : ((event == CS_EVENT_PM_RESUME) ? "pm resume" : "unknown"))))))), priority, args);) 
    switch (event)
    {
    case CS_EVENT_CARD_INSERTION:
      link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
      if (!vnet_config (link))
	        dev->irq = 0;
          Adapter->IsUp = TRUE;
          InitAdapter (&local->Adapter);
          SetParameters (&local->Adapter);
          if (!InitAndStartCard (&local->Adapter))
	        return -1;
      memcpy (dev->dev_addr, Adapter->AssociationInfo.CurrentAddress, 6);
      break;

    case CS_EVENT_CARD_REMOVAL:
      printk(KERN_INFO "Card Removal event called !!!\n");
      del_timer (Adapter->MgmtTimer);
      Adapter->IsUp = FALSE;
      link->state &= ~DEV_PRESENT;
      if (link->state & DEV_CONFIG)
	{
	  netif_stop_queue (dev);
	  netif_device_detach (dev);
#if (LINUX_VERSION_CODE < 0x020600)
	  mod_timer (&link->release, jiffies + HZ / 20);
#else
 	  vnet_release ((u_long) link);
#endif
	}
      break;

    case CS_EVENT_PM_SUSPEND:
      link->state |= DEV_SUSPEND;
    case CS_EVENT_RESET_PHYSICAL:
      Adapter->IsUp = FALSE;
      if (link->state & DEV_CONFIG)
	{
	      netif_device_detach (dev);
#if (LINUX_VERSION_CODE < 0x020602)
	  CardServices (ReleaseConfiguration, link->handle);
#else
          pcmcia_release_configuration(link->handle);
#endif
	}
      break;

    case CS_EVENT_PM_RESUME:
      link->state &= ~DEV_SUSPEND;
      // Fall through
    case CS_EVENT_CARD_RESET:
      if (link->state & DEV_CONFIG)
	{
#if (LINUX_VERSION_CODE < 0x020602)
	  CardServices (RequestConfiguration, link->handle, &link->conf);
#else
          pcmcia_request_configuration(link->handle, &link->conf);
#endif
	  if (link->open)
	    {
	      netif_device_attach (dev);
	      netif_start_queue (dev);
	    }
	}
      break;
    }

  IF_VERY_LOUD (DbgPrint ("<- vnet_event()\n");) 
    return 0;
}

/********************************************************************
 * MODULE INSERTION / REMOVAL
********************************************************************/
#if (LINUX_VERSION_CODE > 0x020542)
static struct pcmcia_driver fastvnet_driver = {
 	.owner		= THIS_MODULE,
 	.drv		= {
 		.name	= dev_info 
  },
 	.attach		= vnet_attach,
 	.detach		= vnet_detach,
};
#endif

extern int
init_module (void)
{
#if (LINUX_VERSION_CODE < 0x020602)
  servinfo_t serv;
#endif

  IF_VERY_LOUD (DbgPrint ("-> init_module()\n");) 
  IF_LOUD (DbgPrint ("%s: ATMEL 11Mbps Wireless PCMCIA LAN driver version %s\n", dev_info, version);)
#if (LINUX_VERSION_CODE < 0x020602)
    // Check CardServices release
    CardServices (GetCardServicesInfo, &serv);

  if (serv.Revision != CS_RELEASE_CODE)
    {
      printk ("%s: CardServices release does not match!\n", dev_info);
      return -1;
    }
#endif
  // Register PCMCIA driver
#if (LINUX_VERSION_CODE > 0x020542)
  pcmcia_register_driver (&fastvnet_driver);
#else
  register_pccard_driver (&dev_info, &vnet_attach, &vnet_detach);
#endif
  IF_VERY_LOUD (DbgPrint ("<- init_module()\n");) return 0;
}

extern void
cleanup_module (void)
{
  IF_VERY_LOUD (DbgPrint ("-> cleanup_module()\n");)
    // Unregister PCMCIA driver
#if (LINUX_VERSION_CODE > 0x020542)
  pcmcia_unregister_driver (&fastvnet_driver);
#else
  unregister_pccard_driver (&dev_info);
#endif

  // Remove leftover devices
  if (dev_list)
    {
    IF_LOUD (DbgPrint ("%s: Removing leftover devices!\n", dev_info);)}
  while (dev_list)
    {
      if (dev_list->state & DEV_CONFIG)
	vnet_release ((u_long) dev_list);
      vnet_detach (dev_list);
    }

IF_LOUD (DbgPrint ("%s: Driver unloaded\n", dev_info);)
    IF_VERY_LOUD (DbgPrint ("<- cleanup_module()\n");)}


// Show CardServices error message (syslog)
void
cs_error (client_handle_t handle, int func, int ret)
{
  error_info_t err = { func, ret };
#if (LINUX_VERSION_CODE < 0x020602)
  CardServices (ReportError, handle, &err);
#else
  pcmcia_report_error(handle, &err);
#endif
}
#endif //ifndef PCI
/********************************************************************
 * NET OPEN / CLOSE
********************************************************************/

#if (LINUX_VERSION_CODE > 0x020213)
static int
vnet_open (struct net_device *dev)
#else
static int
vnet_open (struct device *dev)
#endif
{
#ifndef PCI
  struct net_local *local = (struct net_local *) dev->priv;
  struct dev_link_t *link = local->link;
#endif

  IF_VERY_LOUD (DbgPrint ("-> vnet_open(%s)\n", dev->name);)
  if(!netif_running(dev)){
      if (netif_queue_stopped (dev)){
            IF_LOUD (DbgPrint ("waiking up the net queue - (%s)\n", dev->name);)
            netif_wake_queue (dev);
      }
  }
#ifndef PCI
  link->open++;
#endif

#if (LINUX_VERSION_CODE < 0x020500)
  MOD_INC_USE_COUNT;
#else
  SET_MODULE_OWNER(dev);
#endif

  IF_LOUD (DbgPrint ("<- vnet_open(%s)\n", dev->name);) 
      return 0;
}

#if (LINUX_VERSION_CODE > 0x020213)
static int
vnet_close (struct net_device *dev)
#else
static int
vnet_close (struct device *dev)
#endif
{
#ifndef PCI
  struct net_local *local = (struct net_local *) dev->priv;
  struct dev_link_t *link = local->link;


  // If the device isn't open, then nothing to do
  if (!link->open)
    {
      IF_LOUD (DbgPrint ("<> vnet_close(%s)\n", dev->name);) return 0;
    }

  IF_VERY_LOUD (DbgPrint ("-> vnet_close(%s)\n", dev->name);)
    // Close the device
    link->open--;
#endif
#if (LINUX_VERSION_CODE < 0x020500)
  MOD_DEC_USE_COUNT;
#endif

  // Check if card is still present
  if (netif_running (dev))
    {
      netif_stop_queue (dev);
      netif_device_detach (dev);
    }
#ifndef PCI
  else if (link->state & DEV_STALE_CONFIG)
#if (LINUX_VERSION_CODE < 0x020600)
    mod_timer (&link->release, jiffies + HZ / 20);
#else
 	vnet_release ((u_long) link);
#endif
#endif
  IF_LOUD (DbgPrint ("<- vnet_close(%s)\n", dev->name);) 
      return 0;
}


/********************************************************************
 * NET Statistics
 ********************************************************************/
#if (LINUX_VERSION_CODE > 0x020213)
struct net_device_stats *
vnet_get_stats (struct net_device *dev)
#else
struct net_device_stats *
vnet_get_stats (struct device *dev)
#endif
{
  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;
  IF_VERY_LOUD (DbgPrint ("<> vnet_get_stats(%s)\n", dev->name);) local->
    stats.tx_packets = Adapter->Stats.TxDataPacketsOk;
  local->stats.rx_packets = Adapter->Stats.RxDataPacketsOk;
  local->stats.tx_errors = Adapter->Stats.TxDataPacketsError;
  local->stats.rx_errors = Adapter->Stats.RxDataPacketsError;
  return (&((struct net_local *) dev->priv)->stats);
}

/*******************************************************************
Management Timer
*******************************************************************/

VOID
MgmtTimer (u_long a)
{
#if (LINUX_VERSION_CODE > 0x020213)
  struct net_device *dev = (struct net_device *) a;
#else
  struct device *dev = (struct device *) a;
#endif
  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;

  if (!Adapter->IsUp
#ifndef PCI
      || !(local->link->state & DEV_PRESENT)
#endif
    )
    {
      IF_LOUD (DbgPrint ("** MgmtTimer reject ***\n");) return;
    }


  MgmtTimeOutCallBack (Adapter);
}

/********************************************************************
 * NET TX / RX
 ********************************************************************/

#if (LINUX_VERSION_CODE > 0x020213)
int
vnet_tx (struct sk_buff *skb, struct net_device *dev)
#else
int
vnet_tx (struct sk_buff *skb, struct device *dev)
#endif
{
  struct net_local *local = (struct net_local *) dev->priv;
  ULONG TotalBytes = 0;
  USHORT StartOfTxBuffer = 0;
  BOOLEAN IsBroadcast;
  unsigned long flags;
  //ULONG Status;
  int len;
  UCHAR *p;
  PVNet_ADAPTER Adapter = &local->Adapter;

    if (Adapter->StationState != STATION_STATE_READY
#ifndef PCI
							       || !(local-> link->state & DEV_PRESENT)
#endif
    ) {
          printk("interface not ready yet\n");
          DEV_KFREE_SKB (skb);
          return 0;
    }

#if (KERNEL_VERSION_CODE < KERNEL_VERSION(2,3,42))
  if (netif_queue_stopped (dev))
    {
#ifndef PCI
      IF_VERY_LOUD (DbgPrint ("%s: vnet_tx(%s) called while busy!\n", dev_info, dev->name);)
#endif

      if ((jiffies - dev->trans_start) < TX_TIMEOUT)
      return 1;

      if (!netif_running (dev)) {
#ifndef PCI
	  IF_LOUD (DbgPrint ("%s: %s Tx on stopped device!\n", dev_info, dev->name);)
#endif
	    return 1;
      }
    }
#endif
  skb_tx_check (dev, skb);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,1,79))
  skb->arp = 1;
#endif
  // Disable interrupts
  spin_lock_irqsave (&local->slock, flags);
  // Prepare packet
  p = skb->data;
  len = (ETH_ZLEN < skb->len) ? skb->len : ETH_ZLEN;
  TotalBytes = len;

  if (TxResourcesAvailable (Adapter, TotalBytes + 18, &StartOfTxBuffer) == FALSE) {
      Adapter->Stats.RxLost++;
      netif_stop_queue (dev);
      spin_unlock_irqrestore (&local->slock, flags);
      return 1;
    }
#ifdef ATMEL_WLAN
  if(Adapter->OperatingMode == INFRASTRUCTURE_MODE){ //In ESS the final DA is always the AP
            IsBroadcast = FALSE;
  }else {
    if ( (*((PUCHAR)(skb->data))) & 0x01 ){
      IsBroadcast = TRUE;
    }else {
      IsBroadcast = FALSE;
    }
  }
#endif
  TxEthernetPacket (Adapter, skb->data, len, &TotalBytes, StartOfTxBuffer, IsBroadcast);
  TxUpdateDescriptor (Adapter, TotalBytes, StartOfTxBuffer, TRUE, IsBroadcast);

  // Send packet
  // Remember time transmission and count tx bytes
  dev->trans_start = jiffies;
  add_tx_bytes (&local->stats, len);

  // Re-enable interrupts
  spin_unlock_irqrestore (&local->slock, flags);
  DEV_KFREE_SKB (skb);

  IF_VERY_LOUD (DbgPrint ("<- vnet_tx()\n");) 
    return 0;
}
/********************************************************************
 * INTERRUPT HANDLER
 */
#if (LINUX_VERSION_CODE < 0x020500)
static void
#else
static irqreturn_t
#endif  // Linux v. < 2.5+
vnet_interrupt (int irq, void *dev_id, struct pt_regs *regs)
{
  //UCHAR                         InterruptStatus;        
  //INTERRUPT_TYPE                InterruptType;                  
#if (defined (PCI))

#if (LINUX_VERSION_CODE < 0x020213)
      struct device *dev = (struct device *) dev_id;
#elif (LINUX_VERSION_CODE < 0x020500)
        struct net_device *dev = ((struct pci_dev *) dev_id)->driver_data;
#else // Linux v. >= 2.5
          struct net_device *dev = (struct net_device*) pci_get_drvdata((struct pci_dev*) dev_id);
#endif

#else   // !defined PCI
            struct net_device *dev = (struct net_device *) dev_id;
#endif

  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;
  USHORT gcr, ar; 
  USHORT IntType __attribute__ ((unused));
  UCHAR IntStatus __attribute__ ((unused));

  IF_VERY_LOUD (DbgPrint ("-> vnet_interrupt(%d)\n", irq);)
  spin_lock (&local->slock);
    if (!Adapter->IsUp
#ifndef PCI
		|| !(local->link->state & DEV_PRESENT)
#endif
    )
      goto none;
  // Check device
  if (!dev)
    {
#ifndef PCI
      IF_LOUD (DbgPrint ("%s: IRQ %d for unknown device!\n", dev_info, irq);)
#endif
        goto none;
    }

    CardReadUshort(Adapter, GCR, &gcr);
    if( (gcr & 0x000A) == 0x000A) {
          CardReadUshort(Adapter, AR, &ar);
          DisableInterrupts (Adapter);
          AcknowledgeInterrupt(Adapter);
          ProcessInterrupt (Adapter, dev);
          // Turn back interrupts on (unlock)
          EnableInterrupts (Adapter);
          CardWriteUshort(Adapter, AR, ar);
          IF_VERY_LOUD (DbgPrint ("<- vnet_interrupt()\n");)
          goto handle;
    }
    goto none;

handle :
  spin_unlock (&local->slock);
#if (LINUX_VERSION_CODE < 0x020500)
    return;
#else
    return IRQ_HANDLED;
#endif

none :
  spin_unlock (&local->slock);
#if (LINUX_VERSION_CODE < 0x020500)
    return;
#else
    return IRQ_NONE;
#endif

}

VOID
UpdateSupportAppInfo (PVNet_ADAPTER Adapter, PDEVICE_CONFIGURATION pDevConfig)
{
  int i;
  //DEVICE_CONFIGURATION DevConfig;

  pDevConfig->StationState = Adapter->StationState;
  pDevConfig->OperatingMode = Adapter->OperatingMode;
  pDevConfig->Channel = (Adapter->Channel & 0x7f);


  memcpy(pDevConfig->SSID, Adapter->AssociationInfo.DesiredSSID, Adapter->AssociationInfo.DesiredSSIDsize);
  pDevConfig->SSIDlength = Adapter->AssociationInfo.DesiredSSIDsize;

  VNetMoveMemory (pDevConfig->BSSID, Adapter->AssociationInfo.CurrentBSSID, 6);
  pDevConfig->InternationalRoaming = Adapter->IntRoamingEnabled;
  if (pDevConfig->InternationalRoaming)
    {
      for (i = 0; i < 14; i++)
	{

	  pDevConfig->ChannelVector[i] = Adapter->ChannelVector[i];

	}
    }

  pDevConfig->MgmtErrorCode = Adapter->MgmtErrorCode;
  pDevConfig->RegDomain = Adapter->RegDomain;
  pDevConfig->PreambleType = Adapter->AssociationInfo.PreambleType;
  pDevConfig->PowerMgmtMode = Adapter->PowerMgmtMode;
  pDevConfig->TxPowerLevel = Adapter->TxPowerLevel;
  pDevConfig->RadioIsOn = Adapter->RadioIsOn;

  pDevConfig->TxRate = Adapter->DataFramesTxRate;
  pDevConfig->FragmentationThreshold = Adapter->FragmentationThreshold;
  pDevConfig->RtsCtsThreshold = Adapter->RtsThreshold;
	if(Adapter->RadioIsOn){
	  pDevConfig->Rssi = Adapter->Rssi;
    IF_VERY_LOUD(DbgPrint("Rssi : %d\n", Adapter->Rssi);)
	  pDevConfig->LinkQuality = Adapter->LinkQuality;
	}else{
	  pDevConfig->Rssi = 0;
	  pDevConfig->LinkQuality = 0;
	}
  if ((Adapter->StationState != 0x05) || (Adapter->OperatingMode == AD_HOC_MODE)){
	  pDevConfig->Rssi = 0;
	  pDevConfig->LinkQuality = 0;
  }
}
/* IW-HANDLERS */
static const long frequency_list[] = { 2412, 2417, 2422, 2427, 2432, 2437, 2442,
				2447, 2452, 2457, 2462, 2467, 2472, 2484 };

static int vnet_set_encode(struct net_device *dev,
			    struct iw_request_info *info,
			    struct iw_point *dwrq,
			    char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        int i __attribute__ ((unused));
       // DEVICE_CONFIGURATION DevConfig;
        
    DisableInterrupts(Adapter);

    // Ensure That we only set WEP here !!!
     Adapter->SecurityInfo.InWpaMode = 0;
            
		if ((dwrq->flags & IW_ENCODE_DISABLED) || (dwrq->flags & IW_ENCODE_NOKEY)) {

	  Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = 0;
#ifdef ATMEL_WLAN
      Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = CIPHER_SUITE_NONE;
#endif
	ConfigureWEP (Adapter);
	if(Adapter->StationIsAssociated) { 
	  EnableInterrupts (Adapter);
	  return 0;
  }
  else if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
       		Join(Adapter, BSS_TYPE_INFRASTRUCTURE);
	else
       		Join(Adapter, BSS_TYPE_AD_HOC);

	  EnableInterrupts (Adapter);
	  return 0;
	}
  if (dwrq->flags & IW_ENCODE_ENABLED)
	{
	  Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = 1;
	  EnableInterrupts (Adapter);
	  return 0;
	}

      if (dwrq->flags & IW_ENCODE_OPEN){
          Adapter->SecurityInfo.AuthenticationMode = _802_11AuthModeOpen;
          Adapter->SecurityInfo.EncryptionInfo.AuthenticationType = Adapter->SecurityInfo.AuthenticationMode;
      } else if (dwrq->flags & IW_ENCODE_RESTRICTED) {
          Adapter->SecurityInfo.AuthenticationMode = _802_11AuthModeShared;
          Adapter->SecurityInfo.EncryptionInfo.AuthenticationType = Adapter->SecurityInfo.AuthenticationMode;
      }

      if ((dwrq->length != 5) && (dwrq->length != 13))
	{
	  EnableInterrupts (Adapter);
	  return -EINVAL;
	}
      if (dwrq->length == 5)
	{
	  Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = WEP_64BIT;
#ifdef ATMEL_WLAN
          Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = 1;
#endif
	}
      else
	{
	  Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = WEP_128BIT;
#ifdef ATMEL_WLAN
          Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = 5;
#endif
	}
      IF_LOUD (printk
	       (KERN_DEBUG "Using %dbit key, position %d\n",
		(Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == WEP_64BIT) ? 64 : 128,
		(dwrq->flags & IW_ENCODE_INDEX) - 1);)
      if ((Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == WEP_64BIT)
	  || (Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == WEP_128BIT))
	{
	  if ((dwrq->flags & IW_ENCODE_INDEX) == 0)
	    {
	      Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse = 0;
	    }
	  else{
	    Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse =
	      (dwrq->flags & IW_ENCODE_INDEX) - 1;
          }

	}
      else{
	Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse = 0;
      }
      if (Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse > 3)
	{
	  EnableInterrupts (Adapter);
	  return -EINVAL;
	}

#ifdef ATMEL_WLAN
      memcpy(((PUCHAR) Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse]),
		      extra, dwrq->length);
      Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse][39] = 1;
#else
      memcpy(((PUCHAR) Adapter->SecurityInfo.EncryptionInfo.WepKey1 +
		       ((13 * Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse))),
		      extra, dwrq->length);
#endif

//	ConfigureWEP (Adapter);
  Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex = Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse;
  Adapter->SecurityInfo.PairwiseKeyAvailable = TRUE;
  Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption1Enabled;
  SetEncryptionStatus(Adapter, TRUE, Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse, TRUE);
	if(Adapter->StationIsAssociated) {
            EnableInterrupts(Adapter);
            return 0;
        } else if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
       		Join(Adapter, BSS_TYPE_INFRASTRUCTURE);
	else
       		Join(Adapter, BSS_TYPE_AD_HOC);

        EnableInterrupts(Adapter);
	return 0;	
}

static int vnet_get_encode(struct net_device *dev,
			    struct iw_request_info *info,
			    struct iw_point *dwrq,
			    char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;

      if (Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == 0)
	{
	  dwrq->flags |= IW_ENCODE_DISABLED;
	  dwrq->length = 0;
	}
      else
	{
	  dwrq->flags &= ~IW_ENCODE_DISABLED;
	  dwrq->flags |= (Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse + 1);


	  if (Adapter->SecurityInfo.EncryptionInfo.AuthenticationType == C80211_MGMT_AAN_OPENSYSTEM)
	    dwrq->flags |= IW_ENCODE_OPEN;
          else dwrq->flags |= IW_ENCODE_RESTRICTED;
	  dwrq->length = (Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == WEP_64BIT) ? 5 : 13;
	}

#ifdef ATMEL_WLAN
      memcpy (extra,
		    ((PUCHAR) Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse] +
		     ((Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse * 13))),
		    dwrq->length);
#else
      memcpy (extra,
		    ((PUCHAR) Adapter->SecurityInfo.EncryptionInfo.WepKey1 +
		     ((Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse * 13))),
		    dwrq->length);
#endif
	return 0;
}

static int vnet_set_rts(struct net_device *dev,
			 struct iw_request_info *info,
			 struct iw_param *vwrq,
			 char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;

        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);
        if(vwrq->disabled)
          DevConfig.RtsCtsThreshold = 2347;
        if((DevConfig.RtsCtsThreshold < 255) || (DevConfig.RtsCtsThreshold > 2347)) {
          EnableInterrupts(Adapter);
          return -EINVAL;
        }
	DevConfig.RtsCtsThreshold = vwrq->value;
        ChangeConfiguration(Adapter, &DevConfig);
        EnableInterrupts(Adapter);
	
	return 0;	
}

static int vnet_set_wap(struct net_device *dev,
			 struct iw_request_info *info,
			 struct sockaddr *awrq,
			 char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
	      int i;
	
	if (awrq->sa_family != ARPHRD_ETHER)
		return -EINVAL;
	
        DisableInterrupts(Adapter);
	if (memcmp(Adapter->AssociationInfo.CurrentBSSID, awrq->sa_data, 6) == 0) {
		del_timer(Adapter->MgmtTimer);
    Scan(Adapter, TRUE);
    EnableInterrupts(Adapter);
		return 0;
	}
	for(i=0; i<Adapter->BSSListEntries; i++) {
		if (memcmp(Adapter->BSSinfo[i].BSSID, awrq->sa_data, 6) == 0) {
      memcpy(Adapter->AssociationInfo.CurrentBSSID, Adapter->BSSinfo[i].BSSID, 6);
      JoinSpecificBSS(Adapter, i, TRUE);
      EnableInterrupts(Adapter);
			return 0;
		}
	}
		
  EnableInterrupts(Adapter);

	return 0;
}
static int vnet_get_wap(struct net_device *dev,
			 struct iw_request_info *info,
			 struct sockaddr *awrq,
			 char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
	memcpy(awrq->sa_data, Adapter->AssociationInfo.CurrentBSSID, 6);
	awrq->sa_family = ARPHRD_ETHER;

	return 0;
}

static int vnet_get_rts(struct net_device *dev,
			 struct iw_request_info *info,
			 struct iw_param *vwrq,
			 char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;

        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);
	
	vwrq->value = DevConfig.RtsCtsThreshold;
	vwrq->disabled = (vwrq->value >= 2347);
	vwrq->fixed = 1;
        EnableInterrupts(Adapter);

	return 0;
}
static int vnet_get_frag(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_param *vwrq,
			  char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;

        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);
        vwrq->value = DevConfig.FragmentationThreshold;
        vwrq->disabled = (vwrq->value >= 2346);
        vwrq->fixed = 1;
        EnableInterrupts(Adapter);

	return 0;
}
static int vnet_set_frag(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_param *vwrq,
			  char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;

        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);
        if(vwrq->disabled)
          DevConfig.FragmentationThreshold = 2346;
        if((DevConfig.FragmentationThreshold < 256) || (DevConfig.FragmentationThreshold > 2346)) {
          EnableInterrupts(Adapter);
          return -EINVAL;
        }
        DevConfig.FragmentationThreshold = vwrq->value;
        ChangeConfiguration(Adapter, &DevConfig);
        EnableInterrupts(Adapter);
	
	return 0;
}

static int vnet_set_rate(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_param *vwrq,
			  char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;
	
        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);

			switch (vwrq->value) {
			case  1000000: DevConfig.TxRate = 0; break;
			case  2000000: DevConfig.TxRate = 1; break;
			case  5500000: DevConfig.TxRate = 2; break;
			case 11000000: DevConfig.TxRate = 3; break;
			default: return -EINVAL;
			}
                        ChangeConfiguration(Adapter, &DevConfig);
        EnableInterrupts(Adapter);

	return 0;
}

static int vnet_set_freq(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_freq *fwrq,
			  char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;
	      int rc = 0;
	
	/* If setting by frequency, convert to a channel */
        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);

	if((fwrq->e == 1) &&
	   (fwrq->m >= (int) 241200000) &&
	   (fwrq->m <= (int) 248700000)) {
		int f = fwrq->m / 100000;
		int c = 0;
		while((c < 14) && (f != frequency_list[c]))
			c++;
		fwrq->e = 0;
		fwrq->m = c + 1;
	}
	if((fwrq->m > 1000) || (fwrq->e > 0))
		rc = -EOPNOTSUPP;
	else {
		int channel = fwrq->m;
			Adapter->Channel = (UCHAR) channel;
	}

        ChangeConfiguration(Adapter, &DevConfig);
        EnableInterrupts(Adapter);
	return rc;
}

static int vnet_get_freq(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_freq *fwrq,
			  char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;

        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);

	fwrq->m = frequency_list[(Adapter->Channel) -1] * 100000;
	fwrq->e = 1;

        EnableInterrupts(Adapter);
	return 0;
}

static int vnet_get_range(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_point *dwrq,
			   char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
	struct iw_range *range = (struct iw_range *) extra;
	int k,i,j;

	dwrq->length = sizeof(struct iw_range);
	memset(range, 0, sizeof(range));
	range->min_nwid = 0x0000;
	range->max_nwid = 0x0000;
	range->num_channels = 0;

	for (j = 0; j < sizeof(channel_table)/sizeof(channel_table[0]); j++)
		if (Adapter->RegDomain == channel_table[j].reg_domain) {
			range->num_channels = channel_table[j].max - channel_table[j].min + 1;
			break;
		}
	if (range->num_channels != 0) {
		for(k = 0, i = channel_table[j].min; i <= channel_table[j].max; i++) {
			range->freq[k].i = i; 
			range->freq[k].m = frequency_list[i-1] * 100000;
			range->freq[k++].e = 1;
		}
		range->num_frequency = k;
	}
	
	range->max_qual.qual = 100;
	range->max_qual.level = 100;
	range->max_qual.noise = 0;
	range->sensitivity = 0;

	range->bitrate[0] =  1000000;
	range->bitrate[1] =  2000000;
	range->bitrate[2] =  5500000;
	range->bitrate[3] = 11000000;
	range->num_bitrates = 4;

	range->min_rts = 0;
	range->max_rts = 2347;
	range->min_frag = 256;
	range->max_frag = 2346;

	range->encoding_size[0] = 5;
	range->encoding_size[1] = 13;
	range->num_encoding_sizes = 2;
	range->max_encoding_tokens = 4;
	
	range->pmp_flags = IW_POWER_ON;
	range->pmt_flags = IW_POWER_ON;
	range->pm_capa = 0;
	
	range->we_version_source = WIRELESS_EXT;
	range->we_version_compiled = WIRELESS_EXT;
	range->retry_capa = IW_RETRY_LIMIT ;
	range->retry_flags = IW_RETRY_LIMIT;
	range->r_time_flags = 0;
	range->min_retry = 1;
	range->max_retry = 65535;
	range->avg_qual.qual = 50;
	range->avg_qual.level = 50;
	range->avg_qual.noise = 0;

	return 0;
}
static int vnet_get_name(struct net_device *dev,
			  struct iw_request_info *info,
			  char *cwrq,
			  char *extra)
{
#if defined(R504)
      strcpy (cwrq, "ATMEL 504");
#elif defined(R504_2958)
      strcpy (cwrq, "ATMEL 504+2958");
#elif defined(PCI)
      strcpy (cwrq, "ATMEL PCI");
#elif defined(REV_D)
      strcpy (cwrq, "ATMEL REVD");
#elif defined(REV_E)
      strcpy (cwrq, "ATMEL REVE");
#elif defined(RFMD) || defined(RFMD_3COM)
      strcpy (cwrq, "ATMEL RFMD");
#endif
	return 0;
}
#if 0
static int vnet_config_commit(struct net_device *dev,
			       struct iw_request_info *info,	/* NULL */
			       void *zwrq,			/* NULL */
			       char *extra)			/* NULL */
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
	return 0;
}
#endif

static int vnet_get_rate(struct net_device *dev,
			 struct iw_request_info *info,
			 struct iw_param *vwrq,
			 char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;

        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);
		vwrq->fixed = 1;
		switch(DevConfig.TxRate) {
		case 0: vwrq->value =  1000000; 
                        break;
		case 1: vwrq->value =  2000000; 
                        break;
		case 2: vwrq->value =  5500000; 
                        break;
		case 3: vwrq->value = 11000000; 
                        break;
                }
        EnableInterrupts(Adapter);
	return 0;
}
static int vnet_set_scan(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_param *vwrq,
			  char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;

	
	if (Adapter->StationState == STATION_STATE_NO_CARD)
		return -EAGAIN;

	if (Adapter->SiteSurveyState == SITE_SURVEY_IN_PROGRESS)
		return -EBUSY;
        
  DisableInterrupts(Adapter);
	Adapter->SiteSurveyState = SITE_SURVEY_IN_PROGRESS;
  Scan (Adapter, FALSE);
	EnableInterrupts(Adapter);
	
//	return -EINPROGRESS;
	return 0;
}
static int vnet_set_essid(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_point *dwrq,
			   char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;

        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);
        memcpy (DevConfig.SSID, extra, dwrq->length);
        DevConfig.SSID[dwrq->length] ='\0';
        DevConfig.SSIDlength = strlen(DevConfig.SSID);
        ChangeConfiguration (Adapter, &DevConfig);
        EnableInterrupts(Adapter);
	return 0;
}

static int vnet_get_essid(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_point *dwrq,
			   char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;

        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);
      if (Adapter->AssociationInfo.DesiredSSIDsize != 0) {
	  memcpy (extra, Adapter->AssociationInfo.DesiredSSID, Adapter->AssociationInfo.DesiredSSIDsize);
	  dwrq->length = Adapter->AssociationInfo.DesiredSSIDsize;
	} else {
	  memcpy (extra, Adapter->InitialSSID, Adapter->InitialSSIDsize);
	  dwrq->length = Adapter->InitialSSIDsize;
	}
      dwrq->flags = 1;
	EnableInterrupts(Adapter);

	return 0;
}

static int vnet_set_mode(struct net_device *dev,
			  struct iw_request_info *info,
			  ULONG *uwrq,
			  char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;

	if (*uwrq != IW_MODE_ADHOC && *uwrq != IW_MODE_INFRA)
		return -EINVAL;

        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);
      switch (*uwrq)
	{
	case IW_MODE_ADHOC:
	  DevConfig.OperatingMode = 0;
	  break;
	case IW_MODE_INFRA:
	  DevConfig.OperatingMode = 1;
	  break;
	default:
	  return -EINVAL;
	}
        ChangeConfiguration (Adapter, &DevConfig);
        del_timer(Adapter->MgmtTimer);
        EnableInterrupts(Adapter);
	return 0;  
	//return -EINPROGRESS;  
}
static int vnet_get_mode(struct net_device *dev,
			  struct iw_request_info *info,
			  ULONG *uwrq,
			  char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
	
      switch (Adapter->OperatingMode)
	{
	case AD_HOC_MODE:
	  *uwrq = IW_MODE_ADHOC;
	  break;
	case INFRASTRUCTURE_MODE:
	  *uwrq = IW_MODE_INFRA;
	  break;
	}
	return 0;
}
#define MAX_WPA_IE_LEN 64

void VNetGetSystemTime(PVOID tCurrentTime) 
{
  u64* pTime = (u64*) tCurrentTime; 
  struct timeval timev;
  do_gettimeofday(&timev);
  (u64) (*pTime) = timev.tv_sec * 10000000;
};

int atmel_assoc_info_event (struct net_device *dev, PVNet_ADAPTER Adapter)
{
	union iwreq_data wrqu;
    char buf[255];
    int j;
    UCHAR *p = buf;

    IF_LOUD(DbgPrint("Sending AssocInfo Event\n");)

    p += sprintf(p, "ASSOCINFO(ReqIEs=");
    for (j = 0; j < Adapter->AssociationInfo.AssReassIEInfoBufferLength; j++) {
        
            p += sprintf(p, "%02x", Adapter->AssociationInfo.AssReassIEInfoBuffer[j]);
    }

    memcpy(wrqu.ap_addr.sa_data, Adapter->AssociationInfo.CurrentBSSID, 6);
    wrqu.ap_addr.sa_family = 1;
    wireless_send_event(dev, SIOCGIWAP, &wrqu, NULL);
    memset(&wrqu, 0, sizeof(wrqu));
    wrqu.data.length = strlen(buf);
    wireless_send_event(dev, IWEVCUSTOM, &wrqu, buf);

    return 0;
}
int atmel_scan_info_event (struct net_device *dev, PVNet_ADAPTER Adapter)
{
	union iwreq_data wrqu;

    printk(KERN_INFO "Sending Scan Results Event\n");

    wrqu.data.length = 0;
    wrqu.data.flags = 0;
    wireless_send_event(dev, SIOCGIWSCAN, &wrqu, NULL);
    Adapter->SiteSurveyState = SITE_SURVEY_COMPLETED;
    return 0;
}

static int vnet_get_scan(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_point *dwrq,
			  char *extra)
{
	int i, j;
        struct net_local *local = (struct net_local *) dev->priv;
	char *current_ev = extra;
        PVNet_ADAPTER Adapter = &local->Adapter;
	char buf[MAX_WPA_IE_LEN * 2 + 30];
	struct iw_event	iwe;
	
	if (Adapter->SiteSurveyState != SITE_SURVEY_COMPLETED)
		return -EAGAIN;
	

  //DisableInterrupts(Adapter);
        
	for(i=0; i<Adapter->BSSListEntries; i++) { 
		iwe.cmd = SIOCGIWAP;
		iwe.u.ap_addr.sa_family = ARPHRD_ETHER;
		memcpy(iwe.u.ap_addr.sa_data, Adapter->BSSinfo[i].BSSID, 6);
		current_ev = iwe_stream_add_event(current_ev, extra + IW_SCAN_MAX_DATA, &iwe, IW_EV_ADDR_LEN);

		iwe.u.data.length =  Adapter->BSSinfo[i].SSIDsize;
		if (iwe.u.data.length > 32)
			iwe.u.data.length = 32;
		iwe.cmd = SIOCGIWESSID;
		iwe.u.data.flags = 1;
		current_ev = iwe_stream_add_point(current_ev, extra + IW_SCAN_MAX_DATA, &iwe, Adapter->BSSinfo[i].SSID);
		
		iwe.cmd = SIOCGIWMODE;
		iwe.u.mode = Adapter->BSSinfo[i].BSStype;
		current_ev = iwe_stream_add_event(current_ev, extra + IW_SCAN_MAX_DATA, &iwe, IW_EV_UINT_LEN);
	
		iwe.cmd = SIOCGIWFREQ;
		iwe.u.freq.m = Adapter->BSSinfo[i].Channel;
		iwe.u.freq.e = 0;
		current_ev = iwe_stream_add_event(current_ev, extra + IW_SCAN_MAX_DATA, &iwe, IW_EV_FREQ_LEN);
		
		iwe.cmd = SIOCGIWENCODE;
		if (Adapter->BSSinfo[i].EncryptionIsOn)
			iwe.u.data.flags = IW_ENCODE_ENABLED | IW_ENCODE_NOKEY;
		else
			iwe.u.data.flags = IW_ENCODE_DISABLED;
		iwe.u.data.length = 0;
		current_ev = iwe_stream_add_point(current_ev, extra + IW_SCAN_MAX_DATA, &iwe, NULL);

    if(Adapter->BSSinfo[i].WpaIELen > 0) 
    {
        UCHAR *p = buf;
        p += sprintf(p, "wpa_ie=");
        for (j = 0; j < Adapter->BSSinfo[i].WpaIELen; j++) {
                p += sprintf(p, "%02x", Adapter->BSSinfo[i].WpaIE[j]);
        }

        memset(&iwe, 0, sizeof(iwe));
        iwe.cmd = IWEVCUSTOM;
        iwe.u.data.length = strlen(buf);
        current_ev = iwe_stream_add_point( current_ev, extra + IW_SCAN_MAX_DATA, &iwe, buf);

    
    }
	}

	/* Length of data */
	dwrq->length = (current_ev - extra);
	dwrq->flags = 0;	/* todo */
  //EnableInterrupts(Adapter);
	
	return 0;
}


/*------------------------------------------------------------------*/
/*
 * Wireless Private Handler : get preamble
 */
static int vnet_ioctl_get_preamble(struct net_device *dev,
			    struct iw_request_info *info,
                            union iwreq_data *wrq,
			    char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;
	
        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);
      if (wrq->data.pointer) {
          wrq->data.flags = 1;
          if (DevConfig.PreambleType)
            {
              wrq->data.length = 6;
              if (copy_to_user (wrq->data.pointer, "short", 6)) {
                  EnableInterrupts (Adapter);
                  return -EFAULT;
              }
            } else {
              wrq->data.length = 5;
              if (copy_to_user (wrq->data.pointer, "long", wrq->data.length)) {
                  EnableInterrupts (Adapter);
                  return -EFAULT;
              }
            }
      }
        EnableInterrupts(Adapter);
	return 0;
}

static int vnet_set_txpower(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_param *vwrq,
			   char *extra)
{
  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;
  DEVICE_CONFIGURATION DevConfig;

  DisableInterrupts(Adapter);

  UpdateSupportAppInfo (Adapter, &DevConfig);
  vwrq->flags = IW_TXPOW_DBM;
  IF_LOUD(DbgPrint("Setting tx power level %d\n", vwrq->value);)
  DevConfig.TxPowerLevel = (10 * vwrq->value);

  if (vwrq->flags != IW_TXPOW_DBM) {
        printk("SIOCSIWTXPOW with mW is not supported; use dBm\n");
        return -EOPNOTSUPP;
  }

  ChangeConfiguration(Adapter, &DevConfig);

  EnableInterrupts(Adapter);
  
	return 0;
}

static int vnet_get_txpower(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_param *vwrq,
			   char *extra)
{
  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;
  DEVICE_CONFIGURATION DevConfig;
  
  DisableInterrupts(Adapter);
  UpdateSupportAppInfo (Adapter, &DevConfig);
  if(Adapter->RadioIsOn) {
    vwrq->disabled = 0;
    vwrq->fixed = 1;
    vwrq->value = (Adapter->TxPowerLevel)/(UCHAR)10;
    vwrq->flags = IW_TXPOW_DBM;
  }
  EnableInterrupts(Adapter);
	return 0;
}

static int vnet_set_power(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_param *vwrq,
			   char *extra)
{
  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;
  DEVICE_CONFIGURATION DevConfig;

  DisableInterrupts(Adapter);

  UpdateSupportAppInfo (Adapter, &DevConfig);

	DevConfig.PowerMgmtMode = vwrq->disabled ? 0 : 1;
  
  ChangeConfiguration(Adapter, &DevConfig);

  EnableInterrupts(Adapter);
  
	return 0;
}

static int vnet_get_power(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_param *vwrq,
			   char *extra)
{
  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;
  DEVICE_CONFIGURATION DevConfig;
  
  DisableInterrupts(Adapter);
  UpdateSupportAppInfo (Adapter, &DevConfig);
	vwrq->disabled = DevConfig.PowerMgmtMode ? 0 : 1;
	vwrq->flags = IW_POWER_ON;
  EnableInterrupts(Adapter);
	return 0;
}

/*------------------------------------------------------------------*/
/*
 * Wireless Private Handler : set preamble
 */
static int vnet_ioctl_set_preamble(struct net_device *dev,
			    struct iw_request_info *info,
                            union iwreq_data *wrq,
			    char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
        DEVICE_CONFIGURATION DevConfig;
	
        DisableInterrupts(Adapter);
        UpdateSupportAppInfo (Adapter, &DevConfig);
        DevConfig.PreambleType = *((int *) wrq->name) & 1;
        ChangeConfiguration (Adapter, &DevConfig);
        del_timer(Adapter->MgmtTimer);
        EnableInterrupts(Adapter);
	return 0;
}



static int vnet_ioctl_get_wpa_param(struct net_device *dev,
					  struct iw_request_info *info,
                                          union iwreq_data *wrq,
					  char *extra)
{
        struct net_local *local = (struct net_local *) dev->priv;
        PVNet_ADAPTER Adapter = &local->Adapter;
	      int ret = 0;
        int length = 0;
        char tmpbuff[9];

        DisableInterrupts(Adapter);

   memset(tmpbuff, 0, 9); 

	switch (wrq->data.flags) {
	case ATMEL_PARAM_WPA:
		break;
	case ATMEL_PARAM_PRIVACY_INVOKED:
		break;
	case ATMEL_PARAM_WPA_TYPE:
              switch(Adapter->SecurityInfo.AuthenticationMode) {
              case 0:
                length = 5;
                memcpy(tmpbuff, "Open", 4);
                tmpbuff[5] = '\0';
                break;
              case 1:
                length = 7;
                memcpy(tmpbuff, "Shared", 6);
                tmpbuff[7] = '\0';
                break;
              case 3:
                length = 4;
                memcpy(tmpbuff, "WPA", 3);
                tmpbuff[4] = '\0';
                break;
              case 4:
                length = 8;
                memcpy(tmpbuff, "WPA PSK", 7);
                tmpbuff[8] = '\0';
                break;
              default :
                break;
              }
              if(wrq->data.pointer) {
                  wrq->data.length = length;
                  if(copy_to_user(wrq->data.pointer, tmpbuff, wrq->data.length)) {
                    return -EFAULT; 
                  }
              }
		break;
	default:
		ret = -EOPNOTSUPP;
		break;
	}

        EnableInterrupts(Adapter);
  return 0;
}

static int vnet_ioctl_set_wpa_param(struct net_device *dev,
					  struct iw_request_info *info, union iwreq_data *wrq, char *extra)
{
  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;
	int *i = (int *) extra;
	int param = *i;
	int value = *(i + 1);
	int ret = 0;

  DisableInterrupts(Adapter);
	switch (param) {
	case ATMEL_PARAM_WPA:
    if((Adapter->SecurityInfo.AuthenticationMode == 3) || (Adapter->SecurityInfo.AuthenticationMode == 4))
                Adapter->SecurityInfo.InWpaMode = 1;
		break;
	case ATMEL_PARAM_PRIVACY_INVOKED:
		break;
	case ATMEL_PARAM_WPA_TYPE:
                    Adapter->SecurityInfo.AuthenticationMode = value;
//                Adapter->SecurityInfo.EncryptionInfo.AuthenticationType = value;
                    if(value == 3){
                        IF_LOUD(DbgPrint("Entering WPA Mode - Waiting for the Completion of Key Exchange ...\n");)
                    }else  if(value == 4){
                        IF_LOUD(DbgPrint("Entering WPA PSK Mode - Waiting for the Completion of Key Exchange ...\n");)
                    }
                Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption2KeyAbsent;
		break;
	default:
		printk(KERN_INFO "%s: atmel_param: unknown param %d\n",
		       dev->name, param);
		ret = -EOPNOTSUPP;
		break;
	}

        EnableInterrupts(Adapter);
	return ret;
}

static const iw_handler		vnet_handler[] =
{
	//(iw_handler) vnet_config_commit,	/* SIOCSIWCOMMIT */
	(iw_handler) NULL,	                /* SIOCSIWCOMMIT */
	(iw_handler) vnet_get_name,    	/* SIOCGIWNAME */
	(iw_handler) NULL,			/* SIOCSIWNWID */
	(iw_handler) NULL,			/* SIOCGIWNWID */
	(iw_handler) vnet_set_freq,	        /* SIOCSIWFREQ */
	(iw_handler) vnet_get_freq,		/* SIOCGIWFREQ */
	(iw_handler) vnet_set_mode,		/* SIOCSIWMODE */
	(iw_handler) vnet_get_mode,		/* SIOCGIWMODE */
	(iw_handler) NULL,                  	/* SIOCSIWSENS */
	(iw_handler) NULL,           		/* SIOCGIWSENS */
	(iw_handler) NULL,			/* SIOCSIWRANGE */
	(iw_handler) vnet_get_range,           /* SIOCGIWRANGE */
	(iw_handler) NULL,			/* SIOCSIWPRIV */
	(iw_handler) NULL,			/* SIOCGIWPRIV */
	(iw_handler) NULL,			/* SIOCSIWSTATS */
	(iw_handler) NULL,			/* SIOCGIWSTATS */
	(iw_handler) NULL,			/* SIOCSIWSPY */
	(iw_handler) NULL,			/* SIOCGIWSPY */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) vnet_set_wap,		/* SIOCSIWAP */
	(iw_handler) vnet_get_wap,		/* SIOCGIWAP */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) NULL,       		/* SIOCGIWAPLIST */
	(iw_handler) vnet_set_scan,		/* SIOCSIWSCAN */
	(iw_handler) vnet_get_scan,		/* SIOCGIWSCAN */
	(iw_handler) vnet_set_essid,		/* SIOCSIWESSID */
	(iw_handler) vnet_get_essid,		/* SIOCGIWESSID */
	(iw_handler) NULL,         		/* SIOCSIWNICKN */
	(iw_handler) NULL,          		/* SIOCGIWNICKN */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) vnet_set_rate,		/* SIOCSIWRATE */
	(iw_handler) vnet_get_rate,		/* SIOCGIWRATE */
	(iw_handler) vnet_set_rts,		/* SIOCSIWRTS */
	(iw_handler) vnet_get_rts,		/* SIOCGIWRTS */
	(iw_handler) vnet_set_frag,		/* SIOCSIWFRAG */
	(iw_handler) vnet_get_frag,		/* SIOCGIWFRAG */
	(iw_handler) vnet_set_txpower,         		/* SIOCSIWTXPOW */
	(iw_handler) vnet_get_txpower,          		/* SIOCGIWTXPOW */
	(iw_handler) NULL,		/* SIOCSIWRETRY */
	(iw_handler) NULL,		/* SIOCGIWRETRY */
	(iw_handler) vnet_set_encode,		/* SIOCSIWENCODE */
	(iw_handler) vnet_get_encode,		/* SIOCGIWENCODE */
	(iw_handler) vnet_set_power,		/* SIOCSIWPOWER */
	(iw_handler) vnet_get_power,		/* SIOCGIWPOWER */
};

static const iw_handler		vnet_private_handler[] =
{
	(iw_handler) vnet_ioctl_set_preamble,	/* +0 */
	(iw_handler) vnet_ioctl_get_preamble,	/* +1 */
        NULL,
	(iw_handler) vnet_ioctl_set_wpa_param,	/* +3 */
	(iw_handler) vnet_ioctl_get_wpa_param,	/* +4 */
};

static const struct iw_priv_args vnet_private_args[] = {
/*{ cmd,         set_args,                            get_args, name } */
  { SIOCIWPRIVSPREAM, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "set_preamble" },
  { SIOCIWPRIVGPREAM, 0, IW_PRIV_TYPE_CHAR | 6 | 1, "get_preamble" },
  { ATMEL_WPA_IOCTL_PARAM, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 2, 0, "atmlparam" },
  { ATMEL_WPA_IOCTL_GET_PARAM, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, "get_atmlparam" },
  { ATMEL_WPA_IOCTL_PARAM, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "" },
  { ATMEL_WPA_IOCTL_GET_PARAM, 0, IW_PRIV_TYPE_CHAR | 8 | 1, "" },
  { ATMEL_PARAM_WPA, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "wpa" },
  { ATMEL_PARAM_WPA, 0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, "getwpa" },
  { ATMEL_PARAM_PRIVACY_INVOKED, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "privacy_invoked" },
  { ATMEL_PARAM_PRIVACY_INVOKED, 0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, "getprivacy_invo" },
  { ATMEL_PARAM_WPA_TYPE, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "wpa_mode" },
  { ATMEL_PARAM_WPA_TYPE, 0, IW_PRIV_TYPE_CHAR | 8 | 1, "getwpa_mode" },
};

static const struct iw_handler_def	vnet_handler_def =
{
	.num_standard	= sizeof(vnet_handler)/sizeof(iw_handler),
	.num_private	= sizeof(vnet_private_handler)/sizeof(iw_handler), 
	.num_private_args = sizeof(vnet_private_args)/sizeof(struct iw_priv_args), 
	.standard	= (iw_handler *) vnet_handler,
	.private	= (iw_handler *) vnet_private_handler, 
	.private_args	= (struct iw_priv_args *) vnet_private_args
};

static int atmel_wpa_ioctl(PVNet_ADAPTER Adapter, struct iw_point *p)
{
	struct atmel_param *param;
#ifdef HWWPA
        UCHAR mic_offset[8];
#endif
        BOOLEAN IsTxKey, IsGroupKey;
        UCHAR BroadcastBSSID[] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
        UCHAR EncryptionSuite;
        //int i;
	int ret = 0;
	int ap_ioctl = 0;

	if (p->length < sizeof(struct atmel_param) ||
	    p->length > 1024 || !p->pointer)
		return -EINVAL;

	param = (struct atmel_param *) kmalloc(p->length, GFP_KERNEL);
	if (param == NULL)
		return -ENOMEM;


	if (copy_from_user(param, p->pointer, p->length)) {
		ret = -EFAULT;
		goto out;
	}
        DisableInterrupts(Adapter);

	switch (param->cmd) {
	case SET_WPA_ENCRYPTION:

		if(!param->key_len){ 
        break;
    }

		if((!param->key_idx) && (!Adapter->SecurityInfo.InWpaMode))
		{
			printk("Key For Open 1x\n");
			IsGroupKey = TRUE;
      if(param->set_tx) {
			    IsTxKey = TRUE;
      } else {
			    IsTxKey = FALSE;
      }
    } else if(!param->key_idx) {
			printk("Pairwise Key\n");
			IsGroupKey = FALSE;
			IsTxKey = TRUE;
		} else if((param->key_idx) && (param->set_tx & 0x80)) {
			printk("Pairwise Key for 1X (No-WPA)\n");
			IsGroupKey = FALSE;
			IsTxKey = TRUE;
		} else if((param->key_idx) && (param->set_tx)) {
			printk("Group Key / Tx / Rx\n");
			IsGroupKey = TRUE;
			IsTxKey = TRUE;
		} else if((param->key_idx) && (!param->set_tx)) {
			printk("Group Key Rx only\n");
			IsGroupKey = TRUE;
			IsTxKey = FALSE;
		} else {
      printk(KERN_INFO "Invalid Combination\n");
      break;
    }

		if((Adapter->OperatingMode == AD_HOC_MODE) && IsGroupKey && MemCompare(param->sta_addr, BroadcastBSSID, 6))
		{
			IF_LOUD(DbgPrint("*** ERROR: Invalid BSSID / Group Key combination in AdHoc Mode\n");)
			break;
		}

		if(param->key_len == 5)
		{
			IF_LOUD(DbgPrint("Cipher Suite is WEP-64...\n");)
			EncryptionSuite = CIPHER_SUITE_WEP_64;
		}
		else
		if(param->key_len == 13)
		{
			IF_LOUD(DbgPrint("Cipher Suite is WEP-128...\n");)
			EncryptionSuite = CIPHER_SUITE_WEP_128;
		}
		else
#ifdef AES
		if(param->key_len == 16)
		{
			IF_LOUD(DbgPrint("Cipher Suite is AES...\n");)
			EncryptionSuite = CIPHER_SUITE_AES;
		}
		else
#endif
		if(param->key_len == 32)
		{
			printk(KERN_INFO "Cipher Suite is TKIP...\n");
			EncryptionSuite = CIPHER_SUITE_TKIP;
		}
		else
		{
			IF_LOUD(DbgPrint("Cipher Suite is NONE...\n");)
			EncryptionSuite = CIPHER_SUITE_NONE;
		}
                printk(KERN_INFO "#######################################################\n");
                printk(KERN_INFO "ENCRYPTION -SETTINGS : key_idx=%d set_tx=%d seq_len=%d "
                                              "key_len=%d\n", param->key_idx, param->set_tx, param->seq_len, param->key_len);
                printk(KERN_INFO "#######################################################\n");

		

		memcpy(Adapter->SecurityInfo.EncryptionInfo.KeyRSC, param->seq, param->seq_len);

			if(!memcmp(param->sta_addr, BroadcastBSSID, 6) && (Adapter->OperatingMode == INFRASTRUCTURE_MODE )){
                            
		                memcpy(Adapter->SecurityInfo.EncryptionInfo.BSSID, Adapter->AssociationInfo.CurrentBSSID, 6);
                           }else {
		                memcpy(Adapter->SecurityInfo.EncryptionInfo.BSSID, param->sta_addr, 6);
                           }

#ifdef HWWPA
		memcpy(&mic_offset[0], &param->key[24], 8);
		memcpy(&Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[(param->key_idx)][0], param->key, param->key_len);
		memcpy(&Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[(param->key_idx)][24], &param->key[16], 8);
		memcpy(&Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[(param->key_idx)][16], &mic_offset[0], 8);
#else
		memcpy(&Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[(param->key_idx)][0], param->key, param->key_len);
#endif
		
		if(IsGroupKey)
		{
			if(!((param->key_len == 5 || param->key_len == 13) && !IsTxKey))
				Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex = param->key_idx;

			Adapter->SecurityInfo.EncryptionInfo.GroupKeyLength = param->key_len;
			Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite = EncryptionSuite;

			if(!memcmp(param->sta_addr, BroadcastBSSID, 6) || 
				(Adapter->OperatingMode == INFRASTRUCTURE_MODE && !memcmp(param->sta_addr, Adapter->AssociationInfo.CurrentBSSID, 6)))
			{ //configure immediately
				IF_LOUD(DbgPrint(">>>> Configure Immediately (Group key with Broadcast BSSID or our BSSID in Infrastructure mode)\n");)

			//	if(IsTxKey)
				{
					Adapter->SecurityInfo.GroupKeyAvailable = TRUE;
				}
				if(SetEncryptionStatus(Adapter, TRUE, param->key_idx, FALSE))
				{
					IF_LOUD(DbgPrint("Encryption configured with SUCCESS...\n");)
				}
				
			}
		}
		else //Pairwise Key
		{
			if(!((param->key_len == 5 || param->key_len == 13) && !IsTxKey))
				Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex = param->key_idx;

			Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyLength = param->key_len;
			Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = EncryptionSuite;

                        printk(KERN_INFO "#### PAIRWISE index = %d###\n", param->key_idx);
	//		if(IsTxKey)
			{
				Adapter->SecurityInfo.PairwiseKeyAvailable = TRUE;
			}
			if(SetEncryptionStatus(Adapter, TRUE, param->key_idx, TRUE))
			{
				IF_LOUD(DbgPrint("Encryption configured with SUCCESS...\n");)
			}
		}

		if(EncryptionSuite == CIPHER_SUITE_WEP_64 || EncryptionSuite == CIPHER_SUITE_WEP_128)
		{
			Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption1Enabled;
		}
		else
		if(EncryptionSuite == CIPHER_SUITE_TKIP)
		{
			if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
			{
				if(IsTxKey)
				{
					if(Adapter->SecurityInfo.PairwiseKeyAvailable)
						Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption2Enabled;
					// Maybe this case is wrong, since if we get a group key, without already
					// having a pairwise, with the Tx flag on, we will disable the encryption.
					else
						Adapter->SecurityInfo.EncryptionStatus = _802_11WEPDisabled;
				}
			}
			else
			{
				if(IsTxKey)
					Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption2Enabled;
			}
		}
#ifdef AES
		else
		if(EncryptionSuite == CIPHER_SUITE_AES)
		{
			if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
			{
				if(IsTxKey)
				{
					if(Adapter->SecurityInfo.PairwiseKeyAvailable)
						Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption3Enabled;
					// Maybe this case is wrong, since if we get a group key, without already
					// having a pairwise, with the Tx flag on, we will disable the encryption.
					else
						Adapter->SecurityInfo.EncryptionStatus = _802_11WEPDisabled;
				}
			}
			else
			{
				if(IsTxKey)
					Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption3Enabled;
			}
		}
#endif
		break;
	case SET_CIPHER_SUITES:
                printk(KERN_INFO "SET_CIPHER_SUITES\n");
		break;
	case MLME_STA_DEAUTH:
            printk(KERN_INFO "MLME_STA_DEAUTH\n");
		break;
	case MLME_STA_DISASSOC:
            printk(KERN_INFO "MLME_STA_DISASSOC\n");
            RemoveEncryptionSettings(Adapter);
		break;
	default:
            printk(KERN_INFO "DEFAULT IOCTL COMMAND for Supplicant\n");
		ap_ioctl = 1;
		break;
	}

	if (ret == 1 || !ap_ioctl) {
		if (copy_to_user(p->pointer, param, p->length)) {
			ret = -EFAULT;
			goto out;
		} else if (ap_ioctl)
			ret = 0;
	}

 out:
	if (param != NULL) {
		kfree(param);
  }

  EnableInterrupts(Adapter);
	return ret;
}

// IO CONTROLS
#ifdef WIRELESS_EXT
#define SuP 	1
#define WeP		2

#if (LINUX_VERSION_CODE > 0x020213)
static int
vnet_ioctl (struct net_device *dev, struct ifreq *rq, int cmd)
#else
static int
vnet_ioctl (struct device *dev, struct ifreq *rq, int cmd)
#endif
{
  struct net_local *local = (struct net_local *) dev->priv;
  struct iwreq *wrq = (struct iwreq *) rq;
  USHORT subcmd;
  int wpa_result;
  PVNet_ADAPTER Adapter = &local->Adapter;
  ULONG GenericULong;
  //UCHAR                 ApInfoList[MAX_BSS_ENTRIES*sizeof(BSS_INFO)];
  int rc = 0, update = 0, WEP_OR_SUPPORT = 0, APindex, i;
  DEVICE_CONFIGURATION DevConfig;

  // check the user. if not superuser return no perm.
  if (!capable (CAP_NET_ADMIN))
    {
      return -EPERM;
    }


  DisableInterrupts (Adapter);
  UpdateSupportAppInfo (Adapter, &DevConfig);

    switch (cmd)
    {
    case SIOCGIWPRIV:
      if (wrq->u.data.pointer)
	{
	wrq->u.data.length = sizeof(vnet_private_args) / sizeof(vnet_private_args[0]);
	  if (copy_to_user (wrq->u.data.pointer, vnet_private_args, sizeof (vnet_private_args)))
	    {
	      EnableInterrupts (Adapter);
	      return -EFAULT;
	    }
	}
      break;
    case SIOCGIWNAME:
#if defined(R504)
      strcpy (wrq->u.name, "ATMEL 504");
#elif defined(R504_2958)
      strcpy (wrq->u.name, "ATMEL 504+2958");
#elif defined(PCI)
      strcpy (wrq->u.name, "ATMEL PCI");
#elif defined(REV_D)
      strcpy (wrq->u.name, "ATMEL REVD");
#elif defined(REV_E)
      strcpy (wrq->u.name, "ATMEL REVE");
#elif defined(RFMD) || defined(RFMD_3COM)
      strcpy (wrq->u.name, "ATMEL RFMD");
#endif
      break;
    case SIOCDEVPRIVATE:
      subcmd = wrq->u.data.flags;
      switch (subcmd)
	{
	case GET_SUPPORT_INFO:
	  wrq->u.data.length = sizeof (DEVICE_CONFIGURATION);
	  copy_to_user (wrq->u.data.pointer, &DevConfig, wrq->u.data.length);
	  break;
	case SET_SUPPORT_INFO:
	  copy_from_user (&DevConfig, wrq->u.data.pointer,
			  wrq->u.data.length);
	  ChangeConfiguration (Adapter, &DevConfig);
	  break;
	case GET_ENCRYPTION_INFO:
	  wrq->u.data.length = sizeof (ENCRYPTION_INFO);
	  copy_to_user (wrq->u.data.pointer, &Adapter->SecurityInfo.EncryptionInfo,
			wrq->u.data.length);
	  break;
	case SET_ENCRYPTION_INFO:
	  copy_from_user (&Adapter->SecurityInfo.EncryptionInfo, wrq->u.data.pointer,
			  wrq->u.data.length);
	  ConfigureWEP (Adapter);
    	if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
             Join(Adapter, BSS_TYPE_INFRASTRUCTURE);
        else
             Join(Adapter, BSS_TYPE_AD_HOC);
	  break;
	case GET_VERSION_INFO:
	  Adapter->VersionInfo.DriverMajorVersion = FAST_VNET_MAJOR_VERSION;
	  Adapter->VersionInfo.DriverMinorVersion = FAST_VNET_MINOR_VERSION;
	  Adapter->VersionInfo.DriverSubVersion = FAST_VNET_SUB_VERSION;
	  Adapter->VersionInfo.DriverBuild = FAST_VNET_BUILD;
	  Adapter->VersionInfo.FwMajorVersion = Adapter->HostInfo.MajorVer;
	  Adapter->VersionInfo.FwMinorVersion = Adapter->HostInfo.MinorVer;
	  Adapter->VersionInfo.FwSubVersion = 0;
	  Adapter->VersionInfo.FwBuild = Adapter->HostInfo.BuildVer;
	  wrq->u.data.length = sizeof (VERSION_INFO);
	  copy_to_user (wrq->u.data.pointer, &Adapter->VersionInfo,
			wrq->u.data.length);
	  break;
	case RESET_CARD:
	  ResetAdapter (Adapter, TRUE);
	  break;
	case CLEAR_PACKETS:
	  VNetZeroMemory ((PUCHAR) & Adapter->Stats, sizeof (STATISTICS));
	  update = 0;
	  break;
	case GET_AP_INFO:
	  if (!Adapter->RadioIsOn){
    Adapter->BSSListEntries = 0;
    }
#ifdef INT_ROAM
    if((Adapter->IntRoamingEnabled) && (!(Adapter->MultiDomainMib.DomainInfoValid))){
    Adapter->BSSListEntries = 0;
    }
#endif
	  wrq->u.data.length = sizeof (BSS_INFO_EX) * Adapter->BSSListEntries;
	  copy_to_user (wrq->u.data.pointer, &Adapter->BSSinfo,
			wrq->u.data.length);
	  break;
	case SITE_SURVEY:
	  if (Adapter->RadioIsOn)
	    {
#ifdef INT_ROAM
        if((Adapter->IntRoamingEnabled) && (!Adapter->MultiDomainMib.DomainInfoValid)){
            Adapter->SiteSurveyState = SITE_SURVEY_COMPLETED;
			      EnableInterrupts (Adapter);
            return 0;

        }
#endif
	      if (Adapter->SiteSurveyState == SITE_SURVEY_IDLE || Adapter->SiteSurveyState == SITE_SURVEY_COMPLETED) {
            Adapter->SiteSurveyState = SITE_SURVEY_IN_PROGRESS;
            Scan (Adapter, FALSE);
        }
	      GenericULong = 0;
	      wrq->u.data.length = 0;
      }
	  break;
	case STOP_SITE_SURVEY:
	  if (!Adapter->RadioIsOn){
		  EnableInterrupts (Adapter);
                  return 0;
          }
	  if (Adapter->RadioIsOn)
	    {
	      if (Adapter->OperatingMode != AD_HOC_MODE)
		{
		  if (Adapter->IntRoamingEnabled)
		    {
		      for (i = 0; i < 14; i++)
			{
			  if (Adapter->ChannelVector[i] != 0)
			    break;
			}
		      if (i < 14)
			{
			  if (Adapter->MgmtErrorCode == 0xa0)
			    {
			      Adapter->StationState =
				STATION_STATE_MGMT_ERROR;
			      EnableInterrupts (Adapter);
			      return 0;
			    }
			  else if (Adapter->MgmtErrorCode == 0xa1)
			    {
			      Adapter->StationState =
				STATION_STATE_MGMT_ERROR;
			      EnableInterrupts (Adapter);
			      return 0;
			    }
			}
		      else
			{
			  Adapter->StationState = STATION_STATE_WAIT_ACTION;
			  Scan (Adapter, TRUE);
			}
		    }
		  else
		    {
		      if (Adapter->MgmtErrorCode == 0xa0)
			{
			  Adapter->StationState = STATION_STATE_MGMT_ERROR;
			  EnableInterrupts (Adapter);
			  return 0;
			}
		      else if (Adapter->MgmtErrorCode == 0xa1)
			{
			  Adapter->StationState = STATION_STATE_MGMT_ERROR;
			  EnableInterrupts (Adapter);
			  return 0;
			}
		    }
		}
	      else
		{
		  Adapter->StationState = STATION_STATE_READY;
		}
	    }
	  else
	    {
	      Adapter->SiteSurveyState = SITE_SURVEY_IDLE;
	      SelectBSS (Adapter, 0xff);	//ff indicates no AP selected
	    }
	  break;
	case SELECT_AP_BY_INDEX:
	  if (Adapter->RadioIsOn)
	    {
	      copy_from_user (&APindex, wrq->u.data.pointer,
			      wrq->u.data.length);
	      SelectBSS (Adapter, APindex - 1);
	      Adapter->SiteSurveyState = SITE_SURVEY_IDLE;
	    }
	  break;

	case SITE_SURVEY_STATE:
	  wrq->u.data.length = sizeof (UCHAR);
	  copy_to_user (wrq->u.data.pointer, &Adapter->SiteSurveyState,
			wrq->u.data.length);
	  break;

	case GET_STATISTICS:
	  wrq->u.data.length = sizeof (STATISTICS);
	  copy_to_user (wrq->u.data.pointer, &Adapter->Stats,
			wrq->u.data.length);
	  break;

	case GET_MAC_ADDRESS:
	  wrq->u.data.length = VNet_LENGTH_OF_ADDRESS;
	  copy_to_user (wrq->u.data.pointer, &Adapter->AssociationInfo.CurrentAddress,
			wrq->u.data.length);
	  update = 0;
	  break;
	}
      break;
        case ATMEL_WPA_IOCTL : 
                printk(KERN_INFO "Atmel WPA Ioctl is Called\n");
                wpa_result = atmel_wpa_ioctl(Adapter, &wrq->u.data); 
          break;
    }
  // After Getting Request Check if Updates are Needed and if so
  // Run Configure_WEP(1) or ChangeConfiguration(2)
  if (update == 1)
    {
      switch (WEP_OR_SUPPORT)
	{
	case WeP:
	ConfigureWEP (Adapter);
	if(Adapter->StationIsAssociated) break;
    	else if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
       		Join(Adapter, BSS_TYPE_INFRASTRUCTURE);
	else
       		Join(Adapter, BSS_TYPE_AD_HOC);
	break;
	case SuP:
	  ChangeConfiguration (Adapter, &DevConfig);
	  //Reset ?
	  break;
	default:
	  //Case of error
	  EnableInterrupts (Adapter);
	  return -1;
	  break;
	}
      update = 0;
    }
  EnableInterrupts (Adapter);
  return rc;
}

// Iwconfig statistics...
#if (LINUX_VERSION_CODE >= 0x020213)
struct iw_statistics *
vnet_get_wireless_stats (struct net_device *dev)
#else
struct iw_statistics *
vnet_get_wireless_stats (struct device *dev)
#endif
{
  struct net_local *local = (struct net_local *) dev->priv;
  PVNet_ADAPTER Adapter = &local->Adapter;
  local->wstats.status = 0;
  IF_VERY_LOUD (DbgPrint ("-> vnet_get_wireless_stats(%s)\n", dev->name);)
    if (Adapter->OperatingMode == INFRASTRUCTURE_MODE)
    {
      local->wstats.qual.qual = Adapter->LinkQuality;
      local->wstats.qual.level = Adapter->Rssi;
      local->wstats.qual.noise = 0;	//max(min( , 0x8a), 0x2f) - 0x95;
      local->wstats.qual.updated = 7;
    }
  else
    {
      // Quality levels cannot be determined in ad-hoc mode,
      // because we can 'hear' more that one remote station.
      local->wstats.qual.qual = 0;
      local->wstats.qual.level = 0;
      local->wstats.qual.noise = 0;
      local->wstats.qual.updated = 0;
    }
  // Packets discarded in the wireless adapter due to wireless specific problems
  local->wstats.discard.nwid = 0;
  local->wstats.discard.code = 0;
  local->wstats.discard.misc = 0;
  IF_VERY_LOUD (DbgPrint ("<- vnet_get_wireless_stats()\n");) return (&local->
								      wstats);
}
#endif



// Kernel Specific Functions

VOID
SetMgmtTimer (PVNet_ADAPTER Adapter)
{
#if (LINUX_VERSION_CODE < VERSION(2,3,0))
  if (Adapter->MgmtTimer->prev == (struct timer_list *) NULL)
#elif (LINUX_VERSION_CODE >= VERSION(2,5,0))
  if (Adapter->MgmtTimer->entry.prev != (struct list_head *) NULL)
#else
  if (Adapter->MgmtTimer->list.prev == (struct list_head *) NULL)
#endif
    {
      /* set timer to expire in WATCHDOG_JIFFIES */
      Adapter->MgmtTimer->expires = jiffies + MGMT_JIFFIES;
      add_timer (Adapter->MgmtTimer);
    }
}

VOID
VnetSleep (ULONG stime)
{
  mdelay (stime / 1000);
}

VOID
VNet_netif_present (PVNet_ADAPTER Adapter)
{
  netif_running (Adapter->dev);
}

VOID
VNet_netif_stop_queue (PVNet_ADAPTER Adapter)
{
  netif_stop_queue (Adapter->dev);
}

VOID
VNet_netif_wake_queue (PVNet_ADAPTER Adapter)
{
  netif_wake_queue (Adapter->dev);
}

VOID
RxIndicatePacket (PVNet_ADAPTER Adapter, PUCHAR pRxBuf, ULONG PacketLength,
		  ULONG IndicateLen)
{
#if (LINUX_VERSION_CODE > 0x020213)
  struct net_device *dev = (struct net_device *) Adapter->dev;
#else
  struct device *dev = (struct device *) Adapter->dev;
#endif
  struct sk_buff *skb;

  skb = dev_alloc_skb (PacketLength + 2);
  if (!skb)
    {
      Adapter->Stats.RxLost++;
      return;
    }
  skb_reserve (skb, 2);
  VNetMoveMemory (skb_put (skb, PacketLength), pRxBuf, PacketLength);
  Adapter->dev->last_rx = jiffies;
  skb->dev = dev;
  skb->protocol = eth_type_trans (skb, dev);
  skb->ip_summed = CHECKSUM_NONE;
  netif_rx (skb);
  return;
}

#if defined(REV_D)
#include "fwr_d.h"
#elif defined(REV_E)
#include "fwr_e.h"
#elif defined(R504)
#include "fwr504c.h"
#elif defined(R504_2958)
#include "fwr504_2958.h"
#elif defined(R504A_2958)
#include "fwr504a_2958.h"
#elif defined(PCI)
#include "fwr506.h"
#elif defined(RFMD)
#include "fwr.h"
#elif defined(RFMD_3COM)
#include "fwr_3COM.h"
#endif //REV_D

BOOLEAN
GetFwFromFile (PVNet_ADAPTER Adapter)
{
  return FALSE;
}

BOOLEAN
GetDefaultFw (PVNet_ADAPTER Adapter)
{
  ULONG SoFw = sizeof (BasicFW);

  IF_LOUD (DbgPrint ("--=+>GetFwFile BasicFW : %d, \n", SoFw);) Adapter->
    MappedFirmwareBuffer = (PUCHAR) kmalloc (SoFw, GFP_KERNEL);
  IF_LOUD (DbgPrint
	   ("Fw Mapped in : 0x%08X\n",
	    Adapter->MappedFirmwareBuffer);) VNetZeroMemory (Adapter->
							     MappedFirmwareBuffer,
							     SoFw);
  VNetMoveMemory (Adapter->MappedFirmwareBuffer, BasicFW, SoFw);
  Adapter->FirmwareFileLength = SoFw;
  IF_LOUD (DbgPrint ("<=- Out of GetFwFile\n");) return TRUE;
}

#ifdef PCI
#if (LINUX_VERSION_CODE > 0x020213)
int
vnet_pci_probe (struct pci_dev *dev, const struct pci_device_id *id)
{
  long ioaddr0 = 0, ioaddr1;
  //unsigned short i, tmp;
  int res, ret;
  struct net_local *local;
  //unsigned short basemem = 0;

  local = kmalloc (sizeof (struct net_local), GFP_KERNEL);
  if (!local)
    return -ENOMEM;
  memset (local, 0, sizeof (struct net_local));

#if (LINUX_VERSION_CODE > 0x020414)
  local->Adapter.dev = alloc_etherdev(sizeof(struct net_local));
  if(local->Adapter.dev == NULL)
    return -ENOMEM;
  local->dev= local->Adapter.dev;
#endif

  res = pci_enable_device (dev);

  if (res < 0)
    {
      DbgPrint (" Failed to enable device %d\n", res);
      return res;
    }
//      ioaddr0 = pci_resource_start(dev, 0);
  ioaddr1 = pci_resource_start (dev, 1);
  if (!ioaddr1 || (!(pci_resource_flags (dev, 1) & IORESOURCE_IO)))
    return -ENODEV;

  if (request_region (ioaddr1, VNET_IO_LENGTH, "pcifvnet") == NULL)
    return -EBUSY;

  ret = request_irq (dev->irq, vnet_interrupt, SA_SHIRQ, local->Adapter.dev->name, dev);
  if (ret != 0)
    return ret;

  DbgPrint (__FILE__ " Io0 %08x, Io1 %04x, Irq %d\n", ioaddr0, ioaddr1, dev->irq);

  local->Adapter.pci = dev;
  local->Adapter.IoBase = ioaddr1;
#if (LINUX_VERSION_CODE < 0x020500)
  local->Adapter.dev = init_etherdev (NULL, 0);
  if (local->Adapter.dev == NULL)
    return -1;
#endif
  local->Adapter.pci = dev;

  IF_LOUD (DbgPrint (__FILE__ " Net Device : 0x%08x\n ", local->Adapter.dev);)
    local->Adapter.IsUp = TRUE;
  spin_lock_init (&local->slock);
  spin_lock_init (&local->Adapter.qlock);
  InitAdapter (&local->Adapter);

#if (LINUX_VERSION_CODE < 0x020500)
  local->Adapter.pci->driver_data = local->Adapter.dev;
#else
  pci_set_drvdata(local->Adapter.pci, local->Adapter.dev);
#endif

  local->Adapter.dev->open = vnet_open;
  local->Adapter.dev->stop = vnet_close;
  local->Adapter.dev->hard_start_xmit = vnet_tx;
  local->Adapter.dev->priv = local;
  local->Adapter.dev->addr_len = 6;
  local->Adapter.dev->mtu = mtu;
  local->Adapter.dev->irq = dev->irq;
  local->Adapter.dev->base_addr = ioaddr1;
  strcpy(local->Adapter.dev->name, "atml%d");
  memset (local->Adapter.dev->dev_addr, 0xff, 6);
  IF_LOUD (DbgPrint ("To Set Parameters -->\n");)
  SetParameters (&local->Adapter);
  IF_LOUD (DbgPrint ("Init And Start Card\n");)
    if (!InitAndStartCard (&local->Adapter))
    return -1;
  
  memcpy (local->Adapter.dev->dev_addr, local->Adapter.AssociationInfo.CurrentAddress, 6);
  DbgPrint (" Mac : %02x:%02x:%02x:%02x:%02x:%02x\n",
	    local->Adapter.dev->dev_addr[0], local->Adapter.dev->dev_addr[1],
	    local->Adapter.dev->dev_addr[2], local->Adapter.dev->dev_addr[3],
	    local->Adapter.dev->dev_addr[4], local->Adapter.dev->dev_addr[5]);

  // Initialize specific data
  local->Adapter.MgmtTimer = kmalloc (sizeof (struct timer_list), GFP_KERNEL);
  init_timer (local->Adapter.MgmtTimer);
  local->Adapter.MgmtTimer->function = MgmtTimer;
  local->Adapter.MgmtTimer->data = (unsigned long) local->Adapter.dev;

  local->Adapter.dev->get_stats = vnet_get_stats;

#ifdef WIRELESS_EXT
  local->Adapter.dev->do_ioctl = vnet_ioctl;
  local->Adapter.dev->get_wireless_stats = vnet_get_wireless_stats;
  local->Adapter.dev->wireless_handlers = (struct iw_handler_def *)&vnet_handler_def;
#endif
#if (LINUX_VERSION_CODE > 0x020414)
  if(register_netdev(local->Adapter.dev))
    return -1;
#else
  if(!(register_netdev(local->Adapter.dev)))
    return -1;
#endif
  EnableInterrupts (&local->Adapter);
  printk ("ATMEL Wireless PCI card version %d.%d.%d.%d\n",
	  FAST_VNET_MAJOR_VERSION, FAST_VNET_MINOR_VERSION,
	  FAST_VNET_SUB_VERSION, FAST_VNET_BUILD);
  return 0;
}

void
vnet_pci_remove (struct pci_dev *dev)
{
#if (LINUX_VERSION_CODE < 0x020500)
  struct net_device *net = (struct net_device *) dev->driver_data;
#else
  struct net_device *net = (struct net_device *) pci_get_drvdata(dev);
#endif

  PVNet_ADAPTER Adapter =
    (PVNet_ADAPTER) (&((struct net_local *) net->priv)->Adapter);
  DbgPrint ("Device Disconnecting\n");
  DisableInterrupts (Adapter);
  del_timer (Adapter->MgmtTimer);
  netif_stop_queue (net);
  netif_device_detach (net);
  Adapter->IsUp = FALSE;
  free_irq (dev->irq, dev);
  unregister_netdev (net);
#if (LINUX_VERSION_CODE >= 0x020500)
  pci_release_regions(dev);
  free_netdev(net);
  pci_set_drvdata(dev, NULL);
#else
  net->priv = NULL;
  net->name[0] = '\0';
  kfree (net);
#endif

  release_region (Adapter->IoBase, VNET_IO_LENGTH);
}

struct pci_device_id vnet_pci_id_table[] = {
    { 0x1114, 0x0506, PCI_ANY_ID, PCI_ANY_ID},
    { 0,}
};

struct pci_driver vnet_pci_driver = {
name:"pcifvnet",
id_table:vnet_pci_id_table,
probe:&vnet_pci_probe,
remove:&vnet_pci_remove
};

int __init
vnet_init (void)
{
  DbgPrint (KERN_INFO "registering\n");
  return pci_module_init (&vnet_pci_driver);
}

void __exit
vnet_exit (void)
{
  pci_unregister_driver (&vnet_pci_driver);
}

MODULE_AUTHOR ("DriversGroup");

module_init (vnet_init);
module_exit (vnet_exit);

#endif
#endif

/********************************************************************
 *PCI KERNEL : v2.2 Patch
 ********************************************************************/
#ifdef PCI
#if LINUX_VERSION_CODE <= 0x020213

MODULE_AUTHOR ("DriversGroup");

/* Used to pass the full-duplex flag, etc. */
#define MAX_UNITS 8		/* More are supported, limit only on options */
static int options[MAX_UNITS] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static int full_duplex[MAX_UNITS] = { -1, -1, -1, -1, -1, -1, -1, -1 };

enum pci_flags_bit
{
  PCI_USES_IO = 1, PCI_USES_MEM = 2, PCI_USES_MASTER = 4,
  PCI_ADDR0 = 0x10 << 0, PCI_ADDR1 = 0x10 << 1, PCI_ADDR2 =
    0x10 << 2, PCI_ADDR3 = 0x10 << 3,
};

struct pci_id_info
{
  const char *name;
  u16 vendor_id, device_id, device_id_mask, flags;
  int io_size;
  struct device *(*probe1) (int pci_bus, int pci_devfn, struct device * dev,
			    long ioaddr, int irq, int bug, int fnd_cnt);
};

static struct device *vnet_pci_probe1 (int pci_bus, int pci_devfn,
				       struct device *dev, long ioaddr,
				       int irq, int bug, int fnd_cnt);

static struct pci_id_info pci_tbl[] = { {"pcifvnet",
					 0x1114, 0x0506, 0xffff, PCI_USES_IO,
					 0x20, vnet_pci_probe1},
{0,},				/* 0 terminated list. */
};


/* A list of all installed pci devices, for removing the driver module. */
static struct device *vnet_pci_dev = NULL;


int
vnet_pci_probe (struct device *dev)
{
  int cards_found = 0;
  int pci_index = 0;
  unsigned char pci_bus, pci_device_fn;

  if (!pcibios_present ())
    return -ENODEV;

  for (; pci_index < 0xff; pci_index++)
    {
      u16 vendor, device, pci_command, new_command;
      int chip_idx, irq;
      long ioaddr;

      if (pcibios_find_class (PCI_CLASS_NETWORK_ETHERNET << 8, pci_index,
			      &pci_bus, &pci_device_fn) != PCIBIOS_SUCCESSFUL)
	break;
      pcibios_read_config_word (pci_bus, pci_device_fn,
				PCI_VENDOR_ID, &vendor);
      pcibios_read_config_word (pci_bus, pci_device_fn,
				PCI_DEVICE_ID, &device);

      if (!((vendor == pci_tbl[0].vendor_id) &&
	    ((device & pci_tbl[0].device_id_mask) == pci_tbl[0].device_id)))
	continue;
      {
	u32 pci_ioaddr;
	u32 pci_ioaddr2;
	u8 pci_irq_line;

	pcibios_read_config_byte (pci_bus, pci_device_fn, PCI_INTERRUPT_LINE,
				  &pci_irq_line);

	pcibios_read_config_dword (pci_bus, pci_device_fn, PCI_BASE_ADDRESS_1,
				   &pci_ioaddr);

	ioaddr = pci_ioaddr & ~3;

	irq = pci_irq_line;
      }

      if ((pci_tbl[0].flags & PCI_USES_IO) &&
	  check_region (ioaddr, pci_tbl[0].io_size))
	continue;

      pcibios_read_config_word (pci_bus, pci_device_fn, PCI_COMMAND,
				&pci_command);
      new_command = pci_command | (pci_tbl[0].flags & 7);
      if (pci_command != new_command)
	{
	  printk (KERN_INFO "  The PCI BIOS has not enabled the"
		  " device at %d/%d!  Updating PCI command %4.4x->%4.4x.\n",
		  pci_bus, pci_device_fn, pci_command, new_command);
	  pcibios_write_config_word (pci_bus, pci_device_fn, PCI_COMMAND,
				     new_command);
	}

      dev =
	pci_tbl[0].probe1 (pci_bus, pci_device_fn, dev, ioaddr, irq, 0,
			   cards_found);

      if (dev && (pci_tbl[0].flags & PCI_COMMAND_MASTER))
	{
	  u8 pci_latency;
	  pcibios_read_config_byte (pci_bus, pci_device_fn, PCI_LATENCY_TIMER,
				    &pci_latency);
	  if (pci_latency < 32)
	    {
	      pcibios_write_config_byte (pci_bus, pci_device_fn,
					 PCI_LATENCY_TIMER, 32);
	    }
	}
      dev = 0;
      cards_found++;
    }

  return cards_found ? 0 : -ENODEV;
}

static struct device *
vnet_pci_probe1 (int pci_bus, int pci_devfn,
		 struct device *dev, long ioaddr,
		 int irq, int bug, int found_cnt)
{
  struct net_local *local;
  int ret, res;
  int i, option = found_cnt < MAX_UNITS ? options[found_cnt] : 0;

  local = kmalloc (sizeof (struct net_local), GFP_KERNEL);
  if (local == NULL)
    return 0;
  memset (local, 0, sizeof (struct net_local));
  dev = init_etherdev (dev, 0);

  dev->irq = irq;
  dev->base_addr = ioaddr;
  dev->priv = local;
  vnet_pci_dev = dev;
  local->dev = dev;
  local->pci_bus = pci_bus;
  local->pci_devfn = pci_devfn;

  request_region (ioaddr, pci_tbl[0].io_size, "pcifvnet");
  ret = request_irq (irq, vnet_interrupt, SA_SHIRQ, "pcifvnet", local->dev);

  if (ret != 0)
    return 0;

  DbgPrint (__FILE__ " " __FUNCTION__ " Io1 %04x, Irq %d\n", ioaddr,
	    dev->irq);

  local->Adapter.IoBase = ioaddr;
  local->Adapter.dev = dev;
  IF_LOUD (DbgPrint
	   (__FILE__ " " __FUNCTION__ "Net Device : 0x%08x\n ",
	    local->Adapter.dev);) local->Adapter.IsUp = TRUE;
  spin_lock_init (&local->slock);
  InitAdapter (&local->Adapter);

  local->Adapter.dev->open = vnet_open;
  local->Adapter.dev->stop = vnet_close;
  local->Adapter.dev->hard_start_xmit = vnet_tx;
  local->Adapter.dev->priv = local;
  local->Adapter.dev->addr_len = 6;
  local->Adapter.dev->mtu = mtu;
  local->Adapter.dev->irq = dev->irq;
  local->Adapter.dev->base_addr = ioaddr;
  memset (local->Adapter.dev->dev_addr, 0xff, 6);
  IF_LOUD (DbgPrint ("To Set Parameters -->\n");) SetParameters (&local->
								 Adapter);
  if (!InitAndStartCard (&local->Adapter))
    return 0;

  memcpy (local->Adapter.dev->dev_addr, local->Adapter.CurrentAddress, 6);

  register_netdev (local->Adapter.dev);

  DbgPrint (__FUNCTION__ " Mac : %02x:%02x:%02x:%02x:%02x:%02x\n",
	    local->Adapter.dev->dev_addr[0], local->Adapter.dev->dev_addr[1],
	    local->Adapter.dev->dev_addr[2], local->Adapter.dev->dev_addr[3],
	    local->Adapter.dev->dev_addr[4], local->Adapter.dev->dev_addr[5]);

  // Initialize specific data
  local->Adapter.MgmtTimer = kmalloc (sizeof (struct timer_list), GFP_KERNEL);
  init_timer (local->Adapter.MgmtTimer);
  local->Adapter.MgmtTimer->function = MgmtTimer;
  local->Adapter.MgmtTimer->data = (unsigned long) local->Adapter.dev;

  local->Adapter.dev->get_stats = vnet_get_stats;


#ifdef WIRELESS_EXT
  local->Adapter.dev->do_ioctl = vnet_ioctl;
  local->Adapter.dev->get_wireless_stats = vnet_get_wireless_stats;
#endif
  EnableInterrupts (&local->Adapter);
  printk ("ATMEL Wireless PCI card version %d.%d.%d.%d\n",
	  FAST_VNET_MAJOR_VERSION, FAST_VNET_MINOR_VERSION,
	  FAST_VNET_SUB_VERSION, FAST_VNET_BUILD);

  return dev;
}

int
init_module (void)
{
  return vnet_pci_probe (0);
}

void
cleanup_module (void)
{

  struct net_local *local = (struct net_local *) vnet_pci_dev->priv;

  PVNet_ADAPTER Adapter =
    (PVNet_ADAPTER) (&((struct net_local *) vnet_pci_dev->priv)->Adapter);

  DbgPrint ("Device Disconnecting\n");
  DisableInterrupts (Adapter);
  del_timer (Adapter->MgmtTimer);
  netif_stop_queue (vnet_pci_dev);
  netif_device_detach (vnet_pci_dev);
  Adapter->IsUp = FALSE;
  free_irq (vnet_pci_dev->irq, vnet_pci_dev);
  unregister_netdev (vnet_pci_dev);
  vnet_pci_dev->priv = NULL;
  vnet_pci_dev->name[0] = '\0';
  kfree (vnet_pci_dev);

  release_region (Adapter->IoBase, VNET_IO_LENGTH);
}

#endif
#endif
