/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/


/**************************************************************************/
/**************************************************************************/
/**                                                                       */
/**                                                                       */
/**   Functions Handling Mgmt Frames									  */
/**                                                                       */
/**************************************************************************/
/**************************************************************************/


#include "vnet.h"
#include "tx.h"

extern UCHAR BasicRates[2][4];

BOOLEAN 
CheckWPACapabilities(PVNet_ADAPTER Adapter, PBSS_INFO_EX BssInfo)
{
	BOOLEAN AuthSuiteMatch = FALSE;
	USHORT i;

	if(Adapter->OperatingMode == AD_HOC_MODE)
		return TRUE;

	//WPA capabilities

	if(((BssInfo->Capabilities & BSS_CAPAPILITY_WPA) ^ Adapter->SecurityInfo.InWpaMode)) //mismatch...
	{
		IF_LOUD(DbgPrint("(1) INFO: Mismatching capabilities: %d - %d\n", (BssInfo->Capabilities & BSS_CAPAPILITY_WPA), Adapter->SecurityInfo.InWpaMode);)
		return FALSE;
	}

	if(BssInfo->Capabilities & BSS_CAPAPILITY_WPA)
	{
		// if the mixed mode is disabled we need to have the group cipher suite 
		// set to the driver's WPA encryption type.
		if(Adapter->SecurityInfo.WPA_MixedModeDisabled)
		{
#ifdef AES
			if( !( ((BssInfo->WpaInfo.GroupSuite == 2) && 
				    ((Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption2KeyAbsent) || (Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption2Enabled)) ) ||
				   ((BssInfo->WpaInfo.GroupSuite == 4) &&
				    ((Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption3KeyAbsent) || (Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption3Enabled)) )
				 )
			  )
#else
			if(BssInfo->WpaInfo.GroupSuite != 2)
#endif
			{
				IF_LOUD(DbgPrint("INFO: WPA_MixedModeDisabled: GroupSuite = %d\n", BssInfo->WpaInfo.GroupSuite);)
				return FALSE;
			}
			
		}

		// at least one pairwise cipher suite must be WPA.
		for(i=0; i<BssInfo->WpaInfo.PairwiseSuiteCnt; i++)
		{
#ifdef AES
			if( ((BssInfo->WpaInfo.PairwiseSuiteList[i] == 2) && 
				 ((Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption2KeyAbsent) || (Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption2Enabled) ) ) ||
				((BssInfo->WpaInfo.PairwiseSuiteList[i] == 4) &&
				 ( (Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption3KeyAbsent) || (Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption3Enabled) ) )
			  )
#else
			if(BssInfo->WpaInfo.PairwiseSuiteList[i] == 2)
#endif
				break;

			// if the next check is true this means that we have
			// exhausted the Pairwise Suites without having found 
			// a WPA cipher.
			if(i == BssInfo->WpaInfo.PairwiseSuiteCnt - 1)
			{
				IF_LOUD(DbgPrint("INFO: WPA Mismatching Capabilities: PairwiseSuite != WPA\n");)
				return FALSE;
			}
		}
		IF_LOUD(DbgPrint("PW CS (Retrieved): %d\n", Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite);)
	
		// WPA Authentication Suite

		if(Adapter->SecurityInfo.AuthenticationMode == _802_11AuthModeWPA)
		{
			AuthSuiteMatch = FALSE;
			for(i=0;i<BssInfo->WpaInfo.AuthMgmtSuiteCnt;i++)
			{
				if(BssInfo->WpaInfo.AuthMgmtSuiteList[i] == 1)
				{
					AuthSuiteMatch = TRUE;
					break;
				}
			}
		}
		else
		if(Adapter->SecurityInfo.AuthenticationMode== _802_11AuthModeWPAPSK)
		{
			AuthSuiteMatch = FALSE;
			for(i=0;i<BssInfo->WpaInfo.AuthMgmtSuiteCnt;i++)
			{
				if(BssInfo->WpaInfo.AuthMgmtSuiteList[i] == 2)
				{
					AuthSuiteMatch = TRUE;
					break;
				}
			}
		}
		else
		{
			IF_LOUD(DbgPrint("INFO: EncryptionInfo.AuthenticationType: %d\n", Adapter->SecurityInfo.AuthenticationMode);)
		}
		return AuthSuiteMatch;
	}
	else //in non-WPA mode
		return TRUE;

	
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    SendAuthenticationRequest                                                                               
//
//    This function formats an Authentication Request Frame and calls     
//    TransmitMgmtFrame() to transmit it.                                              
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

BOOLEAN 
SendAuthenticationRequest(PVNet_ADAPTER Adapter)
{
	USHORT							FrameLen;

	s80211_MgmtFrm					MgmtFrame;

	BuildAuthenticationRequestFrame((PUCHAR) &MgmtFrame, &Adapter->AuthenticationInfo, 
									&Adapter->AssociationInfo, &Adapter->SecurityInfo,
									&FrameLen);

	//
	//send frame & return status
	//
	return TransmitMgmtFrame(Adapter, (UCHAR*)&MgmtFrame, (USHORT)FrameLen);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    SendAssociationRequest										      
//
//    This function formats an Association Request Frame and calls		  
//    TransmitMgmtFrame() to transmit it.	                              
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
BOOLEAN
SendAssociationRequest(PVNet_ADAPTER Adapter)
{
	USHORT							FrameLen,offset __attribute__ ((unused)), BytesAdded __attribute__ ((unused));
	UCHAR							j __attribute__ ((unused)), PwCsSet __attribute__ ((unused));
	s80211_MgmtFrm					MgmtFrame;

	PwCsSet=0;
	BuildAssociationRequestFrame((PUCHAR) &MgmtFrame, &Adapter->AssociationInfo,
								 &Adapter->SecurityInfo, &FrameLen);

	IF_LOUD(DbgPrint("Sending Association Request...\n");)

	//
	// send Association Request frame
	//

	Adapter->LastAssReasReqBytes = FrameLen-24;
	VNetMoveMemory(Adapter->LastAssReasReqFrameBody, &MgmtFrame.Body[0], Adapter->LastAssReasReqBytes);

	return TransmitMgmtFrame(Adapter, (PUCHAR)&MgmtFrame, (USHORT)FrameLen);
}

/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    SendReAssociationRequest									
//
//    This function formats a ReAssociation Request Frame and calls		  
//    TransmitMgmtFrame() to transmit it. Based on SendAssociationRequest 
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
BOOLEAN
SendReAssociationRequest(PVNet_ADAPTER Adapter)
{
	USHORT							FrameLen;
	s80211_MgmtFrm					MgmtFrame;

	BuildReAssociationRequestFrame((PUCHAR) &MgmtFrame, &Adapter->AssociationInfo,
									&Adapter->SecurityInfo, &FrameLen);

	IF_LOUD(DbgPrint("Sending ReAssociation Request...\n");)

	Adapter->LastAssReasReqBytes = FrameLen-24;
	VNetMoveMemory(Adapter->LastAssReasReqFrameBody, &MgmtFrame.Body[0], Adapter->LastAssReasReqBytes);

	return TransmitMgmtFrame(Adapter, (PUCHAR)&MgmtFrame, (USHORT)FrameLen);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    TransmitMgmtFrame                                                                                               
//
//    This function submits to MAC a formated MGMT frame for transmission 
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
BOOLEAN 
TransmitMgmtFrame(	PVNet_ADAPTER Adapter, 
					PUCHAR pFrameBuffer, 
					USHORT	FrameSize)
{

	USHORT		StartOfTxBuffer;

	Adapter->MgmtErrorCode = C80211_MGMT_SC_Success;

	if(!TxResourcesAvailable(Adapter, FrameSize, &StartOfTxBuffer))
		return FALSE;
		
	HostMemToCardMem16(Adapter, StartOfTxBuffer, pFrameBuffer, FrameSize);

	TxUpdateDescriptor(Adapter, FrameSize, StartOfTxBuffer, FALSE, FALSE);

    return TRUE;


}

BOOLEAN
CheckWPABlackList(PVNet_ADAPTER Adapter, PUCHAR BSSID)
{
    return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    RetrieveBSS														  
//
//	if a BSSID is specified during the call of this function it returns   
//  the index in the array of the coresponding AP. if no BSSID is		  
//  specified it returns the index of the AP in the list having the same  
//  SSID and the best RSSI. If SSID is NULL and we have not previously fail
//  to connect to this BSS, it returns the index of the AP with the best RSSI.											  
//  If no AP exists or no AP matches the requirements it returns -1.	  
//																		  
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

int
RetrieveBSS(PVNet_ADAPTER Adapter, PUCHAR BSSID)
{
	UCHAR		i;
	CHAR		MaxRssi;
	UCHAR		MaxRssiIndex;
	BOOLEAN		SpecificSSID = TRUE;
	BOOLEAN		APfound = FALSE;

	if(Adapter->BSSListEntries == 0)
	{
		IF_LOUD(DbgPrint(" NO APs in the List...\n");)
		return -1;
	}

	if(BSSID == NULL)
	{
		//
		// Check if a specific SSID is used
		//
		if(Adapter->ConnectToAnyBSS)
			SpecificSSID = FALSE;

		MaxRssi = -128;
		MaxRssiIndex = 0;
		Adapter->CurrentBSSindex = 0;
		
		for(i=0; i<Adapter->BSSListEntries; i++)
		{
			if(SpecificSSID)
			{
				if(Adapter->AssociationInfo.DesiredSSIDsize == Adapter->BSSinfo[i].SSIDsize)
				{
					if(!MemCompare(Adapter->AssociationInfo.DesiredSSID, Adapter->BSSinfo[i].SSID, Adapter->AssociationInfo.DesiredSSIDsize))
					{
						if(	(Adapter->OperatingMode == AD_HOC_MODE && Adapter->BSSinfo[i].BSStype == BSS_TYPE_AD_HOC /*&& Adapter->BSSinfo[i].Channel == Adapter->Channel*/) || 
							(( Adapter->OperatingMode == INFRASTRUCTURE_MODE && Adapter->BSSinfo[i].BSStype == BSS_TYPE_INFRASTRUCTURE && 
							  ValidateChannelForDomain(Adapter, &Adapter->BSSinfo[i].Channel))
							&&
							MatchSecuritySettings(&(Adapter->SecurityInfo), &(Adapter->BSSinfo[i]), Adapter->OperatingMode, &Adapter->Stats))
						  ) 
						{
							APfound = TRUE;

							if(Adapter->BSSinfo[i].RSSI >= MaxRssi) 
							{
								IF_LOUD(DbgPrint("Max Rssi was %d at index %d-->%d at index %d\n", MaxRssi,MaxRssiIndex, Adapter->BSSinfo[i].RSSI,i);)
								MaxRssi = Adapter->BSSinfo[i].RSSI;
								MaxRssiIndex = i;
							}
						}
					}
				}
			}
			else
			{
				//
				// Select a BSS with the max-RSSI but of the same type and of the same WEP mode
				// and that it is not marked as 'bad' (i.e. we had previously failed to connect to
				// this BSS with the settings that we currently use)
				//
				if(
					( (Adapter->OperatingMode == AD_HOC_MODE && Adapter->BSSinfo[i].BSStype == BSS_TYPE_AD_HOC /*&& Adapter->BSSinfo[i].Channel == Adapter->Channel*/) || 
					  (Adapter->OperatingMode == INFRASTRUCTURE_MODE && Adapter->BSSinfo[i].BSStype == BSS_TYPE_INFRASTRUCTURE && 
					   ValidateChannelForDomain(Adapter, &Adapter->BSSinfo[i].Channel))
					) &&
					( MatchSecuritySettings(&(Adapter->SecurityInfo), &(Adapter->BSSinfo[i]), Adapter->OperatingMode, &Adapter->Stats) &&
						!Adapter->BSSinfo[i].FailedToConnect &&	
						Adapter->BSSinfo[i].SSIDsize
					)
				   ) 
				{
					APfound = TRUE;
					if(Adapter->BSSinfo[i].RSSI >= MaxRssi) 
					{
						MaxRssi = Adapter->BSSinfo[i].RSSI;
						MaxRssiIndex = i;
						Adapter->CurrentBSSindex = i;
					}
				}
			}
		}

		if(APfound)
		{
			IF_LOUD(DbgPrint("Max Rssi is %d in index = %d\n", MaxRssi, MaxRssiIndex);)
			
			memcpy(&Adapter->AssociationInfo.CurrentBSSinfo, &Adapter->BSSinfo[MaxRssiIndex], sizeof(BSS_INFO_EX));

			return MaxRssiIndex;
		}
	}
	else  //Specific BSSID
	{
		for(i=0; i<Adapter->BSSListEntries; i++)
		{
			if( (!MemCompare(BSSID, Adapter->BSSinfo[i].BSSID, 6) && 
				 ValidateChannelForDomain(Adapter, &Adapter->BSSinfo[i].Channel) 
				) &&
				(MatchSecuritySettings(&Adapter->SecurityInfo, &Adapter->BSSinfo[i], Adapter->OperatingMode, &Adapter->Stats))
			  )
			{
//				if(Adapter->OperatingMode == AD_HOC_MODE && Adapter->BSSinfo[i].Channel != Adapter->Channel)
//					continue;

				memcpy(&Adapter->AssociationInfo.CurrentBSSinfo, &Adapter->BSSinfo[i], sizeof(BSS_INFO_EX));

				return i;
			}
		}
	}

	return -1;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//  SelectBSS
//                                                                                                                        
//      After a Site Survey the user selected AP index is passed to this          
//      function to Join it.                                                                                              
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VOID
SelectBSS (PVNet_ADAPTER Adapter, int BSSindex)
{
  if (BSSindex == 0xff)
    {
      IF_LOUD (DbgPrint ("Site Survey Stopped. NO AP SELECTED\n");
	)if (Adapter->OperatingMode == INFRASTRUCTURE_MODE)
	{
	  if (Adapter->StationIsAssociated)
	    {
	      IF_LOUD (int i;
		       DbgPrint ("SSID: ");
		       for (i = 0; i < Adapter->AssociationInfo.DesiredSSIDsize; i++)
		       DbgPrint ("%c", Adapter->AssociationInfo.DesiredSSID[i]);
		       DbgPrint ("\n");
		       DbgPrint ("BSSID: ");
		       for (i = 0; i < 6; i++)
		       DbgPrint ("%02X ", Adapter->AssociationInfo.CurrentBSSID[i]);
		       DbgPrint ("\n");)

		Adapter->ApSelected = TRUE;

	      Adapter->StationState = STATION_STATE_JOINNING;
	      Join (Adapter, BSS_TYPE_INFRASTRUCTURE);
	    }
	  else
	    Scan (Adapter, TRUE);
	}
      else			//Ad-Hoc
	{
	  Adapter->StationState = STATION_STATE_JOINNING;
	  Join (Adapter, BSS_TYPE_AD_HOC);
	}

    }
  else
    {
      IF_LOUD (DbgPrint ("Site Survey Stopped. BSS #%d selected\n", BSSindex);
	)if (BSSindex != -1)
	{
	  JoinSpecificBSS (Adapter, BSSindex, TRUE);
	}			//BSS index != -1


    }
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    SendDisAssociationRequest										      
//
//    This function formats a DisAssociation Request Frame and calls		  
//    TransmitMgmtFrame() to transmit it.	                              
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

BOOLEAN
SendDisAssociationRequest(PVNet_ADAPTER Adapter, USHORT ReasonCode)
{
	USHORT							FrameLen, offset __attribute__ ((unused));
	s80211_MgmtFrm					MgmtFrame;

	BuildDisAssociationRequestFrame((PUCHAR) &MgmtFrame, &Adapter->AssociationInfo,
									ReasonCode, &FrameLen);

	IF_LOUD(DbgPrint("Sending DisAssociation Request...\n");)

	//
	// send DisAssociation Request frame
	//

	return TransmitMgmtFrame(Adapter, (PUCHAR)&MgmtFrame, (USHORT)FrameLen);
}


BOOLEAN
SendDeAuthenticationRequest(PVNet_ADAPTER Adapter, USHORT ReasonCode)
{
	USHORT								FrameLen, offset __attribute__ ((unused));
	s80211_MgmtFrm						MgmtFrame;

	BuildDeAuthenticationRequestFrame((PUCHAR) &MgmtFrame, &Adapter->AssociationInfo,
										ReasonCode, &FrameLen);

	IF_LOUD(DbgPrint("Sending DeAuthentication Request...\n");)

	//
	// send DisAssociation Request frame
	//
	
	return TransmitMgmtFrame(Adapter, (PUCHAR)&MgmtFrame, (USHORT)FrameLen);
}



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    DisplayMgmtFrameStatus                                                                                      
//
//      if a BSSID is specified during the call of this function it returns   
//  the index in the array of the coresponding AP. if no BSSID is                 
//  specified it returns the index of the AP in the list having the same  
//  SSID and the best RSSI. If SSID is any it simply returns the index of 
//  the AP with the best RSSI.                                                                                    
//  If no AP exists or no AP matches the requirements it returns -1.      
//                                                                                                                                                
/////////////////////////////////////////////////////////////////////////////
VOID
DisplayMgmtFrameStatus(ULONG StatusCode)
{
	switch(StatusCode){
	case C80211_MGMT_SC_Unspecified:
		IF_LOUD(DbgPrint("Unspecified error in Management Frame...\n");)
		break;
	case C80211_MGMT_SC_AssDenied:
		IF_LOUD(DbgPrint("Association Denied...\n");)
		break;
	case C80211_MGMT_SC_ReassDenied:
		IF_LOUD(DbgPrint("Re-Association Denied...\n");)
		break;
	case C80211_MGMT_SC_AssDeniedHandleAP:
		IF_LOUD(DbgPrint("Association Denied(Access Point can not handle any more stations)...\n");)
		break;
	case C80211_MGMT_SC_AssDeniedBSSRate:
		IF_LOUD(DbgPrint("Association Denied (Supported Rates mismatch)...\n");)
		break;
	case C80211_MGMT_SC_SupportCapabilities:
		IF_LOUD(DbgPrint("SupportCapabilities error in Association Request...\n");)
		break;
	case C80211_MGMT_SC_AuthAlgNotSupported:
		IF_LOUD(DbgPrint("Authentication Algorithm not supported...\n");)
		break;
	case C80211_MGMT_SC_AuthTransSeqNumError:
		IF_LOUD(DbgPrint("Bad Authentication transaction sequence number...\n");)
		break;
	case C80211_MGMT_SC_AuthRejectChallenge:
		IF_LOUD(DbgPrint("Authentication challenge text rejection...\n");)
		break;
	case C80211_MGMT_SC_AuthRejectTimeout:
		IF_LOUD(DbgPrint("Authentication Timeout...\n");)
		break;
	default:
		{
		IF_LOUD(DbgPrint("Error 0x%X\n", StatusCode);)
		}
	}
}
/////////////////////////////////////////////////////////////////////////////


VOID DisplayReasonCode(USHORT ReasonCode)
{
	IF_LOUD(DbgPrint("Reason Code: ");)

	switch(ReasonCode){
	case C80211_MGMT_RC_Unspecified:
		IF_LOUD(DbgPrint("Unspecified Reason...\n");)
		break;
	case C80211_MGMT_RC_AuthPrevNoValid:
		IF_LOUD(DbgPrint("Previous Authentication no longer valid...\n");)
		break;
	case C80211_MGMT_RC_DeAuthLeaving:
		IF_LOUD(DbgPrint("Station is leaving BSS...\n");)
		break;
	case C80211_MGMT_RC_DisassInactivity:
		IF_LOUD(DbgPrint("Disassociated due to inactivity...\n");)
		break;
	case C80211_MGMT_RC_DisassAPHandle:
		IF_LOUD(DbgPrint("AP is unable to handle all currently associated stations...\n");)
		break;
	case C80211_MGMT_RC_AuthClass2Frm:
		IF_LOUD(DbgPrint("Class 2 frame received from non-authenticated station...\n");)
		break;
	case C80211_MGMT_RC_AssClass3Frm:
		IF_LOUD(DbgPrint("Class 3 frame received from non-associated station...\n");)
		break;
	case C80211_MGMT_RC_DisassLeaving:
		IF_LOUD(DbgPrint("Station is leaving BSS...\n");)
		break;
	case C80211_MGMT_RC_AssNotAuth:
		IF_LOUD(DbgPrint("Station requesting association is not authenticated...\n");)
		break;
	case C80211_MGMT_RC_InvalidInformationElement:
		IF_LOUD(DbgPrint("Invalid Information Element...\n");)
		break;
	case C80211_MGMT_RC_MicFailure:
		IF_LOUD(DbgPrint("Mic Failure...\n");)
		break;
	case C80211_MGMT_RC_4WayHandsakeTimeOut:
		IF_LOUD(DbgPrint("4-Way Handsake Time Out...\n");)
		break;
	case C80211_MGMT_RC_GroupKeyUpdateTimeOut:
		IF_LOUD(DbgPrint("Group Key Update Time Out...\n");)
		break;
	case C80211_MGMT_RC_MismatchInformationElement:
		IF_LOUD(DbgPrint("Mismatch Information Element...\n");)
		break;
	case C80211_MGMT_RC_InvalidMulticastCipher:
		IF_LOUD(DbgPrint("Invalid Multicast Cipher...\n");)
		break;
	case C80211_MGMT_RC_InvalidUnicastCipher:
		IF_LOUD(DbgPrint("Invalid Unicast Cipher...\n");)
		break;
	case C80211_MGMT_RC_InvalidAKMP:
		IF_LOUD(DbgPrint("Invalid AKMP...\n");)
		break;
	case C80211_MGMT_RC_UnsupportedWPAVersion:
		IF_LOUD(DbgPrint("UnsupportedWPAVersion...\n");)
		break;
	case C80211_MGMT_RC_InvalidWPACapabilities:
		IF_LOUD(DbgPrint("Invalid WPA Capabilities...\n");)
		break;
	case C80211_MGMT_RC_8021xAuthenticationFailed:
		IF_LOUD(DbgPrint("802.1x Authentication Failed...\n");)
		break;
	default:
		IF_LOUD(DbgPrint("UnKnown Code 0x%X\n", ReasonCode);)
		break;
	}
}
