/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/


/**************************************************************************/
/**************************************************************************/
/**                                                                       */
/**   Functions used for Rx												  */
/**                                                                       */
/**************************************************************************/
/**************************************************************************/

#include "vnet.h"
#include "rx.h"

#define __802_11_AUTH_REQUEST_GROUP_ERROR         0x0E
#define __802_11_AUTH_REQUEST_PAIRWISE_ERROR              0x06


#define MAX_WPA_ELEMENT_SIZE				200 //52+...
#define ELEMENT_ID_SSID		0
#define ELEMENT_ID_RATES	1
#define ELEMENT_ID_FH		2
#define ELEMENT_ID_DS		3
#define ELEMENT_ID_CF		4
#define ELEMENT_ID_TIM		5
#define ELEMENT_ID_IBSS		6
#define ELEMENT_ID_COUNTRY	7
#define ELEMENT_ID_AIRONET	0x85
#define ELEMENT_ID_WPA		0xDD

UCHAR      OUI_WPA[4] = {0x00, 0x50, 0xF2, 0x01};
//UCHAR	STATIC_ENTRIES_ETHERTYPE_8137[] = {0x81, 0x37};
//UCHAR	STATIC_ENTRIES_ETHERTYPE_80F3[] = {0x80, 0xF3};

USHORT Bytes2Ushort(PUCHAR pBuf)
{
	USHORT tmp;

	tmp = *(pBuf) << 8;
	tmp |= (*(pBuf+1));

	return tmp;
}
void GetRxDescKeyIndex(PVNet_ADAPTER Adapter, USHORT Descriptor, PUCHAR index)
{
	CardMemToHostMem8(Adapter, index, (Adapter->HostInfo.RxDescPos + (SIZE_OF_RXDESC * Descriptor) + RX_DESC_KEY_INDEX_OFFSET), 1);
}

void GetRxDescCipherSuite(PVNet_ADAPTER Adapter, USHORT Descriptor, PUCHAR CipherSuite)
{
	CardMemToHostMem8(Adapter, CipherSuite, (Adapter->HostInfo.RxDescPos + (SIZE_OF_RXDESC * Descriptor) + RX_DESC_CIPHER_TYPE_OFFSET), 1);
}

void GetRxDescKeyReplay(PVNet_ADAPTER Adapter, USHORT Descriptor, PUCHAR KeyReplay)
{
	CardMemToHostMem8(Adapter, KeyReplay, (Adapter->HostInfo.RxDescPos + (SIZE_OF_RXDESC * Descriptor) + RX_DESC_KEY_REPLAY), 1);
}



//////////////////////////////////////////////////
//      
//      Rx Descriptor Manipulation Functions
//
//////////////////////////////////////////////////

void
GetRxDescFlags (PVNet_ADAPTER Adapter, USHORT Descriptor, PUCHAR flags)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.RxDescPos +
			     (SIZE_OF_RXDESC * Descriptor) +
			     RX_DESC_FLAGS_OFFSET));
  CardReadUchar (Adapter, DR, flags);
}


void
SetRxDescFlags (PVNet_ADAPTER Adapter, USHORT Descriptor, UCHAR flags)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.RxDescPos +
			     (SIZE_OF_RXDESC * Descriptor) +
			     RX_DESC_FLAGS_OFFSET));
  CardWriteUchar (Adapter, DR, flags);
}


void
GetRxDescStatus (PVNet_ADAPTER Adapter, USHORT Descriptor, PUCHAR status)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.RxDescPos +
			     (SIZE_OF_RXDESC * Descriptor) +
			     RX_DESC_STATUS_OFFSET));
  CardReadUchar (Adapter, DR, status);
}

void
GetRxDescMsduSize (PVNet_ADAPTER Adapter, USHORT Descriptor, PUSHORT msdusize)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.RxDescPos +
			     (SIZE_OF_RXDESC * Descriptor) +
			     RX_DESC_MSDU_SIZE_OFFSET));
  CardReadUshort (Adapter, DR, msdusize);
}


void
GetRxDescMsduPos (PVNet_ADAPTER Adapter, USHORT Descriptor, PUSHORT msdupos)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.RxDescPos +
			     (SIZE_OF_RXDESC * Descriptor) +
			     RX_DESC_MSDU_POS_OFFSET));
  CardReadUshort (Adapter, DR, msdupos);
}


void
GetRxDescRssi (PVNet_ADAPTER Adapter, USHORT Descriptor, PCHAR rssi)
{
  int tmpRSSI = 0;

  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.RxDescPos +
			     (SIZE_OF_RXDESC * Descriptor) +
			     RX_DESC_RSSI_OFFSET));
  CardReadUchar (Adapter, DR, (PUCHAR) & tmpRSSI);

#if defined (REV_E)
  tmpRSSI = (4 * tmpRSSI);

  if (tmpRSSI > 100)
    tmpRSSI = 100;

#elif defined (PCI) || defined (R504) || defined (R504_C) || defined (R504_2958)
  tmpRSSI = -1 * (((98 - tmpRSSI) * 117) / 100);

  if (tmpRSSI < -90)
    tmpRSSI = -90;

#elif defined (RFMD) || defined (RFMD_3COM)
  tmpRSSI = (Adapter->LinkQuality / 2) + (tmpRSSI * 3);
  if (tmpRSSI > 100)
    tmpRSSI = 100;

#elif defined(R504A_2958)
        IOLog("tmpRssi = %d\n", tmpRSSI);
        tmpRSSI -= 110;
        tmpRSSI = (-tmpRSSI*100)/90;
#endif

  *rssi = (CHAR) tmpRSSI;
  Adapter->Rssi = (CHAR) tmpRSSI;

}


void
GetRxDescLinkQuality (PVNet_ADAPTER Adapter, USHORT Descriptor,
		      PUCHAR LinkQuality)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.RxDescPos +
			     (SIZE_OF_RXDESC * Descriptor) +
			     RX_DESC_LINK_QUALITY_OFFSET));
  CardReadUchar (Adapter, DR, LinkQuality);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    StoreBSSinfo														 
//
//	Stores in an array all the AP information from a beacon or probe 	  
//  response frame. If a BSS already exists in the list the entry is 
//  updated if we are processing a Probe Response. 														  
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
BOOLEAN
StoreBSSinfo(PVNet_ADAPTER Adapter, PUCHAR BSSID, PUCHAR pBeaconOrPrope, UCHAR Channel, BOOLEAN IsProbe, USHORT len)
{
	int			i, index=0;
        //int len;
	BOOLEAN		Valid;
	USHORT		CapabilityInformation;
	BOOLEAN		BSSexists;
	CHAR		tmpRSSI;
	UCHAR		ElementBody[MAX_WPA_ELEMENT_SIZE];
	USHORT		ElementBodyLen, pwcnt, authcnt;

	BSSexists = FALSE;
	for(i=0;i<Adapter->BSSListEntries;i++)
	{
		if(!MemCompare(BSSID, Adapter->BSSinfo[i].BSSID, 6))
		{  
			index = i;
			BSSexists = TRUE;
		}
	}	

	if(!BSSexists)
	{
		if(Adapter->BSSListEntries == MAX_BSS_ENTRIES)
		{
			IF_LOUD(DbgPrint("** AP list is full ***\n");)
			return FALSE;
		}

		index = Adapter->BSSListEntries;	
	}
	else //BSS exists in list
	{
		// If we process a probe and an entry from this BSS exists 
		// we will update the BSS entry with the info from this BSS.
		// If we process a beacon we will only update RSSI if the new 
		// beacon has better RSSI.

		if(!IsProbe) 
		{
			GetRxDescRssi(Adapter, Adapter->RxDescHead, &tmpRSSI);
			if(tmpRSSI > Adapter->BSSinfo[index].RSSI)
			{
				Adapter->BSSinfo[index].RSSI = tmpRSSI;
			}
			//always update the first elements
//			VNetMoveMemory(Adapter->BSSinfo[index].BeaconProbeInfo, pBeaconOrPrope, 12);

			return TRUE;
		}
	}

	Adapter->BSSinfo[index].Channel = Channel;

	Valid = GetSSIDfromBeaconOrProbe(Adapter->BSSinfo[index].SSID,
							&Adapter->BSSinfo[index].SSIDsize,
							(PUCHAR)pBeaconOrPrope);

	if(!Valid){
		IF_DEBUG_ERRORS(DbgPrint("*** Invalid Beacon or Probe Response Frame (S)***\n");)
		return FALSE;
	}

	Valid = GetSupportedRatesfromBeaconOrProbe(Adapter->BSSinfo[index].SupportedRates,
											&Adapter->BSSinfo[index].SupportedRatesLength,
											(PUCHAR)pBeaconOrPrope,
											SSID_ELLEMENT_OFFSET + Adapter->BSSinfo[index].SSIDsize);
	if(!Valid)
	{
		IF_DEBUG_ERRORS(DbgPrint("*** Missing Supported Rates in Beacon or Probe Response Frame ***\n");)
	}


	Adapter->BSSinfo[index].BeaconPeriod = *(PUSHORT)(pBeaconOrPrope + BEACON_INTERVAL_OFFSET);

	CapabilityInformation = *(PUSHORT)(pBeaconOrPrope + CAPABILITY_INFO_IN_BEACON_OFFSET);


	if(CapabilityInformation & C80211_MGMT_CAPABILITY_Privacy)
		Adapter->BSSinfo[index].EncryptionIsOn = 1;
	else
		Adapter->BSSinfo[index].EncryptionIsOn = 0;

	if(CapabilityInformation & C80211_MGMT_CAPABILITY_IBSS)
		Adapter->BSSinfo[index].BSStype = BSS_TYPE_AD_HOC;
	else if(CapabilityInformation & C80211_MGMT_CAPABILITY_ESS)
		Adapter->BSSinfo[index].BSStype = BSS_TYPE_INFRASTRUCTURE;


	if(CapabilityInformation & C80211_MGMT_CAPABILITY_ShortPreamble)
		Adapter->BSSinfo[index].PreambleType = PREAMBLE_TYPE_SHORT;
	else
		Adapter->BSSinfo[index].PreambleType = PREAMBLE_TYPE_LONG;
	
	VNetMoveMemory(Adapter->BSSinfo[index].BSSID, BSSID, 6);
		
	Adapter->BSSinfo[index].Capabilities = 0;

	ElementBodyLen = MAX_WPA_ELEMENT_SIZE;

	// If the privacy bit in capabilities is off, we skip the WPA element discovery.
	if(Adapter->BSSinfo[index].EncryptionIsOn)
	{
		if(GetElementFromBeaconOrProbe(pBeaconOrPrope, len, ELEMENT_ID_WPA, OUI_WPA, ElementBody, &ElementBodyLen))
		{
                 //   int j;
                        Adapter->BSSinfo[i].WpaIELen = ElementBodyLen+2;
                        *Adapter->BSSinfo[i].WpaIE = 0xDD;
                        *(Adapter->BSSinfo[i].WpaIE + 1) = ElementBodyLen;
                        memcpy(Adapter->BSSinfo[i].WpaIE + 2, ElementBody, ElementBodyLen);
#if 0
			DbgPrint("\n======================= WPA Info =======================\n");
                        for(j = 0; j < ElementBodyLen; j++){
			    DbgPrint("%02x", Adapter->BSSinfo[i].WpaIE[j]);
                        }
			DbgPrint("\n======================= WPA Info =======================\n");
#endif

			Adapter->BSSinfo[index].Capabilities |= BSS_CAPAPILITY_WPA;

			if(ElementBodyLen >= 10)
				Adapter->BSSinfo[index].WpaInfo.GroupSuite = ElementBody[WPA_ELEMENT_GROUP_SUITE_OFFSET+3];
			else
				Adapter->BSSinfo[index].WpaInfo.GroupSuite = CIPHER_SUITE_TKIP;
			
			
			if(ElementBodyLen >= 12)			
				VNetMoveMemory(&Adapter->BSSinfo[index].WpaInfo.PairwiseSuiteCnt, &ElementBody[WPA_ELEMENT_PW_SUITE_OFFSET], 2);
			else
				Adapter->BSSinfo[index].WpaInfo.PairwiseSuiteCnt = 1;

			pwcnt = Adapter->BSSinfo[index].WpaInfo.PairwiseSuiteCnt;
			
			if(pwcnt > MAX_PAIRWISE_SUITES)
			{
				pwcnt = MAX_PAIRWISE_SUITES;
				Adapter->BSSinfo[index].WpaInfo.PairwiseSuiteCnt = MAX_PAIRWISE_SUITES;
			}

			if(ElementBodyLen >= (12 + 4*pwcnt))			
			{	
				for(i=1;i<=pwcnt;i++)
					VNetMoveMemory(&Adapter->BSSinfo[index].WpaInfo.PairwiseSuiteList[i-1], &ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*i)-1], 1);
			}
			else
			{
				Adapter->BSSinfo[index].WpaInfo.PairwiseSuiteList[0] = 2;
			}

			if(ElementBodyLen >= (16 + (4 * pwcnt)))			
				VNetMoveMemory(&Adapter->BSSinfo[index].WpaInfo.AuthMgmtSuiteCnt, &ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)], 2);
			else
				Adapter->BSSinfo[index].WpaInfo.AuthMgmtSuiteCnt = 1;

			authcnt = Adapter->BSSinfo[index].WpaInfo.AuthMgmtSuiteCnt;

			if(authcnt > MAX_AUTH_MGMT_SUITES)
			{
				authcnt = MAX_AUTH_MGMT_SUITES;
				Adapter->BSSinfo[index].WpaInfo.AuthMgmtSuiteCnt = MAX_AUTH_MGMT_SUITES;
			}
			
			if(ElementBodyLen >= (16 + (4 * pwcnt) + 2))
			{
				for(i=1;i<=authcnt;i++)
					VNetMoveMemory(&Adapter->BSSinfo[index].WpaInfo.AuthMgmtSuiteList[i-1], &ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)+2+(4*i)-1], 1);
			}
			else
			{
				Adapter->BSSinfo[index].WpaInfo.AuthMgmtSuiteList[0] = 1; //802.1x
			}
			
			if(ElementBodyLen >= (16 + (4 * pwcnt) + 2 + ( 4 * authcnt )))
				VNetMoveMemory(&Adapter->BSSinfo[index].WpaInfo.WpaCapabilities, &ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)+2+(4*authcnt)], 2);

                        
#if 0
                        
			DbgPrint("\n======================= WPA Info =======================\n");
                        for(i=0; i<Adapter->BSSinfo[index].SSIDsize; i++)
                                   DbgPrint("%c", Adapter->BSSinfo[index].SSID[i]);
                                   DbgPrint("\n");
			DbgPrint("\n======================= ******** =======================\n");
			DbgPrint("Group Suite: %d\n", Adapter->BSSinfo[index].WpaInfo.GroupSuite);
			DbgPrint("Pairwise Suite Cnt: %d - Pairwise Suites: ", pwcnt);
			for(i=0;i<pwcnt;i++)
				DbgPrint("%02X ", Adapter->BSSinfo[index].WpaInfo.PairwiseSuiteList[i]);
			DbgPrint("\n");
			DbgPrint("Authentication Suite Cnt: %d - Authentication Suites: ", authcnt);
			for(i=0;i<authcnt;i++)
				DbgPrint("%02X ", Adapter->BSSinfo[index].WpaInfo.AuthMgmtSuiteList[i]);
			DbgPrint("\n");
			DbgPrint("WPA Capabilities: %02X-%02X\n", Adapter->BSSinfo[index].WpaInfo.WpaCapabilities[0], Adapter->BSSinfo[index].WpaInfo.WpaCapabilities[1]);

			DbgPrint("\n========================================================\n");
#endif
			}
	}

#ifdef CCX
	// If the privacy bit in capabilities is off, we skip the AIRONET element discovery.
	if(Adapter->BSSinfo[index].EncryptionIsOn)
	{
		ElementBodyLen = MAX_AIROPEEK_ELEMENT_SIZE;
		if(GetElementFromBeaconOrProbe(pBeaconOrPrope, len, ELEMENT_ID_AIRONET, ElementBody, &ElementBodyLen))
		{
			Adapter->BSSinfo[index].Capabilities |= BSS_CAPAPILITY_CCX;

			if( (Adapter->BSSinfo[i].SSIDsize == Adapter->AssociationInfo.DesiredSSIDsize) && (ElementBodyLen > 6) )
			{
				// if we CCX IE is from the current BSSID and the CwMin or the CwMax is new, 
				// we have to set the new value(s).
				if( ( !MemCompare(Adapter->BSSinfo[i].SSID, Adapter->AssociationInfo.DesiredSSID, Adapter->AssociationInfo.DesiredSSIDsize)) &&
					( Adapter->CwMin != *((PUSHORT)(ElementBody + AIRONET_ELEMENT_CWMIN_OFFSET)) ||
					  Adapter->CwMax != *((PUSHORT)(ElementBody + AIRONET_ELEMENT_CWMAX_OFFSET)) )
				  )
				{
					Adapter->CwMin = *((PUSHORT)(ElementBody + AIRONET_ELEMENT_CWMIN_OFFSET));
					Adapter->CwMax = *((PUSHORT)(ElementBody + AIRONET_ELEMENT_CWMAX_OFFSET));

					SetCwWindow(Adapter, Adapter->BSSinfo[index]);
				}
			}
		}
	}
#endif

	if(len > MAX_IE_LENGTH)
		len = MAX_IE_LENGTH;

	VNetMoveMemory(Adapter->BSSinfo[index].BeaconProbeInfo, pBeaconOrPrope, len);
	Adapter->BSSinfo[index].BeaconProbeLen = len;
	
	GetRxDescRssi(Adapter, Adapter->RxDescHead, &tmpRSSI);
	
	if(BSSexists)
	{
		if(tmpRSSI > Adapter->BSSinfo[index].RSSI)
		{
			Adapter->BSSinfo[index].RSSI = tmpRSSI;
		}
	}
	else
	{
		Adapter->BSSinfo[index].RSSI = tmpRSSI;
	}

	if(!BSSexists)
	{
		Adapter->BSSListEntries++;
	}
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    MgmtFrameRxProcessing												  
//
//	  Processes the Rx MGMT frames and implements the necessary actions needed 
//	  until the station becomes associated.
//																		  
//		-Supports ANY ssid												  
//		-Sets time-outs for the mgmt frames								  
//		-Fullly supports WEP (challenge text/authentication transaction	  
//		 sequence numbers).												  	
//		-"Decides" actions according to the received status codes		  
//		-Supports Reassocioation Response frames						  	
//		-Supports Deauthentication & Disassociation Frames				  
//      -Supports Prefeared BSSID										  
//																	  
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VOID
MgmtFrameRxProcessing(	PVNet_ADAPTER Adapter,
                                                struct net_device *dev,
						PUCHAR RxBuf,
						UCHAR FrameSubtype,
						USHORT FrameLength)
{
	UCHAR							AuthenticationFrame[261];//max auth frame size
	USHORT							StatusCode,ReasonCode;
	int								i;
//	UCHAR							ChallengeTextLength;
	//BOOLEAN							CompareSSIDs = TRUE;
	USHORT							BeaconOrProbeRespFrameBytes;
	USHORT							AuthenticationFrameBytesToGet;
	USHORT							CapabilityInformation;
	BOOLEAN							MgmtFrameFromCurrentBSS;
	BOOLEAN							IsProbe;
	BOOLEAN							PreambleChanged = FALSE;
	BOOLEAN							ChannelChanged = FALSE;
	UCHAR							Channel;
	CHAR							tmpRSSI;
	int								BSSindex;
	USHORT							AssociationID;

	MgmtFrameFromCurrentBSS = TRUE;

	if(Adapter->OperatingMode == AD_HOC_MODE)
	{
		for(i=0;i<6;i++)
			if(RxBuf[DATA_FRAME_ADHOC_ADR3_OFFSET + i] != Adapter->AssociationInfo.CurrentBSSID[i])
			{
				MgmtFrameFromCurrentBSS = FALSE;
				break;
			}
	}
	else
	{
		for(i=0;i<6;i++)
			if(RxBuf[MGMT_FRAME_SA_OFFSET + i] != Adapter->AssociationInfo.CurrentBSSID[i])
			{
				MgmtFrameFromCurrentBSS = FALSE;
				break;
			}
	}


	switch(FrameSubtype)
	{
	case C80211_SUBTYPE_MGMT_BEACON :
	case C80211_SUBTYPE_MGMT_ProbeResponse:

		if(FrameSubtype == C80211_SUBTYPE_MGMT_ProbeResponse)
			IsProbe = TRUE;
		else
			IsProbe = FALSE;

		FrameLength -= MGMT_FRAME_BODY_OFFSET; //we need the length of the Frame Body only (not the wireless header)

		if(FrameLength < 16)
		{
			IF_DEBUG_ERRORS(DbgPrint("*** Too short (%d) Beacon or Probe Response ***\n", FrameLength);)
			break;
		}

		BeaconOrProbeRespFrameBytes = (FrameLength > MAX_BEACON_PROBE_BYTES) ? MAX_BEACON_PROBE_BYTES : FrameLength;

		CapabilityInformation = *(RxBuf + MGMT_FRAME_BODY_OFFSET + CAPABILITY_INFO_IN_BEACON_OFFSET);

		if(!GetChannelFromBeaconOrProbe((RxBuf+MGMT_FRAME_BODY_OFFSET), &Channel))
		{
			IF_DEBUG_ERRORS(DbgPrint("*** Invalid Beacon or Probe Response Frame (Ch)***\n");)
			break;
		}

		if(FrameSubtype == C80211_SUBTYPE_MGMT_BEACON)
			Adapter->Stats.MatchingBeacons++;

		//
		// NTXDEC
		//
		// replace with
		// if(Adapter->StationState == STATION_STATE_SCANNING)
		// also you would have to reverse the "if then else" condition (STATION_STATE_READY first)
		//
		if(	(Adapter->StationState == STATION_STATE_SCANNING) || (Adapter->SiteSurveyState == SITE_SURVEY_IN_PROGRESS) )
		{
			if((CapabilityInformation & 0x0001))
				StoreBSSinfo(Adapter, RxBuf+MGMT_FRAME_SA_OFFSET, RxBuf+MGMT_FRAME_BODY_OFFSET, Channel, IsProbe, BeaconOrProbeRespFrameBytes);
			else
				StoreBSSinfo(Adapter, RxBuf+DATA_FRAME_ADHOC_ADR3_OFFSET, RxBuf+MGMT_FRAME_BODY_OFFSET, Channel, IsProbe, BeaconOrProbeRespFrameBytes);
		}
		else
		if(Adapter->StationState == STATION_STATE_READY)
		{
			if(MgmtFrameFromCurrentBSS)
			{
				if(!IsProbe)
				{
					Adapter->PeriodicBeaconsCnt++;

//ahc
/*					if(Adapter->OperatingMode == AD_HOC_MODE && Adapter->StationAloneInIbss)
					{
						IF_LOUD(DbgPrint("**** STATION is NOT ALONE in IBSS ****\n");)
						Adapter->StationAloneInIbss = FALSE;
						Adapter->StationIsAssociated = TRUE;
						IndicateConnectionStatus(Adapter, TRUE);
					}
*/
				}
				if(Adapter->AutoPreambleDetection)
				{
					if( (Adapter->AssociationInfo.PreambleType == PREAMBLE_TYPE_LONG) && (CapabilityInformation & C80211_MGMT_CAPABILITY_ShortPreamble) )
					{
						IF_LOUD(DbgPrint("*** AUTO SWITCH to SHORT preamble ***\n");)

						Adapter->AssociationInfo.PreambleType = PREAMBLE_TYPE_SHORT;
						PreambleChanged = TRUE;
					}
					else
					if( (Adapter->AssociationInfo.PreambleType == PREAMBLE_TYPE_SHORT) && !(CapabilityInformation & C80211_MGMT_CAPABILITY_ShortPreamble) )
					{
						IF_LOUD(DbgPrint("*** AUTO SWITCH to LONG preamble ***\n");)

						Adapter->AssociationInfo.PreambleType = PREAMBLE_TYPE_LONG;
						PreambleChanged = TRUE;
					}	
				}
				
				if( Adapter->Channel != Channel)
				{
					IF_LOUD(DbgPrint("*** CHANNEL CHANGED ***\n");)
					ChannelChanged = TRUE;
					Adapter->Channel = Channel;
					SetChannel(Adapter, Adapter->Channel);
				}

				if(PreambleChanged || ChannelChanged)
				{
					if(PreambleChanged)
					{
						SetPreambleType(Adapter, Adapter->AssociationInfo.PreambleType);
					}

					Adapter->StationIsAssociated = FALSE;
					Adapter->StationState = STATION_STATE_JOINNING;

					if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
						Join(Adapter, BSS_TYPE_INFRASTRUCTURE);
					else 
						Join(Adapter, BSS_TYPE_AD_HOC);
				} 
			}

			GetRxDescRssi(Adapter, Adapter->RxDescHead, &tmpRSSI);

			if(tmpRSSI)
			{
				Adapter->RssiSample[(int)Adapter->RssiIndex] = tmpRSSI;
				Adapter->RssiIndex++;
				if(Adapter->RssiIndex == MAX_RSSI_SAMPLES)
					Adapter->RssiIndex = 0;
			}
		}
			
		break;
		
		case C80211_SUBTYPE_MGMT_Authentication:
	
			AuthenticationFrameBytesToGet = (FrameLength < 261) ? FrameLength : 261;

			VNetMoveMemory(AuthenticationFrame, &RxBuf[MGMT_FRAME_BODY_OFFSET], AuthenticationFrameBytesToGet);
			
			StatusCode = AuthenticationFrame[STATUS_IN_AUTHENTICATION_OFFSET];
	

			if(Adapter->StationState == STATION_STATE_AUTHENTICATING)
			{
				IF_LOUD(DbgPrint("C80211_SUBTYPE_MGMT_Authentication...\n ");)

				if(StatusCode == C80211_MGMT_SC_Success)
				{
					if(IsEncryptionEnabled(&Adapter->SecurityInfo) && (Adapter->SecurityInfo.AuthenticationMode == _802_11AuthModeShared))//WEP
					{
						if(AuthenticationFrame[TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET] == Adapter->AuthenticationInfo.ExpectedAuthentTransactionSeqNum)
						{
							if(AuthenticationFrame[TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET] == 0x0002)
							{
								IF_LOUD(DbgPrint("AUTHENTICATION with TrSeqNo: %d (ELID %d)\n", AuthenticationFrame[TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET], AuthenticationFrame[CHALL_TEXT_EL_ID_IN_AUTHENTICATION_OFFSET]);)
								if(AuthenticationFrame[CHALL_TEXT_EL_ID_IN_AUTHENTICATION_OFFSET] == C80211_MGMT_ElementID_ChallengeText)
								{
									Adapter->AuthenticationInfo.ChallengeTextLength = AuthenticationFrame[CHALL_TEXT_LENGTH_IN_AUTHENTICATION_OFFSET];
									IF_LOUD(DbgPrint("Challenge Text Length = %d\n", Adapter->AuthenticationInfo.ChallengeTextLength);)
									
									VNET_ASSERT(Adapter->AuthenticationInfo.ChallengeTextLength<=253);
                  memcpy(Adapter->AuthenticationInfo.pChallengeText, &AuthenticationFrame[CHALL_TEXT_IN_AUTHENTICATION_OFFSET], Adapter->AuthenticationInfo.ChallengeTextLength);
                         
									SendAuthenticationRequest(Adapter);
								}
							}
							else
							if(AuthenticationFrame[TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET] == 0x0004)
							{
								IF_LOUD(DbgPrint("AUTHENTICATION with TrSeqNo: %d\n", AuthenticationFrame[TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET]);)
								IF_LOUD(DbgPrint("### AUTHENTICATED ###\n");)

#ifdef CCX
								if(Adapter->CCXinfo.FswNetworkEAPToUse == 0x80)
									Adapter->CCXinfo.FswNetworkEAP_Query = FswCcx_NetworkEapOn;
								else
									Adapter->CCXinfo.FswNetworkEAP_Query = FswCcx_NetworkEapOff;
#endif

								if(!Adapter->StationWasAssociated || 
									!((Adapter->LastSSIDSize == Adapter->AssociationInfo.DesiredSSIDsize) &&
									 (!MemCompare(Adapter->LastSSID, Adapter->AssociationInfo.DesiredSSID, Adapter->AssociationInfo.DesiredSSIDsize)))
								  )
								{
									Adapter->StationState = STATION_STATE_ASSOCIATING;
									SendAssociationRequest(Adapter);
								}
								else
								{
									Adapter->StationState = STATION_STATE_REASSOCIATING;
									SendReAssociationRequest(Adapter);
								}

							}
	
						}
					}
					else
					{
						IF_LOUD(DbgPrint("### AUTHENTICATED ###\n");)
#ifdef CCX
						if(Adapter->CCXinfo.FswNetworkEAPToUse == 0x80)
							Adapter->CCXinfo.FswNetworkEAP_Query = FswCcx_NetworkEapOn;
						else
							Adapter->CCXinfo.FswNetworkEAP_Query = FswCcx_NetworkEapOff;
#endif

						if(!Adapter->StationWasAssociated || 
							!((Adapter->LastSSIDSize == Adapter->AssociationInfo.DesiredSSIDsize) &&
							 (!MemCompare(Adapter->LastSSID, Adapter->AssociationInfo.DesiredSSID, Adapter->AssociationInfo.DesiredSSIDsize)))
						  )
						{
							Adapter->StationState = STATION_STATE_ASSOCIATING;
							SendAssociationRequest(Adapter);
						}
						else
						{
							Adapter->StationState = STATION_STATE_REASSOCIATING;
							SendReAssociationRequest(Adapter);
						}
					}
				}
				else
				{
					Adapter->MgmtErrorCode = StatusCode;

					DisplayMgmtFrameStatus(StatusCode);
					
					if(StatusCode == C80211_MGMT_SC_AuthAlgNotSupported)
					{
						IF_LOUD(DbgPrint("####### Authentication Algorithm Not Supported #######\n");)

#ifdef CCX
						// if the EAP is allowed and you have used Open, try EAP.
						if(Adapter->CCXinfo.FswNetworkEAP_Set == FswCcx_NetworkEapAllowed && Adapter->CCXinfo.FswCcx_AutoSwitchAuthAlgorithm && Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt < MAX_AUTHENTICATION_RETRIES)
						{
							Adapter->CCXinfo.FswNetworkEAPToUse = 0x80;
							Adapter->CCXinfo.FswCcx_AutoSwitchAuthAlgorithm = FALSE;

							IF_LOUD(DbgPrint("FSW: Retry with other Authentication Type 1...\n");)
							Adapter->StationState = STATION_STATE_AUTHENTICATING;
							Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt++;
							SetMgmtTimer(Adapter);
					
							Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum = 0x0001;
							SendAuthenticationRequest(Adapter);

							break;
						}
						// if the Open is allowed and you have used EAP, try Open.
						else if(Adapter->CCXinfo.FswNetworkEAP_Set == FswCcx_NetworkEapPreferred && Adapter->CCXinfo.FswCcx_AutoSwitchAuthAlgorithm && Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt < MAX_AUTHENTICATION_RETRIES)
						{
							Adapter->CCXinfo.FswNetworkEAPToUse = 0;
							Adapter->CCXinfo.FswCcx_AutoSwitchAuthAlgorithm = FALSE;
			
							IF_LOUD(DbgPrint("FSW: Retry with other Authentication Type 2...\n");)
							Adapter->StationState = STATION_STATE_AUTHENTICATING;
							Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt++;
							SetMgmtTimer(Adapter);
					
							Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum = 0x0001;
							SendAuthenticationRequest(Adapter);

							break;
						}
						// if we have mixed algorithm and we have exhausted the two cases or 
						// the retries marked the AP as rogue.
						// Also do this when the non-mixed algorithm has failed
						else if( ( (Adapter->CCXinfo.FswNetworkEAP_Set == FswCcx_NetworkEapPreferred && 
									!Adapter->CCXinfo.FswCcx_AutoSwitchAuthAlgorithm
								   ) ||
								   (Adapter->CCXinfo.FswNetworkEAP_Set == FswCcx_NetworkEapAllowed &&
									!Adapter->CCXinfo.FswCcx_AutoSwitchAuthAlgorithm)
								   ) ||
									Adapter->CCXinfo.FswNetworkEAP_Set == FswCcx_NetworkEapOn
							   )
						{
									IF_LOUD(DbgPrint("FSW: Authentication Algorithm FAILED (1)...\n Mark AP as ROGUE\n");)
									
									RogueAP.FailureReason = 0x0100;	//big endian
									VNetMoveMemory(RogueAP.RogueAPMacAddress, Adapter->AssociationInfo.CurrentBSSID, 6);
									VNetZeroMemory(RogueAP.RogueAPName, 16);

									CCX_AddRogueAPToList( &Adapter->CCXinfo, &RogueAP );

									Scan(Adapter, TRUE);
								
									break;
						}
						
#endif

						if(Adapter->SecurityInfo.AutoSwitchAuthAlgorithm && Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt < MAX_AUTHENTICATION_RETRIES)
						{
							IF_LOUD(DbgPrint(">>>>>>> Retry with other Authentication Type...\n");)
							Adapter->StationState = STATION_STATE_AUTHENTICATING;

							if(Adapter->SecurityInfo.AuthenticationMode == _802_11AuthModeShared)
								Adapter->SecurityInfo.AuthenticationMode = _802_11AuthModeOpen;
							else
							if(Adapter->SecurityInfo.AuthenticationMode == _802_11AuthModeOpen)
								Adapter->SecurityInfo.AuthenticationMode = _802_11AuthModeShared;

							Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt ++ ;
							SetMgmtTimer(Adapter);
					
							Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum = 0x0001;
							SendAuthenticationRequest(Adapter);
						}
						else
						{
							if(Adapter->ConnectToAnyBSS)
							{
								//
								//BL: (4.0.2.28) invalidate list entry...
								//
								IF_LOUD(DbgPrint("**** Try Another BSS-AU (invalidate BSS #%d)...\n", Adapter->CurrentBSSindex);)
								Adapter->BSSinfo[(int) Adapter->CurrentBSSindex].FailedToConnect = 1;

								BSSindex = RetrieveBSS(Adapter, NULL);
								if(BSSindex != -1)
								{
									IF_LOUD(DbgPrint("**** Try BSS #%d in channel %d...\n", BSSindex, Adapter->BSSinfo[BSSindex].Channel);)
									JoinSpecificBSS(Adapter, BSSindex, FALSE);
								}
								else
								{
									if(Adapter->StationIsOutOfRange)
										IndicateConnectionStatus(Adapter, FALSE);

									Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt = 0;
									Adapter->StationState = STATION_STATE_MGMT_ERROR;
									Adapter->StationIsAssociated		   = FALSE;
								}
							}
							else
							{
								if(Adapter->StationIsOutOfRange)
									IndicateConnectionStatus(Adapter, FALSE);

								Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt = 0;
								Adapter->StationState = STATION_STATE_MGMT_ERROR;
								Adapter->StationIsAssociated		   = FALSE;
							}
						}
					}
				}
			}
			break;

		case C80211_SUBTYPE_MGMT_ASS_RESPONSE:
		case C80211_SUBTYPE_MGMT_REASS_RESPONSE:
					
			FrameLength -= MGMT_FRAME_BODY_OFFSET; //we need the length of the Frame Body only (not the wireless header)

			if(Adapter->StationState == STATION_STATE_ASSOCIATING || Adapter->StationState == STATION_STATE_REASSOCIATING)
			{

				StatusCode = RxBuf[MGMT_FRAME_BODY_OFFSET + STATUS_IN_ASSOCIATION_RESP_OFFSET];

				if(StatusCode == C80211_MGMT_SC_Success)
				{
					Adapter->SecurityInfo.MicErrorCnt = 0;
					
//					Adapter->SecurityInfo.MicErrorDetected = FALSE;		Here we get current time..
//////////////////////////////////////////
					
					VNetZeroMemory(Adapter->AssociationRspnsFrameBody, MAX_ASSOCIATION_FR_BODY_BYTES);
					Adapter->AssociationInfo.AssociationRespFrameBodyBytes = (FrameLength > MAX_ASSOCIATION_FR_BODY_BYTES) ? MAX_ASSOCIATION_FR_BODY_BYTES : FrameLength;
					VNetMoveMemory(Adapter->AssociationRspnsFrameBody, &RxBuf[MGMT_FRAME_BODY_OFFSET], Adapter->AssociationInfo.AssociationRespFrameBodyBytes);

					StatusCode = Adapter->AssociationRspnsFrameBody[STATUS_IN_ASSOCIATION_RESP_OFFSET];
					AssociationID = (Adapter->AssociationRspnsFrameBody[ASS_ID_IN_ASSOCIATION_RESP_OFFSET] & 0x3fff);

					Adapter->SupportedRatesLength = Adapter->AssociationRspnsFrameBody[SUP_RATES_LENGTH_IN_ASSOCIATION_RESP_OFFSET];

					if(Adapter->SupportedRatesLength > OPER_RATE_SIZE){ //for safety...
						Adapter->SupportedRatesLength = OPER_RATE_SIZE;
						IF_DEBUG_ERRORS(DbgPrint("*** Invalid Supported Rates Length ***\n");)
					}

					VNetMoveMemory(Adapter->SupportedRates, &Adapter->AssociationRspnsFrameBody[SUP_RATES_IN_ASSOCIATION_RESP_OFFSET], Adapter->SupportedRatesLength);
/////////////////////////////////
					Adapter->AssociationInfo.AssociationRequestRetryCnt = 0;
					Adapter->AssociationInfo.ReAssociationRequestRetryCnt = 0;

					VNetMoveMemory(Adapter->LastSSID, Adapter->AssociationInfo.DesiredSSID, Adapter->AssociationInfo.DesiredSSIDsize);
					Adapter->LastSSIDSize = Adapter->AssociationInfo.DesiredSSIDsize;
					
					Adapter->StationIsOutOfRange = FALSE;
					Adapter->ReConnectRetry = 0;

					if(MemCompare(Adapter->AssociationInfo.CurrentBSSID, Adapter->AssociationInfo.LastBSSID, 6)) //if associated to new BSS keep it in LastBSSID
					{
						VNetMoveMemory(Adapter->AssociationInfo.LastBSSID, Adapter->AssociationInfo.CurrentBSSID, 6);

//						IndicateConnectionStatus(Adapter, TRUE);
					}

					IndicateConnectionStatus(Adapter, TRUE);

#ifdef CCX
					IF_LOUD(DbgPrint("\n\n\nEL ID: 0x%X len = %d\n\n\n", Adapter->AssociationRspnsFrameBody[SUP_RATES_IN_ASSOCIATION_RESP_OFFSET + Adapter->SupportedRatesLength], Adapter->AssociationRspnsFrameBody[SUP_RATES_IN_ASSOCIATION_RESP_OFFSET + Adapter->SupportedRatesLength+1]);)
					pCCXelement = &Adapter->AssociationRspnsFrameBody[SUP_RATES_IN_ASSOCIATION_RESP_OFFSET + Adapter->SupportedRatesLength];

					ret = CCX_ParseElement(&Adapter->CCXinfo, &Adapter->SecurityInfo.EncryptionInfo, pCCXelement);

					IF_LOUD(DbgPrint("CCX_ParseElement returns %d\n",ret);)

					if(ret == Set_keys_CKIP_ON) 
					{
						IF_LOUD(DbgPrint("CCX_ParseElement :=> CKIP is ON and set any available keys.(%d)\n",ret);)
						if(Adapter->SecurityInfo.PairwiseKeyAvailable)
							SetEncryptionStatus(Adapter, TRUE, Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex);

						if(Adapter->SecurityInfo.GroupKeyAvailable)
							SetEncryptionStatus(Adapter, TRUE, Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex);
					}
					else if (ret == Prepare_MIC_key)
					{
						IF_LOUD(DbgPrint("CCX_ParseElement :=> CMIC is ON and CKIP is off and prepare the MIC key of any key available.(%d)\n",ret);)
						if(Adapter->SecurityInfo.PairwiseKeyAvailable && Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite != CIPHER_SUITE_NONE)
							SetEncryptionStatus(Adapter, TRUE, Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex);

						if(Adapter->SecurityInfo.GroupKeyAvailable && Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite != CIPHER_SUITE_NONE)
							SetEncryptionStatus(Adapter, TRUE, Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex);
					}
					else if(ret == RESTORE_BOTH_KEYS)
					{
						IF_LOUD(DbgPrint("CCX_ParseElement :=> restore pairwise key.(%d)\n",ret);)
						SetEncryptionStatus(Adapter, TRUE, Adapter->CCXinfo.KeyToRestore.PairwiseKeyIndex);
						IF_LOUD(DbgPrint("CCX_ParseElement :=> restore group key.(%d)\n",ret);)
						SetEncryptionStatus(Adapter, TRUE, Adapter->CCXinfo.KeyToRestore.GroupKeyIndex);
					}
					else if(ret == RESTORE_PAIRWISE_KEY)
					{
						IF_LOUD(DbgPrint("CCX_ParseElement :=> restore pairwise key.(%d)\n",ret);)
						SetEncryptionStatus(Adapter, TRUE, Adapter->CCXinfo.KeyToRestore.PairwiseKeyIndex);
					}
					else if(ret == RESTORE_GROUP_KEY)
					{
						IF_LOUD(DbgPrint("CCX_ParseElement :=> restore group key.(%d)\n",ret);)
						SetEncryptionStatus(Adapter, TRUE, Adapter->CCXinfo.KeyToRestore.GroupKeyIndex);
					}
					else if(ret == RESTORE_NONE)
					{
						IF_LOUD(DbgPrint("CCX_RestoreEncryptionSettings nothing to do....\n");)
					}

#endif
			
					Adapter->StationWasAssociated = TRUE;

					IF_LOUD(DbgPrint("Association Success (Assoc-ID = 0x%X - Supported Rates = %02X-%02X-%02X-%02X)\n", AssociationID, Adapter->SupportedRates[0],Adapter->SupportedRates[1],Adapter->SupportedRates[2],Adapter->SupportedRates[3]);)

					SetAssociationID(Adapter, AssociationID);

					SetPowerMgmtMode(Adapter, Adapter->PowerMgmtMode);

#ifdef CCX
					BSSindex = RetrieveBSS(Adapter, Adapter->AssociationInfo.LastBSSID);
					if(BSSindex != -1)
						SetCwWindow(Adapter, Adapter->BSSinfo[BSSindex]);
#endif

					Adapter->PeriodicBeaconsCnt = 0;

					Adapter->StationIsAssociated = TRUE;
          //For Linkquality
				  //SetMgmtTimer(Adapter);
					Adapter->StationState = STATION_STATE_READY;
          atmel_assoc_info_event (dev, Adapter);
					
				}
				else
				{
					Adapter->MgmtErrorCode = StatusCode;
					DisplayMgmtFrameStatus(StatusCode);




					if(FrameSubtype == C80211_SUBTYPE_MGMT_ASS_RESPONSE)
					{
						if(	(StatusCode != C80211_MGMT_SC_AssDeniedBSSRate) &&
							(StatusCode != C80211_MGMT_SC_SupportCapabilities) &&
							(Adapter->AssociationInfo.AssociationRequestRetryCnt < MAX_ASSOCIATION_RETRIES))
						{
							SetMgmtTimer(Adapter);
							Adapter->AssociationInfo.AssociationRequestRetryCnt++;
							SendAssociationRequest(Adapter);
						}
						else
						{
							Adapter->StationState = STATION_STATE_MGMT_ERROR;
							Adapter->StationIsAssociated		   = FALSE;
						}
					}
					else
					{
						if(	(StatusCode != C80211_MGMT_SC_AssDeniedBSSRate) &&
							(StatusCode != C80211_MGMT_SC_SupportCapabilities) &&
							(Adapter->AssociationInfo.ReAssociationRequestRetryCnt < MAX_ASSOCIATION_RETRIES))
						{
							SetMgmtTimer(Adapter);
							Adapter->AssociationInfo.AssociationRequestRetryCnt++;
							SendAssociationRequest(Adapter);
						}
						else
						{
							Adapter->StationState = STATION_STATE_MGMT_ERROR;
							Adapter->StationIsAssociated		   = FALSE;
						}
					}
					if(Adapter->StationState == STATION_STATE_MGMT_ERROR)
					{
						if(Adapter->ConnectToAnyBSS)
						{
							//
							//BL: (4.0.2.28) invalidate list entry...
							//
							IF_LOUD(DbgPrint("**** Try Another BSS-AS invalidate BSS #%d...\n", Adapter->CurrentBSSindex);)
							Adapter->BSSinfo[(int) Adapter->CurrentBSSindex].FailedToConnect = 1;

							BSSindex = RetrieveBSS(Adapter, NULL);
							if(BSSindex != -1)
							{
								IF_LOUD(DbgPrint("**** Try BSS #%d in channel %d...\n", BSSindex, Adapter->BSSinfo[BSSindex].Channel);)
								JoinSpecificBSS(Adapter, BSSindex, FALSE);
							}
						}
					}
				}
			}
		
		break;

		case C80211_SUBTYPE_MGMT_DISASSOSIATION:
			if(Adapter->StationIsAssociated && (Adapter->OperatingMode == INFRASTRUCTURE_MODE) && MgmtFrameFromCurrentBSS)
			{

				IF_LOUD(DbgPrint("C80211_SUBTYPE_MGMT_DISASSOSIATION...\n");)

				// Drop any pending Tx packets, since we are not associated any more.
				// Do not remove this line otherwise if there are pending packets
				//RejectPendingTxPackets(Adapter);
								
				ReasonCode = RxBuf[MGMT_FRAME_BODY_OFFSET];

				DisplayReasonCode(ReasonCode);

				Adapter->StationWasAssociated = FALSE;
				Adapter->StationIsAssociated = FALSE;

				if(Adapter->SecurityInfo.InWpaMode)
				{
					RemoveEncryptionSettings(Adapter);
					IndicateConnectionStatus(Adapter, FALSE);
					Adapter->StationIsAssociated = FALSE;
					Adapter->StationState = STATION_STATE_WAIT_ACTION;
//					Scan(Adapter, TRUE);
				}
				else
				{
					Adapter->StationState = STATION_STATE_JOINNING;
					Join(Adapter, BSS_TYPE_INFRASTRUCTURE);
				}

			}
		break;

		case C80211_SUBTYPE_MGMT_Deauthentication:
			if(Adapter->OperatingMode == INFRASTRUCTURE_MODE  && MgmtFrameFromCurrentBSS)
			{
				IF_LOUD(DbgPrint("C80211_SUBTYPE_MGMT_Deauthentication...\n");)
				
				// Drop any pending Tx packets, since we are not associated any more.
				// Do not remove this line otherwise if there are pending packets
				//RejectPendingTxPackets(Adapter);
								
				ReasonCode = RxBuf[MGMT_FRAME_BODY_OFFSET];
				
				DisplayReasonCode(ReasonCode);

				Adapter->StationWasAssociated = FALSE;
				Adapter->StationIsAssociated = FALSE;

				if(Adapter->SecurityInfo.InWpaMode)
				{
					RemoveEncryptionSettings(Adapter);
					IndicateConnectionStatus(Adapter, FALSE);
					Adapter->StationIsAssociated = FALSE;
					Adapter->StationState = STATION_STATE_WAIT_ACTION;
//					Scan(Adapter, TRUE);
				}
				else
				{
					Adapter->StationState = STATION_STATE_JOINNING;
					Join(Adapter, BSS_TYPE_INFRASTRUCTURE);
				}

			}

		break;
		}
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    RxInit                                                                                                                     
//
//    This function initializes Rx related parameters in the Adapter      
//        object and enables Rx operation on the MAC through the coresponding 
//        field of the IFACE.                                                                                             
//                                                                        
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


VOID
RxInit (PVNet_ADAPTER Adapter)
{
  UCHAR configuration;

  IF_VERY_LOUD (DbgPrint ("Initializing Rx Section....\n");
    )
    //
    //      Initialize Rx Descriptors counters and pointers
    //
    Adapter->RxDescHead = 0;

  Adapter->RxBuffStart = Adapter->HostInfo.RxBuffPos;
  Adapter->RxBuffEnd =
    Adapter->HostInfo.RxBuffPos + Adapter->HostInfo.RxBuffSize;


  Adapter->Receptions = 0;

  //
  //      Enable Rx Operation on the card.
  //
  GetFuncCtrl (Adapter, &configuration);
  configuration |= FUNC_CTRL_RxENABLE;
  SetFuncCtrl (Adapter, &configuration);

}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    RxCompleteDPC														  
//
//		Indicates any received packets to the wrapper. Any errored	  
//		frames or Frames that are not of type DATA are dropped by the	  
//		driver, and not indicated to the OS.														  
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VOID 
RxCompleteDPC(PVNet_ADAPTER Adapter, struct net_device *dev)
{

	ULONG							IndicateLen;		
	ULONG							PacketLength;
	UCHAR							status;
	USHORT							msduSize;
	UCHAR							flags;
	UCHAR							FrameType, FrameSubtype;
	UCHAR							loopcnt, InWPAMode = 0;

	ULONG							RxPacketLoc;
	UCHAR							SourceAddr[6];
	UCHAR							StartOfEthPacket;
	UCHAR							KeyIndex,CipherSuite,KeyReplay;
	UCHAR							RxMicKeyOffset __attribute__ ((unused));
        ULONG64                                                 TimeStamp;

//	int								i;

	UCHAR							RxBuf[MAX_WIRELESS_FRAME_SIZE];


	Adapter->IndicateReceiveDone = FALSE;

	GetRxDescFlags(Adapter, Adapter->RxDescHead, &flags );	

	loopcnt = 0;
	while( flags == RX_DESC_FLAG_VALID && flags != 0xff && loopcnt < Adapter->HostInfo.RxDescCount )
	{
		loopcnt++;
		Adapter->Receptions++;

		GetRxDescStatus(Adapter, Adapter->RxDescHead, &status );	
		GetRxDescMsduSize(Adapter, Adapter->RxDescHead, &msduSize );

		// We need to check if the msduSize exceeds the size of RxBuf or WirelessPacket.  
		// If the packet is bigger we drop it. We make the comparison before the possible
		// decrement since the actual packet that we copy to the RxBuf is always 
		// msduSize bytes long.
		// In this step we secure the RxBuf. Later we will check the WirelessPacket.
		// This check covers the cases that the packet reasembly is done in f/w or it is a
		// complete wireless packet or this is the first fragment of the packet.
		if(msduSize > MAX_WIRELESS_FRAME_SIZE)
		{
			status = 0xff;
		}

		if( (status == RX_STATUS_SUCCESS) )	// RX with SUCCESS
		{

			IF_DEBUG_RX(DbgPrint("Rx #%d (Status = 0x%X - Bytes = %d\n", Adapter->Receptions, status, msduSize);)
				
			//
			// Get packets location in RxPacketLoc
			//

			CardMemToHostMem16(Adapter, (PUCHAR)&RxPacketLoc, (Adapter->HostInfo.RxDescPos + (SIZE_OF_RXDESC * Adapter->RxDescHead) + RX_DESC_MSDU_POS_OFFSET), 2);

			CardMemToHostMem16(Adapter, RxBuf, RxPacketLoc, 1);

			FrameType = RxBuf[0] & C80211_TYPE_MASK;

			GetRxDescCipherSuite(Adapter, Adapter->RxDescHead, &CipherSuite);
			GetRxDescKeyIndex(Adapter, Adapter->RxDescHead, &KeyIndex);

			// NEW
			
			// if Managment Frame
			if(FrameType == C80211_TYPE_MGMT)
			{
				Adapter->Stats.RxMgmtPacketsOk++;

				// NEW
				FrameSubtype = RxBuf[0] & C80211_SUBTYPE_MASK;
				CardMemToHostMem16(Adapter, RxBuf, RxPacketLoc, msduSize);
				// ENDNEW
				
				MgmtFrameRxProcessing(Adapter, dev,RxBuf, FrameSubtype, msduSize);
			}
			else if((FrameType == C80211_TYPE_DATA))	
			{
				// if we do not have the key to decrypt the packet, drop it.
				if(Adapter->SecurityInfo.InWpaMode && Adapter->SecurityInfo.PairwiseKeyAvailable && (CipherSuite == CIPHER_SUITE_NONE))
				{
					SetRxDescFlags(Adapter, Adapter->RxDescHead, RX_DESC_FLAG_CONSUMED);
					
					if( Adapter->RxDescHead < (Adapter->HostInfo.RxDescCount - 1 ))
						Adapter->RxDescHead++;					
					else
						Adapter->RxDescHead = 0;
					
					GetRxDescFlags(Adapter, Adapter->RxDescHead, &flags);

					continue;
				}

				// examine Key Replay or double MIC error first
				if( ( (CipherSuite == CIPHER_SUITE_TKIP) &&
					  (Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_TKIP)
					)
#ifdef AES
					||
					( (CipherSuite == CIPHER_SUITE_AES) &&
					  (Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_AES)
					)
#endif
				  )
				  	InWPAMode = TRUE;

				if( (InWPAMode && (Adapter->SecurityInfo.MicErrorCnt == 2)))
				{
					KeyReplay = 0;
					GetRxDescKeyReplay(Adapter, Adapter->RxDescHead, &KeyReplay);

					if(KeyReplay || Adapter->SecurityInfo.MicErrorCnt == 2)
					{
						IF_LOUD(DbgPrint("Rx Packet Dropped \n");)

						if(InWPAMode && KeyReplay)
							Adapter->Stats.RxWpaReplayErrors ++;

						SetRxDescFlags(Adapter, Adapter->RxDescHead, RX_DESC_FLAG_CONSUMED);
						
						if( Adapter->RxDescHead < (Adapter->HostInfo.RxDescCount - 1 ))
							Adapter->RxDescHead++;					
						else
							Adapter->RxDescHead = 0;
						
						GetRxDescFlags(Adapter, Adapter->RxDescHead, &flags);

						continue;
					}
				}
		
				// Since all the initial checks are passed, copy the rest of the packet
				CardMemToHostMem16(Adapter, RxBuf, RxPacketLoc, msduSize);
				
				if(msduSize > 33){ // 30:header + 3:eap+version
					IF_LOUD(ParseEAPOLPacket(NULL, NULL, &RxBuf[30], TRUE, TRUE);)
         }
			

				// if the WPA cipher suite is not implemented in h/w check MIC errors
#ifndef HWWPA
				if( ( (CipherSuite == CIPHER_SUITE_TKIP) &&
					  (Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_TKIP)
					)
#ifdef AES
					||
					( (CipherSuite == CIPHER_SUITE_AES) &&
					  (Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_AES)
					)
#endif
				  )
				{
					if(Adapter->OperatingMode == AD_HOC_MODE)
						RxMicKeyOffset = 16;
					else
						RxMicKeyOffset = 24;

					if(!WPA_validateMic(RxBuf, msduSize, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[KeyIndex][RxMicKeyOffset]))
					{
						IF_LOUD(DbgPrint("### MIC Error ### (KeyIndex = %d - CipherSuite = %d (DA: %02X:%02X:%02X:%02X:%02X:%02X - BSSID: %02X:%02X:%02X:%02X:%02X:%02X - SA: %02X:%02X:%02X:%02X:%02X:%02X)))\n", KeyIndex, CipherSuite, RxBuf[4],
							RxBuf[5],RxBuf[6],RxBuf[7],RxBuf[8],RxBuf[9],
							RxBuf[10], RxBuf[11],RxBuf[12],RxBuf[13],RxBuf[14],RxBuf[15],
							RxBuf[16], RxBuf[17],RxBuf[18],RxBuf[19],RxBuf[20],RxBuf[21]);)
						
						Adapter->Stats.RxDataPacketsError++;
						Adapter->SecurityInfo.MicErrorCnt++;
						Adapter->Stats.RxWpaMicErrors ++;


						if(RxBuf[4]  & 0x01 )
							SendAuthenticationEvent(Adapter, __802_11_AUTH_REQUEST_GROUP_ERROR);
						else
							SendAuthenticationEvent(Adapter, __802_11_AUTH_REQUEST_PAIRWISE_ERROR);							

						if(Adapter->SecurityInfo.MicErrorCnt == 1)
						{
							IF_LOUD(DbgPrint("\n\n\n FIRST MIC ERROR\n");)
//							Adapter->SecurityInfo.MicErrorDetected = TRUE;								
						}
						else
						if(Adapter->SecurityInfo.MicErrorCnt == 2)
						{
							IF_LOUD(DbgPrint("\n\n\n SECOND MIC ERROR\n");)

							Adapter->Stats.RxWpaCountermeasuresCount ++;

							if( WPA_AddAPInBlackList(Adapter->SecurityInfo.MicBlackList, MAX_BSS_ENTRIES, &Adapter->SecurityInfo.MicBlackListEntries, Adapter->AssociationInfo.CurrentBSSID, (PVOID)&TimeStamp) )
								Adapter->Stats.WpaBlackListEntries ++;
							else
								IF_LOUD(DbgPrint("BLACK LIST: New entry ignored\n");)

						}

						SetRxDescFlags(Adapter, Adapter->RxDescHead, RX_DESC_FLAG_CONSUMED);
						
						if( Adapter->RxDescHead < (Adapter->HostInfo.RxDescCount - 1 ))
							Adapter->RxDescHead++;					
						else
							Adapter->RxDescHead = 0;
						
						GetRxDescFlags(Adapter, Adapter->RxDescHead, &flags);

						continue;

					}
				}
#endif //HWWPA

				// find and cut off the extra bytes of MIC
				if( 
					(CipherSuite == CIPHER_SUITE_TKIP)
#ifdef AES
					|| (CipherSuite == CIPHER_SUITE_AES)
#endif
				  )
				  	msduSize-=(USHORT)8;

				//
				// Source Address resides at offset 10 or 16 according to OpMode
				// Destination Address is alway at offset 4 and we do not overwrite it
				//
				if(Adapter->OperatingMode == AD_HOC_MODE)
					VNetMoveMemory(SourceAddr, &RxBuf[10], 6);
				else
					VNetMoveMemory(SourceAddr, &RxBuf[16], 6);

				if (( !MemCompare(&RxBuf[24], SNAP_ETHERNETII, 6) &&
					  !MemCompare(&RxBuf[30], STATIC_ENTRIES_ETHERTYPE_8137, 2)
					) ||
					( RxBuf[24] == 0xaa &&
					   RxBuf[25] == 0xaa &&
					   MemCompare(&RxBuf[30], STATIC_ENTRIES_ETHERTYPE_8137, 2) &&
					   MemCompare(&RxBuf[30], STATIC_ENTRIES_ETHERTYPE_80F3, 2)
					)
				   )
				{
					
					//
					// We will add 12 bytes (DA(6)+SA(6)) and since the Addr4 is NOT used,
					// 802.3 FrameBody starts at an offset of +30 in the 802.11 Frame. 
					// We must start creating the 802.3 header at an offset of +18 (18 + 12 = 30)
					//
					StartOfEthPacket = 18;

					PacketLength = msduSize-18;

					RxPacketLoc += 18;

					//
					//DA
					//
					VNetMoveMemory(&RxBuf[StartOfEthPacket], &RxBuf[4], 6);				
					//
					//SA
					//
					VNetMoveMemory(&RxBuf[StartOfEthPacket+6], SourceAddr, 6);

				}
				else //IP packet
				{
					//
					// We will add 14 bytes (DA(6)+SA(6)+Length(2)) and since
					// the Addr4 is used, 802.3 FrameBody starts at an offset of +24 in the 802.11
					// Frame. We must start creating the 802.3 header at an offset of +10 (10+14 = 24) 
					//
					StartOfEthPacket = 10;  

					PacketLength = msduSize-10;

					RxPacketLoc += 10;
					//
					//DA
					//
					VNetMoveMemory(&RxBuf[StartOfEthPacket], &RxBuf[4], 6);				
					//
					//SA
					//
					VNetMoveMemory(&RxBuf[StartOfEthPacket+6], SourceAddr, 6);
					//
					// Length
					//
					RxBuf[StartOfEthPacket + 12]= (UCHAR) ((msduSize-24)>>8);
					RxBuf[StartOfEthPacket + 13]= (UCHAR) (msduSize-24);
				}

				Adapter->RxPacketLen = PacketLength;
		

				// Lookahead amount to indicate			
				IndicateLen = PacketLength;

				RxIndicatePacket(Adapter, (RxBuf+StartOfEthPacket), PacketLength, IndicateLen);
				Adapter->Stats.RxDataPacketsOk++;

				Adapter->IndicateReceiveDone = TRUE;

			}
			//ENDNEW
		}
		else
		{
#ifdef HWWPA
			if(status == 0xC7)
			{
				IF_LOUD(DbgPrint("### MIC Error ###\n");)
			
				CardMemToHostMem16(Adapter, (PUCHAR)&RxPacketLoc, (Adapter->HostInfo.RxDescPos + (SIZE_OF_RXDESC * Adapter->RxDescHead) + RX_DESC_MSDU_POS_OFFSET), 2);

				CardMemToHostMem16(Adapter, RxBuf, RxPacketLoc, msduSize);

				Adapter->Stats.RxDataPacketsError++;
				Adapter->SecurityInfo.MicErrorCnt++;
				Adapter->Stats.RxWpaMicErrors ++;


				if(RxBuf[4]  & 0x01 )
					SendAuthenticationEvent(Adapter, 1);
				else
					SendAuthenticationEvent(Adapter, 2);							

				if(Adapter->SecurityInfo.MicErrorCnt == 1)
				{
					IF_LOUD(DbgPrint("\n\n\n FIRST MIC ERROR\n");)
//					Adapter->SecurityInfo.MicErrorDetected = TRUE;								
				}
				else
				if(Adapter->SecurityInfo.MicErrorCnt == 2)
				{
					IF_LOUD(DbgPrint("\n\n\n SECOND MIC ERROR\n");)

					Adapter->Stats.RxWpaCountermeasuresCount ++;

					if( WPA_AddAPInBlackList(Adapter->SecurityInfo.MicBlackList, MAX_BSS_ENTRIES, &Adapter->SecurityInfo.MicBlackListEntries, Adapter->AssociationInfo.CurrentBSSID, (PVOID)&TimeStamp) )
						Adapter->Stats.WpaBlackListEntries ++;
					else
						IF_LOUD(DbgPrint("BLACK LIST: New entry ignored\n");)
				}

				SetRxDescFlags(Adapter, Adapter->RxDescHead, RX_DESC_FLAG_CONSUMED);
				
				if( Adapter->RxDescHead < (Adapter->HostInfo.RxDescCount - 1 ))
					Adapter->RxDescHead++;					
				else
					Adapter->RxDescHead = 0;
				
				GetRxDescFlags(Adapter, Adapter->RxDescHead, &flags);

				continue;

			}
			else if(Adapter->SecurityInfo.MicErrorCnt == 2)
			{ // drop rx packets
				IF_LOUD(DbgPrint("Rx Packet Dropped\n");)
				SetRxDescFlags(Adapter, Adapter->RxDescHead, RX_DESC_FLAG_CONSUMED);
				
				if( Adapter->RxDescHead < (Adapter->HostInfo.RxDescCount - 1 ))
					Adapter->RxDescHead++;					
				else
					Adapter->RxDescHead = 0;
				
				GetRxDescFlags(Adapter, Adapter->RxDescHead, &flags);

				continue;
			}
#endif
		}
		
		//
		// Release Descriptor
		//

		SetRxDescFlags(Adapter, Adapter->RxDescHead, RX_DESC_FLAG_CONSUMED);

		if( Adapter->RxDescHead < (Adapter->HostInfo.RxDescCount - 1 ))
			Adapter->RxDescHead++;					
		else
			Adapter->RxDescHead = 0;

		GetRxDescFlags(Adapter, Adapter->RxDescHead, &flags);
	}


	//
	// Indicate ReceiveComplete
	//
	if (Adapter->IndicateReceiveDone)
	{
		Adapter->IndicateReceiveDone = FALSE;
	}
}



/////////////////////////////////////////////////



#ifdef RX_CRC
// CRC
VOID
init_CRCtable (PVNet_ADAPTER Adapter)
{
  ULONG crc;
  int i, j;

  for (i = 0; i < 256; i++)
    {
      crc = i;
      for (j = 8; j > 0; j--)
	{
	  if (crc & 1)
	    crc = (crc >> 1) ^ 0xedb88320;
	  else
	    crc >>= 1;
	}
      Adapter->CrcTable[i] = crc;
    }
}

ULONG
Calculate_CRC32 (PUCHAR p, ULONG FrameLength, PULONG crctab)
{
  ULONG crcv, t;
  ULONG nr = 0;

  crcv = 0xffffffff;
  while (nr < FrameLength)
    {
      t = (crcv ^ p[nr]) & 0xff;
      crcv = ((crcv >> 8) & 0x00ffffff) ^ crctab[t];
      nr++;
    }
  crcv = (crcv ^ 0xffffffff);

  return (ULONG) crcv;
}
#endif
