/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/


/**************************************************************************/
/**************************************************************************/
/**                                                                       */
/**   Functions used for Tx												  */
/**                                                                       */
/**************************************************************************/
/**************************************************************************/

#include "vnet.h"
#include "tx.h"

UCHAR	SNAP_RFC1024[6] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00};
extern UCHAR	SNAP_ETHERNETII[6];
USHORT STATIC_ENTRIES_ETHERTYPES[2] = {0x8137, 0x80F3};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    TxInit                                                                                                                      
//                                                                                                                                            
//    This function initializes Tx related parameters in the Adapter      
//        object and enables Tx operation on the MAC through the coresponding 
//        field of the IFACE.                                                                                             
//                                                                         
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VOID
TxInit (PVNet_ADAPTER Adapter)
{
  UCHAR configuration;
  ULONG next;

  IF_VERY_LOUD (DbgPrint ("Initializing Tx Section....\n");
    )next = (ULONG) 0x80000000;

  SetTxDescFlags (Adapter, 0, 0);
  SetTxDescNext (Adapter, 0, next);
  SetTxDescMsduPos (Adapter, 0, 0);
  SetTxDescMsduSize (Adapter, 0, 0);

  //
  //      Initialize Tx Descriptors counters and pointers
  //
  Adapter->TxDescFree = Adapter->HostInfo.TxDescCount;
  Adapter->TxDescHead = 0;
  Adapter->TxDescTail = 0;
  Adapter->TxDescPrevious = 0;
  //
  //      Initialize TxBuffers counters and pointers.
  //
  Adapter->TxFreeBuffMem = Adapter->HostInfo.TxBuffSize;
  Adapter->TxBuffHead = 0;
  Adapter->TxBuffTail = 0;

  Adapter->TxBuffStart = Adapter->HostInfo.TxBuffPos;
  Adapter->TxBuffEnd =
    Adapter->HostInfo.TxBuffPos + Adapter->HostInfo.TxBuffSize;

  Adapter->CompletedTransmissions = 0;
  Adapter->Transmissions = 0;
  //
  //      Enable Tx Operation on the card.
  //
  GetFuncCtrl (Adapter, &configuration);
  configuration |= FUNC_CTRL_TxENABLE;
  SetFuncCtrl (Adapter, &configuration);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    TxCompleteDPC                                                                                                               
//
//    Called after a Tx Complete interrupt. It checks the status of the   
//        transmission and completes the send.                                                            
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VOID
TxCompleteDPC (PVNet_ADAPTER Adapter)
{
  UCHAR flags, status;
  USHORT MsduSize;
  UCHAR PacketType;
  UCHAR loopcnt;
  UCHAR                           ReportReason = 0;
          USHORT                          ReasonCode;



  if (Adapter->StationState == STATION_STATE_RESETTING ||
      Adapter->StationState == STATION_STATE_RESET_PENDING)
    {
      return;
    }
  GetTxDescFlags (Adapter, Adapter->TxDescHead, &flags);

  loopcnt = 0;
  while (flags == TX_DONE && flags != 0xff
	 && loopcnt < Adapter->HostInfo.TxDescCount)
    {
      loopcnt++;

      Adapter->CompletedTransmissions++;
      GetTxDescStatus (Adapter, Adapter->TxDescHead, &status);

      GetTxDescHostMsduSize (Adapter, Adapter->TxDescHead, &MsduSize);
      GetTxDescPacketType (Adapter, Adapter->TxDescHead, &PacketType);

#ifndef ATMEL_WLAN
      Adapter->TxFreeBuffMem += (USHORT) MsduSize;
#endif
      VNET_ASSERT (Adapter->TxFreeBuffMem <= Adapter->HostInfo.TxBuffSize);

#ifndef ATMEL_WLAN
      if (Adapter->TxBuffHead + MsduSize > Adapter->TxBuffEnd)
	Adapter->TxBuffHead = 0;
      else
	Adapter->TxBuffHead += (USHORT) MsduSize;
#endif
      SetTxDescFlags (Adapter, Adapter->TxDescHead, 0);

      if (Adapter->TxDescHead < (Adapter->HostInfo.TxDescCount - 1))
	Adapter->TxDescHead++;
      else
	Adapter->TxDescHead = 0;

      Adapter->TxDescFree++;

      VNET_ASSERT (Adapter->TxDescFree <= Adapter->HostInfo.TxDescCount);


      if (PacketType == TX_PACKET_TYPE_DATA)
	{
	  if (status == TX_STATUS_SUCCESS)
	    Adapter->Stats.TxDataPacketsOk++;
	  else
	    Adapter->Stats.TxDataPacketsError++;

	    TxIndicateCompletion (Adapter);
			if(ReportReason == EAPOL_MIC_ERROR && Adapter->SecurityInfo.MicErrorCnt == 2)
			{
				RemoveEncryptionSettings(Adapter);

				ReasonCode = 0x0E;
				SendDeAuthenticationRequest(Adapter, ReasonCode);

				Adapter->StationState = STATION_STATE_WAIT_ACTION;
				Adapter->StationIsAssociated = FALSE;
			}
	}
      else if (PacketType == TX_PACKET_TYPE_MGMT)
	{
	  if (status == TX_STATUS_SUCCESS)
	    Adapter->Stats.TxMgmtPacketsOk++;
	  else
	    Adapter->Stats.TxMgmtPacketsError++;
	}
      GetTxDescFlags (Adapter, Adapter->TxDescHead, &flags);
    }
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    TxEthernetPacket                                                                                                    
//                                                                                                                                            
//    This function gets an Ethernet (802.3) packet (at once or in                
//        fragments, converts it to 802.11 and copies it on the card to be    
//    transmitted and then calls TxUpdateDescriptor to initiate                   
//    thr transmission.                                                                                                   
//                                                                         
/////////////////////////////////////////////////////////////////////////////
UCHAR 
TxEthernetPacket(PVNet_ADAPTER Adapter,
					 PUCHAR pTxBuf,
					 ULONG CurrentBytes,
					 PULONG TotalBytes,
					 USHORT StartOfTxBuf,
					 BOOLEAN IsBroadcast)
{
	s80211Header	MacHeader;
	UCHAR			WirelessPacket[MAX_WIRELESS_FRAME_SIZE];
	USHORT			EtherType;
	ULONG			offset;
	UCHAR			KeyID;
	//ULONG			tmp;
	UCHAR			ParseEapStatus;

#ifdef CCX
	BOOLEAN			IsPairwise;
#endif

	IF_DEBUG_TX_VERBOSE(DbgPrint("Tx: %d/%d Bytes to 0x%X \n", CurrentBytes, *TotalBytes, StartOfTxBuf);)
	//
	// The Wireless Header is 30 bytes. In the Ethernet packet we "cut" the
	// 12 first bytes (containing DA/SA) and put them in the appropriate fields of
	// the Wireless Header. Thus the packet length now is the initial + 18 (+30-12)
	//

	offset = 0;

	MacHeader.FrameControl[0] =  C80211_TYPE_DATA;
	MacHeader.DurationID = 0x00;		

	if (Adapter->OperatingMode	== AD_HOC_MODE)
	{
		// do not set the encryption flag in EAP packets
		if( IsEncryptionEnabled(&Adapter->SecurityInfo) && !(pTxBuf[12] == 0x88 && pTxBuf[13] == 0x8E) )
			MacHeader.FrameControl[1] = C80211_FRMCTRL_ENCRYPTION;
		else			
			MacHeader.FrameControl[1] = 0x00;

		VNetMoveMemory(MacHeader.uAddress1, pTxBuf, VNet_LENGTH_OF_ADDRESS);			
		VNetMoveMemory(MacHeader.uAddress2, Adapter->AssociationInfo.CurrentAddress, VNet_LENGTH_OF_ADDRESS);
		VNetMoveMemory(MacHeader.uAddress3, Adapter->BSSID, VNet_LENGTH_OF_ADDRESS);
	}
	else
	{
		IsBroadcast = FALSE;

		if( IsEncryptionEnabled(&Adapter->SecurityInfo) && !(pTxBuf[12] == 0x88 && pTxBuf[13] == 0x8E) ) 
			MacHeader.FrameControl[1] = C80211_FRMCTRL_TO_DS | C80211_FRMCTRL_ENCRYPTION ;
		else
			MacHeader.FrameControl[1] = C80211_FRMCTRL_TO_DS;
///////////////
		if(Adapter->SecurityInfo.InWpaMode && Adapter->SecurityInfo.PairwiseKeyAvailable && IsEncryptionEnabled(&Adapter->SecurityInfo))
			MacHeader.FrameControl[1] = C80211_FRMCTRL_TO_DS | C80211_FRMCTRL_ENCRYPTION;
///////////////////
		VNetMoveMemory(MacHeader.uAddress1, Adapter->AssociationInfo.CurrentBSSID, VNet_LENGTH_OF_ADDRESS);
		VNetMoveMemory(MacHeader.uAddress2, Adapter->AssociationInfo.CurrentAddress, VNet_LENGTH_OF_ADDRESS);
		VNetMoveMemory(MacHeader.uAddress3, pTxBuf, VNet_LENGTH_OF_ADDRESS);
		MacHeader.SequenceControl = 0;
	}

	VNetMoveMemory(WirelessPacket, (PUCHAR)&MacHeader, 24);
	offset += 24;

	//EtherType = ((UINT16) pTxBuf[12]) << 8;
	//EtherType |= (UINT16) pTxBuf[13];
	EtherType = ((USHORT) pTxBuf[12]) << 8;
	EtherType |= (USHORT) pTxBuf[13];

#ifdef CCX
	// fill in the wireless packet after the wireless header.
	// if the CMIC is disable we handle the packet as WEP packet.
	if( Adapter->CCXinfo.CkipEnabled && Adapter->CCXinfo.CkipMicEnabled)
	{
		if ( pTxBuf[12] == 0x88 && pTxBuf[13] == 0x8E ) 
		{
			VNetMoveMemory(&WirelessPacket[offset], SNAP_RFC1024, 6);
			offset += 6;

			VNetMoveMemory(&WirelessPacket[offset], &pTxBuf[12], CurrentBytes-12);

			offset += (CurrentBytes-12);
			*TotalBytes	+= (USHORT)18;	//24+6-12
		}
		else
		{
			
			VNetMoveMemory(&WirelessPacket[offset], SNAP_CCX, 8);
			offset += 8;

			// leave 4 bytes for MIC
			offset += 4; //MIC

			// SEQ
			tmp = ReverseUlong(Adapter->CCXinfo.CkipSeqCnt);
			VNetMoveMemory(&WirelessPacket[offset], &tmp, 4);
			offset += 4;

			VNetMoveMemory(&WirelessPacket[offset], &pTxBuf[12], CurrentBytes-12);

			offset += (CurrentBytes-12);
			*TotalBytes	+= (USHORT)28;	//24+8+4+4-12
			
		}
	}
	else
#endif

	
	if ((EtherType == (USHORT)STATIC_ENTRIES_ETHERTYPES[0]) ||(EtherType == (USHORT)STATIC_ENTRIES_ETHERTYPES[1])) 
	{
		VNetMoveMemory(&WirelessPacket[offset], SNAP_ETHERNETII, 6);
		offset += 6;
	
		VNetMoveMemory(&WirelessPacket[offset], &pTxBuf[12], CurrentBytes-12);

		offset += (CurrentBytes-12);
		*TotalBytes	+= (USHORT)18; //24+6-12

	}
	else if (EtherType > 1536) 
	{
		
		VNetMoveMemory(&WirelessPacket[offset], SNAP_RFC1024, 6);
		offset += 6;

		VNetMoveMemory(&WirelessPacket[offset], &pTxBuf[12], CurrentBytes-12);

		offset += (CurrentBytes-12);
		*TotalBytes	+= (USHORT)18;	//24+6-12

	}
	else
	{
		VNetMoveMemory(&WirelessPacket[offset], &pTxBuf[14], EtherType);

		offset += EtherType;

		*TotalBytes += 10;  //24+0-12-2
	}
		

#ifdef CCX
	// add the 4 bytes of MIC, if the CMIC is enable and 
	// the packet is not EAP
	if( Adapter->CCXinfo.CkipEnabled && Adapter->CCXinfo.CkipMicEnabled )
	{
		if ( pTxBuf[12] != 0x88 || pTxBuf[13] != 0x8E )
		{
			// We decide if we are going to use the pairwise or the group key for CMIC addition.
			// The check of the key availability is necessary since the indexs of the
			// pairwise and group key could contain a valid value even if they are not available.
			if(!IsBroadcast && (Adapter->SecurityInfo.PairwiseKeyAvailable == 1))
				IsPairwise = TRUE;
			else
				IsPairwise = FALSE;

			CCX_AddMIC(WirelessPacket, (USHORT)(*TotalBytes), IsPairwise);
			Adapter->CCXinfo.CkipSeqCnt+=2;
		}
	}
	else
#endif
	
	if(Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_TKIP 
#ifdef AES
	   || Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_AES
#endif
	  )
	{
		KeyID = 0;
		if(Adapter->OperatingMode == AD_HOC_MODE)
		{
			if(Adapter->SecurityInfo.GroupKeyAvailable && !Adapter->SecurityInfo.PairwiseKeyAvailable)
				KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex & 0xff);
			else
			if(!Adapter->SecurityInfo.GroupKeyAvailable && Adapter->SecurityInfo.PairwiseKeyAvailable)
			{
				if(IsBroadcast)
					KeyID = 1; //Pairwise is always to index 0 otherwise add_key oid will fail
				else
					KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex & 0xff);
			}
			else
			{
				if(IsBroadcast)
					KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex & 0xff);
				else
					KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex & 0xff);
			}
		}
		else
		{

			if(Adapter->SecurityInfo.PairwiseKeyAvailable)
				KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex & 0xff);
			else
			{
				if(Adapter->SecurityInfo.GroupKeyAvailable)
					KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex & 0xff);
			}

		}
		
#ifdef AES
	   if(Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite != CIPHER_SUITE_AES)
#endif
		*TotalBytes	+= (USHORT)8;

#ifndef HWWPA
		if(Adapter->OperatingMode == AD_HOC_MODE)
			WPA_addMICToPacket(WirelessPacket, (USHORT)*TotalBytes, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[KeyID][16]);
		else
			WPA_addMICToPacket(WirelessPacket, (USHORT)*TotalBytes, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[KeyID][16]);
#endif
		offset += 8;

	}

#ifndef UNDER_CE	
		if((*TotalBytes) > 33) // 30:header + 3:eap+version
		{
			ParseEAPOLPacket(&ParseEapStatus, NULL, &WirelessPacket[30], FALSE, TRUE);
			if(ParseEapStatus == EAPOL_MIC_ERROR)
				SetTxDescReportReason(Adapter, Adapter->TxDescTail, ParseEapStatus);
		}
#endif

	// if not in WPA mode, we must send the EAP packets unencrypted.
	// we mark these packets here and in TxUpdateDescriptor we set the 
	// cipher suite with no encryption.
	if ( pTxBuf[12] == 0x88 && pTxBuf[13] == 0x8E && !Adapter->SecurityInfo.InWpaMode)
	{
		IF_DEBUG_TX_VERBOSE(DbgPrint("Tx: EAP Packet pending\n");)
		Adapter->TxEapPending = 1;
	}
	
	// copy the packet to the card.
	if(TxResourcesAvailable(Adapter, *TotalBytes, &StartOfTxBuf))
		HostMemToCardMem16(Adapter, StartOfTxBuf, WirelessPacket, *TotalBytes);

	return VNET_TX_STATUS_SUCCESS;
}
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//  TxResourcesAvailable                                                                                     
//                                                                                                                  
//  Returns TRUE if there is enough space & descriptors for a frame to be 
//      transmitted and also the starting address that this frame must be copied.
//     If there are no resources available returns FALSE                                             
//                                                                         
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

#if 0
BOOLEAN
TxResourcesAvailable (PVNet_ADAPTER Adapter, ULONG Bytes,
		      PUSHORT StartOfTxBuf)
{
  USHORT FreeBytesAtBottom;

  FreeBytesAtBottom = Adapter->HostInfo.TxBuffSize - Adapter->TxBuffTail;

  if (Adapter->TxDescFree == 3)
    {
      return FALSE;
    }

  if (Adapter->TxFreeBuffMem < Bytes)	//no space at all
    {
      return FALSE;
    }
  else				// Free space but where?
  if (FreeBytesAtBottom < Bytes)	//No Space at bottom
    {
      if (Adapter->TxFreeBuffMem - FreeBytesAtBottom >= (USHORT) Bytes)	//Ok at top
	{
	  Adapter->TxBuffTail = 0;
	  *StartOfTxBuf = Adapter->HostInfo.TxBuffPos;
	  return TRUE;
	}
      else			// no space at bottom & at top
	{
	  return FALSE;
	}
    }
  else
    {
      *StartOfTxBuf = Adapter->TxBuffStart + Adapter->TxBuffTail;
      return TRUE;
    }
}
#endif
BOOLEAN 
TxResourcesAvailable(PVNet_ADAPTER Adapter, ULONG Bytes, PUSHORT StartOfTxBuf)
{
	
	if(Bytes > 1600)
	{
		IF_LOUD(DbgPrint("*** Packet Rejected (size = %d) ***\n", Bytes);)
		return FALSE;
	}

	if(Adapter->TxDescFree == 1)
	{
//		IF_LOUD(DbgPrint("*** Packet Rejected (descriptors) ***\n");)

		return FALSE;
	}

	*StartOfTxBuf = Adapter->HostInfo.TxBuffPos + (Adapter->TxDescTail * 1600);
	
	return TRUE;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    TxUpdateDescriptor                                                                                                  
//
//    This function updates the Tx descriptors to transmit a frame                
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


VOID
TxUpdateDescriptor (PVNet_ADAPTER Adapter, ULONG TotalBytes,
		    USHORT StartOfTxBuffer, BOOLEAN IsData, BOOLEAN IsBroadcast)
{

#ifdef ATMEL_WLAN
        UCHAR	CipherType, CipherLength;
#endif
		IF_VERY_LOUD(DbgPrint(
		       "TxUpdateDescriptor %d to %d TxDescCount : %d\n", TotalBytes, StartOfTxBuffer, Adapter->HostInfo.TxDescCount);)
		       //"TxUpdateDescriptor %d to 0x%X\n", TotalBytes, StartOfTxBuffer);)
  SetTxDescMsduPos (Adapter, Adapter->TxDescTail, StartOfTxBuffer);
  SetTxDescMsduSize (Adapter, Adapter->TxDescTail, (USHORT) TotalBytes);
//#ifndef ATMEL_WLAN
  SetTxDescHostMsduSize (Adapter, Adapter->TxDescTail, (USHORT) TotalBytes);
//#endif

  if (IsData)
    SetTxDescPacketType (Adapter, Adapter->TxDescTail, TX_PACKET_TYPE_DATA);
  else
    SetTxDescPacketType (Adapter, Adapter->TxDescTail, TX_PACKET_TYPE_MGMT);

  if (Adapter->DataFramesTxRate != 4)
    SetTxDescRate (Adapter, Adapter->TxDescTail, Adapter->DataFramesTxRate);
  else
    SetTxDescRate (Adapter, Adapter->TxDescTail, 3);

  SetTxDescRetryCount (Adapter, Adapter->TxDescTail, 0);

        // We send the unicast packets with the pairwise key if available, else we send 
	// them using the group key. Also, we send the multicast packets with the group key
	// if available , else if the encryption is WEP we use the pairwise key.

#ifdef ATMEL_WLAN
	if(IsBroadcast) {

		CipherType = Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite;
		if(Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite == CIPHER_SUITE_WEP_64 || Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite == CIPHER_SUITE_WEP_128)
			CipherLength = 8;
		else if(Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite == CIPHER_SUITE_TKIP)
			CipherLength = 12;
		else {
			CipherType = CIPHER_SUITE_NONE;
			CipherLength = 0;
		}
    
		if( CipherType == CIPHER_SUITE_NONE && 
			(Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_WEP_64 || 
			 Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_WEP_128)
		  )
		{
			CipherType = Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite;
			CipherLength = 8;
		}
		
	}
	else {
		CipherType = Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite;
		if(Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_WEP_64 || Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_WEP_128){
			CipherLength = 8;
    }
		else if(Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_TKIP)
			CipherLength = 12;
		else {
			CipherType = CIPHER_SUITE_NONE;
			CipherLength = 0;
		}

		if(CipherType == CIPHER_SUITE_NONE)
		{
			CipherType = Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite;
			if(Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite == CIPHER_SUITE_WEP_64 || Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite == CIPHER_SUITE_WEP_128)
				CipherLength = 8;
			else
			if(Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite == CIPHER_SUITE_TKIP)
				CipherLength = 12;
			else
			{
				CipherType = CIPHER_SUITE_NONE;
				CipherLength = 0;
			}
		}

	}

        SetTxDescCipherType(Adapter, Adapter->TxDescTail, CipherType);
	SetTxDescCipherLength(Adapter, Adapter->TxDescTail, CipherLength);

#endif // ATMEL_WLAN

  SetTxDescNext (Adapter, Adapter->TxDescTail, 0x80000000);
  SetTxDescFlags (Adapter, Adapter->TxDescTail, TX_FIRM_OWN);

  if (Adapter->TxDescPrevious != Adapter->TxDescTail)
    SetTxDescNext (Adapter, Adapter->TxDescPrevious, 0x00);


  Adapter->TxDescPrevious = Adapter->TxDescTail;

  if (Adapter->TxDescTail < (Adapter->HostInfo.TxDescCount - 1))
    Adapter->TxDescTail++;
  else
    Adapter->TxDescTail = 0;


  Adapter->TxDescFree--;

 // Adapter->TxFreeBuffMem -= (USHORT) TotalBytes;

  //VNET_ASSERT (Adapter->TxFreeBuffMem <= Adapter->HostInfo.TxBuffSize);
  Adapter->Transmissions++;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    TxIndicateCompletion                                                                                                
//
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VOID
TxIndicateCompletion (PVNet_ADAPTER Adapter)
{
  VNet_netif_wake_queue (Adapter);
}

//////////////////////////////////////////////////
//
// Tx Descriptors manipulation functions
//
//////////////////////////////////////////////////

void
GetTxDescFlags (PVNet_ADAPTER Adapter, USHORT Descriptor, PUCHAR flags)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_FLAGS_OFFSET));
  CardReadUchar (Adapter, DR, flags);
}


void
GetTxDescStatus (PVNet_ADAPTER Adapter, USHORT Descriptor, PUCHAR status)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_STATUS_OFFSET));
  CardReadUchar (Adapter, DR, status);
}


void
SetTxDescFlags (PVNet_ADAPTER Adapter, USHORT Descriptor, UCHAR flags)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_FLAGS_OFFSET));
  CardWriteUchar (Adapter, DR, flags);
}



void
SetTxDescRetryCount (PVNet_ADAPTER Adapter, USHORT Descriptor, UCHAR cnt)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_RETRY_OFFSET));
  CardWriteUchar (Adapter, DR, cnt);
}

void
SetTxDescNext (PVNet_ADAPTER Adapter, USHORT Descriptor, ULONG next)
{
  HostMemToCardMem16 (Adapter,
		      Adapter->HostInfo.TxDescPos +
		      (SIZE_OF_TXDESC * Descriptor) + TX_DESC_NEXT_OFFSET,
		      (PUCHAR) & next, 4);
}

#ifdef ATMEL_WLAN

void SetTxDescCipherType(PVNet_ADAPTER Adapter, USHORT Descriptor, UCHAR CipherType)
{
	HostMemToCardMem8(Adapter, (Adapter->HostInfo.TxDescPos + (SIZE_OF_TXDESC * Descriptor) + TX_DESC_CIPHER_TYPE_OFFSET), &CipherType, 1);
}

void SetTxDescCipherLength(PVNet_ADAPTER Adapter, USHORT Descriptor, UCHAR CipherLen)
{
	HostMemToCardMem8(Adapter, (Adapter->HostInfo.TxDescPos + (SIZE_OF_TXDESC * Descriptor) + TX_DESC_CIPHER_LENGTH_OFFSET), &CipherLen, 1);
}
 
void GetTxDescKeyIndex(PVNet_ADAPTER Adapter, USHORT Descriptor, PUCHAR Index)
{
	CardMemToHostMem8(Adapter, Index, (Adapter->HostInfo.TxDescPos + (SIZE_OF_TXDESC * Descriptor) + TX_DESC_KEY_INDEX_OFFSET), 1);
}

void GetTxDescCipherType(PVNet_ADAPTER Adapter, USHORT Descriptor, PUCHAR CipherType)
{
	CardMemToHostMem8(Adapter, CipherType, (Adapter->HostInfo.TxDescPos + (SIZE_OF_TXDESC * Descriptor) + TX_DESC_CIPHER_TYPE_OFFSET), 1);
}
#endif // ATMEL_WLAN

void
SetTxDescMsduPos (PVNet_ADAPTER Adapter, USHORT Descriptor, USHORT msdupos)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_POS_OFFSET));
  CardWriteUshort (Adapter, DR, msdupos);
}

void
GetTxDescMsduSize (PVNet_ADAPTER Adapter, USHORT Descriptor, PUSHORT msdusize)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_SIZE_OFFSET));
  CardReadUshort (Adapter, DR, msdusize);
}


void
SetTxDescMsduSize (PVNet_ADAPTER Adapter, USHORT Descriptor, USHORT msdusize)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_SIZE_OFFSET));
  CardWriteUshort (Adapter, DR, msdusize);
}


void
SetTxDescRate (PVNet_ADAPTER Adapter, USHORT Descriptor, UCHAR RateIndex)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_RATE_OFFSET));
  CardWriteUchar (Adapter, DR, RateIndex);
}


void
GetTxDescHostMsduSize (PVNet_ADAPTER Adapter, USHORT Descriptor,
		       PUSHORT msdusize)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_HOST_LENGTH_OFFSET));
  CardReadUshort (Adapter, DR, msdusize);
}


void
SetTxDescHostMsduSize (PVNet_ADAPTER Adapter, USHORT Descriptor,
		       USHORT msdusize)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_HOST_LENGTH_OFFSET));
  CardWriteUshort (Adapter, DR, msdusize);
}


void
GetTxDescPacketType (PVNet_ADAPTER Adapter, USHORT Descriptor,
		     PUCHAR PacketType)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_PACKET_TYPE_OFFSET));
  CardReadUchar (Adapter, DR, PacketType);
}


void
SetTxDescPacketType (PVNet_ADAPTER Adapter, USHORT Descriptor,
		     UCHAR PacketType)
{
  CardWriteUshort (Adapter, AR,
		   (USHORT) (Adapter->HostInfo.TxDescPos +
			     (SIZE_OF_TXDESC * Descriptor) +
			     TX_DESC_PACKET_TYPE_OFFSET));
  CardWriteUchar (Adapter, DR, PacketType);
}

void SetTxDescReportReason(PVNet_ADAPTER Adapter, USHORT Descriptor, UCHAR Reason)
{
	HostMemToCardMem8(Adapter, (Adapter->HostInfo.TxDescPos + (SIZE_OF_TXDESC * Descriptor) + TX_DESC_REPORT_REASONS_OFFSET), &Reason, 1);
}

void GetTxDescReportReason(PVNet_ADAPTER Adapter, USHORT Descriptor, PUCHAR pReason)
{
	CardMemToHostMem8(Adapter, pReason, (Adapter->HostInfo.TxDescPos + (SIZE_OF_TXDESC * Descriptor) + TX_DESC_REPORT_REASONS_OFFSET), 1);
}
