/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/


/**************************************************************************/
/**************************************************************************/
/** 																	  */ 
/**   VNet Functions													  */
/** 																	  */
/**************************************************************************/
/**************************************************************************/

#include "vnet.h"
#include "rx.h"
#include "interrupt.h"

 
extern UCHAR	BasicRates[2][4];
ULONG Lefty;
ULONG Righty;

#if defined (RFMD) || defined (RFMD_3COM)
UCHAR	RegDomainTable[8] = {	REG_DOMAIN_FCC,
								REG_DOMAIN_DOC,
								REG_DOMAIN_ETSI,
								REG_DOMAIN_SPAIN,
								REG_DOMAIN_FRANCE,
								REG_DOMAIN_MKK,
								REG_DOMAIN_MKK1,
								REG_DOMAIN_ISRAEL };

#endif

UCHAR	TxPowerLevels[96] = {	 70,  80,  85,  90,  95,
								100, 105, 105, 110, 115,
								115, 120, 120, 125, 125,
								130, 130, 135, 135, 135,
								140, 140, 140, 145, 145,
								145, 150, 150, 150, 150,
								155, 155, 155, 155, 155,
								160, 160, 160, 160, 165,
								165, 165, 165, 165, 165,
								170, 170, 170, 170,	170,
								170, 175, 175, 175,	175,
								175, 175, 175, 180, 180,
								180, 180, 180, 180, 180,
								180, 185, 185, 185, 185,
								185, 185, 185, 185, 185,
								190, 190, 190, 190,	190,
								190, 190, 190, 190, 195,
								195, 195, 195, 195, 195,
								195, 195, 195, 195, 195,
								200 };

BOOLEAN RemoveEncryptionSettings(PVNet_ADAPTER Adapter)
{
	RemoveKey(Adapter, 0, TRUE);

	if(Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption2Enabled || Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption2KeyAbsent)
		Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption2KeyAbsent;
#ifdef AES
	else if(Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption3Enabled || Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption3KeyAbsent)
		Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption3KeyAbsent;
#endif
	else
		Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption1KeyAbsent; 

	Adapter->SecurityInfo.PairwiseKeyAvailable = FALSE;
	Adapter->SecurityInfo.GroupKeyAvailable = FALSE;

	Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = CIPHER_SUITE_NONE;
	Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite = CIPHER_SUITE_NONE;
	
	return SetEncryptionStatus(Adapter, FALSE, 0, TRUE);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  ConfigureWEP														  
//
//	  Sets in MIBs all the WEP related parameters						 
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


BOOLEAN 
ConfigureWEP(PVNet_ADAPTER Adapter)
{
	GET_SET_MIB_STRUCT		sSetMIB;
	UCHAR				status;
	UCHAR				WepIsOn __attribute__ ((unused));
	UCHAR				ExcludeUnEncrypted = (Adapter->SecurityInfo.EncryptionInfo.WepMode == WEP_MODE_MANDATORY) ? 1 : 0;
	UCHAR				KeyLength;
  	int 				i;
#ifdef ATMEL_WLAN
    int         j;
  UCHAR                         index;
  UCHAR				BytesToSet, offset;
  //UCHAR                         InWpaMode = FALSE;
  BOOLEAN                       SetKey ;
  MAC_ENCRYPTION_MIB 		MacEncryptionMib;
#else 
  MAC_WEP_MIB	 		MacWepMib;
#endif	// ATMEL_WLAN

#if defined (RFMD) || defined (RFMD_3COM)
	//UCHAR				MaxWepKeySize=13;
#else
	//UCHAR				MaxWepKeySize=5;
#endif

	// override Authentication Type to OpenSystem when WEP is disabled

	WepIsOn = (Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel != 0) ? 1 : 0;

	if(Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == WEP_DISABLED) {
    printk("ENCRYPTION STATUS DISABLED\n");
                Adapter->SecurityInfo.EncryptionStatus = _802_11EncryptionDisabled;
		Adapter->SecurityInfo.AuthenticationMode = _802_11AuthModeOpen;
		Adapter->SecurityInfo.EncryptionInfo.AuthenticationType = Adapter->SecurityInfo.AuthenticationMode;
#ifdef ATMEL_WLAN
    Adapter->SecurityInfo.PairwiseKeyAvailable = FALSE;
    Adapter->SecurityInfo.GroupKeyAvailable = FALSE;
#endif
  }

		Adapter->SecurityInfo.AuthenticationMode = Adapter->SecurityInfo.EncryptionInfo.AuthenticationType;
	IF_LOUD(DbgPrint("Configuring WEP (%s-%s-%s)\n", (WepIsOn) ? "Enabled" : "Disabled", 
													(ExcludeUnEncrypted) ? "Mandatory" : "Optional",
													(Adapter->SecurityInfo.EncryptionInfo.AuthenticationType == C80211_MGMT_AAN_SHAREDKEY) ? "Shared Key" : "Open System");)
	

	IF_LOUD(DbgPrint("WEP Key to use: %d -%d\n", Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse, Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel);)

	if(Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == WEP_64bit) {
		KeyLength = 5;
      IF_LOUD(DbgPrint("WEP 64-BIT\n");)
#ifdef ATMEL_WLAN
    MacEncryptionMib.PrivacyInvoked = 1;
#endif
	}
	else if(Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == WEP_128bit) {
		KeyLength = 13;
      IF_LOUD(DbgPrint("WEP 128-BIT\n");)
#ifdef ATMEL_WLAN
    MacEncryptionMib.PrivacyInvoked = 1;
#endif
        } else {
		KeyLength = 0;
      IF_LOUD(DbgPrint("NO WEP\n");)
#ifdef ATMEL_WLAN
    MacEncryptionMib.PrivacyInvoked = 0;
#endif
        }
	

#ifdef ATMEL_WLAN
	SetKey = (MacEncryptionMib.PrivacyInvoked ? TRUE : FALSE);
  Adapter->SecurityInfo.PairwiseKeyAvailable = FALSE;
  Adapter->SecurityInfo.GroupKeyAvailable = FALSE;
#endif

	if(KeyLength) {
#ifndef ATMEL_WLAN
		IF_LOUD(
			DbgPrint("WEP Key #1: ");
			for(i=0;i<KeyLength;i++)
				DbgPrint("%02X ", Adapter->SecurityInfo.EncryptionInfo.WepKey1[i]);
			DbgPrint("\n");
			DbgPrint("WEP Key #2: ");
			for(i=0;i<KeyLength;i++)
				DbgPrint("%02X ", Adapter->SecurityInfo.EncryptionInfo.WepKey2[i]);
			DbgPrint("\n");

			DbgPrint("WEP Key #3: ");
			for(i=0;i<KeyLength;i++)
				DbgPrint("%02X ", Adapter->SecurityInfo.EncryptionInfo.WepKey3[i]);
			DbgPrint("\n");

			DbgPrint("WEP Key #4: ");
			for(i=0;i<KeyLength;i++)
				DbgPrint("%02X ", Adapter->SecurityInfo.EncryptionInfo.WepKey4[i]);
			DbgPrint("\n");
		)

#else
		IF_LOUD(
			DbgPrint("WEP Key #1: ");
			for(i=0;i<KeyLength;i++)
				DbgPrint("%02X ", Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[0][i]);
			DbgPrint("\n");
			DbgPrint("WEP Key #2: ");
			for(i=0;i<KeyLength;i++)
				DbgPrint("%02X ", Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[1][i]);
			DbgPrint("\n");
			DbgPrint("WEP Key #3: ");
			for(i=0;i<KeyLength;i++)
				DbgPrint("%02X ", Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[2][i]);
			DbgPrint("\n");

			DbgPrint("WEP Key #4: ");
			for(i=0;i<KeyLength;i++)
				DbgPrint("%02X ", Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[3][i]);
			DbgPrint("\n");
		)
#endif		// ATMEL_WLAN

	}

#ifdef ATMEL_WLAN
	index = Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse;
	MacEncryptionMib.ExcludeUnencrypted = ExcludeUnEncrypted;
  Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite = 0;
  for (i = 0; i < 4; i++){
      if( Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[i][39] == 1 ){
        if(i == index) {
          MacEncryptionMib.CipherDefaultKeyID = index;
          Adapter->SecurityInfo.PairwiseKeyAvailable = TRUE;
        } else {
              MacEncryptionMib.CipherDefaultGroupKeyID = i;
              Adapter->SecurityInfo.GroupKeyAvailable = TRUE;
              Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite = Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite;
        }
      }
  }
	
	if((Adapter->SecurityInfo.PairwiseKeyAvailable && !Adapter->SecurityInfo.GroupKeyAvailable)) {
			MacEncryptionMib.CipherDefaultGroupKeyID = MacEncryptionMib.CipherDefaultKeyID;
	} else if(!Adapter->SecurityInfo.PairwiseKeyAvailable && Adapter->SecurityInfo.GroupKeyAvailable) {
			MacEncryptionMib.CipherDefaultKeyID = MacEncryptionMib.CipherDefaultGroupKeyID;
	} else {
		MacEncryptionMib.CipherDefaultKeyID = index;
		MacEncryptionMib.CipherDefaultGroupKeyID = index;
	}	
        
	VNetMoveMemory(MacEncryptionMib.ReceiverAddress, Adapter->AssociationInfo.CurrentBSSID, 6);
        
        BytesToSet = ENCRYPTION_PARAMETERS_BYTES;
        offset = (MAX_ENCRYPTION_KEYS * MAX_ENCRYPTION_KEY_SIZE);


	sSetMIB.Type = Mac_Wep_Mib_Type;
	sSetMIB.Size = BytesToSet;
	sSetMIB.Index = offset; 
	VNetMoveMemory(sSetMIB.Data, ((PUCHAR)&MacEncryptionMib)+offset, BytesToSet);
        
	status = SendCommand(Adapter, CMD_Set_MIB_Vars, (PUCHAR)&sSetMIB, NULL, 0, FALSE);
#else

	MacWepMib.ExcludeUnencrypted = ExcludeUnEncrypted;
	
	MacWepMib.WEPDefaultKeyID = Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse;

#if defined (RFMD) || defined (RFMD_3COM)
	MacWepMib.EncryptionLevel = Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel;
#endif
	VNetMoveMemory((PUCHAR)MacWepMib.WEPDefaultKeyValue, Adapter->SecurityInfo.EncryptionInfo.WepKey1, KeyLength);
	VNetMoveMemory((PUCHAR)MacWepMib.WEPDefaultKeyValue + MaxWepKeySize, Adapter->SecurityInfo.EncryptionInfo.WepKey2, KeyLength);
	VNetMoveMemory((PUCHAR)MacWepMib.WEPDefaultKeyValue + (MaxWepKeySize * 2), Adapter->SecurityInfo.EncryptionInfo.WepKey3, KeyLength);
	VNetMoveMemory((PUCHAR)MacWepMib.WEPDefaultKeyValue + (MaxWepKeySize * 3), Adapter->SecurityInfo.EncryptionInfo.WepKey4, KeyLength);

	sSetMIB.Type = Mac_Encryption_Mib_Type;
	sSetMIB.Size = sizeof(MAC_WEP_MIB);
	sSetMIB.Index = 0;
	VNetMoveMemory(sSetMIB.Data, &MacWepMib, sizeof(MAC_WEP_MIB));

	status = SendCommand(Adapter, CMD_Set_MIB_Vars, (PUCHAR)&sSetMIB, NULL, 0, FALSE);
#endif	// ATMEL_WLAN

	if(status != CMD_STATUS_COMPLETE){
		IF_DEBUG_ERRORS(DbgPrint("FAILED to set WEP settings (0x%X)...\n",status);)
		return FALSE;
	}

#ifdef ATMEL_WLAN

        if(SetKey) {
            //KeyRSC. In the case of WEP encryption we can ignore it(set it to zero).

            VNetZeroMemory(Adapter->SecurityInfo.EncryptionInfo.KeyRSC[0], 32);

            sSetMIB.Type = Mac_Encryption_Mib_Type;
            sSetMIB.Size = 6;
            sSetMIB.Index = ENCRYPTION_KEY_RSC_BYTES + (index * 8); // * 8 since the array is 4*8 although we use 6 bytes
            VNetMoveMemory(sSetMIB.Data, (PUCHAR)(&Adapter->SecurityInfo.EncryptionInfo.KeyRSC[index][0]), 6);
            
	status = SendCommand(Adapter, CMD_Set_MIB_Vars, (PUCHAR)&sSetMIB, NULL, 0, FALSE);
            if(status != CMD_STATUS_COMPLETE){
                return FALSE;
            }
            
            
            //Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-1] = 0x01;
  for (i = 0; i < 4; i++){
      if( Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[i][39] == 1 ){
        if(i == index) {
                    Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-1] |= 0x02;
                    Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-1] |= 0x04;
                    Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-2] = Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite;
        } else {
                    Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[i][MAX_ENCRYPTION_KEY_SIZE-2] = Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite;
        }
        SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[i][0], i, MAX_ENCRYPTION_KEY_SIZE);
      }  
            for(j=32;j<MAX_ENCRYPTION_KEY_SIZE; j++) {
                Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[i][j] = 0;
            }
  }
    
#if 0
            if(1/*KeyIndex & 0x40000000*/) // Pairwise key
                    Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-1] |= 0x02;
    
            if((1/*KeyIndex & 0x80000000*/)) // key that can be used for Tx
                    Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-1] |= 0x04;

            if((1/*KeyIndex & 0x40000000*/)) // Pairwise key
                    Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-2] = Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite;
            else
                    Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-2] = Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite;
            
            //SetEncryptionKey(&EncryptionInfo.KeyMaterial[index][0], index, MAX_ENCRYPTION_KEY_SIZE);
            SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[0][0], 0, MAX_ENCRYPTION_KEY_SIZE);
            SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[1][0], 1, MAX_ENCRYPTION_KEY_SIZE);
            SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[2][0], 2, MAX_ENCRYPTION_KEY_SIZE);
            SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[3][0], 3, MAX_ENCRYPTION_KEY_SIZE);
#endif

    }
    else {
        for(i=0; i<MAX_ENCRYPTION_KEY_SIZE; i++) {
            Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[0][i] = 0;
            Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[1][i] = 0;
            Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[2][i] = 0;
            Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[3][i] = 0;
        }
            
        SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[0][0], 0, MAX_ENCRYPTION_KEY_SIZE);
        SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[1][0], 1, MAX_ENCRYPTION_KEY_SIZE);
        SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[2][0], 2, MAX_ENCRYPTION_KEY_SIZE);
        SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[3][0], 3, MAX_ENCRYPTION_KEY_SIZE);
    }


#endif // ATMEL_WLAN

	return TRUE;
}

#ifdef ATMEL_WLAN
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  SetEncryptionKey 
//	  Sets Encryption Key						 
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
BOOLEAN
SetEncryptionKey(PVNet_ADAPTER Adapter, PUCHAR Key, UCHAR index, ULONG length)
{
	GET_SET_MIB_STRUCT		sSetMIB;
	UCHAR				status;

	//////////////////////////////
	//	MAC_ENCRYPTION_MIB 
	//	Keys
	//////////////////////////////

	sSetMIB.Type = Mac_Encryption_Mib_Type;
	sSetMIB.Size = (UCHAR)length;
	sSetMIB.Index = (index*MAX_ENCRYPTION_KEY_SIZE);
	VNetMoveMemory(sSetMIB.Data, Key, length);

	status = SendCommand(Adapter, CMD_Set_MIB_Vars, (PUCHAR)&sSetMIB, NULL, 0, FALSE);
        if(status != CMD_STATUS_COMPLETE){
            return FALSE;
	}
        	
	return TRUE;
}

#endif // ATMEL_WLAN



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  ResetAdapter														  
//
//	  Causes a H/W reset and re-initializes whatever is necessery for	  
//	  the device to start working normaly after the reset
//	  It will initiate a SCAN after Reset according the InitiateScan parameter						
//																		  
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

BOOLEAN
ResetAdapter(PVNet_ADAPTER Adapter, BOOLEAN InitiateScan)
{
	USHORT AddressReg;

	if(Adapter->StationState == STATION_STATE_NO_CARD)
		return FALSE;


	CardReadUshort(Adapter, AR, &AddressReg);

	DisableInterrupts(Adapter); 		//will be enabled again in CardStartUp

	Adapter->BSSListEntries = 0;

	Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt		= 0;
	Adapter->AssociationInfo.AssociationRequestRetryCnt 		= 0;
	Adapter->AssociationInfo.ReAssociationRequestRetryCnt		= 0;

	Adapter->InterruptMask						= ISR_TxCOMPLETE | ISR_RxCOMPLETE | ISR_FATAL_ERROR | ISR_RxFRAMELOST | ISR_COMMAND_COMPLETE | ISR_OUT_OF_RANGE | ISR_IBSS_MERGE | ISR_GENERIC_IRQ;
	Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum	= 0x0001;
	Adapter->AuthenticationInfo.ExpectedAuthentTransactionSeqNum	= 0x0002;

  Adapter->SecurityInfo.InWpaMode = 0;
	Adapter->StationState					= STATION_STATE_INITIALIZING;
	
	Adapter->StationIsAssociated				= FALSE;
	Adapter->CmdPending = FALSE;
	Adapter->CmdQEntries = 0;
	Adapter->CmdQHead = Adapter->CmdQTail = 0;

	memset(&Adapter->CmdQ, 0, (sizeof(CMD_Q)*MAX_CMD_Q_ENTRIES));

	Adapter->ApSelected 						= FALSE;

#ifdef INT_ROAM
	VNetZeroMemory((PUCHAR)&Adapter->MultiDomainMib, sizeof(MDOMAIN_MIB));
#endif

#if defined (RFMD) || defined (RFMD_3COM)
	if(Adapter->CardType == CARD_TYPE_EEPROM)
		CopyCodeToInternalMemory(Adapter);
#endif

	while(1)
	{
		if(!CardReset(Adapter)) 		// Resets card and waits for MAC to be initialized
			break;
		
		if(!CardGetHostInfo(Adapter))	// Get HOSTINFO struct from MAC
			break;
		
		if(!CardStartUp(Adapter))
			break;

		if(!SetMibValues(Adapter))
			break;
				
		if(InitiateScan)
		{
			Adapter->StationState = STATION_STATE_SCANNING;
			Scan(Adapter, TRUE);
		}
		CardWriteUshort(Adapter, AR, AddressReg);
		return TRUE;
	}

	CardWriteUshort(Adapter, AR, AddressReg);
	return FALSE;
}



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  InitAdapter														  
//																		  
//	  Initializes members of our Adapter Object 					
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VOID InitAdapter(PVNet_ADAPTER Adapter)
{
	//int 		i = 0;
	//
	// Initialize parameters.
	//
	Adapter->StationState						= STATION_STATE_INITIALIZING;

	VNetZeroMemory(Adapter->BSSID,6);

	VNetZeroMemory(Adapter->AssociationInfo.LastBSSID, 6);
	Adapter->AssociationInfo.CurrentBSSID[0] = 0xFF;		//Initialize it to something invalid....

	VNetZeroMemory((PUCHAR)&Adapter->Stats, sizeof(STATISTICS));
	
	Adapter->MulticastListMax					= DEFAULT_MULTICASTLISTMAX;
	Adapter->CurrentLookAhead					= MAX_WIRELESS_FRAME_SIZE;
	Adapter->InterruptMask						= ISR_TxCOMPLETE | ISR_RxCOMPLETE | ISR_FATAL_ERROR | ISR_RxFRAMELOST | ISR_COMMAND_COMPLETE | ISR_OUT_OF_RANGE | ISR_IBSS_MERGE | ISR_GENERIC_IRQ;

	Adapter->PendingTxPackets					= 0;

	// No resource have been assigned yet

	Adapter->IOregistered						= FALSE;

	Adapter->StationWasAssociated				= FALSE;
	Adapter->StationIsAssociated				= FALSE;
	Adapter->SecurityInfo.AuthenticationMode				= _802_11AuthModeOpen;

	Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum	= 0x0001;
	Adapter->AuthenticationInfo.ExpectedAuthentTransactionSeqNum	= 0x0002;

	Adapter->ApSelected 						= FALSE;

	Adapter->BSSListEntries 					= 0;

	Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt		= 0;
	Adapter->AssociationInfo.AssociationRequestRetryCnt 		= 0;
	Adapter->AssociationInfo.ReAssociationRequestRetryCnt		= 0;

	Adapter->MiniportResetRequest				= FALSE;
	Adapter->MiniportResetPending				= FALSE;

	Adapter->SiteSurveyState					= SITE_SURVEY_IDLE;

	Adapter->IntRoamingEnabled					= 0;

	Adapter->UseWzcs							= 0;

	Adapter->InSnifferMode						= FALSE;

	Adapter->PeriodicBeaconsCnt 				= 0;		
//
// KeyAbsent seems to cause problems with 802.1x so we use only WepDisabled
	Adapter->DeviceIsUp = TRUE;
#ifdef RX_CRC
	init_CRCtable(Adapter);
#endif

}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  ChangeConfiguration												  
//
//	  Is called when the application has changed some configuration 	  
//	  parameters & adopts them. Depending to the parameters that have	  
//	  changed ant the current station state & mode it may reset the card, 
//	  or re-scan or just set some MIBs									  
//																		  
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

BOOLEAN
ChangeConfiguration(PVNet_ADAPTER Adapter, PDEVICE_CONFIGURATION pDevConfig)
{
	UCHAR				i;
	BOOLEAN 			SSIDmatch,AdapterMustBeReset = FALSE;
	UCHAR				bytes;
	//UCHAR				Status;

	//
	// Channel
	//
	if( (Adapter->Channel != pDevConfig->Channel) &&  (Adapter->OperatingMode == AD_HOC_MODE) )
	{

		IF_LOUD(DbgPrint("Changing From Channel %d to %d\n", Adapter->Channel, pDevConfig->Channel);)
		Adapter->Channel = pDevConfig->Channel;
		SetChannel(Adapter, pDevConfig->Channel);
		AdapterMustBeReset = TRUE; //in AdHoc after changing Channel reset
	}
	//
	// Data Frames Tx Rate
	//
	if(Adapter->DataFramesTxRate != pDevConfig->TxRate)
	{
		IF_LOUD(DbgPrint("Change SpeedIndex from %d to %d\n",Adapter->DataFramesTxRate,pDevConfig->TxRate);)
		Adapter->DataFramesTxRate = pDevConfig->TxRate;
		
		SetDataFramesTxRateMode(Adapter, pDevConfig->TxRate);
	}

	//
	//	International Roaming
	//

#ifdef INT_ROAM

	if(Adapter->IntRoamingEnabled != pDevConfig->InternationalRoaming)
	{
		IF_LOUD(DbgPrint("Changing International Roaming : %d to %d\n",Adapter->IntRoamingEnabled, pDevConfig->InternationalRoaming);)
      if(Adapter->OperatingMode == INFRASTRUCTURE_MODE){
		    Adapter->IntRoamingEnabled = pDevConfig->InternationalRoaming;
      }else {
		    Adapter->IntRoamingEnabled = 0;
      }

                SetIntRoaming(Adapter);
                Adapter->StationState = STATION_STATE_SCANNING;
                Scan(Adapter, TRUE);
	}

#endif


	//
	//	RtsThreshold
	//

	if(Adapter->RtsThreshold != pDevConfig->RtsCtsThreshold)
	{
		IF_LOUD(DbgPrint("Changing RTS/CTS Threshold from %d to %d\n",Adapter->RtsThreshold, pDevConfig->RtsCtsThreshold);)
		Adapter->RtsThreshold = pDevConfig->RtsCtsThreshold;

		SetRtsThreshold(Adapter, pDevConfig->RtsCtsThreshold);
	}


	//
	//	Fragmentation Threshold->>
	//
	if(Adapter->FragmentationThreshold != pDevConfig->FragmentationThreshold)
	{
		IF_LOUD(DbgPrint("Changing FragmentationThreshold from %d to %d\n", Adapter->FragmentationThreshold, pDevConfig->FragmentationThreshold);)

		if ((pDevConfig->FragmentationThreshold)%2){
		    pDevConfig->FragmentationThreshold -= 1;
    		
                }
        	
                Adapter->FragmentationThreshold = pDevConfig->FragmentationThreshold ;
    		
    	        SetFragmentationThreshold(Adapter, pDevConfig->FragmentationThreshold );
                
               
	}
	//
	// Preamble
	//
	if(Adapter->AssociationInfo.PreambleType != pDevConfig->PreambleType || Adapter->AutoPreambleDetection)
	{
		if(pDevConfig->PreambleType == AUTO_PREAMBLE && !Adapter->AutoPreambleDetection)
		{
			IF_LOUD(DbgPrint("AUTO Preamble detection enabled...\n");)
			Adapter->AssociationInfo.PreambleType = PREAMBLE_TYPE_LONG;
			Adapter->AutoPreambleDetection = TRUE;

			SetPreambleType(Adapter, Adapter->AssociationInfo.PreambleType);
		}
		else if(pDevConfig->PreambleType != AUTO_PREAMBLE)
		{
			Adapter->AutoPreambleDetection = FALSE;

			IF_LOUD(DbgPrint("Changing Preamble Type from %s to %s\n", (Adapter->AssociationInfo.PreambleType == PREAMBLE_TYPE_LONG) ? "LONG" : "SHORT", (pDevConfig->PreambleType == AUTO_PREAMBLE) ? "AUTO" : (pDevConfig->PreambleType == PREAMBLE_TYPE_LONG) ? "LONG" :"SHORT");)

			Adapter->AssociationInfo.PreambleType = pDevConfig->PreambleType;
			
			SetPreambleType(Adapter, Adapter->AssociationInfo.PreambleType);
		}
	}
	
	//
	//	RADIO MODE
	//
	if(Adapter->RadioIsOn != pDevConfig->RadioIsOn)
	{
    Adapter->RadioIsOn = pDevConfig->RadioIsOn;
		if(Adapter->RadioIsOn == TRUE)
		{
			IF_LOUD(DbgPrint("Changing RADIO Mode -> ON ...\n");)
                        EnableRadio(Adapter);
                        ResetAdapter(Adapter, TRUE);
		}
		else
		{
			IF_LOUD(DbgPrint("Changing Radio Mode -> OFF ...\n");)
		        Adapter->StationState	= STATION_STATE_RADIO_OFF;
		        Adapter->RejectPendingTxPackets = TRUE;
            DisableRadio(Adapter);
		}

	}

	//
	// SSID
	//

        if(Adapter->RadioIsOn == TRUE){
            SSIDmatch = TRUE;

            // (4.0.2.28) Whenever NULL SSID is selected force a SCAN

                if((Adapter->InitialSSIDsize != pDevConfig->SSIDlength) || (pDevConfig->SSIDlength == 0))
                {
                        IF_LOUD(DbgPrint("### SSID size changed (or NULL SSID) ### (%d-%d)\n", Adapter->InitialSSIDsize, pDevConfig->SSIDlength);)
                        SSIDmatch = FALSE;
                }
                else
                {
                        bytes = (Adapter->InitialSSIDsize > pDevConfig->SSIDlength) ? Adapter->InitialSSIDsize : pDevConfig->SSIDlength;
                        for(i=0;i<bytes;i++)
                        {
                                if(Adapter->InitialSSID[i] != pDevConfig->SSID[i])
                                {
                                        IF_LOUD(DbgPrint("### SSID Changed ###\n");)
                                        SSIDmatch = FALSE;
                                        break;
                                }
                        }
                }

                if(!SSIDmatch)
                {
                        IF_LOUD(DbgPrint("Changing SSID...\n");)
                        
                        AdapterMustBeReset = TRUE;

                        memcpy(Adapter->AssociationInfo.DesiredSSID, pDevConfig->SSID, pDevConfig->SSIDlength);
                        Adapter->AssociationInfo.DesiredSSIDsize = pDevConfig->SSIDlength;
                        
                        memcpy(Adapter->InitialSSID, pDevConfig->SSID, pDevConfig->SSIDlength);
                        Adapter->InitialSSIDsize = pDevConfig->SSIDlength;

                        if( Adapter->AssociationInfo.DesiredSSIDsize == 0 )
                                Adapter->ConnectToAnyBSS = TRUE;
                        else
                                Adapter->ConnectToAnyBSS = FALSE;

    #if ((!defined RFMD) && (!defined RFMD_3COM))
                        SetSSIDsize(Adapter, Adapter->AssociationInfo.DesiredSSIDsize);
    #endif

                }
            //
            //	OPERATING MODE
            //
            if(Adapter->OperatingMode != pDevConfig->OperatingMode)
            {
                    AdapterMustBeReset		= TRUE;
                    Adapter->StationState	= STATION_STATE_INITIALIZING;
                    if(Adapter->OperatingMode == AD_HOC_MODE)
                    {
                            IF_LOUD(DbgPrint("Changing Operating Mode to INFRASTRUCTURE...\n");)
                            Adapter->OperatingMode = pDevConfig->OperatingMode;
                    }
                    else
                    {
                            IF_LOUD(DbgPrint("Changing Operating Mode to AD-HOC in channel %d...\n", pDevConfig->Channel);)
                            Adapter->OperatingMode = pDevConfig->OperatingMode;
                            //restore channel
                            IF_LOUD(DbgPrint("Changing From Channel %d to %d\n", Adapter->Channel, pDevConfig->Channel);)
                            Adapter->Channel = pDevConfig->Channel; 
                            if(Adapter->IntRoamingEnabled){
                              Adapter->IntRoamingEnabled = 0;
                            }
                    }

            }

            //
            // Tx Power Level
            //
            if(Adapter->TxPowerLevel != pDevConfig->TxPowerLevel)
            {
              // Hack till we sync winter
              pDevConfig->TxPowerLevel = 200;
              Adapter->TxPowerLevel = pDevConfig->TxPowerLevel;
              SetTxPowerLevel(Adapter, Adapter->TxPowerLevel);
            }
            //
            // Power Mgmt Mode
            //
            if(Adapter->PowerMgmtMode != pDevConfig->PowerMgmtMode)
            {
                    //	
                    //(4.1.2.30): While in AdHoc we MUST NOT enable PS. The selected Power Mgmt
                    // mode is now tested after setting the Operating Mode in Infrastructure.
                    //
                    if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
                    {
                            IF_LOUD(DbgPrint("Changing Power Mgmt Mode from %d to %d\n", Adapter->PowerMgmtMode, pDevConfig->PowerMgmtMode);)

                            if(Adapter->PowerMgmtMode) //from PS->normal needs Join
                                    Join(Adapter, BSS_TYPE_INFRASTRUCTURE);
                            else
                                    SetPowerMgmtMode(Adapter, pDevConfig->PowerMgmtMode);

                            Adapter->PowerMgmtMode = pDevConfig->PowerMgmtMode;
                    }
                    else
                    {
                            if(Adapter->PowerMgmtMode)
                            {
                                    IF_LOUD(DbgPrint("Power Save selection is ignored in AdHoc Mode...\n");)
                    
                                    Adapter->StationIsAssociated = FALSE;
                                    Adapter->RejectPendingTxPackets = TRUE;
                                		ResetAdapter(Adapter, TRUE);

#if 0
                                    AdapterMustBeReset = TRUE; //from PS->normal needs reset
                                    Adapter->PowerMgmtMode = 0;
#endif
                            }
                    }
            }
        }


	//
	//BL: (4.0.2.28) reset the 'black list'...
	//
	for(i=0; i<Adapter->BSSListEntries;i++)
		Adapter->BSSinfo[i].Channel = (Adapter->BSSinfo[i].Channel & 0x7f);
	
	if(AdapterMustBeReset)
	{
		Adapter->StationIsAssociated = FALSE;
		Adapter->RejectPendingTxPackets = TRUE;

		Adapter->StationState = STATION_STATE_SCANNING;
		Scan(Adapter, TRUE);
	
//		ResetAdapter(Adapter, TRUE);
	}

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  InitAndStartCard
//
//	  Initializes the card and starts operation. Is called from the 	  
//	  Initialize handler												  
//																		  
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

BOOLEAN InitAndStartCard(PVNet_ADAPTER Adapter)
{
	BOOLEAN 				ParallelFlash;
	USHORT					temp;
	USHORT					gcr;


	CardReadUshort(Adapter, GCR, &gcr);

#ifndef PCI
	CardWriteUshort(Adapter, GCR, (USHORT)(gcr|0x0060));
	CardReadUshort(Adapter, GCR, &gcr);
	CardWriteUshort(Adapter, GCR, (USHORT)(gcr&0xFF9F));
#endif

	VnetSleep(500000);				

	CardReadUshort(Adapter, MR2, &temp);

	if (!temp) {
		ParallelFlash = FALSE;
		Adapter->CardType = CARD_TYPE_EEPROM;

		CopyCodeToInternalMemory(Adapter);
	}
	else {
		CardReadUshort(Adapter, MR4, &temp);
		if (temp) {
			Adapter->CardType = CARD_TYPE_SPI_FLASH;
			IF_LOUD(DbgPrint("SPI FLASH\n");)
			ParallelFlash = FALSE;
		}
		else {
			Adapter->CardType = CARD_TYPE_PARALLEL_FLASH;
			IF_LOUD(DbgPrint("PARALLEL FLASH\n");)
			ParallelFlash = TRUE;
		}
	}

	while(1) {
		if(!CardGetHostInfo(Adapter))	// waits for MAC to be initialized and gets IFACE struct from MAC
			break;

		if(!CardStartUp(Adapter))
			break;

		//
		// MAC Address
		//
		GetMACaddress(Adapter);
		if(!Adapter->NetAddressOverride)
			VNetMoveMemory(Adapter->AssociationInfo.CurrentAddress, Adapter->PermanentAddress, 6);

#if defined (RFMD) || defined (RFMD_3COM)
		//
		//RegDomain: (4.0.2.28) Get the regulatory domain of this card
		// if the domain is invalid select by default the MKK1
		//
	
                if(!GetRegulatoryDomain(Adapter, &Adapter->RegDomain)) 
                {
                        IF_LOUD(DbgPrint("FAILED to get Regulatory Domain. MKK1 is selected (Default)\n");)
                        Adapter->RegDomain = REG_DOMAIN_MKK1;
                }
#else
		IF_LOUD(DbgPrint("Default Regulatory Domain assigned (MKK1)\n");)
		Adapter->RegDomain = REG_DOMAIN_MKK1;
#endif	//RFMD

		//
		//RegDomain: (4.0.2.28) Check if the currently selected channel is valid for 
		// the selected domain. If not, this function will select the first valid channel
		// for this domain
		//
		ValidateChannelForDomain(Adapter, &Adapter->Channel);

		if(!SetMibValues(Adapter))
		{
			break;
		}

		Adapter->StationState = STATION_STATE_SCANNING;
		Scan(Adapter, TRUE);
		return TRUE;
	}

	return FALSE;
}


#ifdef R504A_2958
#define INTERNAL_MEMORY_BANK_SIZE	0x8000
#else
#define INTERNAL_MEMORY_BANK_SIZE	0x6000
#endif
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  CopyCodeToInternalMemory											  
//
//	  Downloads F/W to the internal memory								  
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VOID CopyCodeToInternalMemory(PVNet_ADAPTER Adapter)
{
	USHORT					gcr;
#ifdef 	R504A_2958
	USHORT					i,Data;
	BOOLEAN					f504B;
#endif
	
	IF_LOUD(DbgPrint("Copy Code To Internal Memory (** EEPROM **)\n");)

	// Download firmware
	IF_LOUD(DbgPrint("*** Downloading Firmware to internal memory ***\n");)

//#if defined(UNDER_CE)
	Adapter->WorkingWithDefaultFW = TRUE;
//	Adapter->WorkingWithDefaultFW = TRUE;
	GetDefaultFw(Adapter);
//#endif //UNDER_CE

	CardStop(Adapter);

	if(Adapter->FirmwareFileLength <= INTERNAL_MEMORY_BANK_SIZE)
	{
#ifdef R504A_2958
		// Remap 
		CardReadUshort(Adapter, GCR, &gcr);
		CardWriteUshort(Adapter, GCR, (USHORT)(gcr|0xc00));
		CardReadUshort(Adapter, GCR, &gcr);

		// memory test for 504B
		//
		f504B = FALSE;
		CardWriteUshort(Adapter, BSR, 0x2002);
		
		CardWriteUshort(Adapter, AR, 0);
		for(i=7500; i<7505; i++)
			CardWriteUshort(Adapter, DR, i);

		CardWriteUshort(Adapter, AR, 0);
		for(i=7500; i<7505; i++)
		{
			CardReadUshort(Adapter, DR, &Data);
			if(Data != i)
				f504B = TRUE;
		}
		//
		//

		if(f504B)
		{
//			Adapter->AdapterSubType = ADAPTER_SUBTYPE_504AL;
			CardWriteUshort(Adapter, BSR, 0x2000);

			HostMemToCardMem8(Adapter, 0x8000, Adapter->MappedFirmwareBuffer, Adapter->FirmwareFileLength);
			IF_LOUD(DbgPrint("DOWNLOADING F/W for 504AL\n");)
		}
		else
		{
			CardWriteUshort(Adapter, BSR, 0x2002);
			HostMemToCardMem8(Adapter, 0, Adapter->MappedFirmwareBuffer, Adapter->FirmwareFileLength);
		}

		CardWriteUshort(Adapter, BSR, 0x2000);

		CardReadUshort(Adapter, GCR, &gcr);
		CardWriteUshort(Adapter, GCR, (USHORT)(gcr&0xf7ff));
		CardReadUshort(Adapter, GCR, &gcr);
#else
		// Remap 
		CardReadUshort(Adapter, GCR, &gcr);
		CardWriteUshort(Adapter, GCR, (USHORT)(gcr|GCR_REMAP));
		CardReadUshort(Adapter, GCR, &gcr);

		CardWriteUshort(Adapter, BSR, BSS_IRAM);

		HostMemToCardMem8(Adapter, 0, Adapter->MappedFirmwareBuffer, Adapter->FirmwareFileLength);

		// Remap 
		CardReadUshort(Adapter, GCR, &gcr);
		CardWriteUshort(Adapter, GCR, (USHORT)(gcr|GCR_REMAP));
		CardReadUshort(Adapter, GCR, &gcr);
#endif

	}
	else //scatter loading
	{
		IF_LOUD(DbgPrint("**** SCATTER LOADING ****\n");)

#ifdef R504A_2958
		// Remap 
		CardReadUshort(Adapter, GCR, &gcr);
		CardWriteUshort(Adapter, GCR, (USHORT)(gcr|0xc00));
		CardReadUshort(Adapter, GCR, &gcr);

		// memory test for 504B
		//
		f504B = FALSE;
		CardWriteUshort(Adapter, BSR, 0x2002);
		
		CardWriteUshort(Adapter, AR, 0);
		for(i=7500; i<7505; i++)
			CardWriteUshort(Adapter, DR, i);

		CardWriteUshort(Adapter, AR, 0);
		for(i=7500; i<7505; i++)
		{
			CardReadUshort(Adapter, DR, &Data);
			if(Data != i)
				f504B = TRUE;
		}
		//
		//

		if(f504B)
		{
	//		Adapter->AdapterSubType = ADAPTER_SUBTYPE_504AL;
			CardWriteUshort(Adapter, BSR, 0x2000);

			HostMemToCardMem8(Adapter, 0x8000, Adapter->MappedFirmwareBuffer, 0x8000);
			IF_LOUD(DbgPrint("DOWNLOADING F/W for 504AL\n");)
		}
		else
		{
			CardWriteUshort(Adapter, BSR, 0x2002);
			HostMemToCardMem8(Adapter, 0, Adapter->MappedFirmwareBuffer, 0x8000);
		}

#else
		// Remap 
		CardReadUshort(Adapter, GCR, &gcr);
		CardWriteUshort(Adapter, GCR, (USHORT)(gcr|GCR_REMAP));
		CardReadUshort(Adapter, GCR, &gcr);

		CardWriteUshort(Adapter, BSR, BSS_IRAM);

		HostMemToCardMem8(Adapter, 0, Adapter->MappedFirmwareBuffer, 0x6000);

#endif

#ifdef R504A_2958

		if(f504B)
		{
			CardWriteUshort(Adapter, BSR, 0x2000);
			HostMemToCardMem8(Adapter, 0x6000, (Adapter->MappedFirmwareBuffer+0x8000), Adapter->FirmwareFileLength-0x8000);
		}
		else
		{
			CardWriteUshort(Adapter, BSR, 0x2001);
			HostMemToCardMem8(Adapter, 0x8000, (Adapter->MappedFirmwareBuffer+0x8000), Adapter->FirmwareFileLength-0x8000);
		}
		
		CardWriteUshort(Adapter, BSR, 0x2000);

		CardReadUshort(Adapter, GCR, &gcr);
		CardWriteUshort(Adapter, GCR, (USHORT)(gcr&0xf7ff));
		CardReadUshort(Adapter, GCR, &gcr);



#else
		CardWriteUshort(Adapter, BSR, 0x2ff);
		HostMemToCardMem8(Adapter, 0x8000, (Adapter->MappedFirmwareBuffer+0x6000), Adapter->FirmwareFileLength-0x6000);

#endif

	}
	//
	// get ARM from reset state. Code will start running
	//
	CardWriteUshort(Adapter, GCR, (USHORT)(gcr&0xFFBF));

}



/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  GetMACaddress 													  
//
//	 Gets (reads) the MAC address from the FLASH. if the MAC address is   
//	 not present, the default MAC address is assigned and the application 
//	 will prompt the user to upgrade his F/W							  
//																		  
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VOID
GetMACaddress(PVNet_ADAPTER Adapter)
{
	UCHAR					DefaultMAC[] = {0x00,0x04, 0x25, 0x00, 0x00, 0x00};
	GET_SET_MIB_STRUCT		sGetMIB;
	UCHAR					status = 0;

	VnetSleep(500000);

	if (Adapter->CardType == CARD_TYPE_PARALLEL_FLASH) {

		CardWriteUshort(Adapter,  BSR, 1);	
		CardMemToHostMem8(Adapter, Adapter->PermanentAddress, 0xc000, 6);
		CardWriteUshort(Adapter,  BSR, BSS_SRAM);
	}
	else 
	{

		sGetMIB.Type = (UCHAR)Mac_Address_Mib_Type;
		sGetMIB.Size = (UCHAR)6;
		sGetMIB.Index = (UCHAR)0;

	status = SendCommand(Adapter, CMD_Get_MIB_Vars, (PUCHAR)&sGetMIB, NULL, 0, FALSE);

		if(status != CMD_STATUS_COMPLETE)
		{
			IF_DEBUG_ERRORS(DbgPrint("******** FAILED to Get Mac Address (0x%X)...\n",status);)
			VNetMoveMemory(Adapter->PermanentAddress, DefaultMAC, 6);
		}

		CardMemToHostMem16(Adapter, sGetMIB.Data, Adapter->HostInfo.CommandPos + CMD_BLOCK_PARAMETERS_OFFSET + 4, 6);
		VNetMoveMemory(Adapter->PermanentAddress, sGetMIB.Data, 6);
	}

	if(Adapter->PermanentAddress[0] == 0xFF)
	{
		IF_DEBUG_ERRORS(
		DbgPrint("*** Invalid MAC address. UPGRADE Firmware ****");
		DbgPrint("Assigning default MAC address...\n");)
		Adapter->InvalidMACaddress = TRUE;
		VNetMoveMemory(Adapter->PermanentAddress, DefaultMAC, 6);
	}
	else 
		Adapter->InvalidMACaddress = FALSE;


	IF_LOUD(
		int 					i;

		DbgPrint("MAC Address: ");
		for(i=0;i<6;i++)
			DbgPrint("%02X ", Adapter->PermanentAddress[i]);
		DbgPrint("\n");
	)
}



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  AsciiToByte														  
//
//	  This function gets two ASCII values and returns a byte with the	  
//	  4 MSbits set to the value that coresponds to the first ASCII and	  
//	  the 4 LSbits set to the value that coresponds to the second ASCII.  
//	  This function is required because the WEP keys are get as a 10	  
//	  characters string and we need to extract the 5 HEX Bytes of the KEY 
//	  There are no range checks 										  
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

UCHAR
AsciiToByte(UCHAR Ascii1, UCHAR Ascii2)
{
	UCHAR digit1= 0;
	UCHAR digit2= 0;

	if(Ascii1 >= 0x30 && Ascii1 <= 0x39)
			digit1 = Ascii1 - 0x30;
	else if(Ascii1 >= 0x41 && Ascii1 <= 0x46)
			digit1 = Ascii1 - 0x31 - 6;


	if(Ascii2 >= 0x30 && Ascii2 <= 0x39)
			digit2 = Ascii2 - 0x30;
	else if(Ascii2 >= 0x41 && Ascii2 <= 0x46)
			digit2 = Ascii2 - 0x31 - 6;


	return (digit1<<4) + digit2;
}

VOID
ByteToAscii(UCHAR Byte, PUCHAR Ascii1, PUCHAR Ascii2)
{
	if( (( Byte & 0xf0 ) >> 4) <=9 )
		*Ascii1 = (( Byte & 0xf0 ) >> 4) + 0x30;
	else
		*Ascii1 = (( Byte & 0xf0 ) >> 4) + 0x57;

	if( ( Byte & 0x0f )  <=9 )
		*Ascii2 = ( Byte & 0x0f ) + 0x30;
	else
		*Ascii2 = ( Byte & 0x0f ) + 0x57;	
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  MgmtTimeOutCallBack												  
//																		  
//	  This is the OS indipendent callback routine of a mgmt timer set when
//	  a management frame is being transmited.							  
//																		   
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VOID MgmtTimeOutCallBack(PVNet_ADAPTER Adapter)
{
	BOOLEAN 	TakeAction = FALSE;
	int 		BSSindex;
  UCHAR MaxCnt;

//
// (4.0.2.28) if we must go in STATION_STATE_MGMT_ERROR we decide if we will do 
// a SCAN, JOIN or nothing (ie just stay in STATION_STATE_MGMT_ERROR) according to the
// OS (XP or not) and to the SSID (NULL or not)
//
	if(Adapter->StationState == STATION_STATE_READY) {
		if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
		{
      IF_VERY_LOUD(DbgPrint("Periodic Beacons Count : %d\n", Adapter->PeriodicBeaconsCnt);)
			MaxCnt = 2000 / Adapter->AssociationInfo.BeaconPeriod;
		
			// To have right indication of LinkQuality when Power Save is used we use
			// the ListenInterval. In Light Power Save we keep the normal calculation.
			if(Adapter->PowerMgmtMode == 1) {
				MaxCnt = MaxCnt / Adapter->AssociationInfo.ListenInterval;
      }

      IF_VERY_LOUD(DbgPrint("MAX Count : %d\n", MaxCnt);)
			Adapter->LinkQuality = (UCHAR)(((Adapter->PeriodicBeaconsCnt*100)/MaxCnt));

			if(Adapter->LinkQuality > 100) {
				Adapter->LinkQuality = 100;
      }
      IF_VERY_LOUD(DbgPrint("LINK QUALITY : %d\n", Adapter->LinkQuality);)
		}
      IF_VERY_LOUD(DbgPrint("BEACON PERIOD : %d\n", Adapter->AssociationInfo.BeaconPeriod);)
      IF_VERY_LOUD(DbgPrint("LISTEN INTERVAL : %d\n", Adapter->AssociationInfo.ListenInterval);)
      mod_timer(Adapter->MgmtTimer, jiffies + MGMT_JIFFIES);
  }

	if(Adapter->StationState == STATION_STATE_AUTHENTICATING )
	{
		IF_LOUD(DbgPrint("#### Authentication Request Timed Out (%d)...\n", Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt);)
		
		if(Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt >= MAX_AUTHENTICATION_RETRIES){
			Adapter->StationState				   = STATION_STATE_MGMT_ERROR;
			Adapter->StationIsAssociated		   = FALSE;
			Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt = 0;

			TakeAction = TRUE;
		}
		else
		{
			IF_LOUD(DbgPrint("Retry to Authenticate...\n");)
			Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt++;
			Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum = 0x0001;
      mod_timer(Adapter->MgmtTimer, jiffies + MGMT_JIFFIES);
			//SetMgmtTimer(Adapter);
			SendAuthenticationRequest(Adapter);
		}

	}
	else
	if(Adapter->StationState == STATION_STATE_ASSOCIATING)
	{
		IF_LOUD(DbgPrint("#### Association Request Timed Out...\n");)
		
		if(Adapter->AssociationInfo.AssociationRequestRetryCnt == MAX_ASSOCIATION_RETRIES){
			Adapter->StationState				= STATION_STATE_MGMT_ERROR;
			Adapter->StationIsAssociated		= FALSE;
			Adapter->AssociationInfo.AssociationRequestRetryCnt = 0;

			TakeAction = TRUE;
		}
		else
		{
			IF_LOUD(DbgPrint("Retry to Associate...\n");)
			Adapter->AssociationInfo.AssociationRequestRetryCnt++;
      mod_timer(Adapter->MgmtTimer, jiffies + MGMT_JIFFIES);
			//SetMgmtTimer(Adapter);
			SendAssociationRequest(Adapter);
		}

	}
	else
	if(Adapter->StationState == STATION_STATE_REASSOCIATING)
	{
		IF_LOUD(DbgPrint("#### ReAssociation Request Timed Out...\n");)
		
		if(Adapter->AssociationInfo.ReAssociationRequestRetryCnt == MAX_ASSOCIATION_RETRIES){
			Adapter->StationState = STATION_STATE_MGMT_ERROR;
			Adapter->StationIsAssociated		   = FALSE;
			Adapter->AssociationInfo.ReAssociationRequestRetryCnt = 0;

			TakeAction = TRUE;
		}
		else
		{
			IF_LOUD(DbgPrint("Retry to ReAssociate...\n");)
			Adapter->AssociationInfo.ReAssociationRequestRetryCnt++;
      mod_timer(Adapter->MgmtTimer, jiffies + MGMT_JIFFIES);
			//SetMgmtTimer(Adapter);
			SendReAssociationRequest(Adapter);
		}

	}

	if(TakeAction)
	{
		if(Adapter->ConnectToAnyBSS)
		{
			//
			//BL: (4.0.2.28) invalidate list entry...
			//
			IF_LOUD(DbgPrint("**** Try Another BSS-T (channel |= 0x80 for BSS #%d)...\n", Adapter->CurrentBSSindex);)
			Adapter->BSSinfo[(int)(Adapter->CurrentBSSindex)].Channel |= 0x80;

			BSSindex = RetrieveBSS(Adapter, NULL);
			if(BSSindex != -1)
			{
				IF_LOUD(DbgPrint("**** Try BSS #%d in Channel %d...\n", BSSindex, Adapter->BSSinfo[BSSindex].Channel);)
				JoinSpecificBSS(Adapter, BSSindex, FALSE);
			}
			else
			{
				IF_LOUD(DbgPrint("**** No BSS to try...\n");)

				if(Adapter->UseWzcs)
					IndicateConnectionStatus(Adapter, FALSE);
				else
					Scan(Adapter, FALSE);
			}
		}
		else
		{
			IF_LOUD(DbgPrint("#### Re Scan after Mgmt Frame Time Out ####\n");)
			Scan(Adapter, TRUE);
		}
	}

}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  JoinSpecificBSS												  
//																		  
//	  Joins a specific BSS from the list by index.
//																		   
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

BOOLEAN 
JoinSpecificBSS(PVNet_ADAPTER Adapter, int BSSindex, BOOLEAN ChangeInitialSSID)
{
	int 	i __attribute__ ((unused));

	i = 0;
	if(BSSindex<0 || BSSindex > MAX_BSS_ENTRIES)
	{
		IF_LOUD(DbgPrint("*** Invalid BSS index ***\n");)
		return FALSE;
	}

	VNetMoveMemory(Adapter->AssociationInfo.CurrentBSSID, Adapter->BSSinfo[BSSindex].BSSID, 6);
	Adapter->AssociationInfo.DesiredSSIDsize = Adapter->BSSinfo[BSSindex].SSIDsize;
	VNetMoveMemory(Adapter->AssociationInfo.DesiredSSID, (PUCHAR)Adapter->BSSinfo[BSSindex].SSID, Adapter->AssociationInfo.DesiredSSIDsize);

#if ((!defined RFMD) && (!defined RFMD_3COM))
	SetSSIDsize(Adapter, Adapter->AssociationInfo.DesiredSSIDsize);
#endif

	if(ChangeInitialSSID)
	{
		VNetZeroMemory(Adapter->InitialSSID, MAX_SSID_LENGTH);
		VNetMoveMemory(Adapter->InitialSSID, (PUCHAR)Adapter->BSSinfo[BSSindex].SSID, Adapter->BSSinfo[BSSindex].SSIDsize);
		Adapter->InitialSSIDsize = Adapter->BSSinfo[BSSindex].SSIDsize;
		
		//BL: (4.0.2.28) for safety since MSbit is used from driver
		Adapter->Channel = (Adapter->BSSinfo[BSSindex].Channel & 0x7f); 	
	}

	//When switching to AdHoc turn OFF International Roaming / Power Save if needed

	if((Adapter->BSSinfo[BSSindex].BSStype == BSS_TYPE_AD_HOC) &&
		(Adapter->OperatingMode != AD_HOC_MODE) && 
		(Adapter->IntRoamingEnabled || Adapter->PowerMgmtMode))
	{
#ifdef INT_ROAM
		if(Adapter->IntRoamingEnabled)
		{
			IF_LOUD(DbgPrint("*** Turn OFF Int Roaming (switching to AdHoc) ***\n");)
			Adapter->IntRoamingEnabled = 0;
			SetIntRoaming(Adapter);
		}
#endif
		if(Adapter->PowerMgmtMode)
		{
			IF_LOUD(DbgPrint("*** Turn OFF Power Save (switching to AdHoc) ***\n");)
			Adapter->PowerMgmtMode = 0;
			SetPowerMgmtMode(Adapter, Adapter->PowerMgmtMode);
		}
	}
  
  printk("TX POWER LEVEL : %d\n", Adapter->TxPowerLevel);
	SetTxPowerLevel(Adapter, Adapter->TxPowerLevel);
					
	if(Adapter->BSSinfo[BSSindex].BSStype == BSS_TYPE_AD_HOC)
		Adapter->OperatingMode = AD_HOC_MODE;
	else
	if(Adapter->BSSinfo[BSSindex].BSStype == BSS_TYPE_INFRASTRUCTURE)
		Adapter->OperatingMode = INFRASTRUCTURE_MODE;

	//BL: (4.0.2.28) for safety since MSbit is used from driver
	Adapter->Channel = (Adapter->BSSinfo[BSSindex].Channel & 0x7f); 	

	if(Adapter->AutoPreambleDetection && Adapter->AssociationInfo.PreambleType != Adapter->BSSinfo[BSSindex].PreambleType)
	{
		IF_LOUD(DbgPrint("Changing Preamble Type from %s to %s\n", (Adapter->AssociationInfo.PreambleType == PREAMBLE_TYPE_LONG) ? "LONG" : "SHORT", (Adapter->BSSinfo[BSSindex].PreambleType == PREAMBLE_TYPE_LONG) ? "LONG" : "SHORT");)
		Adapter->AssociationInfo.PreambleType = Adapter->BSSinfo[BSSindex].PreambleType;

		SetPreambleType(Adapter, Adapter->AssociationInfo.PreambleType);
	}
	
	Adapter->AssociationInfo.BeaconPeriod = Adapter->BSSinfo[BSSindex].BeaconPeriod;
	
  if((Adapter->BSSinfo[BSSindex].EncryptionIsOn) && (!Adapter->SecurityInfo.InWpaMode)) {
    Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption1Enabled;
  } else if((!Adapter->BSSinfo[i].EncryptionIsOn) && (Adapter->SecurityInfo.InWpaMode)){
    Adapter->SecurityInfo.EncryptionStatus = _802_11EncryptionDisabled;
  } else {
    Adapter->SecurityInfo.EncryptionStatus = _802_11EncryptionDisabled;
  }

	IF_LOUD(
		DbgPrint("*** BSS Selected ***\n");
		DbgPrint("Channel = %d\n", Adapter->BSSinfo[BSSindex].Channel);
		DbgPrint("BSSID = %02X-%02X-%02X-%02X-%02X-%02X\n", Adapter->AssociationInfo.CurrentBSSID[0],Adapter->AssociationInfo.CurrentBSSID[1],Adapter->AssociationInfo.CurrentBSSID[2], Adapter->AssociationInfo.CurrentBSSID[3],Adapter->AssociationInfo.CurrentBSSID[4],Adapter->AssociationInfo.CurrentBSSID[5]);
		DbgPrint("SSID: ");
		for(i=0;i<Adapter->AssociationInfo.DesiredSSIDsize;i++)
			DbgPrint("%c", Adapter->AssociationInfo.DesiredSSID[i]);
		DbgPrint("\n");
		DbgPrint("Beacon Period = %d\n", Adapter->AssociationInfo.BeaconPeriod);
		DbgPrint("Preamble Type is %s(%d)\n", (Adapter->AssociationInfo.PreambleType == 0)?"LONG":"SHORT", Adapter->AssociationInfo.PreambleType);)


#if 0
		if(!Adapter->UseWzcs)
		{
			if(Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == WEP_DISABLED && Adapter->BSSinfo[BSSindex].EncryptionIsOn)
			{
				Adapter->StationState = STATION_STATE_MGMT_ERROR;
				Adapter->MgmtErrorCode = MGMT_ERROR_WEP_REQUIRED;
				
				Adapter->StationIsAssociated		   = FALSE;
				
				IF_LOUD(DbgPrint("**** WEP must be enabled for this BSS ***\n");)
				return FALSE;
			}

			if(Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel != WEP_DISABLED && !Adapter->BSSinfo[BSSindex].UsingWEP)
			{
				Adapter->StationState = STATION_STATE_MGMT_ERROR;
				Adapter->MgmtErrorCode = MGMT_ERROR_WEP_NOT_REQUIRED;

				Adapter->StationIsAssociated		   = FALSE;

				IF_LOUD(DbgPrint("**** WEP must be disabled for this BSS ***\n");)
				return FALSE;
			}
		}
#endif
		Adapter->ApSelected = TRUE;

		Adapter->StationState = STATION_STATE_JOINNING;
		
		if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
			Join(Adapter, BSS_TYPE_INFRASTRUCTURE);
		else 
			Join(Adapter, BSS_TYPE_AD_HOC);


		return TRUE;
}




/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  SetMibValues
//
//	  Initializes the minimum MIB fields, necessary for the FW to operate
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

BOOLEAN
SetMibValues(PVNet_ADAPTER Adapter)
{
	UCHAR			status;

	// Tx Power Level
	status = SetTxPowerLevel(Adapter, Adapter->TxPowerLevel);

	if(!status)
		return FALSE;

	if(Adapter->RadioIsOn && !Adapter->InSnifferMode)
	{
		status = EnableRadio(Adapter);
		
		if(status == CMD_STATUS_REJECTED_RADIO_OFF)
		{
			IF_LOUD(DbgPrint("*** Radio ON command rejected (H/W Radio selection is OFF\n");)
#ifndef RFMD_3COM                        
			Adapter->RadioIsOn = FALSE;
#endif                        
			Adapter->HwRadioIsOn = FALSE;
		}
		else
		if(status != CMD_STATUS_COMPLETE)
		{
			IF_DEBUG_ERRORS(DbgPrint("Failed to turn Radio ON (0x%X)\n", status);)
			return FALSE;
		}
		else
		{
			IF_LOUD(DbgPrint("Radio ON command completed with SUCCESS (0x%X)...\n", status);)
		}

	}

	//	Data Frames Tx Rate Mode

	status = SetDataFramesTxRateMode(Adapter, Adapter->DataFramesTxRate);

	if(!status)
		return FALSE;

#if ((!defined RFMD) && (!defined RFMD_3COM))
// SSID
	status = SetSSIDsize(Adapter, Adapter->AssociationInfo.DesiredSSIDsize);
	if(!status)
		return FALSE;
#endif

	// Promiscuous
		
	status = SetPromiscuousMode(Adapter, PROM_MODE_OFF);
	if(!status)
		return FALSE;


	// RtsThreshold

	status = SetRtsThreshold(Adapter, Adapter->RtsThreshold);

	if(!status)
		return FALSE;


	//	Fragmentation Threshold

	status = SetFragmentationThreshold(Adapter, Adapter->FragmentationThreshold);

	if(!status)
		return FALSE;

	
	//Preamble	

	status = SetPreambleType(Adapter, Adapter->AssociationInfo.PreambleType);

	if(!status)
		return FALSE;


	if(!SetMACaddress(Adapter, Adapter->AssociationInfo.CurrentAddress))
		return FALSE;
	
	
	// PS MODE (always Active until the station becomes associated )
	
	status = SetPowerMgmtMode(Adapter, 0);

	if(!status)
		return FALSE;
	

	status = SetBeaconPeriod(Adapter, Adapter->DefaultBeaconPeriod);

	if(!status)
		return FALSE;

	status = SetSupportedRates(Adapter, &BasicRates[Adapter->BasicRatesIndex][0], 4);

	if(!status)
		return FALSE;

#ifdef INT_ROAM
	//Int Roaming

	status = SetIntRoaming(Adapter);

	if(!status)
		return FALSE;
#endif

	// WEP
	
	status = ConfigureWEP(Adapter);

	if(!status)
	{
		return FALSE;
	}

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//    ValidateChannelForDomain														  
//	
//    If International Roaming is enabled it returns TRUE if the channel is 
//    valid for the domain otherwise it returns FALSE. 															         
//	  If International Roaming is NOT enabled, according to the current 
//	  Regulatory Domain validates the channel. If the current channel is not 
//    valid for the domain, a valid channel is automatically selected and 
//	  returns always TRUE
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

BOOLEAN ValidateChannelForDomain(PVNet_ADAPTER Adapter, PUCHAR pChannel)
{
	if(Adapter->IntRoamingEnabled)
	{
#ifdef INT_ROAM
		if(Adapter->MultiDomainMib.DomainInfoValid && 
			Adapter->MultiDomainMib.ChannelList[(*pChannel)-1])
			return TRUE;
		else if(Adapter->MultiDomainMib.DomainInfoValid)		
		{
			IF_LOUD(DbgPrint("*** Channel %d is invalid (802.11d) ***\n", *pChannel);)
			return FALSE;
		}
		else 
			return TRUE;
#else
		IF_LOUD(DbgPrint("*** This Driver build does not support Int Roaming !!!***\n");)
		return FALSE;
#endif
	}
	else
	{
		switch(Adapter->RegDomain)
		{
		case REG_DOMAIN_FCC: //Channels 1-11
		case REG_DOMAIN_DOC: 
			if ((*pChannel<1) || (*pChannel>11))
			{
				*pChannel = 1;
			}
			break;

		case REG_DOMAIN_ETSI: //Channels 1-13
			if ((*pChannel<1) || (*pChannel>13))
			{
				*pChannel = 1;
			}
			break;

		case REG_DOMAIN_SPAIN: //Channels 10-11
			if ((*pChannel<10) || (*pChannel>11))
			{
				*pChannel = 10;
			}
			break;

		case REG_DOMAIN_FRANCE: //Channels 10-13
			if ((*pChannel<10) || (*pChannel>13)) 
			{
				*pChannel = 10;
			}
			break;

		case REG_DOMAIN_MKK: //Channel 14
			if (*pChannel!=14)
			{
				*pChannel = 14;
			}
			break;

		case REG_DOMAIN_MKK1: //Channels 1-14
			if ((*pChannel<1) || (*pChannel>14)) 
			{
				*pChannel = 1;
			}
			break;

		case REG_DOMAIN_ISRAEL: //Channels 3-9
			if ((*pChannel<3) || (*pChannel>9)) 
			{
				*pChannel = 3;
			}
			break;

		default:
			Adapter->RegDomain = REG_DOMAIN_MKK1; //Channels 1-14
			if ((*pChannel<1) || (*pChannel>14))
			{
				*pChannel = 1;
			}
		}

		return TRUE;
	}
}


VOID PrepareBSSList2View(PVNet_ADAPTER Adapter)
{
	BOOLEAN							SkipEntry = FALSE;
	int								i,j;

	Adapter->BSSListEntries2App = 0;
	for(i=0; i<Adapter->BSSListEntries; i++)
	{
		SkipEntry = FALSE;
		
		if(Adapter->BSSinfo[i].SSIDsize == 0)
			continue;
		else
		{
			for(j=0; j<Adapter->BSSinfo[i].SSIDsize; j++)
			{
				if(Adapter->BSSinfo[i].SSID[j] < 0x20)
				{
					SkipEntry = TRUE;
					break;
				}
			}

			if(!SkipEntry)
			{
				for(j=0; j<Adapter->BSSinfo[i].SSIDsize; j++)
				{
					if(Adapter->BSSinfo[i].SSID[j] != 0x20)
						break;
				}
				if(j == Adapter->BSSinfo[i].SSIDsize)
					SkipEntry = TRUE;
			}

			if(!SkipEntry)
			{
				VNetMoveMemory(Adapter->BSSinfo2App + Adapter->BSSListEntries2App, &Adapter->BSSinfo[i], sizeof(BSS_INFO));
				Adapter->BSSListEntries2App++;
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
//	  MemCompare														  
//																		 
//	  Compares a number of bytes of two buffers. Returns 0 if they are 
//	  identical 1 otherwise
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

int MemCompare(PUCHAR pMem1, PUCHAR pMem2, ULONG bytes)
{

	ULONG i;

	for(i=0; i<bytes; i++)
		if(pMem1[i] != pMem2[i])
			break;

	if(i == bytes)	
		return 0;
	else
		return 1;
}

#if 0
VOID VNetMoveMemory(PVOID pDest, PVOID pSrc, ULONG bytes)
{
	memcpy(pDest, pSrc, bytes);
}


VOID VNetZeroMemory(PUCHAR buf, ULONG bytes)
{
	memset(buf, 0, bytes);
}
#endif

VOID IndicateConnectionStatus(PVNet_ADAPTER Adapter, BOOLEAN Connected)
{
    if(Connected==TRUE){
        netif_wake_queue(Adapter->dev);
    }else{
        netif_stop_queue(Adapter->dev);
    }
}

VOID
SendAuthenticationEvent(PVNet_ADAPTER Adapter, ULONG flags)
{
    printk(KERN_INFO "Not Yet Implemented\n");
}

#if (!defined(PCMCIA_DEBUG) && !defined(PCI_DEBUG))
VOID DbgPrint(const char *fmt,...)
{
}
#endif

VOID DbgPrintBssListentries(PVNet_ADAPTER Adapter)
{
	UCHAR							index, len, i;

	DbgPrint("%d Entries in BSS List:\n", Adapter->BSSListEntries);
	DbgPrint("\n----------------------------------------------------\n");
	if(Adapter->BSSListEntries)
	{
		for(index=0; index<Adapter->BSSListEntries; index++)
		{
			DbgPrint("#%d: ", index);
			DbgPrint("Ch: %d - ", Adapter->BSSinfo[index].Channel );
			DbgPrint("BSSID: %02X:%02X:%02X:%02X:%02X:%02X - ", Adapter->BSSinfo[index].BSSID[0],Adapter->BSSinfo[index].BSSID[1],
				Adapter->BSSinfo[index].BSSID[2],Adapter->BSSinfo[index].BSSID[3],Adapter->BSSinfo[index].BSSID[4],Adapter->BSSinfo[index].BSSID[5]);
			for(len=0; len<Adapter->BSSinfo[index].SSIDsize; len++)
				DbgPrint("%c", Adapter->BSSinfo[index].SSID[len]);
			if(Adapter->BSSinfo[index].EncryptionIsOn && Adapter->BSSinfo[index].Capabilities & BSS_CAPAPILITY_WPA)
			{
				DbgPrint(" - WPA - Gr: %d - Pw: ", Adapter->BSSinfo[index].WpaInfo.GroupSuite);
//				DbgPrint("Pairwise Suite Cnt: %d - Pairwise Suites: ", pwcnt);
				for(i=0;i<Adapter->BSSinfo[index].WpaInfo.PairwiseSuiteCnt;i++)
					DbgPrint("%d ", Adapter->BSSinfo[index].WpaInfo.PairwiseSuiteList[i]);
				DbgPrint(" - Auth: ");
				for(i=0;i<Adapter->BSSinfo[index].WpaInfo.AuthMgmtSuiteCnt;i++)
					DbgPrint("%d ", Adapter->BSSinfo[index].WpaInfo.AuthMgmtSuiteList[i]);
			}
			else
			if(Adapter->BSSinfo[index].Capabilities & BSS_CAPAPILITY_CCX)
			{
				DbgPrint(" - CCX");
			}

			DbgPrint(" - WEP is %s", (Adapter->BSSinfo[index].EncryptionIsOn) ? "ON" : "OFF");
			
			DbgPrint(" - %s Pr", (Adapter->BSSinfo[index].PreambleType)  ? "SHORT" : "LONG");
			DbgPrint(" - RSSI %d\n", Adapter->BSSinfo[index].RSSI);
			DbgPrint("\n");
		}
	}
	else
	{
		DbgPrint("**** BSS List is empty ****\n");
	}
	DbgPrint("----------------------------------------------------\n");
}


BOOLEAN
RemoveKey(PVNet_ADAPTER Adapter, UCHAR index, BOOLEAN RemoveAll)
{
	ULONG					i,j,k;
	BOOLEAN					status;


	if(Adapter->SecurityInfo.AllKeysRemoved) // to avoid unnecessary key removals
	{
		IF_LOUD(DbgPrint("RemoveKey(): Nothing to do...All keys are already removed\n");)
		return TRUE;
	}

	if(RemoveAll)
	{
		IF_LOUD(DbgPrint("Remove ALL Keys\n");)
			
		Adapter->SecurityInfo.AllKeysRemoved = TRUE;

		k=0;
		for(i=0;i<MAX_ENCRYPTION_KEYS;i++)
		{
			for(j=0;j<32;j++)
			{
				Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[i][j] = Adapter->AssociationInfo.CurrentAddress[k];
				k++;
				if(k==6)
					k=0;
			}
			for(j=32;j<40;j++)
			{
				Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[i][j] = 0;
			}

		}

		status = SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[0][0], 0, (MAX_ENCRYPTION_KEYS * MAX_ENCRYPTION_KEY_SIZE));
	}
	else
	{
		if(index > 3)
		{
			IF_DEBUG_ERRORS(DbgPrint("ERROR: Invalid Index (0x%x). No Key Removed.\n");)
			return FALSE;
		}
		k=0;
		for(i=0;i<32;i++)
		{
			Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][i] = Adapter->AssociationInfo.CurrentAddress[k];
			k++;
			if(k==6)
				k=0;
		}
		for(i=32;i<40;i++)
		{
			Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][i] = 0;
		}


		status = SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][0], index, MAX_ENCRYPTION_KEY_SIZE);
	}

	return status;
}


#define ROL32( A, n )  ( ((A) << (n)) | ( ((A)>>(32-(n)))  & ( (1UL << (n)) - 1 ) ) )
#define ROR32( A, n ) ROL32( (A), 32-(n) )


void WPA_MicCalculation(UCHAR *msg,UINT msgLength )
{
	UINT i;
	ULONG *M;
	
	M=(ULONG *) msg;
	for(i=0;i<msgLength;i++)
	{
		Lefty ^= M[i];
		Righty ^= ROL32( Lefty, 17 );
		Lefty += Righty;
		Righty ^= ((Lefty & 0xff00ff00) >> 8) | ((Lefty & 0x00ff00ff) << 8);
		Lefty += Righty;
		Righty ^= ROL32( Lefty, 3 );
		Lefty += Righty;
		Righty ^= ROR32( Lefty, 2 );
		Lefty += Righty;
	}
}

void setAddressesField(UCHAR *pPacket,UCHAR *addressesField)
{
	UCHAR tmpSA,tmpDA, i;

	tmpDA = (!(pPacket[1] & 0x01)) ? 4 : 16; //To DS
	tmpSA = (pPacket[1] & 0x02) ? 16 : 10;  //From DS

	for(i=0;i<6;i++)
	{
		addressesField[i]=pPacket[tmpDA+i];
		addressesField[6+i]=pPacket[tmpSA+i];
	}
	for(i=12;i<16;i++)
		addressesField[i]=0;

}

void WPA_michael(UCHAR *packet, UCHAR *micKey, UINT msgLength)
{
	UCHAR   numOfBytes2Add, i, addressesField[16];	

	Lefty=((ULONG *) micKey)[0];
	Righty=((ULONG *) micKey)[1];

	//Michael operation on address fields: DA and SA.
	setAddressesField(packet,addressesField);
	WPA_MicCalculation(addressesField,4);
	//Padding
	packet[msgLength]=0x5a;
	numOfBytes2Add=(msgLength+1)%4;
	if(numOfBytes2Add==0) numOfBytes2Add=4;
	else numOfBytes2Add=8-numOfBytes2Add;
	
	for(i=1;i<=numOfBytes2Add;i++) packet[msgLength+i]=0;

	//Calc Mic for raw data			
	WPA_MicCalculation(&packet[24],(UINT)((msgLength+1+numOfBytes2Add-24)>>2));
}

UCHAR WPA_validateMic(UCHAR *packet,UINT msgLength,UCHAR *micKey)
{

	UCHAR	rxMic[8];
	UCHAR	i, *mic;

	//Keep the MIC value of the received packet
	mic = packet + msgLength - 8;
	for(i=0;i<8;i++) rxMic[i]=mic[i];
	//Calculate MIC from the payload of the received packet
	WPA_michael(packet, micKey, (UINT)(msgLength-8) );
	//Append received mic value at the end of the packet
	for(i=0;i<8;i++) mic[i]=rxMic[i];
	
	//Check if mic value received and calculated match
	if( (Lefty!=(((ULONG *) rxMic)[0]))||(Righty!=(((ULONG *) (rxMic+4) )[0]) ) )	
	{
		return 0;
	}
	return 1;
}

void WPA_addMICToPacket(UCHAR *packet,UINT msgLength,UCHAR *micKey)
{
	UCHAR i, *mic;
			
	WPA_michael(packet, micKey, (UINT)(msgLength-8) );

	mic = packet + msgLength - 8;
	for(i=0;i< 4;i++)
	{
		mic[i]=(UCHAR) (Lefty>>(8*i) );
		mic[i+4]=(UCHAR)(Righty>>(8*i) );
	}
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//		Compare_Array()  checks if x and y are identical 
//
//	returns:	TRUE  : x and y are identical
//				FALSE : x and y are not identical
/////////////////////////////////////////////////////////////////////////////
BOOLEAN Compare_Array(PUCHAR x,
                     PUCHAR y,
                     USHORT l)
{
    USHORT  i;

    for (i=0; i<l; i++)
    {
        if (x[i]!=y[i]) return FALSE;
    }

    return TRUE;
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	WPA_AddAPInBlackList()
//	Adds an AP to the MIC Black List
//
//	Parameters
//		MIC_BLACK_LIST*	Pointer to the black list
//		UCHAR			Size of the black list (entries, not bytes)
//		PUCHAR			Pointer to a parameter that keeps the entries that are in use
//		PUCHAR			Pointer to a 6-byte buffer that holds the AP's BSSID
//		PVOID			Pointer to a parameter that contains the timestamp (OS depended)
//		
//
//	returns:	TRUE  : if the AP was added in the list
//				FALSE : if the list was full
/////////////////////////////////////////////////////////////////////////////
BOOLEAN WPA_AddAPInBlackList(MIC_BLACK_LIST* MicBlackList, unsigned char ListEntries, unsigned char* pValidListEntries, unsigned char* pBSSID, void* pTime)
{
	UCHAR	i;
	UCHAR	ReservedBSSID[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

	// limit the search for empty entry to the ListEntries
	for(i=0; i<ListEntries; i++)
	{
		// if an entry is not used, use it
		if(!MicBlackList[i].InUse)
		{
			VNetMoveMemory(MicBlackList[i].Bssid, pBSSID, 6);
			MicBlackList[i].TimeStamp = *((ULONG64*)pTime);
			// mark it as used
			MicBlackList[i].InUse = TRUE;

			// increase the valid list entries and return success
			(*pValidListEntries) ++;
			return TRUE;
		}
	}

	// if we are here it means that all the list members are in use.
	// we have to check if there are any entries that can be removed.
	// we use the InvalidBSSID to force the WPA_IsOkToConnect to invalidate
	// all unnecessary entries.
	WPA_IsOkToConnect(MicBlackList, ListEntries, pValidListEntries, ReservedBSSID, pTime);

	// we might have found an empty place, so try again to add the AP to the list
	for(i=0; i<ListEntries; i++)
	{
		if(!MicBlackList[i].InUse)
		{
			VNetMoveMemory(MicBlackList[i].Bssid, pBSSID, 6);
			MicBlackList[i].TimeStamp = *((ULONG64*)pTime);
			MicBlackList[i].InUse = TRUE;

			(*pValidListEntries) ++;
			return TRUE;
		}
	}

	// could not add the AP in the list, return failure
	return FALSE;
}

MAC_ENCRYPTION_MIB		gMacEncryptionMib;

BOOLEAN
GetEncryptionMIB(PVNet_ADAPTER Adapter)
{
	GET_SET_MIB_STRUCT		sGetMIB;
	UCHAR					status = 0;

	sGetMIB.Type = (UCHAR) Mac_Encryption_Mib_Type;
	sGetMIB.Size = (UCHAR)sizeof(MAC_ENCRYPTION_MIB);
	sGetMIB.Index = (UCHAR)0;

	status = SendCommand(Adapter, CMD_Get_MIB_Vars, (PUCHAR)&sGetMIB, (PUCHAR)&gMacEncryptionMib, sizeof(MAC_ENCRYPTION_MIB), FALSE);

	if(status != CMD_STATUS_COMPLETE)
	{
		IF_LOUD(DbgPrint("******** FAILED to get MAC_ENCRYPTION_MIB (0x%X)...\n",status);)
		return FALSE;
	}

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	WPA_IsOkToConnect()
//	Checks if an AP is in the black list for more or less than 60 seconds
//	If it is more than 60 secondos, it invalidates that entry.
//
//	Parameters
//		MIC_BLACK_LIST*	Pointer to the black list
//		UCHAR			Size of the black list (entries, not bytes)
//		PUCHAR			Pointer to a parameter that keeps the entries that are in use
//		PUCHAR			Pointer to a 6-byte buffer that holds the AP's BSSID
//		PVOID			Pointer to a parameter that contains the timestamp (OS depended)
//		
//
//	returns:	TRUE  : if the AP was added in the list
//				FALSE : if the list was full
//
//	NOTE: The ReservedBSSID indicates that we should go throught the whole list
//		  until we find an AP that could be removed from the list, in order to
//		  create space for a new entry.
/////////////////////////////////////////////////////////////////////////////
BOOLEAN WPA_IsOkToConnect(MIC_BLACK_LIST* MicBlackList, unsigned char ListEntries, unsigned char* pValidListEntries, unsigned char* pBSSID, void* pCurrentTime)
{
	ULONG	TimeInterval;
	UCHAR	i;
	UCHAR	ValidEntriesInBlackListCounter = 0;
	UCHAR	ReservedBSSID[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

	// limit the search to the ListEntries
	for(i=0;i<ListEntries;i++)
	{
		// first check if this entry is valid
		if(MicBlackList[i].InUse)
		{
			ValidEntriesInBlackListCounter ++;

			// compare the requested BSSID with the list BSSID and the ReservedBSSID.
			if(Compare_Array(MicBlackList[i].Bssid, pBSSID, 6) || 
			   Compare_Array(ReservedBSSID, pBSSID, 6)
			  )
			{
				TimeInterval = (ULONG)((*(ULONG64*)pCurrentTime) - MicBlackList[i].TimeStamp);
				// check if the AP is more than 60 seconds
				if(TimeInterval > 600000000)
				{
					// invalidate the entry and reduce the valid list entries
					MicBlackList[i].InUse = FALSE;
					(*pValidListEntries) --;
					return TRUE;			
				}

				// if we are not going throught the whole list, indicate
				// that we have found an bad AP.
				if( !Compare_Array(ReservedBSSID, pBSSID, 6) )
					return FALSE;
			}

			if(ValidEntriesInBlackListCounter == *pValidListEntries)
				return TRUE;
		}
	}	
	return TRUE;
}

BOOLEAN
SetEncryptionStatus(PVNet_ADAPTER Adapter, BOOLEAN SetKey, ULONG KeyIndex, BOOLEAN IsPairWise)
{
	GET_SET_MIB_STRUCT		sSetMIB;
	UCHAR					status, index;
	MAC_ENCRYPTION_MIB		MacEncryptionMib;
	UCHAR					i;
	UCHAR					BytesToSet, offset;
#ifdef CCX
	UCHAR					CkipExpandedKey[16];
#endif

	
	VNetZeroMemory((PUCHAR)&MacEncryptionMib, sizeof(MAC_ENCRYPTION_MIB));

	if(Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_NONE && Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite == CIPHER_SUITE_NONE)
	{
		MacEncryptionMib.PrivacyInvoked = 0;
	}
	else
	{
		MacEncryptionMib.PrivacyInvoked = 1;
	}

	index = (UCHAR)(KeyIndex & 0xff);

	MacEncryptionMib.ExcludeUnencrypted = 1;
	
	if(Adapter->SecurityInfo.PairwiseKeyAvailable && Adapter->SecurityInfo.GroupKeyAvailable)
	{
		IF_LOUD(DbgPrint("Both Keys available (G: %d - P: %d)\n", (UCHAR)Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex, (UCHAR)Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex);)
		MacEncryptionMib.CipherDefaultKeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex & 0xff);
		MacEncryptionMib.CipherDefaultGroupKeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex & 0xff);
	}
	else
	if((Adapter->SecurityInfo.PairwiseKeyAvailable && !Adapter->SecurityInfo.GroupKeyAvailable)) 
	{
		IF_LOUD(DbgPrint("Pairwise Key ONLY available (%d)\n", (UCHAR)Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex);)

		MacEncryptionMib.CipherDefaultKeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex & 0xff);

		if(Adapter->SecurityInfo.InWpaMode)
			MacEncryptionMib.CipherDefaultGroupKeyID = 0xff;
		else
			MacEncryptionMib.CipherDefaultGroupKeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex & 0xff);
	}
	else
	if(!Adapter->SecurityInfo.PairwiseKeyAvailable && Adapter->SecurityInfo.GroupKeyAvailable)
	{
		IF_LOUD(DbgPrint("Group Key ONLY available (%d)\n", (UCHAR)Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex);)

		MacEncryptionMib.CipherDefaultGroupKeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex & 0xff);

		if(Adapter->SecurityInfo.InWpaMode)
			MacEncryptionMib.CipherDefaultKeyID = 0xff;
		else
			MacEncryptionMib.CipherDefaultKeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex & 0xff);
	}
	else
	{
		IF_LOUD(DbgPrint("NO Keys available (G: %d - P: %d)\n", (UCHAR)Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex, (UCHAR)Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex);)

		MacEncryptionMib.CipherDefaultKeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex & 0xff);
		MacEncryptionMib.CipherDefaultGroupKeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex & 0xff);
	}	
	
	VNetMoveMemory(MacEncryptionMib.ReceiverAddress, Adapter->SecurityInfo.EncryptionInfo.BSSID, 6);

#ifdef CCX
	if(Adapter->CCXinfo.CkipEnabled)
	{

		IF_LOUD(DbgPrint("KEY PERMUTATION (%d)\n", Adapter->CCXinfo.CkipKeyPermutation);)
		MacEncryptionMib.CkipKeyPermutation = Adapter->CCXinfo.CkipKeyPermutation;
	}
#endif

	BytesToSet = ENCRYPTION_PARAMETERS_BYTES;
	offset = (MAX_ENCRYPTION_KEYS * MAX_ENCRYPTION_KEY_SIZE);

	//////////////////////////////
	//	MAC_ENCRYPTION_MIB 
	//	
	//////////////////////////////

	sSetMIB.Type = Mac_Encryption_Mib_Type;
	sSetMIB.Size = BytesToSet;
	sSetMIB.Index = offset; 
	VNetMoveMemory(sSetMIB.Data, (PUCHAR)&MacEncryptionMib+offset, BytesToSet);


	status = SendCommand(Adapter, CMD_Set_MIB_Vars, (PUCHAR)&sSetMIB, NULL, 0, FALSE);

	if(status != CMD_STATUS_COMPLETE){
		IF_DEBUG_ERRORS(DbgPrint("FAILED to set keys (0x%X)...\n",status);)
		return FALSE;
	}

	if(SetKey)
	{
		Adapter->SecurityInfo.AllKeysRemoved = FALSE;

		//KeyRSC
		if(KeyIndex/* & 0x20000000*/)
		{
			IF_LOUD(DbgPrint("KeyRSC must be used (0x%X)\n", Adapter->SecurityInfo.EncryptionInfo.KeyRSC[index]);)
		}
		else
		{
			IF_LOUD(DbgPrint("KeyRSC must be reset\n");)

			VNetZeroMemory(Adapter->SecurityInfo.EncryptionInfo.KeyRSC[0], 32);
		}

		sSetMIB.Type = Mac_Encryption_Mib_Type;
		sSetMIB.Size = 6;
		sSetMIB.Index = ENCRYPTION_KEY_RSC_BYTES + (index * 8); // * 8 since the array is 4*8 although we use 6 bytes
		VNetMoveMemory(sSetMIB.Data, (PUCHAR)(&Adapter->SecurityInfo.EncryptionInfo.KeyRSC[index][0]), 6);

		status = SendCommand(Adapter, CMD_Set_MIB_Vars, (PUCHAR)&sSetMIB, NULL, 0, FALSE);

		if(status != CMD_STATUS_COMPLETE){
			IF_DEBUG_ERRORS(DbgPrint("FAILED to set KeyRSC (0x%X)...\n",status);)
			return FALSE;
		}


		for(i=32;i<MAX_ENCRYPTION_KEY_SIZE; i++)
			Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][i] = 0;
		
		Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-1] = 0x01;

		if(!KeyIndex) // Pairwise key
			Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-1] |= 0x02;

//		if((!Adapter->SecurityInfo.PairwiseKeyAvailable) || (KeyIndex & 0x40000000))
//		{
			if(!KeyIndex) // key that can be used for Tx
				Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-1] |= 0x04;
//		}

#ifdef CCX	
		if(Adapter->CCXinfo.CkipEnabled && Adapter->CCXinfo.CkipKeyPermutation)
		{
			Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = CIPHER_SUITE_CCX;
			Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite = CIPHER_SUITE_CCX;
		}
#endif 
		if(IsPairWise) // Pairwise key
			Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-2] = Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite;
		else
			Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][MAX_ENCRYPTION_KEY_SIZE-2] = Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite;
		
		SetEncryptionKey(Adapter, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[index][0], index, MAX_ENCRYPTION_KEY_SIZE);

	}

  GetEncryptionMIB(Adapter);
	return TRUE;
}


