/* $Id: vnetusba.h,v 1.21 2004/09/22 14:50:17 tmpetsos Exp $ */
/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/
#ifndef __vnetusba_h_OK__
#define __vnetusba_h_OK__
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/wireless.h>
#include <linux/etherdevice.h>
#include <linux/version.h>
#include <linux/usb.h>
#include <asm/uaccess.h>
#include "general.h"
#include "appint.h"

#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#endif

// Kernel 2.6 Related definitions
#define AT76C503A_USB_CONFDESCR_LEN     0x20
#define ATMEL_DEVDESCR_SIZE             0x12

#define HUB_ROOT_RESET_TIME	50	/* times are in msec */
#define PORT_RESET_TRIES	5
#define SET_ADDRESS_TRIES	2
#define GET_DESCRIPTOR_TRIES	2
#define SET_CONFIG_TRIES	2
#define HUB_RESET_TRIES			5
#define HUB_LONG_RESET_TIME		200
#define USB_PORT_FEAT_ENABLE		1
#define USB_MAXCONFIG			8
#define USB_RT_PORT	(USB_TYPE_CLASS | USB_RECIP_OTHER)
#define HUB_RESET_TIMEOUT	500
#define USB_PORT_FEAT_CONNECTION	0
#define USB_PORT_FEAT_ENABLE		1
#define USB_PORT_FEAT_SUSPEND		2
#define USB_PORT_FEAT_OVER_CURRENT	3
#define USB_PORT_FEAT_RESET		4
#define USB_PORT_FEAT_POWER		8
#define USB_PORT_FEAT_LOWSPEED		9
#define USB_PORT_FEAT_HIGHSPEED		10
#define USB_PORT_FEAT_C_CONNECTION	16
#define USB_PORT_FEAT_C_ENABLE		17
#define USB_PORT_FEAT_C_SUSPEND		18
#define USB_PORT_FEAT_C_OVER_CURRENT	19
#define USB_PORT_FEAT_C_RESET		20
#define USB_PORT_FEAT_TEST              21
#define USB_PORT_FEAT_INDICATOR         22
#define USB_PORT_STAT_CONNECTION	0x0001
#define USB_PORT_STAT_ENABLE		0x0002
#define USB_PORT_STAT_SUSPEND		0x0004
#define USB_PORT_STAT_OVERCURRENT	0x0008
#define USB_PORT_STAT_RESET		0x0010
/* bits 5 to 7 are reserved */
#define USB_PORT_STAT_POWER		0x0100
#define USB_PORT_STAT_LOW_SPEED		0x0200
#define USB_PORT_STAT_HIGH_SPEED        0x0400
#define USB_PORT_STAT_TEST              0x0800
#define USB_PORT_STAT_INDICATOR         0x1000
#define USB_PORT_STAT_C_CONNECTION	0x0001
#define USB_PORT_STAT_C_ENABLE		0x0002
#define USB_PORT_STAT_C_SUSPEND		0x0004
#define USB_PORT_STAT_C_OVERCURRENT	0x0008
#define USB_PORT_STAT_C_RESET		0x0010
#define USB_MAXALTSETTING		128
#if (LINUX_VERSION_CODE > 0x020500)
#define USB_MAXENDPOINTS		30
#endif
#define USB_MAXINTERFACES		32
#define HUB_SHORT_RESET_TIME	10
//End of Kernel Specific definitions

/* Kernel 2.4.20 Changes */
#if LINUX_VERSION_CODE > 132115
typedef struct devrequest {
    u8  requesttype;
    u8  request;
    u16  value;
    u16  index;
    u16 length;
}__attribute__((packed)) devrequest;

typedef struct urb *purb_t;
#endif	// Linux-2.4+

/* kernel 2.5.xx changes */
#ifndef USB_ST_NOERROR		// if noerror is defined, all are defined
 
#define USB_ST_NOERROR		0
#define USB_ST_CRC		(-EILSEQ)
#define USB_ST_BITSTUFF		(-EPROTO)
#define USB_ST_NORESPONSE	(-ETIMEDOUT)			/* device not responding/handshaking */
#define USB_ST_DATAOVERRUN	(-EOVERFLOW)
#define USB_ST_DATAUNDERRUN	(-EREMOTEIO)
#define USB_ST_BUFFEROVERRUN	(-ECOMM)
#define USB_ST_BUFFERUNDERRUN	(-ENOSR)
#define USB_ST_INTERNALERROR	(-EPROTO) 			/* unknown error */
#define USB_ST_SHORT_PACKET	(-EREMOTEIO)
#define USB_ST_PARTIAL_ERROR	(-EXDEV)			/* ISO transfer only partially completed */
#define USB_ST_URB_KILLED	(-ENOENT)			/* URB canceled by user */
#define USB_ST_URB_PENDING	(-EINPROGRESS)
#define USB_ST_REMOVED		(-ENODEV) 			/* device not existing or removed */
#define USB_ST_TIMEOUT		(-ETIMEDOUT)			/* communication timed out, also in urb->status**/
#define USB_ST_NOTSUPPORTED	(-ENOSYS)			
#define USB_ST_BANDWIDTH_ERROR	(-ENOSPC)			/* too much bandwidth used */
#define USB_ST_URB_INVALID_ERROR  (-EINVAL)			/* invalid value/transfer type */
#define USB_ST_URB_REQUEST_ERROR  (-ENXIO)			/* invalid endpoint */
#define USB_ST_STALL		(-EPIPE) 			/* pipe stalled, also in urb->status*/
#endif //USB_ST_*		
 
#ifndef FILL_BULK_URB
#define FILL_BULK_URB usb_fill_bulk_urb
#endif
 
#ifndef FILL_CONTROL_URB
#define FILL_CONTROL_URB usb_fill_control_urb
#endif
 
#if (LINUX_VERSION_CODE >= 0x020500)
#define usb_inc_dev_use		usb_get_dev		
#define usb_dec_dev_use		usb_put_dev		
#endif



/* the version of this driver */
#define VERSION_STRING "2.0.1.0"
// reawaken network queue this soon after stopping; else watchdog barks
#define TX_TIMEOUT_JIFFIES      (5*HZ)


/* when we got so many pending tx packets in TxList we stop the netdev queue */
#define THRE_STOP_NET_TX 20

/* and restart it below this value */
#define THRE_START_NET_TX 5

#define VNetVID				0x03EB

#define VNetPID_RFMD		0x7605
#define VNetPID_R505    	0x7606
#define VNetPID_R505A    	0x7614
#define VNetPID_R505AMX    	0x7615
#define VNetPID_R505AS    	0x7617
#define VNetPID_R505_2958    	0x7613

#define	VNET_TX_TIMEOUT		(HZ*10)

#define	PERIODIC_TIMER_OUT		2000

// version.....
#define USB_VNET_MAJOR_VERSION  2
#define USB_VNET_MINOR_VERSION  1
#define USB_VNET_SUB_VERSION	2
#define USB_VNET_BUILD	        0


#define RFMD_Board		1
#define R505_Board    	        4
#define R505_2958_Board    	5
#define R505A_Board    	        6
#define R505AMX_Board  	        7

#define VNET_MAX_WIRELESS_PACKET 1554

#define ACTIVE_MODE				0
#define PS_MODE					1
#define SPS_MODE				2

#define	USB_ALIGN(x)				x __attribute__((aligned(L1_CACHE_BYTES)))

// various flags
#define RX_BUSY					0x00000002	// receiver working
#define TX_SLEEP				0x00000004	// bulk out endpoint in use
#define STOPPED_RX				0x00000008	// used to stop Rx path....

//functionallity controls
#define	CTRL_URB_SLEEP			0x00000010	// ctrl endpoint in use
#define VNET_RUNNING			0x00000020	// driver up and running
#define NEED_CLEAR_HALT			0x00000040	// in case of usb_st_stall
#define INIT_PENDING			0x00000080	// flag for initialization
#define TX_STOPPED				0x00000100	// flag for tx queue status(system's queue)
#define INTERNAL_FW_DNLD		0x00000200	// internal f/w downloading (DFU)
#define USB_RESET_DEVICE     	0x00000400	// RESET NEEDED..
#define IOCTL_SLEEPING			0x00000800	// Ioctl Pending....
#define ChangePending			0x00001000
#define GetCommandPedding		0x00002000
// mgmt status
#define HeardBeacons			0x00004000	// check for out of range....
#define StationIsAuthenticated	0x00008000	// check if the station is authenticated                                                                                // (needed in Association Responce)
#define StationWasAssociated	0x00010000	// check if the station was associated
#define BSSIDfound				0x00020000	// in selectBSS flag for BSSID state
#define StationIsAssociated		0x00040000

#define SITE_SURVEY_REQUEST		0x00080000
// wep status
#define WepIsSet				0x00100000
//SetPreamble Status, Fragmentation, Rts/Cts
#define	PreambleChanged			0x00200000
#define FragChanged			0x00400000
#define RtsCtsChanged			0x00800000
#define AutoRFChanged			0x01000000

#define TimerReScheduled		0x02000000
#define WAKE_UP_WITH_ERROR		0x04000000

#define FIRST_SCAN_DONE			0x08000000
#define RadioOnOffChanged		0x10000000


// WEP key values
#define	WEP_DISABLED			0
#define	WEP_64BIT				1
#define	WEP_128BIT				2

#define WEP_MODE_OPTIONAL		0
#define WEP_MODE_MANDATORY		1

// Privacy
#define C80211_MGMT_AAN_OPENSYSTEM		0x0000
#define C80211_MGMT_AAN_SHAREDKEY		0x0001

#define	AD_HOC_MODE						1
#define	INFRASTRUCTURE_MODE				2


#define	PREAMBLE_TYPE_LONG				0
#define PREAMBLE_TYPE_SHORT				1

//Regulatory Domains
#define REG_DOMAIN_FCC          0x10    //Channels      1-11    USA
#define REG_DOMAIN_DOC          0x20    //Channel       1-11    Canada
#define REG_DOMAIN_ETSI         0x30    //Channel       1-13    Europe (ex Spain/France)
#define REG_DOMAIN_SPAIN        0x31    //Channel       10-11   Spain
#define REG_DOMAIN_FRANCE       0x32    //Channel       10-13   France
#define REG_DOMAIN_MKK          0x40    //Channel       14              Japan
#define REG_DOMAIN_MKK1         0x41    //Channel       1-14    Japan(MKK1)
#define REG_DOMAIN_ISRAEL       0x50    //Channel       3-9     ISRAEL

///////////////////////
// StationState defines
///////////////////////
#define STATION_STATE_INITIALIZING              0x00
#define STATION_STATE_SCANNING					0x01
#define STATION_STATE_JOINING 					0x02
#define STATION_STATE_AUTHENTICATING			0x03
#define STATION_STATE_ASSOCIATING				0x04
#define STATION_STATE_READY                     0x05
#define STATION_STATE_OUT_OF_RANGE				0x06
#define STATION_STATE_DISASSOCIATING			0x07
#define STATION_STATE_DEAUTHENTICATING			0x08
#define STATION_STATE_REASSOCIATING				0x09
#define STATION_STATE_RECONFIGURING				0x0A
#define STATION_STATE_FW_DOWNLOAD				0x0B
#define STATION_STATE_IDLE						0x0C

#define STATION_STATE_RESET						0x0F
#define STATION_STATE_EXITING					0x10
#define STATION_STATE_RADIO_OFF						0x13

#define STATION_STATE_MGMT_ERROR				0xf1
#define STATION_STATE_INIT_FAILED				0xf2	// if initialization failed
	// and a function sleeps(vnet_probe) when it wakes up it has to know what happened
/////////////////////////
// Device Operating Modes
/////////////////////////
#define DeviceNetCardMode					1
#define DeviceConfigurationMode				2
#define DeviceDFUMode						3
#define DeviceNOFlashNetCardMode			4

//#define UNDER_CONSTRUCTION    0x01
//#define TX_DATA                               0x02
//#define TX_MGMT                               0x04
#define UNDER_SUBMIT		0x08

//Setup Packet used in Ctrl urb's filler....

#define FILL_REQUEST(a,aa,ab,ac,ad,ae)\
  do {\
      (a)->devreq->request = aa;\
      (a)->devreq->requesttype = ab;\
      (a)->devreq->value = cpu_to_le16(ac);\
      (a)->devreq->index = cpu_to_le16(ad);\
      (a)->devreq->length = cpu_to_le16(ae);\
  }while(0)


typedef struct __DFU {
	PUCHAR ExternalRom;
	PUCHAR InternalRom;
	USHORT BlockCnt;
	UCHAR State;
	UCHAR StatusBuf[6];
	ULONG TimeOut;
	UCHAR ReSendState;
	USHORT ExternalFWSizeLeft;
	USHORT InternalFWSizeLeft;
} DFU, *PDFU;

#ifdef ATMEL_WLAN
#define HEADER_SIZE           14
#define MAXIMUM_PACKET_SIZE       1500
#define WIRELESS_HEADER_OVERHEAD    18
#define MAX_TX_PADDING_BYTES      50
#define MAX_RX_PADDING_BYTES      66
#define CRC32_BYTES           4
#define MIC_OVERHEAD          8
#define CCX_OVERHEAD          10

#define MAX_WIRELESS_SIZE     WIRELESS_HEADER_OVERHEAD + HEADER_SIZE + MAXIMUM_PACKET_SIZE + MIC_OVERHEAD + CRC32_BYTES + CCX_OVERHEAD

#define MAX_ASSOC_BUFF_LEN      255

#define MAX_ELEMENT_SIZE                                                       255
#define MAX_AP_ENTRIES                                                       64
#define ESS_ID_SIZE                                                       32

#define WPA_MODE_OFF  0
#define WPA_MODE_1    1
#define WPA_MODE_2    2

#define SSID_ELEMENT_OFFSET             14      //12+2
#define SSID_ELEMENT_LENGTH_OFFSET          13      //12+1
#define SSID_ELEMENT_ID_OFFSET            12      //12+1

#define SUP_RATES_ELEMENT_LENGTH_REL_OFFSET     15      //12+2+1
#define DS_PARAM_SET_ELEMENT_REL_OFFSET       18      //12+2+2+2
#define BSSID_HEADER_OFFSET             16
#define MGMT_FRAME_BODY_OFFSET            24

#define SUPPORTED_RATES_ELEMENT_ID                                 1
#define MIN_BEACON_PROBE_BYTES        60

#define MAX_WPA_ELEMENT_SIZE        200 //52+...
#define WPA_ELEMENT_GROUP_SUITE_OFFSET    6
#define WPA_ELEMENT_PW_SUITE_OFFSET     10
#define WPA_ELEMENT_PW_LIST_OFFSET      12
#define MIN_WPA_ELEMENT_DATA_SIZE     6
#define MAX_RSSI_SAMPLES          32
#define MAX_RECONNECT_RETRIES       9

#define MAX_EAPOL_PACKETS         5

#define EAPOL_KEY_INFO_TYPE   0x0008
#define EAPOL_KEY_INFO_ACK    0x0080
#define EAPOL_KEY_INFO_INDEX  0x0030

#define EAPOL_KEY_INFO_REQUEST  0x0800
#define EAPOL_KEY_INFO_ERROR  0x0400
#define EAPOL_KEY_INFO_SECURE 0x0200
#define EAPOL_KEY_INFO_MIC    0x0100

typedef struct __MIC_BLACK_LIST{
    UCHAR   Bssid[6];
    UCHAR   InUse;
    ULONG64   TimeStamp;
}MIC_BLACK_LIST, *PMIC_BLACK_LIST;


typedef enum _802_11_AUTHENTICATION_MODE
{
    _802_11AuthModeOpen,
    _802_11AuthModeShared,
    _802_11AuthModeAutoSwitch,
    _802_11AuthModeWPA,
    _802_11AuthModeWPAPSK,
    _802_11AuthModeWPANone,
    _802_11AuthModeWPA2,
    _802_11AuthModeWPA2PSK,
    _802_11AuthModeMax
} __802_11_AUTHENTICATION_MODE, *P_802_11_AUTHENTICATION_MODE;

typedef enum _802_11_WEP_STATUS
{
    _802_11WEPEnabled,  //0
    _802_11Encryption1Enabled = _802_11WEPEnabled, //0
    _802_11WEPDisabled,  //1
    _802_11EncryptionDisabled = _802_11WEPDisabled, //1
    _802_11WEPKeyAbsent, //2
    _802_11Encryption1KeyAbsent = _802_11WEPKeyAbsent, //2
    _802_11WEPNotSupported, //3
    _802_11EncryptionNotSupported = _802_11WEPNotSupported, //3
    _802_11Encryption2Enabled, //4
    _802_11Encryption2KeyAbsent,//5
    _802_11Encryption3Enabled, //6
    _802_11Encryption3KeyAbsent //7
} __802_11_ENCRYPTION_STATUS, *P_802_11_ENCRYPTION_STATUS;
//

#define MAX_PMKID_LIST_ENTRIES		16
#define MAX_PKMID_INFO_ENTRIES  16
#define MAX_PMKID_COUNT     4

#define RSN_ELEMENT_GROUP_SUITE_OFFSET    2
#define RSN_ELEMENT_PW_SUITE_OFFSET     6
#define RSN_ELEMENT_PW_LIST_OFFSET      8
#define MIN_RSN_ELEMENT_DATA_SIZE     2

// PMKID Structures
typedef UCHAR   _802_11_PMKID_VALUE[16];

typedef struct _BSSID_INFO
{
    MACADDR BSSID;
    _802_11_PMKID_VALUE PMKID;
} BSSID_INFO, *PBSSID_INFO;

typedef struct _802_11_PMKID_LIST
{
    ULONG Length;
    ULONG BSSIDInfoCount;
    BSSID_INFO BSSIDInfo[MAX_PMKID_LIST_ENTRIES];
} __802_11_PMKID_LIST, *P__802_11_PMKID_LIST;

typedef struct _PMKID_CANDIDATE {
    MACADDR BSSID;
    ULONG Flags;
} PMKID_CANDIDATE, *PPMKID_CANDIDATE;

typedef struct _PMKID_CANDIDATELIST
{
    ULONG Version;       // Version of the structure
    ULONG NumCandidates; // No. of pmkid candidates
    PMKID_CANDIDATE CandidateList[MAX_PKMID_INFO_ENTRIES];
} PMKID_CANDIDATELIST, *PPMKID_CANDIDATELIST;

typedef enum _802_11_PRIVACY_FILTER
{
    _802_11PrivFilterAcceptAll,
    _802_11PrivFilter8021xWEP
} __802_11_PRIVACY_FILTER, *P__802_11_PRIVACY_FILTER;


typedef struct __SECURITY_INFO
{
	__802_11_AUTHENTICATION_MODE 	AuthenticationMode;
	__802_11_PRIVACY_FILTER			PrivacyFilter;
  __802_11_ENCRYPTION_STATUS		EncryptionStatus;
	ENCRYPTION_INFO					EncryptionInfo;
	BOOLEAN 						AutoSwitchAuthAlgorithm;
	BOOLEAN							PairwiseKeyAvailable;
	BOOLEAN							GroupKeyAvailable;
	UCHAR							InWpaMode;
	UCHAR							MicErrorCnt;
	BOOLEAN							AllKeysRemoved;
	BOOLEAN							WPA_MixedModeDisabled;
	UCHAR							MicBlackListEntries;
	MIC_BLACK_LIST					MicBlackList[MAX_BSS_ENTRIES];

	unsigned long long				MicErrorDetected;

#ifdef WPA2
	__802_11_PMKID_LIST				PMKIDList;

	PMKID_CANDIDATELIST				PMKIDCandidateList;
	PMKID_CANDIDATELIST				PMKIDCandidateList2App;
#endif

#ifdef CCX
	CCX_INFO								CCXinfo;
#endif

}SECURITY_INFO, *PSECURITY_INFO;

#define MAX_EAPOL_PACKETS             5
//######################################################
//			common				
//
//			
typedef enum __ASSOCIATION_METHOD
{
    AssocMethodNoRequest,
    AssocMethodAssociationRequest,
    AssocMethodReAssociationRequest
}ASSOCIATION_METHOD, *PASSOCIATION_METHOD;

typedef struct  __EAPOL_BUFFER
{
    BOOLEAN   InUse;
      ULONG   Length;
        UCHAR   Packet[1600];
} EAPOL_BUFFER, *PEAPOL_BUFFER;


#define MAX_ASSOC_BUFF_LEN      255
typedef struct __ASSOCIATION_INFO{

UCHAR			CurrentBSSID[6];
USHORT 			AssocCapabilities;
UCHAR			LastBSSID[6];
USHORT 			ListenInterval;
UCHAR 			CurrentAddress[6];
USHORT 			BeaconPeriod;
UCHAR 			DesiredSSID[MAX_SSID_LENGTH];
UCHAR			IPAddress[4];
USHORT			CwMin;
USHORT			CwMax;
UCHAR 			PreambleType;
UCHAR         DesiredSSIDsize;
UCHAR			AssociationRequestRetryCnt;
UCHAR			ReAssociationRequestRetryCnt;
USHORT			AssociationRespFrameBodyBytes;
ASSOCIATION_METHOD	AssociationMethod;
SITE_SURVEY_INFO SiteS;
ULONG                                           LastAssReasReqBytes;
UCHAR                                           LastAssReasReqFrameBody[1500];
UCHAR                                           AssReassIEInfoBuffer[MAX_ASSOC_BUFF_LEN];
USHORT                                           AssReassIEInfoBufferLength;
}ASSOCIATION_INFO, *PASSOCIATION_INFO;

typedef struct __AUTHENTICATION_INFO{
	USHORT			CurrentAuthentTransactionSeqNum;
  UCHAR			AuthenticationRequestRetryCnt;
	USHORT			ExpectedAuthentTransactionSeqNum;	
	UCHAR 			pChallengeText[254];
	UCHAR 			ChallengeTextLength;
}AUTHENTICATION_INFO, *PAUTHENTICATION_INFO;

typedef struct __RX_BUFFER {
	USHORT WLength;
	UCHAR RxRate;
	UCHAR newbss;
	UCHAR KeyIndex;
	char RSSI;
	UCHAR LinkQuality;
	UCHAR CipherType;
	ULONG Reserved;
	UCHAR WirelessPacket[MAX_WIRELESS_SIZE];
}__attribute__((packed)) RX_BUFFER , *PRX_BUFFER ;

#else

typedef struct __RX_BUFFER {
	USHORT WLength;
	UCHAR RxRate;
	UCHAR newbss;
	UCHAR Fragmentation;
	UCHAR RSSI;
	UCHAR LinkQuality;
	UCHAR NoiseLevel;
	UCHAR RxTime[4];
	UCHAR WirelessPacket[VNET_MAX_WIRELESS_PACKET];
}__attribute__((packed)) RX_BUFFER, *PRX_BUFFER;
#endif

typedef struct __VNET_ADAPTER {
	struct usb_device *usb;
#if (LINUX_VERSION_CODE < 0x020500)
	struct usb_config_descriptor *config;
#else
 	struct usb_host_config *config;
#endif
	struct net_device *net;
	struct net_device_stats *netstats;
#ifdef WIRELESS_EXT
	struct iw_statistics *w_stats;
#endif
	struct list_head *TxList;
	struct list_head *freeTx;
	struct timer_list *CtrlTimer;
	devrequest *devreq;
	wait_queue_head_t *ctrl_wait;
	wait_queue_head_t *remove_wait;
	RX_BUFFER RxInBuff;
	RX_BUFFER RxBuffUp;
	STATISTICS Stats;
  SECURITY_INFO           SecurityInfo;
  ASSOCIATION_INFO          AssociationInfo;
  AUTHENTICATION_INFO         AuthenticationInfo;
	PDFU pDfu;
	purb_t ctrl_urb, rx_urb, tx_urb;
	spinlock_t lock;
	ULONG CrcTable[256];
	unsigned long flags;
	ULONG dbg_mask;		/* mask to control debug msgs, see DBG_* below */
	int (*cmdhandler) (struct __VNET_ADAPTER * Adapter);
	int (*Enchandler) (struct __VNET_ADAPTER * Adapter, BOOLEAN SetKey, ULONG KeyIndex);
        int received;
        int dropped;
        int transmitted;
	USHORT AssocID;
	USHORT RtsThreshold;
	USHORT FragmentationThreshold;
//	UCHAR ctrl_buff[200];
	UCHAR ctrl_buff[300];
	UCHAR CmdStatus[40];
	UCHAR SelectedSSID[ESSID_SIZE];
  UCHAR ChannelVector[14];
	UCHAR StationAddress[6];
	UCHAR tmpMAC[6];
	UCHAR FwVersion[4];
	UCHAR OperationalRates[4];
	UCHAR LastCommand[2];
	UCHAR BulkInAddr;
	UCHAR BoardType;
	UCHAR TxPacketsPending;

// Device Specific
	// Addresses
	// AP's Info
	UCHAR SSID_size;
	UCHAR Channel;
	// CardInfo
	UCHAR DevMode;		//Parallel, ....
	UCHAR TxRate;
	UCHAR PowerMgmtMode;
  UCHAR       PeriodicBeaconsCnt;
	UCHAR InternationalRoaming;
	UCHAR RadioIsOn;
	UCHAR OperatingMode;
	UCHAR Rssi;
	UCHAR LinkQuality;
	UCHAR StationState;
	UCHAR TempState;
	UCHAR PrivacyInvoked;
  UCHAR IsInternationalScanEnabled;
  UCHAR               CurrentTXpowerIndbm;

	// General Info
	UCHAR RegDomain;
	//driver info
	UCHAR CallbacksPending;


	UCHAR ErrorCode;
	//Mgmt Info


 	BOOLEAN					        SetKey;
  UCHAR               EapolMsg;
  UCHAR               LeapChallenge;
  UCHAR                 TmpKeyIndex;
  UCHAR                 KeyIndex;
  BOOLEAN               TxEapPending;
  BOOLEAN               IsBroadcast;
  BOOLEAN               IsTxKey;
  BOOLEAN               IsPairwiseKey;
  BOOLEAN               IsGroupKey;

}VNet_ADAPTER, *PVNet_ADAPTER;

#ifdef ATMEL_WLAN
typedef struct __TX_BUFFER {
	USHORT WLength;
	UCHAR TxRate;
	UCHAR PaddingBytes;
	UCHAR		KeyID;	
	UCHAR		CipherType;
	UCHAR		CipherLength;
	UCHAR		Reserved;	
	UCHAR WirelessPacket[MAX_WIRELESS_SIZE];
}__attribute__((packed)) TX_BUFFER , *PTX_BUFFER ;

#else

typedef struct __TX_BUFFER {
	USHORT WLength;
	UCHAR TxRate;
	UCHAR PaddingBytes;
	UCHAR Reserved[4];
	UCHAR WirelessPacket[1534];
}__attribute__((packed)) TX_BUFFER, *PTX_BUFFER;

#endif

typedef struct __TxList {
	struct list_head tx_list;
	struct sk_buff *skb;
	PVNet_ADAPTER Adapter;
	PTX_BUFFER tx_buff;
	USHORT length;
	UCHAR status;
}__attribute__((packed)) TX_LIST, *PTX_LIST;

#ifdef ATMEL_WLAN

#define MAX_ENCRYPTION_KEY_SIZE 40
#define CIPHER_SUITE_NONE 0
#define CIPHER_SUITE_WEP_64 1
#define CIPHER_SUITE_TKIP 2
#define CIPHER_SUITE_AES  3     
#define CIPHER_SUITE_CCX  4
#define CIPHER_SUITE_WEP_128  5

#endif


#ifdef USBDBG
//# define dbgusb(format, arg...) printk(KERN_INFO format, ## arg);

/* some flags to control dbgcond */
#define DBG_STATE             0x00000001	/* show state changes */
#define DBG_IOCTL             0x00000002	/* show ioctl calls */
#define DBG_FW_DL             0x00000004	/* firmware download */
#define DBG_INIT              0x00000008	/* initialization */
#define DBG_SCAN              0x00000010	/* scanning (I)BSS */
#define DBG_CTRL_CB           0x00000020	/* ctrl_callback */
#define DBG_CMD_CB            0x00000040	/* GetCmd_callback */
#define DBG_RX_CB             0x00000080	/* RxCallback */
#define DBG_DFU_CB            0x00000100	/* DFU callback */
#define DBG_CRC_CHECK         0x00000200	/* CRC check in RxCallback */
#define DBG_CTRL_URB          0x00000400	/* SendCtrlUrb */
#define DBG_CMDS              0x00000800	/* debug all commands to adapter */
#define DBG_JOIN              0x00001000	/* Join request and answer */
#define DBG_START_IBSS        0x00002000	/* debug only the START_IBSS cmds */
#define DBG_WEP               0x00004000	/* debug WEP keys */
#define DBG_BEACON            0x00008000	/* debug all incoming beacon msgs - many ! */
#define DBG_AUTH              0x00010000	/* debug AUTH msgs */
#define DBG_TXDATA            0x00020000	/* debug first bytes of wireless tx data - for every packet ! */
#define DBG_TX                0x00040000	/* debug other TX stuff, e.g. en/dequeueing */
#define DBG_SCHEDULE          0x00080000	/* debug scheduling */
#define DBG_ASSOC             0x00100000	/* debug association */
#define DBG_DISASS            0x00200000	/* debug disassociation */
#define DBG_CONFIG            0x00400000	/* debug configuration */
#define DBG_RXDATA            0x00800000	/* debug rx data packets - every packet ! */
#define DBG_TX_CB             0x01000000	/* debug the TxOk callback */

# define dbgcond(flag, format, arg...) \
  do {\
    if ((flag) & Adapter->dbg_mask) \
      printk(format, ## arg);\
  } while (0)

# ifdef INSANEDBG
#define dbgusbnl(format, arg...) {\
	printk(KERN_INFO format "\n" , ## arg);\
	udelay(1000);udelay(1000);udelay(1000);}
# else
#  define dbgusbnl(format, arg...) printk(KERN_INFO format "\n" , ## arg);
# endif

#else
# define dbgcond(flag, format, arg...)
# define dbgusb(format, arg...) {}
# define dbgusbnl(format, arg...) {}
#endif

#include "stdinclude.h"
#endif				//__vnetusba_h_OK__
