/* $Id: rx.c,v 1.13 2004/09/16 16:26:18 tmpetsos Exp $ */
/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/
#include "vnetusba.h"
#include "rx.h"
#include "frame.h"
#include "mib.h"
#include "stdinclude.h"

#define MAX_ELEMENT_SIZE                255
#define SUPPORTED_RATES_ELEMENT_ID      1
#define SSID_ELEMENT_OFFSET             14
#define ELEMENT_ID_WPA             0xDD
#define C80211_MGMT_ElementID_WPA                               221 //0xDD
#define MAX_BEACON_PROBE_BYTES                                         MAX_IE_LENGTH
#define WPA_ELEMENT_GROUP_SUITE_OFFSET                         6
#define        WPA_ELEMENT_PW_SUITE_OFFSET                                     10
#define WPA_ELEMENT_PW_LIST_OFFSET                                     12
#define        MIN_WPA_ELEMENT_DATA_SIZE                                       6
UCHAR   OUI_WPA[4] = {0x00, 0x50, 0xF2, 0x01};

/*************************************************
	UsbRxInit : If Bulk In endpoint polling has
	started return or else start polling on it.
*************************************************/
int UsbRxInit(PVNet_ADAPTER Adapter)
{
	if (Adapter->rx_urb->status == -EINPROGRESS)
		return 0;	// bulk in polling ok return
	FILL_BULK_URB(Adapter->rx_urb,
		      Adapter->usb,
		      usb_rcvbulkpipe(Adapter->usb, Adapter->BulkInAddr),
		      (PVOID)&Adapter->RxInBuff,
		      MAX_WIRELESS_SIZE + 2, RxCallback, Adapter);

#if (LINUX_VERSION_CODE < 0x020500)
	if (usb_submit_urb(Adapter->rx_urb) != 0) {
#else
	if (usb_submit_urb(Adapter->rx_urb, GFP_KERNEL) != 0) {
#endif
		err("bulk-in polling initialization failed");
		return -EINVAL; //-1;
	}

	dbgcond(DBG_INIT, "bulk-in polling started\n");
	Adapter->flags &= ~STOPPED_RX;
	return 0;
}

//
// ParseBeaconOrProbeRespone
//
VOID ParseBeaconOrProbeResponse(PVNet_ADAPTER Adapter, PSITE_SURVEY_INFO BSSList, PUCHAR FrameBody, UCHAR tmpChannel, PUCHAR BSSID, 
							   USHORT CapabilityInformation, USHORT FrameLength, USHORT i)
{
	USHORT		ElementBodyLen, len, pwcnt, authcnt;
	UCHAR		ElementBody[MAX_ELEMENT_SIZE];
	USHORT		m, n __attribute__ ((unused));

	// SSID
    GetSSIDfromBeaconOrProbe((PUCHAR)&BSSList->BssInfo[i].SSID,
                             (PUCHAR)&BSSList->BssInfo[i].SSIDsize,
                             FrameBody);
    // Channel
	BSSList->BssInfo[i].Channel = tmpChannel;
	// BSSID
    memcpy((PUCHAR)&BSSList->BssInfo[i].BSSID, (PUCHAR)BSSID, 6);
	// Preamble Type
    if (CapabilityInformation & C80211_MGMT_CAPABILITY_ShortPreamble)
        BSSList->BssInfo[i].PreambleType = 1;
    else
        BSSList->BssInfo[i].PreambleType = 0;
	// Encryption
    if (CapabilityInformation & C80211_MGMT_CAPABILITY_Privacy)
        BSSList->BssInfo[i].EncryptionIsOn = 1;
    else
        BSSList->BssInfo[i].EncryptionIsOn = 0;
	// Beacon Period
    BSSList->BssInfo[i].BeaconPeriod = *(PUSHORT)(FrameBody+8);
	// BSS Type
    if (CapabilityInformation & C80211_MGMT_CAPABILITY_ESS)
        BSSList->BssInfo[i].BSStype = 2; //Infrastructure
    else
        BSSList->BssInfo[i].BSStype = 1; //Ad-Hoc

	// Supported Rates
	GetSupportedRatesfromBeaconOrProbe(BSSList->BssInfo[i].SupportedRates, 
									   &BSSList->BssInfo[i].SupportedRatesLength,
									   FrameBody,SSID_ELEMENT_OFFSET+ BSSList->BssInfo[i].SSIDsize);

	BSSList->BssInfo[i].Capabilities = 0;
	len = (FrameLength > MAX_BEACON_PROBE_BYTES) ? MAX_BEACON_PROBE_BYTES : FrameLength;

	// RSN Element
#ifdef WPA2
	ElementBodyLen = MAX_ELEMENT_SIZE;
	
	if(GetElementFromBeaconOrProbe(FrameBody, len, ELEMENT_ID_RSN, OUI_NONE, ElementBody, &ElementBodyLen))
	{
		BSSList->BssInfo[i].Capabilities |= BSS_CAPAPILITY_RSN;

		// GROUP CIPHER SUITE
		if(ElementBodyLen >= 6)
			BSSList->BssInfo[i].Wpa2Info.GroupSuite = ElementBody[RSN_ELEMENT_GROUP_SUITE_OFFSET+3];
		else
			BSSList->BssInfo[i].Wpa2Info.GroupSuite = CIPHER_SUITE_AES;//default value
		
		// PAIRWISE CIPHER SUITE COUNT
		if(ElementBodyLen >= 8)			
			NdisMoveMemory(&BSSList->BssInfo[i].Wpa2Info.PairwiseSuiteCnt, &ElementBody[RSN_ELEMENT_PW_SUITE_OFFSET], 2);
		else
			BSSList->BssInfo[i].Wpa2Info.PairwiseSuiteCnt = 1;//default value

		pwcnt = BSSList->BssInfo[i].Wpa2Info.PairwiseSuiteCnt;
		
		if(pwcnt > MAX_PAIRWISE_SUITES)
		{
			pwcnt = MAX_PAIRWISE_SUITES;
			BSSList->BssInfo[i].Wpa2Info.PairwiseSuiteCnt = MAX_PAIRWISE_SUITES;
		}

		// PAIRWISE CIPHER SUITES
		if(ElementBodyLen >= (8 + 4*pwcnt))			
		{	
			for(k=1;k<=pwcnt;k++)
				NdisMoveMemory(&BSSList->BssInfo[i].Wpa2Info.PairwiseSuiteList[k-1], &ElementBody[RSN_ELEMENT_PW_LIST_OFFSET+(4*k)-1], 1);
		}
		else
			BSSList->BssInfo[i].Wpa2Info.PairwiseSuiteList[0] = CIPHER_SUITE_AES;//default value

		// AUTHENTICATION AND KEY MGMT SUITE COUNT
		if(ElementBodyLen >= (12 + (4 * pwcnt)))			
			NdisMoveMemory(&BSSList->BssInfo[i].Wpa2Info.AuthMgmtSuiteCnt, &ElementBody[RSN_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)], 2);
		else
			BSSList->BssInfo[i].Wpa2Info.AuthMgmtSuiteCnt = 1;//default value

		authcnt = BSSList->BssInfo[i].Wpa2Info.AuthMgmtSuiteCnt;

		if(authcnt > MAX_AUTH_MGMT_SUITES)
		{
			authcnt = MAX_AUTH_MGMT_SUITES;
			BSSList->BssInfo[i].Wpa2Info.AuthMgmtSuiteCnt = MAX_AUTH_MGMT_SUITES;
		}
		
		// AUTHENTICATION AND KEY MGMT SUITES
		if(ElementBodyLen >= (12 + (4 * pwcnt) + 2))
		{
			for(k=1;k<=authcnt;k++)
				memcpy(&BSSList->BssInfo[i].Wpa2Info.AuthMgmtSuiteList[k-1], &ElementBody[RSN_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)+2+(4*k)-1], 1);
		}
		else
		{
			BSSList->BssInfo[i].Wpa2Info.AuthMgmtSuiteList[0] = 1; //802.1x //default value
		}
		
		// RSN CAPABILITIES
		if(ElementBodyLen >= (12 + (4 * pwcnt) + 2 + ( 4 * authcnt )))
			memcpy(&BSSList->BssInfo[i].Wpa2Info.WpaCapabilities, &ElementBody[RSN_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)+2+(4*authcnt)], 2);
		else
			memset(&BSSList->BssInfo[i].Wpa2Info.WpaCapabilities, 0, 2);

		dbgcond(DBG_BEACON,("\n======================= RSN Info =======================\n"));
		dbgcond(DBG_BEACON,("Group Suite: %d", BSSList->BssInfo[i].Wpa2Info.GroupSuite));
		dbgcond(DBG_BEACON,("Pairwise Suite Cnt: %d - Pairwise Suites: ", pwcnt));
		for(k=0;k<pwcnt;k++)
			dbgcond(DBG_BEACON,("%02X ", BSSList->BssInfo[i].Wpa2Info.PairwiseSuiteList[k]));
      dbgcond(DBG_BEACON, "\n");
		dbgcond(DBG_BEACON,("Authentication Suite Cnt: %d - Authentication Suites: ", authcnt));
		for(k=0;k<authcnt;k++)
			dbgcond(DBG_BEACON,("%02X ", BSSList->BssInfo[i].Wpa2Info.AuthMgmtSuiteList[k]));
      dbgcond(DBG_BEACON, "\n");
		dbgcond(DBG_BEACON,("RSN Capabilities: %02X-%02X", BSSList->BssInfo[i].Wpa2Info.WpaCapabilities[0], BSSList->BssInfo[i].Wpa2Info.WpaCapabilities[1]));
		dbgcond(DBG_BEACON,("\n========================================================\n"));
	}
#endif

	// WPA element
	ElementBodyLen = MAX_ELEMENT_SIZE;
	if(GetElementFromBeaconOrProbe(FrameBody, len, ELEMENT_ID_WPA, OUI_WPA, ElementBody, &ElementBodyLen))
	{
		BSSList->BssInfo[i].Capabilities |= BSS_CAPAPILITY_WPA;

		// Group Suite
		if(ElementBodyLen >= 10)
			BSSList->BssInfo[i].WpaInfo.GroupSuite = ElementBody[WPA_ELEMENT_GROUP_SUITE_OFFSET+3];
		else
			BSSList->BssInfo[i].WpaInfo.GroupSuite = CIPHER_SUITE_TKIP;
	
		// Pairwise Suite count	
		if(ElementBodyLen >= 12)
			memcpy(&BSSList->BssInfo[i].WpaInfo.PairwiseSuiteCnt, &ElementBody[WPA_ELEMENT_PW_SUITE_OFFSET], 2);
		else
			BSSList->BssInfo[i].WpaInfo.PairwiseSuiteCnt = 1;

		pwcnt = BSSList->BssInfo[i].WpaInfo.PairwiseSuiteCnt;
		if(pwcnt > MAX_PAIRWISE_SUITES)
			pwcnt = MAX_PAIRWISE_SUITES;
		
		// Pairwise Suite List
		if(ElementBodyLen >= 16)
		{
			for(m=1;m<=pwcnt;m++)   
				memcpy(&BSSList->BssInfo[i].WpaInfo.PairwiseSuiteList[m-1], &ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*m)-1], 1);
		}
		else
			BSSList->BssInfo[i].WpaInfo.PairwiseSuiteList[0] = 2;

		// Authentication Mgmt Suite Count
		if(ElementBodyLen >= (16 + (4 * pwcnt)))
			memcpy(&BSSList->BssInfo[i].WpaInfo.AuthMgmtSuiteCnt, &ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)], 2);
		else
			BSSList->BssInfo[i].WpaInfo.AuthMgmtSuiteCnt = 1;

		authcnt = BSSList->BssInfo[i].WpaInfo.AuthMgmtSuiteCnt;
		if(authcnt > MAX_AUTH_MGMT_SUITES)
			authcnt = MAX_AUTH_MGMT_SUITES;
		
		// Authentication Mgmt Suite List
		if(ElementBodyLen >= (16 + (4 * pwcnt) + 2))
		{
			for(m=1;m<=authcnt;m++) 
			{
#ifdef CCX2
				if((ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)+2+4*(m-1)]   == OUI_CCKM[0]) &&
				   (ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)+2+4*(m-1)+1] == OUI_CCKM[1]) &&
				   (ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)+2+4*(m-1)+2] == OUI_CCKM[2]) &&
				   (ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)+2+4*(m-1)+3] == OUI_CCKM[3]))
				{
					BSSList->BssInfo[i].Capabilities |= BSS_CAPAPILITY_CCKM;
				}
#endif
				memcpy(&BSSList->BssInfo[i].WpaInfo.AuthMgmtSuiteList[m-1], &ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)+2+(4*m)-1], 1);
			}
		}
		else
			BSSList->BssInfo[i].WpaInfo.AuthMgmtSuiteList[0] = 1; //802.1x

		// WPA Capabilities
		if(ElementBodyLen >= (16 + (4 * pwcnt) + 2 + ( 4 * authcnt )))
			memcpy(&BSSList->BssInfo[i].WpaInfo.WpaCapabilities, &ElementBody[WPA_ELEMENT_PW_LIST_OFFSET+(4*pwcnt)+2+(4*authcnt)], 2);
	
		dbgcond(DBG_SCAN, "\n======================= WPA Info =======================\n");
		//for (n=0;n<BSSList->BssInfo[i].SSIDsize;n++)
    {
      char buf[64 + 1] __attribute__ ((unused));
			dbgcond(DBG_SCAN,"%s", ssid2str(buf, BSSList->BssInfo[i].SSID, BSSList->BssInfo[i].SSIDsize));
    }
		dbgcond(DBG_SCAN, "\n======================= WPA Info Element=======================\n");
    
                        BSSList->BssInfo[i].WpaIELen = ElementBodyLen+2;
                        *BSSList->BssInfo[i].WpaIE = 0xDD;
                        *(BSSList->BssInfo[i].WpaIE + 1) = ElementBodyLen;
                        memcpy(BSSList->BssInfo[i].WpaIE + 2, ElementBody, ElementBodyLen);
    for ( n = 0; n < BSSList->BssInfo[i].WpaIELen; n++) { 
			dbgcond(DBG_SCAN,"%2x", BSSList->BssInfo[i].WpaIE[n]);
    }
		dbgcond(DBG_SCAN, "\n========================================================\n");
	
		dbgcond(DBG_SCAN, "Group Suite: %d\n", BSSList->BssInfo[i].WpaInfo.GroupSuite);
		dbgcond(DBG_SCAN, "Pairwise Suite Cnt: %d - Pairwise Suites: ", pwcnt);
		for(i=0;i<pwcnt;i++)
			dbgcond(DBG_SCAN, "%02X", BSSList->BssInfo[i].WpaInfo.PairwiseSuiteList[i]);
    dbgcond(DBG_SCAN, "\n");
		dbgcond(DBG_SCAN, "Authentication Suite Cnt: %d - Authentication Suites: ", authcnt);
		for(i=0;i<authcnt;i++)
			dbgcond(DBG_SCAN, "%02X ", BSSList->BssInfo[i].WpaInfo.AuthMgmtSuiteList[i]);
    dbgcond(DBG_SCAN, "\n");

		dbgcond(DBG_SCAN, "WPA Capabilities: %02X-%02X", BSSList->BssInfo[i].WpaInfo.WpaCapabilities[0], BSSList->BssInfo[i].WpaInfo.WpaCapabilities[1]);
		dbgcond(DBG_SCAN, "\n========================================================\n");
	}

#ifdef CCX
	// Aironet element
	if(BSSList->BssInfo[i].EncryptionIsOn && GetElementFromBeaconOrProbe(FrameBody, len, ELEMENT_ID_AIRONET, OUI_NONE, ElementBody, &ElementBodyLen))
		BSSList->BssInfo[i].Capabilities |= BSS_CAPAPILITY_CCX;
#endif
#if(CCX2)
	// Cell Power Limit Element 
	if(GetElementFromBeaconOrProbe(FrameBody, len, ELEMENT_ID_CELL_POWER_INFO, OUI_CCX_TPC, ElementBody, &ElementBodyLen))
	{
		BSSList->BssInfo[i].Capabilities |= BSS_CAPAPILITY_CCX2;
		BSSList->BssInfo[i].CellPowerLimit = ElementBody[CELL_POWER_ELEMENT_TPC_OFFSET];									
	}
	// Version Number CCX Element	
	if(GetElementFromBeaconOrProbe(FrameBody, len, ELEMENT_ID_VERSION_NUM, OUI_CCX_VERSION, ElementBody, &ElementBodyLen))
	{
		BSSList->BssInfo[i].Capabilities |= BSS_CAPAPILITY_CCX2;
		BSSList->BssInfo[i].VersionNumber = ElementBody[VERSION_NUMBER_OFFSET];									
	}
	else
	{
		BSSList->BssInfo[i].VersionNumber = 0x00;
	}
#endif

	if(len > MAX_IE_LENGTH)
		len = MAX_IE_LENGTH;

	memcpy(&BSSList->BssInfo[i].BeaconProbeInfo, FrameBody, len);
	BSSList->BssInfo[i].BeaconProbeLen = len;

}

/************************************************************
	MgmtFrameProcessing : Processing of Received management
	frames. From here we get information for any Access Point
	or station in the area.	
************************************************************/
VOID MgmtFrameProcessing(PVNet_ADAPTER Adapter, UCHAR FrameSubType)
{
	PUCHAR DestinationAddress;
	PUCHAR SourceAddress;
	PUCHAR BSSID;
	PUCHAR FrameBody;
	PUCHAR AuthenticationFrame;
	USHORT CapabilityInformation;	// 2 bytes
	USHORT FrameLength;
	USHORT AssocID;
	UCHAR SupportedRatesLength;
	UCHAR SupportedRates[4];
	int i; //k, l;
	UCHAR tmpSSIDLen, tmpSRatesLen, tmpChannel;
//	UCHAR ChallengeTextLength;
	UCHAR AssociationRespFrameBytesToGet;
	PUCHAR AssociationRspnsFrame;

	Adapter->Stats.RxMgmtPacketsOk++;

	DestinationAddress =
	    Adapter->RxInBuff.WirelessPacket + MGMT_FRAME_DA_OFFSET;
	SourceAddress =
	    Adapter->RxInBuff.WirelessPacket + MGMT_FRAME_SA_OFFSET;
	BSSID =
	    Adapter->RxInBuff.WirelessPacket +
	    WIRELESS_HEADER_ADDRESS3_OFFSET;
	FrameBody =
	    Adapter->RxInBuff.WirelessPacket + MGMT_FRAME_BODY_OFFSET;

  FrameLength = Adapter->RxInBuff.WLength - MGMT_FRAME_BODY_OFFSET;
	switch (FrameSubType) {
	case C80211_SUBTYPE_MGMT_BEACON:
	case C80211_SUBTYPE_MGMT_ProbeResponse:

#ifdef ATMEL_WLAN
		if (Adapter->RxInBuff.WLength < 37)
			break;
#else
		if (Adapter->RxInBuff.WLength < 41)
			break;
#endif

		/* the cap info is in little endian ... */
		CapabilityInformation =
		    (USHORT) FrameBody[CAPABILITY_INFO_IN_BEACON_OFFSET] |
		    ((USHORT)
		     FrameBody[CAPABILITY_INFO_IN_BEACON_OFFSET + 1]) << 8;

		tmpSSIDLen = *(FrameBody + SSID_ELLEMENT_LENGTH_OFFSET);
		tmpSRatesLen = *(FrameBody + tmpSSIDLen +
				 SUP_RATES_ELLEMENT_LENGTH_REL_OFFSET);
		tmpChannel = *(FrameBody + tmpSSIDLen +
			       tmpSRatesLen +
			       DS_PARAM_SET_ELLEMENT_REL_OFFSET);

		{
			char buf[2 * ESSID_SIZE + 1] __attribute__ ((unused));
#if 0
			dbgcond(DBG_BEACON,
				"got beacon on ch %d, ssid %s, capa x%x, "
				"rssi %d, link qual %d\n",
				tmpChannel, ssid2str(buf,
						     &FrameBody
						     [SSID_ELLEMENT_OFFSET],
						     MIN(ESSID_SIZE,
							 FrameBody
							 [SSID_ELLEMENT_LENGTH_OFFSET])),
				CapabilityInformation,
				Adapter->RxInBuff.RSSI,
				Adapter->RxInBuff.LinkQuality);
#endif
		}

		if (tmpSSIDLen > 32) {
			Adapter->Stats.RxMgmtPacketsError++;
			break;
		}
		switch (Adapter->StationState) {
		case STATION_STATE_SCANNING:

			dbgcond(DBG_CMDS, "found BSS ch %d, bssid "
				"%02x:%02x:%02x:%02x:%02x:%02x, "
				"cap x%04x, RSSI %d\n",
				tmpChannel, BSSID[0], BSSID[1], BSSID[2],
				BSSID[3], BSSID[4], BSSID[5],
				CapabilityInformation,
				Adapter->RxInBuff.RSSI);

			if (Adapter->AssociationInfo.SiteS.BSSInList < MAX_BSS_ENTRIES) {
				for (i = 0; i < Adapter->AssociationInfo.SiteS.BSSInList;
				     i++) {
					if (memcmp (BSSID, Adapter->AssociationInfo.SiteS.BssInfo[i].BSSID, 6) == 0) {
						//dbgcond(DBG_SCAN, "already in list\n");
						break;
					}
				}

				if (i == Adapter->AssociationInfo.SiteS.BSSInList) {
					/* BSS not in list yet */
           ParseBeaconOrProbeResponse(Adapter, &Adapter->AssociationInfo.SiteS,FrameBody, tmpChannel, BSSID, CapabilityInformation, FrameLength, i);
           {
					char buf[ESSID_SIZE + 1] __attribute__ ((unused));

					GetSSIDfromBeaconOrProbe(Adapter->AssociationInfo.SiteS.BssInfo[Adapter->AssociationInfo.SiteS.BSSInList].SSID, &Adapter->AssociationInfo.SiteS.BssInfo[Adapter->AssociationInfo.SiteS.BSSInList].SSIDsize,
								 FrameBody);
					dbgcond(DBG_SCAN, "[ %d ] : SSID: %s\n", i,
						ssid2str(buf,
							 Adapter->AssociationInfo.SiteS.
							 BssInfo[Adapter->
								 AssociationInfo.SiteS.
								 BSSInList].
							 SSID,
							 Adapter->AssociationInfo.SiteS.
							 BssInfo[Adapter->
								 AssociationInfo.SiteS.
								 BSSInList].
							 SSIDsize));

					Adapter->AssociationInfo.SiteS.BssInfo[Adapter->
							       AssociationInfo.SiteS.
							       BSSInList].
					    Channel = tmpChannel;
					memcpy(Adapter->AssociationInfo.SiteS.
					       BssInfo[Adapter->AssociationInfo.SiteS.
						       BSSInList].BSSID,
					       BSSID, 6);
					Adapter->AssociationInfo.SiteS.BssInfo[Adapter->
							       AssociationInfo.SiteS.
							       BSSInList].
					    RSSI = Adapter->RxInBuff.RSSI;
					if (CapabilityInformation &
					    C80211_MGMT_CAPABILITY_ShortPreamble)
						Adapter->AssociationInfo.SiteS.
						    BssInfo[Adapter->AssociationInfo.SiteS.
							    BSSInList].
						    PreambleType = 1;
					else
						Adapter->AssociationInfo.SiteS.
						    BssInfo[Adapter->AssociationInfo.SiteS.
							    BSSInList].
						    PreambleType = 0;

					if (CapabilityInformation &
					    C80211_MGMT_CAPABILITY_Privacy)
						Adapter->AssociationInfo.SiteS.
						    BssInfo[Adapter->AssociationInfo.SiteS.
							    BSSInList].EncryptionIsOn = 1;
					else
						Adapter->AssociationInfo.SiteS.
						    BssInfo[Adapter->AssociationInfo.SiteS.
							    BSSInList].EncryptionIsOn = 0;

					Adapter->AssociationInfo.SiteS.BssInfo[Adapter->AssociationInfo.SiteS.BSSInList].BeaconPeriod = *(PUSHORT) (FrameBody + 8);

					if (CapabilityInformation &
					    C80211_MGMT_CAPABILITY_ESS)
						//Infrastructure
						Adapter->AssociationInfo.SiteS.
						    BssInfo[Adapter->AssociationInfo.SiteS.
							    BSSInList].
						    BSStype = 2;
					else
						//Ad-Hoc
						Adapter->AssociationInfo.SiteS.
						    BssInfo[Adapter->AssociationInfo.SiteS.
							    BSSInList].
						    BSStype = 1;
           }

					Adapter->AssociationInfo.SiteS.BSSInList++;
				}
			} else
				//dbgcond(DBG_SCAN, "BSS list is full\n");
			break;

		case STATION_STATE_READY:
			if (FrameSubType == C80211_SUBTYPE_MGMT_BEACON) {
				if (memcmp(BSSID, Adapter->AssociationInfo.CurrentBSSID, 6)
				    != 0) {
					Adapter->Stats.RxMgmtPacketsError++;
					return;
				}
        Adapter->PeriodicBeaconsCnt ++;
				Adapter->Stats.MatchingBeacons++;
				Adapter->flags |= HeardBeacons;
			}
			break;
		default:
			break;
		}
		break;

	case C80211_SUBTYPE_MGMT_Authentication:

		dbgcond(DBG_AUTH, "Auth response from BSSID "
			"%02x:%02x:%02x:%02x:%02x:%02x, status %d, trans_seq x%x\n",
			BSSID[0], BSSID[1], BSSID[2], BSSID[3], BSSID[4],
			BSSID[5],
			Adapter->RxInBuff.
			WirelessPacket[MGMT_FRAME_BODY_OFFSET +
				       STATUS_IN_AUTHENTICATION_OFFSET],
			Adapter->RxInBuff.
			WirelessPacket[MGMT_FRAME_BODY_OFFSET +
				       TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET]);
		VnetTimer(Adapter, 0);
		if (memcmp(BSSID, Adapter->AssociationInfo.CurrentBSSID, 6) != 0) {
			dbgcond(DBG_AUTH, "wrong BSSID, expected: "
				"%02x:%02x:%02x:%02x:%02x:%02x\n",
				Adapter->AssociationInfo.CurrentBSSID[0],
				Adapter->AssociationInfo.CurrentBSSID[1],
				Adapter->AssociationInfo.CurrentBSSID[2],
				Adapter->AssociationInfo.CurrentBSSID[3],
				Adapter->AssociationInfo.CurrentBSSID[4],
				Adapter->AssociationInfo.CurrentBSSID[5]);
			Adapter->Stats.RxMgmtPacketsError++;
			return;
		}

		AuthenticationFrame =
		    Adapter->RxInBuff.WirelessPacket +
		    MGMT_FRAME_BODY_OFFSET;

		//jal: Is this not an _u16 value as with assoc response ???
		Adapter->ErrorCode =
		    AuthenticationFrame[STATUS_IN_AUTHENTICATION_OFFSET];

		if (Adapter->StationState == STATION_STATE_AUTHENTICATING) {

			switch (Adapter->ErrorCode) {

			case C80211_MGMT_SC_Success:
#if 0
        // NEED TO CHANGE THIS IN ORDER TO BE COMPATIBLE WITH THE
        // OS W.....
				if ((IsEncryptionEnabled(&Adapter->SecurityInfo))
				    && (Adapter->SecurityInfo.
					AuthenticationMode ==
					_802_11AuthModeShared)) {
#endif
				//if ((Adapter->SecurityInfo.EncryptionInfo.AuthenticationMode == _802_11AuthModeShared)) {
				if ((Adapter->SecurityInfo.AuthenticationMode == _802_11AuthModeShared)) {
					if (AuthenticationFrame
					    [TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET]
					    == Adapter->AuthenticationInfo.ExpectedAuthentTransactionSeqNum)
					{
						if (AuthenticationFrame
						    [TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET]
						    == 0x0002) {
							if (AuthenticationFrame[CHALL_TEXT_EL_ID_IN_AUTHENTICATION_OFFSET] == C80211_MGMT_ElementID_ChallengeText) {
								Adapter->AuthenticationInfo.ChallengeTextLength = AuthenticationFrame [CHALL_TEXT_LENGTH_IN_AUTHENTICATION_OFFSET];
                memcpy(Adapter->AuthenticationInfo.pChallengeText, &AuthenticationFrame[CHALL_TEXT_IN_AUTHENTICATION_OFFSET], Adapter->AuthenticationInfo.ChallengeTextLength);
								SendAuthRequest (Adapter);
							}
						} else
						    if (AuthenticationFrame
							[TRANS_SEQ_NO_IN_AUTHENTICATION_OFFSET]
							== 0x0004) {
							Adapter->flags |=
							    StationIsAuthenticated;
							if (!
							    (Adapter->
							     flags &
							     StationWasAssociated))
							{
								Adapter->AssociationInfo.AssociationRequestRetryCnt = 0;
								ChangeState
								    (Adapter,
								     STATION_STATE_ASSOCIATING);
								SendAssocRequest
								    (Adapter);
							} else {
								Adapter->AssociationInfo.ReAssociationRequestRetryCnt
								    = 0;
								ChangeState
								    (Adapter,
								     STATION_STATE_REASSOCIATING);
								SendReAssocRequest
								    (Adapter);
							}
						}
					}
				} else {
					Adapter->flags |=
					    StationIsAuthenticated;
					dbgcond(DBG_AUTH,
						"Authentication Response Success (no WEP)\n");
					if (!(Adapter->flags & StationWasAssociated))
					{
						Adapter->AssociationInfo.AssociationRequestRetryCnt = 0;
						ChangeState(Adapter, STATION_STATE_ASSOCIATING);
						SendAssocRequest(Adapter);
					} else {
						Adapter->AssociationInfo.ReAssociationRequestRetryCnt = 0;
						ChangeState(Adapter,
							    STATION_STATE_REASSOCIATING);
						SendReAssocRequest
						    (Adapter);
					}
				}
				break;


			case C80211_MGMT_SC_Unspecified:
				DropPendingTxPackets(Adapter);
				ChangeState(Adapter,
					    STATION_STATE_SCANNING);
				UsbScan(Adapter);
				break;

			case C80211_MGMT_SC_AuthAlgNotSupported:	//WEP....
				Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum =
				    0x0001;
				if (++Adapter->
				    AuthenticationInfo.AuthenticationRequestRetryCnt >=
				    MAX_AUTHENTICATION_RETRIES) {
					dbgcond(DBG_AUTH,
						"max auth retries (%d) reached - giving up\n",
						MAX_AUTHENTICATION_RETRIES);
					DropPendingTxPackets(Adapter);
					ChangeState(Adapter,
						    STATION_STATE_SCANNING);
					UsbScan(Adapter);
				} else
					SendAuthRequest(Adapter);
				break;

			case C80211_MGMT_SC_AuthTransSeqNumError:
				Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum =
				    0x0001;

				if (++Adapter->
				    AuthenticationInfo.AuthenticationRequestRetryCnt >=
				    MAX_AUTHENTICATION_RETRIES) {
					dbgcond(DBG_AUTH,
						"max auth retries (%d) reached - giving up\n",
						MAX_AUTHENTICATION_RETRIES);
					DropPendingTxPackets(Adapter);
					ChangeState(Adapter,
						    STATION_STATE_SCANNING);
					UsbScan(Adapter);
				} else
					SendAuthRequest(Adapter);
				break;

			case C80211_MGMT_SC_AuthRejectTimeout:
				//jal: why not: 
				//Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum = 0x0001;
				if (++Adapter->
				    AuthenticationInfo.AuthenticationRequestRetryCnt >=
				    MAX_AUTHENTICATION_RETRIES) {
					dbgcond(DBG_AUTH,
						"max auth retries (%d) reached - giving up",
						MAX_AUTHENTICATION_RETRIES);
					DropPendingTxPackets(Adapter);
					ChangeState(Adapter,
						    STATION_STATE_SCANNING);
					UsbScan(Adapter);
				} else
					SendAuthRequest(Adapter);
				break;

			case C80211_MGMT_SC_AuthRejectChallenge:
			default:
				ChangeState(Adapter,
					    STATION_STATE_MGMT_ERROR);

			}
		}		/* if(Adapter->StationState == STATION_STATE_AUTHENTICATING) */
		break;

	case C80211_SUBTYPE_MGMT_ASS_RESPONSE:
	case C80211_SUBTYPE_MGMT_REASS_RESPONSE:
		dbgcond(DBG_ASSOC, "%sAssoc response from BSSID "
			"%02x:%02x:%02x:%02x:%02x:%02x, status %u, assoc id %u\n",
			FrameSubType ==
			C80211_SUBTYPE_MGMT_REASS_RESPONSE ? "Re" : "",
			BSSID[0], BSSID[1], BSSID[2], BSSID[3], BSSID[4],
			BSSID[5],
			/* status is a _u16 in little endian format from arbitrary address 
			   jal TODO: find a nicer way to port the driver to big endian processors
			   and processors which cannot access _u16 at arbitrary addresses */
			Adapter->RxInBuff.
			WirelessPacket[MGMT_FRAME_BODY_OFFSET +
				       STATUS_IN_ASSOCIATION_RESP_OFFSET] |
			(Adapter->RxInBuff.
			 WirelessPacket[MGMT_FRAME_BODY_OFFSET +
					STATUS_IN_ASSOCIATION_RESP_OFFSET +
					1] << 8),
			Adapter->RxInBuff.
			WirelessPacket[MGMT_FRAME_BODY_OFFSET +
				       ASS_ID_IN_ASSOCIATION_RESP_OFFSET]);

		VnetTimer(Adapter, 0);

		if (memcmp(BSSID, Adapter->AssociationInfo.CurrentBSSID, 6) != 0) {
			dbgcond(DBG_ASSOC, "wrong BSSID, expected: "
				"%02x:%02x:%02x:%02x:%02x:%02x\n",
				Adapter->AssociationInfo.CurrentBSSID[0],
				Adapter->AssociationInfo.CurrentBSSID[1],
				Adapter->AssociationInfo.CurrentBSSID[2],
				Adapter->AssociationInfo.CurrentBSSID[3],
				Adapter->AssociationInfo.CurrentBSSID[4],
				Adapter->AssociationInfo.CurrentBSSID[5]);
			Adapter->Stats.RxMgmtPacketsError++;
			return;
		}
		if ((Adapter->StationState == STATION_STATE_ASSOCIATING) ||
		    (Adapter->StationState ==
		     STATION_STATE_REASSOCIATING)) {
#ifdef ATMEL_WLAN
			AssociationRespFrameBytesToGet =
			    MIN(Adapter->RxInBuff.WLength, 16);
#else
			AssociationRespFrameBytesToGet =
			    MIN(Adapter->RxInBuff.WLength - 4, 16);
#endif
			AssociationRspnsFrame =
			    Adapter->RxInBuff.WirelessPacket +
			    MGMT_FRAME_BODY_OFFSET;

			(UCHAR) Adapter->ErrorCode =
			    (UCHAR) *
			    (PUSHORT) &
			    AssociationRspnsFrame
			    [STATUS_IN_ASSOCIATION_RESP_OFFSET];

			AssocID = *((PUSHORT) &(AssociationRspnsFrame [ASS_ID_IN_ASSOCIATION_RESP_OFFSET]));
			SupportedRatesLength =
			    AssociationRspnsFrame
			    [SUP_RATES_LENGTH_IN_ASSOCIATION_RESP_OFFSET];

			if (SupportedRatesLength > OPER_RATE_SIZE) {
				SupportedRatesLength = OPER_RATE_SIZE;
			}

			memcpy(SupportedRates,
			       &AssociationRspnsFrame
			       [SUP_RATES_IN_ASSOCIATION_RESP_OFFSET],
			       SupportedRatesLength);

			switch (Adapter->ErrorCode) {
			case C80211_MGMT_SC_Success:
				Adapter->flags |=
				    (StationWasAssociated |
				     StationIsAssociated | HeardBeacons);
				AssocID = AssocID & 0x3fff;	/* ??? */
				Adapter->AssocID = AssocID;
				Adapter->AssociationInfo.AssociationRequestRetryCnt = 0;
				Adapter->AssociationInfo.ReAssociationRequestRetryCnt = 0;
				VnetTimer(Adapter, 20);
				ChangeState(Adapter, STATION_STATE_READY);
        if(Adapter->SecurityInfo.InWpaMode) {
          printk("We Are Inside Supplicant Zone !\n");
          atmel_assoc_info_event (Adapter);
        }
				if(Adapter->PowerMgmtMode){
					SetAssocID(Adapter);
				}
				break;
			case C80211_MGMT_SC_Unspecified:
				//jal: If Access is denied, shouldn't we better go for another BSS ???
			case C80211_MGMT_SC_AssDenied:
			case C80211_MGMT_SC_AssDeniedHandleAP:
				if (FrameSubType ==
				    C80211_SUBTYPE_MGMT_ASS_RESPONSE) {
					assert(Adapter->StationState ==
					       STATION_STATE_ASSOCIATING);
					if (++Adapter->AssociationInfo.AssociationRequestRetryCnt >
					    MAX_ASSOCIATION_RETRIES) {
						DropPendingTxPackets
						    (Adapter);
						ChangeState(Adapter, STATION_STATE_SCANNING);
						UsbScan(Adapter);
					} else {
						SendAssocRequest(Adapter);
					}
				} else {
					assert(Adapter->StationState ==
					       STATION_STATE_REASSOCIATING);
					if (++Adapter->AssociationInfo.ReAssociationRequestRetryCnt >
					    MAX_ASSOCIATION_RETRIES) {
						DropPendingTxPackets
						    (Adapter);
						ChangeState(Adapter,
							    STATION_STATE_SCANNING);
						UsbScan(Adapter);
					} else {
						SendReAssocRequest
						    (Adapter);
					}
				}
				break;

			case C80211_MGMT_SC_AssDeniedBSSRate:
			case C80211_MGMT_SC_SupportCapabilities:
			default:
				ChangeState(Adapter,
					    STATION_STATE_MGMT_ERROR);
			}	/* switch (Adapter->ErrorCode) */
		}		/* if ((Adapter->StationState == STATION_STATE_ASSOCIATING) ||
				   (Adapter->StationState == STATION_STATE_REASSOCIATING)) */
		break;

	case C80211_SUBTYPE_MGMT_DISASSOSIATION:
		dbgcond(DBG_DISASS, "Disassociation from BSSID "
			"%02x:%02x:%02x:%02x:%02x:%02x, flags x%lx, state %d, op mode %d\n",
			BSSID[0], BSSID[1], BSSID[2], BSSID[3], BSSID[4],
			BSSID[5], Adapter->flags, Adapter->StationState,
			Adapter->OperatingMode);

		if (!(Adapter->flags & StationIsAssociated) ||
		    Adapter->OperatingMode == AD_HOC_MODE)
			return;
		if (memcmp(BSSID, Adapter->AssociationInfo.CurrentBSSID, 6) != 0) {
			dbgcond(DBG_AUTH, "wrong BSSID, expected: "
				"%02x:%02x:%02x:%02x:%02x:%02x\n",
				Adapter->AssociationInfo.CurrentBSSID[0],
				Adapter->AssociationInfo.CurrentBSSID[1],
				Adapter->AssociationInfo.CurrentBSSID[2],
				Adapter->AssociationInfo.CurrentBSSID[3],
				Adapter->AssociationInfo.CurrentBSSID[4],
				Adapter->AssociationInfo.CurrentBSSID[5]);
			Adapter->Stats.RxMgmtPacketsError++;
			return;
		}

    if(Adapter->SecurityInfo.InWpaMode){
        RemoveEncryptionSettings(Adapter);
    }
		ChangeState(Adapter, STATION_STATE_ASSOCIATING);
		Adapter->AssociationInfo.AssociationRequestRetryCnt = 0;
		SendAssocRequest(Adapter);
		break;

	case C80211_SUBTYPE_MGMT_Deauthentication:

		dbgcond(DBG_DISASS, "Deauthentication from BSSID "
			"%02x:%02x:%02x:%02x:%02x:%02x, flags x%lx, state %d, op mode %d\n",
			BSSID[0], BSSID[1], BSSID[2], BSSID[3], BSSID[4],
			BSSID[5], Adapter->flags, Adapter->StationState,
			Adapter->OperatingMode);

		if (!(Adapter->flags & StationIsAuthenticated) ||
		    Adapter->OperatingMode == AD_HOC_MODE)
			return;

		if (memcmp(BSSID, Adapter->AssociationInfo.CurrentBSSID, 6) != 0) {
			dbgcond(DBG_AUTH, "wrong BSSID, expected: "
				"%02x:%02x:%02x:%02x:%02x:%02x\n",
				Adapter->AssociationInfo.CurrentBSSID[0],
				Adapter->AssociationInfo.CurrentBSSID[1],
				Adapter->AssociationInfo.CurrentBSSID[2],
				Adapter->AssociationInfo.CurrentBSSID[3],
				Adapter->AssociationInfo.CurrentBSSID[4],
				Adapter->AssociationInfo.CurrentBSSID[5]);
			Adapter->Stats.RxMgmtPacketsError++;
			return;
		}

    if(Adapter->SecurityInfo.InWpaMode){
        RemoveEncryptionSettings(Adapter);
    }
		Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum = 0x0001;
		Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt = 0;
		ChangeState(Adapter, STATION_STATE_AUTHENTICATING);
		SendAuthRequest(Adapter);
		break;

	default:
		err("unknown frame subtype %d\n", FrameSubType);
	}			/* switch(FrameSubType) */
}

static const UCHAR SNAP_RFC1024[6] =
    { 0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00 };
static const UCHAR SNAP_ETHERNETII[6] =
    { 0xaa, 0xaa, 0x03, 0x00, 0x00, 0xF8 };
static const UCHAR STATIC_ETHERTYPE_8137[2] = { 0x81, 0x37 };
static const UCHAR STATIC_ETHERTYPE_80F3[2] = { 0x80, 0xF3 };

void RxData(PVNet_ADAPTER Adapter)
{
  PUCHAR        pEapolMsg = NULL;
	struct sk_buff *skb;
	USHORT PacketLength;
#ifndef ATMEL_WLAN
	UCHAR MoreFragments = 0;
	UCHAR PacketFragmentNo;
	UCHAR ProcessPacket = FALSE;
	USHORT PacketSequenceNo;
#endif
  UCHAR       RxMicKeyOffset, KeyIndex;
	BOOLEAN IP_packet;
	UCHAR DataOffset = 0;
	char buf[2 * 32 + 1] __attribute__ ((unused));

	//jal: RSSI, LinkQuality and NoiseLevel in RxInBuff contains zero ???
#ifdef ATMEL_WLAN
	dbgcond(DBG_RXDATA,
		"Wlen x%x, rate %d, newbss %d, frag %d, rssi %d,"
		" linkqual %d,"
		" data %s (Adapter->flags %lx, netdev->state %lx)",
		Adapter->RxInBuff.WLength, Adapter->RxInBuff.RxRate,
		Adapter->RxInBuff.newbss, Adapter->RxInBuff.KeyIndex,
		Adapter->RxInBuff.RSSI, Adapter->RxInBuff.LinkQuality, 
		buf2str(buf, Adapter->RxInBuff.WirelessPacket, (sizeof(buf) -1) /2),
		Adapter->flags, Adapter->net->state);
#else 
	dbgcond(DBG_RXDATA,
		"Wlen x%x, rate %d, newbss %d, frag %d, rssi %d,"
		" linkqual %d, noiselevel %d,"
		" data %s (Adapter->flags %lx, netdev->state %lx)",
		Adapter->RxInBuff.WLength, Adapter->RxInBuff.RxRate,
		Adapter->RxInBuff.newbss, Adapter->RxInBuff.Fragmentation,
		Adapter->RxInBuff.RSSI, Adapter->RxInBuff.LinkQuality, Adapter->RxInBuff.NoiseLevel,
		buf2str(buf, Adapter->RxInBuff.WirelessPacket, (sizeof(buf) -1) /2),
		Adapter->flags, Adapter->net->state);
#endif

#ifndef ATMEL_WLAN
	MoreFragments = Adapter->RxInBuff.WirelessPacket[1] & 0x04;
	PacketFragmentNo = Adapter->RxInBuff.WirelessPacket[22] & 0x0f;
	PacketSequenceNo = (Adapter->RxInBuff.WirelessPacket[23] << 8) |
	    (Adapter->RxInBuff.WirelessPacket[22] & 0xf0);

	if ((MoreFragments == 0) && (PacketFragmentNo == 0)) {
		memcpy(&Adapter->RxBuffUp, &Adapter->RxInBuff,
		       12 + Adapter->RxInBuff.WLength);
		ProcessPacket = TRUE;
	} else {
		if ((MoreFragments != 0) && (PacketFragmentNo == 0)) {
			memcpy(&Adapter->RxBuffUp, &Adapter->RxInBuff,
			       Adapter->RxInBuff.WLength + 12 - 4);
		} else if ((MoreFragments != 0) &&
			   (memcmp(Adapter->RxBuffUp.WirelessPacket,
				   Adapter->RxInBuff.WirelessPacket + 16,
				   6) == 0)) {
			memcpy(Adapter->RxBuffUp.WirelessPacket +
			       Adapter->RxBuffUp.WLength,
			       Adapter->RxInBuff.WirelessPacket + 24,
			       Adapter->RxInBuff.WLength - 28);
			Adapter->RxBuffUp.WLength +=
			    Adapter->RxInBuff.WLength - 28;
		} else if ((MoreFragments == 0)
			   &&
			   (memcmp
			    (Adapter->RxBuffUp.WirelessPacket,
			     Adapter->RxInBuff.WirelessPacket + 16,
			     6) == 0)) {
			ProcessPacket = TRUE;
			memcpy(Adapter->RxBuffUp.WirelessPacket +
			       Adapter->RxBuffUp.WLength,
			       Adapter->RxInBuff.WirelessPacket + 24,
			       Adapter->RxInBuff.WLength - 28);
			Adapter->RxBuffUp.WLength +=
			    Adapter->RxInBuff.WLength - 28;
		}
	}

	if (!ProcessPacket) {
		dbgcond(DBG_RXDATA, "fragmented packet\n");
		return;
	}
#endif


						//
						// Validate MIC bytes - TKIP case
						//

						if(((Adapter->RxBuffUp.CipherType==CIPHER_SUITE_TKIP)&&
							(Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_TKIP))	
#ifdef AES
							||((Adapter->RxBuffUp.CipherType==CIPHER_SUITE_AES)&&
							(Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_AES))
#endif
							)
						{

							if(Adapter->OperatingMode == AD_HOC_MODE)
								RxMicKeyOffset = 24;
							else
								RxMicKeyOffset = 16;

							KeyIndex = Adapter->RxBuffUp.KeyIndex ; 
							
#if ((RFMD)||(R505)||(R505_2958))
							// In this case the 4 CRC bytes	are not received from fw				
							if(!WPA_validateMic(Adapter->RxBuffUp.WirelessPacket , Adapter->RxBuffUp.WLength , &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[KeyIndex][RxMicKeyOffset]) )
							{
								printk(KERN_INFO "### MIC Error ### ( (DA: %02X:%02X:%02X:%02X:%02X:%02X - BSSID: %02X:%02X:%02X:%02X:%02X:%02X - SA: %02X:%02X:%02X:%02X:%02X:%02X))\n", /*KeyIndex,*/ 
										Adapter->RxBuffUp.WirelessPacket[4],	 Adapter->RxBuffUp.WirelessPacket[5], Adapter->RxBuffUp.WirelessPacket[6], Adapter->RxBuffUp.WirelessPacket[7], Adapter->RxBuffUp.WirelessPacket[8], Adapter->RxBuffUp.WirelessPacket[9],
										Adapter->RxBuffUp.WirelessPacket[10], Adapter->RxBuffUp.WirelessPacket[11],Adapter->RxBuffUp.WirelessPacket[12], Adapter->RxBuffUp.WirelessPacket[13],Adapter->RxBuffUp.WirelessPacket[14],Adapter->RxBuffUp.WirelessPacket[15],
										Adapter->RxBuffUp.WirelessPacket[16], Adapter->RxBuffUp.WirelessPacket[17],Adapter->RxBuffUp.WirelessPacket[18], Adapter->RxBuffUp.WirelessPacket[19],Adapter->RxBuffUp.WirelessPacket[20],Adapter->RxBuffUp.WirelessPacket[21]);


								Adapter->SecurityInfo.MicErrorCnt++;

						//		Adapter->Statistics.RxDataPacketsError++;
						//		Adapter->Statistics.RxWpaMicErrors ++;								
								
#if 0
								//
								// Event Indication in MIC ERROR case
								//
								if(Adapter->RxBuffUp.WirelessPacket[4]  & 0x01 )
									SendAuthenticationEvent(Adapter, Ndis802_11StatusType_Authentication, NDIS_802_11_AUTH_REQUEST_GROUP_ERROR);
								else
									SendAuthenticationEvent(Adapter, Ndis802_11StatusType_Authentication, NDIS_802_11_AUTH_REQUEST_PAIRWISE_ERROR);
								
								//
								// For first MIC Error set up a timer and
								// For second MIC Error check the time passed from first error 
								// and if necessary add the BSS in black list
								//
								if (Adapter->SecurityInfo.MicErrorCnt == 1)
								{
									DbgINFO(1,("\n\n\nFIRST MIC ERROR\n"));
									
									NdisGetCurrentSystemTime(&Adapter->SecurityInfo.MicErrorDetected);
									NdisMSetTimer(&Adapter->MgmtTimeOutTimer, 60000);
								}
								else
								if (Adapter->SecurityInfo.MicErrorCnt == 2)
								{
									DbgINFO(1,("\n\n\nSECOND MIC ERROR\n"));

									NdisGetCurrentSystemTime((PLARGE_INTEGER)&TimeMicError2);
									NdisMSetTimer(&Adapter->MgmtTimeOutTimer, 60000);

									if (WPA_AddAPInBlackList(Adapter->SecurityInfo.MicBlackList, black_list_entries , &Adapter->SecurityInfo.MicBlackListEntries, (PUCHAR)&Adapter->AssociationInfo.CurrentBSSID , (PVOID)&TimeMicError2))
										DbgINFO(1,("AP is Added in Black List with success.\n"));

									Adapter->Statistics.RxWpaCountermeasuresCount++;
								}

#endif
							}
							else if(Adapter->SecurityInfo.MicErrorCnt == 2)
							{ // drop rx packets
                return;
							}							
              Adapter->RxBuffUp.WLength -= (USHORT) 8;
#else
							if (Adapter->RxBuffUp.LinkQuality & 0x02)
							{
								printk(KERN_INFO "### MIC Error ### ( (DA: %02X:%02X:%02X:%02X:%02X:%02X - BSSID: %02X:%02X:%02X:%02X:%02X:%02X - SA: %02X:%02X:%02X:%02X:%02X:%02X))\n", /*KeyIndex,*/ 
										Adapter->RxBuffUp.WirelessPacket[4],	 Adapter->RxBuffUp.WirelessPacket[5], Adapter->RxBuffUp.WirelessPacket[6], Adapter->RxBuffUp.WirelessPacket[7], Adapter->RxBuffUp.WirelessPacket[8], Adapter->RxBuffUp.WirelessPacket[9],
										Adapter->RxBuffUp.WirelessPacket[10], Adapter->RxBuffUp.WirelessPacket[11],Adapter->RxBuffUp.WirelessPacket[12], Adapter->RxBuffUp.WirelessPacket[13],Adapter->RxBuffUp.WirelessPacket[14],Adapter->RxBuffUp.WirelessPacket[15],
										Adapter->RxBuffUp.WirelessPacket[16], Adapter->RxBuffUp.WirelessPacket[17],Adapter->RxBuffUp.WirelessPacket[18], Adapter->RxBuffUp.WirelessPacket[19],Adapter->RxBuffUp.WirelessPacket[20],Adapter->RxBuffUp.WirelessPacket[21]);


								Adapter->SecurityInfo.MicErrorCnt++;
              }
							else if(Adapter->SecurityInfo.MicErrorCnt == 2)
							{ // drop rx packets
                return;
							}							
#endif
						}
  
	if (((memcmp
	      (Adapter->RxBuffUp.WirelessPacket + 24, SNAP_ETHERNETII,
	       6) == 0)
	     &&
	     (memcmp
	      (Adapter->RxBuffUp.WirelessPacket + 30,
	       STATIC_ETHERTYPE_8137, 2) == 0))
	    || ((Adapter->RxBuffUp.WirelessPacket[24] == 0xaa)
		&& (Adapter->RxBuffUp.WirelessPacket[25] == 0xaa)
		&&
		(memcmp
		 ((PUCHAR) Adapter->RxBuffUp.WirelessPacket + 30,
		  STATIC_ETHERTYPE_8137, 2) != 0)
		&&
		(memcmp
		 ((PUCHAR) Adapter->RxBuffUp.WirelessPacket + 30,
		  STATIC_ETHERTYPE_80F3, 2) != 0))) {
		IP_packet = TRUE;
	} else {
#ifdef ATMEL_WLAN
		Adapter->RxBuffUp.WirelessPacket[23] =
		    ((Adapter->RxBuffUp.WLength - 24) & 0x00FF);
		Adapter->RxBuffUp.WirelessPacket[22] =
		    ((Adapter->RxBuffUp.WLength - 24) >> 8);
#else
		Adapter->RxBuffUp.WirelessPacket[23] =
		    ((Adapter->RxBuffUp.WLength - 24 - 4) & 0x00FF);
		Adapter->RxBuffUp.WirelessPacket[22] =
		    ((Adapter->RxBuffUp.WLength - 24 - 4) >> 8);
#endif
		IP_packet = FALSE;
	}

	if (Adapter->OperatingMode == AD_HOC_MODE) {
		if (IP_packet) {
			memcpy(Adapter->RxBuffUp.WirelessPacket + 18,
			       Adapter->RxBuffUp.WirelessPacket + 4, 6);
			memcpy(Adapter->RxBuffUp.WirelessPacket + 24,
			       Adapter->RxBuffUp.WirelessPacket + 10, 6);
#ifdef ATMEL_WLAN
			PacketLength = Adapter->RxBuffUp.WLength - 18;
#else
			PacketLength = Adapter->RxBuffUp.WLength - 18 - 4;
#endif
			DataOffset = 18;
							//
							// EAPOL packet
							//
							if((Adapter->RxBuffUp.WirelessPacket[30] == 0x88) &&	// 10+12
								(Adapter->RxBuffUp.WirelessPacket[31] == 0x8E))  // 10+13
							{
								ParseEAPOLPacket(pEapolMsg, &Adapter->LeapChallenge, Adapter->RxBuffUp.WirelessPacket + 30, TRUE, TRUE);
								//StoreEapolPacket(Adapter->EapolBuffer, Adapter->RxBuffUp.WirelessPacket+32 , PacketLength);
							}
		} else {
			memcpy(Adapter->RxBuffUp.WirelessPacket + 10,
			       Adapter->RxBuffUp.WirelessPacket + 4, 6);
			memcpy(Adapter->RxBuffUp.WirelessPacket + 16,
			       Adapter->RxBuffUp.WirelessPacket + 10, 6);
#ifdef ATMEL_WLAN
			PacketLength = Adapter->RxBuffUp.WLength - 10;
#else
			PacketLength = Adapter->RxBuffUp.WLength - 10 - 4;
#endif
			DataOffset = 10;
							//
							// EAPOL packet
							//
							if((Adapter->RxBuffUp.WirelessPacket[22] == 0x88) &&	// 10+12
								(Adapter->RxBuffUp.WirelessPacket[23] == 0x8E))  // 10+13
							{
								ParseEAPOLPacket(pEapolMsg, &Adapter->LeapChallenge, Adapter->RxBuffUp.WirelessPacket + 22, TRUE, TRUE);
								//StoreEapolPacket(Adapter->EapolBuffer, Adapter->RxBuffUp.WirelessPacket+24 , PacketLength);
							}
		}

	} else if (Adapter->OperatingMode == INFRASTRUCTURE_MODE) {
		if (IP_packet) {
			memcpy(Adapter->RxBuffUp.WirelessPacket + 24,
			       Adapter->RxBuffUp.WirelessPacket + 16, 6);
			memcpy(Adapter->RxBuffUp.WirelessPacket + 18,
			       Adapter->RxBuffUp.WirelessPacket + 4, 6);
#ifdef ATMEL_WLAN
			PacketLength = Adapter->RxBuffUp.WLength - 18;	//18;
#else
			PacketLength = Adapter->RxBuffUp.WLength - 22;	//18 - 4;
#endif
			DataOffset = 18;
							//
							// EAPOL packet
							//
							if((Adapter->RxBuffUp.WirelessPacket[30] == 0x88) &&	// 10+12
								(Adapter->RxBuffUp.WirelessPacket[31] == 0x8E))  // 10+13
							{
								ParseEAPOLPacket(pEapolMsg, &Adapter->LeapChallenge, Adapter->RxBuffUp.WirelessPacket + 30, TRUE, TRUE);
								//StoreEapolPacket(Adapter->EapolBuffer, Adapter->RxBuffUp.WirelessPacket+32 , PacketLength);
							}

		} else {
			memcpy(Adapter->RxBuffUp.WirelessPacket + 10,
			       Adapter->RxBuffUp.WirelessPacket + 4, 6);
#ifdef ATMEL_WLAN
			PacketLength = Adapter->RxBuffUp.WLength - 10;	//10;
#else
			PacketLength = Adapter->RxBuffUp.WLength - 14;	//10 - 4;
#endif
			DataOffset = 10;
							//
							// EAPOL packet
							//
							if((Adapter->RxBuffUp.WirelessPacket[22] == 0x88) &&	// 10+12
								(Adapter->RxBuffUp.WirelessPacket[23] == 0x8E))  // 10+13
							{
								ParseEAPOLPacket(pEapolMsg, &Adapter->LeapChallenge, Adapter->RxBuffUp.WirelessPacket + 22, TRUE, TRUE);
								//StoreEapolPacket(Adapter->EapolBuffer, Adapter->RxBuffUp.WirelessPacket+24 , PacketLength);
							}
		}
	} else {
		assert(FALSE);
		return;
	}

	if (rx_kernel_part(Adapter, &skb, PacketLength, DataOffset) != 0) {
		Adapter->Stats.RxLost++;
		return;
	}

	Adapter->flags |= HeardBeacons;

	Adapter->net->last_rx = jiffies;
	Adapter->netstats->rx_packets++;
	Adapter->Stats.RxDataPacketsOk++;
	Adapter->netstats->rx_bytes += PacketLength;
	skb->dev = Adapter->net;
	skb->protocol = eth_type_trans(skb, Adapter->net);
	skb->ip_summed = CHECKSUM_NONE;
	netif_rx(skb);
}
