/* $Id: tx.c,v 1.11 2004/09/16 16:26:18 tmpetsos Exp $ */
/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

		Atmel wireless lan drivers is free software; you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation; either version 2 of the License, or
		(at your option) any later version.

		Atmel wireless lan drivers is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with Atmel wireless lan drivers; if not, write to the Free Software
		Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	02111-1307	USA

**************************************************************************************/
#include "vnetusba.h"
#include "rx.h"
#include "frame.h"
#include "stdinclude.h"


/* == PROC AllocPacket ==
	 alloc an TX_LIST struct and fill some fields */
TX_LIST *AllocPacket(PVNet_ADAPTER Adapter, USHORT length)
{
	int kmflag = in_interrupt()? GFP_ATOMIC : GFP_KERNEL;
	TX_LIST *retval;

	if (!(retval = kmalloc(sizeof(TX_LIST), kmflag))) {
		err("cannot kmalloc %d byte", sizeof(TX_LIST));
		return NULL;
	}

	/* jal: check if we really need to alloc a complete TX_BUFFER ... */
	if (!(retval->tx_buff = kmalloc(sizeof(TX_BUFFER), kmflag))) {
		err("cannot kmalloc %d byte\n", sizeof(TX_BUFFER));
		kfree(retval);
		return NULL;
	}

	/* fill some fields */
	retval->length = length + 8 + 18;
	retval->tx_buff->WLength = length;
	retval->tx_buff->TxRate = 1;	/* fixed to 1 ??? */
	retval->Adapter = Adapter;
	retval->skb = NULL;	// skb is filled by caller, shows that this buffer contains tx data 

	if (Adapter->TxPacketsPending > THRE_STOP_NET_TX &&
	    !(Adapter->flags & TX_STOPPED))
		VNetIf_queue(Adapter->net, 0);	/* sets TX_STOPPED inside */
	return retval;
}

int SendAssocRequest(PVNet_ADAPTER Adapter)
{
	USHORT FrameLen;
	s80211_MgmtFrm MgmtFrame;
#ifdef USBDBG
	s80211_MgmtBody_AssRqst *AssReqFrameBody =
	    (s80211_MgmtBody_AssRqst *) MgmtFrame.Body;
#endif
	PTX_LIST TxList;
	UCHAR i __attribute__ ((unused));

	if (Adapter->AssociationInfo.AssociationRequestRetryCnt ==
	    MAX_ASSOCIATION_RETRIES) {
		dbgcond(DBG_ASSOC,
			"max assoc retries (%d) reached - giving up\n",
			MAX_ASSOCIATION_RETRIES);
		DropPendingTxPackets(Adapter);
		ChangeState(Adapter, STATION_STATE_SCANNING);
		UsbScan(Adapter);
		return 0;
	}

  BuildAssociationRequestFrame((PUCHAR) &MgmtFrame, &Adapter->AssociationInfo, &Adapter->SecurityInfo, &FrameLen);
	/* Create Association Request Frame Header */
	{
		char buf[2 * ESSID_SIZE + 1] __attribute__ ((unused));
		char buf2[2 * 4 + 1] __attribute__ ((unused));	/* dbg buffer */
		dbgcond(DBG_ASSOC,
			"AssocReq to bssid %02x:%02x:%02x:%02x:%02x:%02x, "
			"ssid %s, capa x%x, rates %s\n", MgmtFrame.uDA[0],
			MgmtFrame.uDA[1], MgmtFrame.uDA[2],
			MgmtFrame.uDA[3], MgmtFrame.uDA[4],
			MgmtFrame.uDA[5], ssid2str(buf,
						   AssReqFrameBody->SSID.
						   Information,
						   AssReqFrameBody->SSID.
						   Length),
			AssReqFrameBody->CapabilityInformation,
			buf2str(buf2,
				AssReqFrameBody->SSID.Information +
				AssReqFrameBody->SSID.Length + 2,
				*(AssReqFrameBody->SSID.Information +
				  AssReqFrameBody->SSID.Length + 1)));
	}
  Adapter->AssociationInfo.AssociationRequestRetryCnt++;
	if ((TxList = AllocPacket(Adapter, FrameLen))) {
		memcpy(TxList->tx_buff->WirelessPacket, &MgmtFrame, FrameLen);
		DeQueueTxPacket(Adapter, TxList);
	}
	VnetTimer(Adapter, 25);
	return 0;
}				/* SendAssocRequest */

int SendAuthRequest(PVNet_ADAPTER Adapter)
{
	USHORT FrameLen;
	s80211_MgmtFrm MgmtFrame;
#ifdef USBDBG
	s80211_MgmtBody_Authentication *AuthFrameBody = (s80211_MgmtBody_Authentication *) MgmtFrame.Body;
#endif
	TX_LIST *TxList;


    BuildAuthenticationRequestFrame((PUCHAR) &MgmtFrame, &Adapter->AuthenticationInfo, &Adapter->AssociationInfo, &Adapter->SecurityInfo, &FrameLen);

	dbgcond(DBG_AUTH,
		"AuthReq to bssid %02x:%02x:%02x:%02x:%02x:%02x, "
		"WEP %s, algorithm nr %d, seq nr %d, challenge text len %d\n",
		MgmtFrame.uDA[0], MgmtFrame.uDA[1], MgmtFrame.uDA[2],
		MgmtFrame.uDA[3], MgmtFrame.uDA[4], MgmtFrame.uDA[5],
		MgmtFrame.FrameControl[1] ? "yes" : "no",
		AuthFrameBody->AuthenticationAlgorithmNumber,
		AuthFrameBody->AuthenticationTransactionSequenceNumber,
		AuthFrameBody->ChallengeText.Length);

          Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt++;

	//send frame & return status
	if ((TxList = AllocPacket(Adapter, FrameLen))) {
		memcpy(TxList->tx_buff->WirelessPacket, &MgmtFrame,
		       FrameLen);
		DeQueueTxPacket(Adapter, TxList);
	}			/* AllocPacket prints error if TxList is empty ... */
	VnetTimer(Adapter, 25);
	return 0;
}

int SendReAssocRequest(PVNet_ADAPTER Adapter)
{
	USHORT FrameLen;
	s80211_MgmtFrm MgmtFrame;
#ifdef USBDBG
	s80211_MGMT_REASS_REQUEST *ReAssReqFrameBody =
	    (s80211_MGMT_REASS_REQUEST *) MgmtFrame.Body;
#endif
	int i __attribute__ ((unused));
	PTX_LIST TxList;


	if (++Adapter->AssociationInfo.ReAssociationRequestRetryCnt ==
	    MAX_REASSOCIATION_RETRIES) {
		dbgcond(DBG_ASSOC,
			"max assoc retries (%d) reached - giving up\n",
			MAX_REASSOCIATION_RETRIES);
		DropPendingTxPackets(Adapter);
		ChangeState(Adapter, STATION_STATE_SCANNING);
		UsbScan(Adapter);
		return 0;
	}

	/* Create Association Request Frame Header */
  BuildReAssociationRequestFrame((PUCHAR) &MgmtFrame, &Adapter->AssociationInfo, &Adapter->SecurityInfo, &FrameLen);


	{
		char buf[2 * ESSID_SIZE + 1] __attribute__ ((unused));
		char buf2[2 * 4 + 1] __attribute__ ((unused));	/* dbg buffer */
		dbgcond(DBG_ASSOC,
			"ReAssocReq to bssid %02x:%02x:%02x:%02x:%02x:%02x "
			"ssid %s, capa x%x, rates %s\n", MgmtFrame.uDA[0],
			MgmtFrame.uDA[1], MgmtFrame.uDA[2],
			MgmtFrame.uDA[3], MgmtFrame.uDA[4],
			MgmtFrame.uDA[5], ssid2str(buf,
						   ReAssReqFrameBody->SSID.
						   Information,
						   ReAssReqFrameBody->SSID.
						   Length),
			ReAssReqFrameBody->CapabilityInformation,
			buf2str(buf2,
				ReAssReqFrameBody->SSID.Information +
				ReAssReqFrameBody->SSID.Length + 2,
				*(ReAssReqFrameBody->SSID.Information +
				  ReAssReqFrameBody->SSID.Length + 1)));
	}

	if ((TxList = AllocPacket(Adapter, FrameLen))) {
		memcpy(TxList->tx_buff->WirelessPacket, &MgmtFrame,
		       FrameLen);
		DeQueueTxPacket(Adapter, TxList);
	}
	VnetTimer(Adapter, 25);
	return 0;
}				/* SendReAssocRequest */

/***************************************************************************
	In Ethernet2(to)Wireless the ethernet packet we got from above is
	mutated to a wireless one. the ethernet's first 12 bytes (source
	and destination address) will be overwritten by the wireless
	header.
***************************************************************************/
void Ethernet2Wireless(PVNet_ADAPTER Adapter, PTX_LIST NewTxL)
{
#ifdef ATMEL_WLAN
  UCHAR           KeyID __attribute__ ((unused));
	USHORT EtherTypeOrLength;
	UCHAR           tmpWirelessPacket[MAX_WIRELESS_SIZE];
	UCHAR           SNAP_RFC1024[6] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00};
        UCHAR           SNAP_ETHERNETII[6] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0xF8};
        USHORT          STATIC_ENTRIES_ETHERTYPES[2] = {0x8137, 0x80F3};
   //     UCHAR           STATIC_ETHERTYPE_8137[2] = {0x81, 0x37};
    //    UCHAR           STATIC_ETHERTYPE_80F3[2] = {0x80, 0xF3};
#endif
        UCHAR TmpBuff[6];
	// Byte 0. Packet Type (DATA)
	NewTxL->tx_buff->WirelessPacket[0] = C80211_TYPE_DATA;
	// byte 2-3 : DurationID
	*(PUSHORT) ((PUCHAR) NewTxL->tx_buff->WirelessPacket + 2) =
	    (USHORT) 0x00;
        
		NewTxL->tx_buff->WirelessPacket[1] = 0x00;
		if (Adapter->PrivacyInvoked != 0){
      if(!(( NewTxL->tx_buff->WirelessPacket[30] == 0x88) && ( NewTxL->tx_buff->WirelessPacket[31] == 0x8E))){
			       NewTxL->tx_buff->WirelessPacket[1] |= C80211_FRMCTRL_WEP;
      }
    }

      if(Adapter->SecurityInfo.PairwiseKeyAvailable) {
          if(( NewTxL->tx_buff->WirelessPacket[30] == 0x88) && ( NewTxL->tx_buff->WirelessPacket[31] == 0x8E)){
                 NewTxL->tx_buff->WirelessPacket[1] |= C80211_FRMCTRL_ENCRYPTION;
          }
      }
	if (Adapter->OperatingMode == AD_HOC_MODE) {

		memcpy(&NewTxL->tx_buff->WirelessPacket[4], &NewTxL->tx_buff->WirelessPacket[18], 6);	//Destination
		memcpy(&NewTxL->tx_buff->WirelessPacket[10], Adapter->StationAddress, 6);	//MACADDRESS
		memcpy(&NewTxL->tx_buff->WirelessPacket[16], Adapter->AssociationInfo.CurrentBSSID, 6);	//BSSID Created

	} else {

		/* infrastructure mode */
			NewTxL->tx_buff->WirelessPacket[1] |= C80211_FRMCTRL_TO_DS;

		memcpy(&TmpBuff[0], &NewTxL->tx_buff->WirelessPacket[18], 6);	//Destination
		memcpy(&NewTxL->tx_buff->WirelessPacket[16], &TmpBuff[0], 6);	//Destination
		memcpy(&NewTxL->tx_buff->WirelessPacket[4], Adapter->AssociationInfo.CurrentBSSID, 6);	//Associated BSSID
		memcpy(&NewTxL->tx_buff->WirelessPacket[10], Adapter->StationAddress, 6);	//MAC
		// Sequence Control.....
		*(PUSHORT) ((PUCHAR) NewTxL->tx_buff->WirelessPacket +
			    22) = (USHORT) 0;
	}

#ifdef ATMEL_WLAN
        //Ethernet Address Translation
	EtherTypeOrLength = ((USHORT)*(NewTxL->tx_buff->WirelessPacket + 30)) << 8 ;
	EtherTypeOrLength |= ((USHORT)*(NewTxL->tx_buff->WirelessPacket + 31));

        if ((EtherTypeOrLength == (USHORT)STATIC_ENTRIES_ETHERTYPES[0]) || (EtherTypeOrLength == (USHORT)STATIC_ENTRIES_ETHERTYPES[1]))
        {
            memcpy(NewTxL->tx_buff->WirelessPacket+24, SNAP_ETHERNETII, 6);
	          NewTxL->tx_buff->WLength = NewTxL->length - 8;
        }
        else if (EtherTypeOrLength >1536)
        {
            memcpy(NewTxL->tx_buff->WirelessPacket+24, SNAP_RFC1024, 6);
	          NewTxL->tx_buff->WLength = NewTxL->length - 8;

        }
        else
        {
            NewTxL->tx_buff->WLength = EtherTypeOrLength + 24;
            memcpy(tmpWirelessPacket,NewTxL->tx_buff->WirelessPacket+32, EtherTypeOrLength);
            memcpy(NewTxL->tx_buff->WirelessPacket+24, tmpWirelessPacket, EtherTypeOrLength);
        }

#if((!R505A)) && ((!R505AMX))
	if (Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite == CIPHER_SUITE_TKIP)
	{
		NewTxL->tx_buff->WLength += (USHORT) 8;
		KeyID = 0;
		if(Adapter->OperatingMode == AD_HOC_MODE)
		{
			if(Adapter->SecurityInfo.GroupKeyAvailable && !Adapter->SecurityInfo.PairwiseKeyAvailable)
				KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex & 0xff);
			else
			if(!Adapter->SecurityInfo.GroupKeyAvailable && Adapter->SecurityInfo.PairwiseKeyAvailable)
			{
				if(Adapter->IsBroadcast)
					KeyID = 1; //Pairwise is always to index 0 otherwise add_key oid will fail
				else
					KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex & 0xff);
			}
			else
			{
				if(Adapter->IsBroadcast)
					KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex & 0xff);
				else
					KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex & 0xff);
			}
		}
		else
		{
			if(Adapter->SecurityInfo.PairwiseKeyAvailable) {
				KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex & 0xff);
      } else {
					KeyID = (UCHAR)(Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex & 0xff);
			}
		}


		if(Adapter->OperatingMode == AD_HOC_MODE){
			WPA_addMICToPacket(NewTxL->tx_buff->WirelessPacket, (USHORT) NewTxL->tx_buff->WLength, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[KeyID][24]);
    } else {

			WPA_addMICToPacket(NewTxL->tx_buff->WirelessPacket, (USHORT) NewTxL->tx_buff->WLength, &Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[KeyID][24]);

    }
	}
#endif
  
#else

	NewTxL->tx_buff->WLength = NewTxL->length - 8;

#endif

	NewTxL->tx_buff->TxRate = Adapter->TxRate;

	{
		//char buf[2 * 32 + 1] __attribute__ ((unused));	/* dump the first 32 byte of the WirelessPacket */
		char buf[NewTxL->tx_buff->WLength +1 ] __attribute__ ((unused));	/* dump the first 32 byte of the WirelessPacket */
		dbgcond(DBG_TXDATA, "TxData: %s\n",
			buf2str(buf, NewTxL->tx_buff->WirelessPacket,
				(sizeof(buf) - 1)/2));
	}

}				/* Ethernet2Wireless */

/***************************************************************************

***************************************************************************/
/* == PROC DeQueueTxPacket ==
	 check if the TxList contains another pending Tx packet and, if yes, 
	 send it. If new_tx is != NULL, it points to a new tx packet which is queued
	 at the end of TxList.
   Must be called inside the spin_lock Adapter->lock !*/
void DeQueueTxPacket(PVNet_ADAPTER Adapter, TX_LIST * new_tx)
{
	UCHAR LastPacket = 0;	//Adapter->TxList.length%64;
	USHORT txLen;
  UCHAR           KeyID __attribute__ ((unused));
	PTX_LIST TxOut;
	BOOLEAN			TxEapPacketUnencrypted=FALSE;
#ifdef CCX
	BOOLEAN			IsPairwise;
	UCHAR           tmpWirelessPacket2[MAX_WIRELESS_SIZE];
	BOOLEAN			EAP_packet;
	ULONG			tmp;
#endif

	if (new_tx) {
		list_add_tail(&new_tx->tx_list, Adapter->TxList);
		Adapter->TxPacketsPending++;
	}
	// the device isn't running.
	if (!(Adapter->flags & VNET_RUNNING)) {
		err("device not running");
		return;
	}

	/* We have already sent a tx packet and will 
	   dequeue next packet in the completion handler. */
	if (Adapter->flags & TX_SLEEP) {
		dbgcond(DBG_TXDATA, "tx on\n");
		return;
	}

	if ((Adapter->TxPacketsPending < THRE_START_NET_TX) &&
	    (Adapter->flags & TX_STOPPED)) {
		VNetIf_queue(Adapter->net, 1);	/* resets TX_STOPPED inside */
	}

	/* nothing to be sent */
	if (list_empty(Adapter->TxList))
		return;

	TxOut = list_entry(Adapter->TxList->next, TX_LIST, tx_list);

  //Smarkou
	txLen = TxOut->tx_buff->WLength + 8;

#if 0
	LastPacket = TxOut->length % 64;
#endif 
	LastPacket = txLen % 64;
	Adapter->flags |= TX_SLEEP;
#if 0
	txLen = TxOut->length;
#endif


	// the 4th byte of the data_buff is the padding_bytes entry in the TX_BUFFER structure
	if (LastPacket < 50)
		TxOut->tx_buff->PaddingBytes = 50 - LastPacket;
	else if (LastPacket == 63)
		TxOut->tx_buff->PaddingBytes = 51;
	else if (LastPacket == 62)
		TxOut->tx_buff->PaddingBytes = 52;
	else if (LastPacket == 61)
		TxOut->tx_buff->PaddingBytes = 53;


	txLen += TxOut->tx_buff->PaddingBytes;

#ifdef ATMEL_WLAN
	if (((TxOut->tx_buff->WirelessPacket[0] == C80211_SUBTYPE_MGMT_Authentication) && Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite && 
			(Adapter->SecurityInfo.EncryptionInfo.AuthenticationMode == C80211_MGMT_AAN_SHAREDKEY) && (Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum > 1)) || (TxOut->tx_buff->WirelessPacket[0] == C80211_TYPE_DATA)){
	 // TxOut->tx_buff->KeyID = Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex;
    TxOut->tx_buff->CipherType = Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite;
    if((Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite  == CIPHER_SUITE_WEP_64) || (Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite  == CIPHER_SUITE_WEP_128))
        TxOut->tx_buff->CipherLength = 8;
    else
        TxOut->tx_buff->CipherLength = 0;
	} else{
	
    	TxOut->tx_buff->CipherType = 0;
      TxOut->tx_buff->CipherLength = 0;
	}
#endif  

	//
	// CipherType and CipherLength
	//

	if ((*(TxOut->tx_buff->WirelessPacket + 30) == 0x88) && 
		(*(TxOut->tx_buff->WirelessPacket + 31) == 0x8E) )
	{
		ParseEAPOLPacket(&Adapter->EapolMsg, &Adapter->LeapChallenge, TxOut->tx_buff->WirelessPacket+30, FALSE, TRUE);

		if (Adapter->EapolMsg == EAPOL_TX_GROUP_UPDATE_MSG2) {
      printk("Key Exchange Completed Succesfully - Now we are ready to obtain IP\n");
    }
    
		if (((!Adapter->SecurityInfo.InWpaMode)||(Adapter->EapolMsg == EAPOL_TX_HANDSHAKE_MSG4)))
		{
      printk("TxEapPacketUnencrypted\n");
			TxEapPacketUnencrypted = TRUE;
		}
	}
    
	SelectCiphers(&Adapter->SecurityInfo, Adapter->IsBroadcast, &Adapter->TxEapPending, TxEapPacketUnencrypted, &TxOut->tx_buff->CipherType, &TxOut->tx_buff->CipherLength);


	FILL_BULK_URB(Adapter->tx_urb,
		      Adapter->usb,
		      usb_sndbulkpipe(Adapter->usb, 2),
		      TxOut->tx_buff, txLen, TxOkCB, TxOut);

#if (LINUX_VERSION_CODE < 0x020500)
	if (usb_submit_urb(Adapter->tx_urb) != 0) {
#else
	if (usb_submit_urb(Adapter->tx_urb, GFP_ATOMIC) != 0) {
#endif
		err("tx not submitted (%d)", Adapter->tx_urb->status);
		Adapter->flags &= ~TX_SLEEP;
	}

	mod_use(1, Adapter);
	return;
}
