/* $Id: vnet.c,v 1.18 2004/09/16 16:26:18 tmpetsos Exp $ */
/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

		Atmel wireless lan drivers is free software; you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation; either version 2 of the License, or
		(at your option) any later version.

		Atmel wireless lan drivers is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with Atmel wireless lan drivers; if not, write to the Free Software
		Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	02111-1307	USA

**************************************************************************************/
#include "vnetusba.h"
#include <linux/wireless.h>
#include "command.h"
#include "rx.h"
#include "stdinclude.h"

//static        int             mtu = 1500;
//static        int             eth = 1;
static UCHAR channel = 10;
static UCHAR TxRate = 3;
static USHORT RTSThreshold = 2347;
static USHORT FragThreshold = 2346;
static UCHAR OpMode = INFRASTRUCTURE_MODE;
static char ESSID[ESSID_SIZE + 1] = "";
static UCHAR WepKeyToUse = 0;
static char WepKey1[(WEP_KEY_SIZE * 2) + 1] =
    "00000000000000000000000000\0";
static char WepKey2[(WEP_KEY_SIZE * 2) + 1] =
    "00000000000000000000000000\0";
static char WepKey3[(WEP_KEY_SIZE * 2) + 1] =
    "00000000000000000000000000\0";
static char WepKey4[(WEP_KEY_SIZE * 2) + 1] =
    "00000000000000000000000000\0";
static UCHAR WepMode = WEP_MODE_MANDATORY;
static UCHAR EncryptionLevel = 0;
static UCHAR AuthenticationType = C80211_MGMT_AAN_OPENSYSTEM;
static UCHAR PreambleType = 0;
static UCHAR PwrMgmtMode = ACTIVE_MODE;
static UCHAR TxPwrLevel = 200;
static UCHAR InternationalRoaming = 0;
static UCHAR RadioIsOn = 0;

ULONG Lefty;
ULONG Righty;

typedef UCHAR UINT8;
typedef USHORT  UINT16;


void SetParameters(PVNet_ADAPTER Adapter)
{
	int i;
#ifndef ATMEL_WLAN
  int j;
#endif
// Channel
	Adapter->Channel = channel;
// TxRate
	Adapter->TxRate = TxRate;
// OperatingMode
	Adapter->OperatingMode = OpMode;
// Preamble Type
	Adapter->AssociationInfo.PreambleType = PreambleType;
// International Roaming 
	Adapter->InternationalRoaming = InternationalRoaming;
// Radio On - Off
	Adapter->RadioIsOn = RadioIsOn;
// Power Mgmt Mode
	Adapter->PowerMgmtMode = PwrMgmtMode;
// Tx Power Level
	Adapter->CurrentTXpowerIndbm = TxPwrLevel;
  Adapter->PeriodicBeaconsCnt = 0;

  if( Adapter->PowerMgmtMode )
    Adapter->AssociationInfo.ListenInterval = 2;
  else
    Adapter->AssociationInfo.ListenInterval = 1;

// SSID
	Adapter->SSID_size = Adapter->AssociationInfo.DesiredSSIDsize = strlen(ESSID);

	memcpy((PUCHAR) Adapter->AssociationInfo.DesiredSSID, (PUCHAR) ESSID,
	       Adapter->SSID_size);
	memset((PUCHAR) Adapter->SelectedSSID, 0, ESSID_SIZE);
	Adapter->SSID_size = 0;
// FragmentationThreshold
	Adapter->FragmentationThreshold = FragThreshold;
  Adapter->TxEapPending = FALSE;
  Adapter->SecurityInfo.InWpaMode = 0;
  Adapter->SecurityInfo.WPA_MixedModeDisabled = FALSE;
			Adapter->AssociationInfo.AssociationRequestRetryCnt = 0;
			Adapter->AssociationInfo.ReAssociationRequestRetryCnt = 0;
	Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum = 0;
Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt = 0;
	Adapter->AuthenticationInfo.ExpectedAuthentTransactionSeqNum = 0;	
        Adapter->SecurityInfo.EncryptionStatus = _802_11EncryptionDisabled;
	Adapter->SecurityInfo.PairwiseKeyAvailable = FALSE;
	Adapter->SecurityInfo.GroupKeyAvailable = FALSE;
// RtsThreshold
	Adapter->RtsThreshold = RTSThreshold;
// Wep Key Information
	Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse = WepKeyToUse;

	if (!EncryptionLevel) {
		Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = 1;
		Adapter->PrivacyInvoked = 0;
	} else {
		Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = EncryptionLevel;
		Adapter->PrivacyInvoked = 1;
	}
#ifdef ATMEL_WLAN
	Adapter->SecurityInfo.EncryptionInfo.WepMode = WepMode;
#else
	Adapter->SecurityInfo.EncryptionInfo.ExcludeUnencrypted = WepMode;
#endif
	Adapter->SecurityInfo.EncryptionInfo.AuthenticationMode = AuthenticationType;
	Adapter->InternationalRoaming = InternationalRoaming;

#ifdef ATMEL_WLAN

  for (i = 0; i < MAX_ENCRYPTION_KEY_SIZE; i += 2)
    {
      Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[0][i] =
	AsciiToByte ((UCHAR) WepKey1[i], (UCHAR) WepKey1[i + 1]);
    }
  // Wep Key 2
  for (i = 0; i < MAX_ENCRYPTION_KEY_SIZE; i += 2)
    {
      Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[1][i] =
	AsciiToByte ((UCHAR) WepKey2[i], (UCHAR) WepKey2[i + 1]);
    }
  // Wep Key 3
  for (i = 0; i < MAX_ENCRYPTION_KEY_SIZE; i += 2)
    {
      Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[2][i] =
	AsciiToByte ((UCHAR) WepKey3[i], (UCHAR) WepKey3[i + 1]);
    }
  // Wep Key 4
  for (i = 0; i < MAX_ENCRYPTION_KEY_SIZE; i += 2)
    {
      Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[3][i] =
	AsciiToByte ((UCHAR) WepKey4[i], (UCHAR) WepKey4[i + 1]);
    }
#else
  
	// Wep Key 1
	j = 0;
	for (i = 0; i < WEP_KEY_SIZE * 2; i += 2) {
		Adapter->SecurityInfo.EncryptionInfo.WepKey1[j] =
		    AsciiToByte((UCHAR) WepKey1[i],
				(UCHAR) WepKey1[i + 1]);
		j++;
	}
	// Wep Key 2
	j = 0;
	for (i = 0; i < WEP_KEY_SIZE * 2; i += 2) {
		Adapter->SecurityInfo.EncryptionInfo.WepKey2[j] =
		    AsciiToByte((UCHAR) WepKey2[i],
				(UCHAR) WepKey2[i + 1]);
		j++;
	}
	// Wep Key 3
	j = 0;
	for (i = 0; i < WEP_KEY_SIZE * 2; i += 2) {
		Adapter->SecurityInfo.EncryptionInfo.WepKey3[j] =
		    AsciiToByte((UCHAR) WepKey3[i],
				(UCHAR) WepKey3[i + 1]);
		j++;
	}
	// Wep Key 4
	j = 0;
	for (i = 0; i < WEP_KEY_SIZE * 2; i += 2) {
		Adapter->SecurityInfo.EncryptionInfo.WepKey4[j] =
		    AsciiToByte((UCHAR) WepKey4[i],
				(UCHAR) WepKey4[i + 1]);
		j++;
	}
#endif
// TxRate       
	Adapter->TxRate = TxRate;
	Adapter->flags &= ~StationWasAssociated;
	Adapter->flags |= (TX_STOPPED | STOPPED_RX);
}


/*********************************************************/
//              Converts two Characters in One Byte                              //
/*********************************************************/
UCHAR AsciiToByte(UCHAR a1, UCHAR a2)
{
	return (HEX2BIN(a1) << 4) | (HEX2BIN(a2));
}

/************************************************************************
	Values initialization (through InitAdapter).
	Start Device initialization. Urb error handling (since in
	interrupt time (every callback runs in interrupt time)
	we cannot call usb_clear_halt, usb_reset_device etc).
************************************************************************/

int InitDevice(PVNet_ADAPTER Adapter)
{

	if (InitAdapter(Adapter) != 0) {
		err("parameters initialization failed");
		return -1;
	}

	Adapter->cmdhandler = GetOpMode;
	Adapter->flags |= INIT_PENDING;

	dbgcond(DBG_INIT, "flags %08lx\n", Adapter->flags);

	do {
		if (Adapter->flags & NEED_CLEAR_HALT) {
			Adapter->flags &= ~NEED_CLEAR_HALT;
			usb_clear_halt(Adapter->usb, usb_pipeendpoint(Adapter->ctrl_urb->pipe));
		}
		if (Adapter->flags & USB_RESET_DEVICE) {
			if (VnetUsbReset(Adapter) < 0) {
				err("Device Reset failed");
				Adapter->flags &= ~INIT_PENDING;
				Adapter->cmdhandler = NULL;
				return -1;
			}
                        
			Adapter->flags &= ~USB_RESET_DEVICE;
		}
		if (Adapter->ctrl_urb->status != -EINPROGRESS) {
			Adapter->cmdhandler(Adapter);
                        VNetSchedule(Adapter, 1);
			if (Adapter->flags & WAKE_UP_WITH_ERROR) {
			dbgcond(DBG_INIT, "init procedure - ctrl_wait %p Adapter %p\n", Adapter->ctrl_wait, Adapter);
				return -1;
			}
		}
	} while (Adapter->flags & INIT_PENDING);

  
	if (Adapter->CmdStatus[5] != CMD_STATUS_COMPLETE) {
		err("StartUp failed");
		return -1;
	}

	memcpy(Adapter->AssociationInfo.CurrentAddress, Adapter->StationAddress, 6);
	{
		extern struct usb_driver vnet_driver;
		printk(KERN_INFO "%s: driver version " VERSION_STRING
		       ", compiled " __DATE__ " " __TIME__
		       " (dbg_mask x%lx)\n", vnet_driver.name,
		       Adapter->dbg_mask);
	}
	printk(KERN_INFO "MAC addr %02X:%02X:%02X:%02X:%02X:%02X "
	       "firmware %d.%d.%d.%d\n",
	       Adapter->StationAddress[0], Adapter->StationAddress[1],
	       Adapter->StationAddress[2], Adapter->StationAddress[3],
	       Adapter->StationAddress[4], Adapter->StationAddress[5],
	       Adapter->FwVersion[0], Adapter->FwVersion[1],
	       Adapter->FwVersion[2], Adapter->FwVersion[3]);
	return 0;
}				/* InitDevice */

int ChangeConfiguration(PVNet_ADAPTER Adapter,
			DEVICE_CONFIGURATION * pDevConfig)
{
	UCHAR ScanOnChange = 0;
        unsigned long flags __attribute__ ((unused));

        int i;


	if ((Adapter->Channel != pDevConfig->Channel) ||
	    (Adapter->OperatingMode == AD_HOC_MODE)) {

		Adapter->Channel = pDevConfig->Channel;
		ScanOnChange = 1;
	}

        if(Adapter->InternationalRoaming != pDevConfig->InternationalRoaming){
           Adapter->InternationalRoaming = pDevConfig->InternationalRoaming;
	   if (Adapter->OperatingMode == AD_HOC_MODE){
		   Adapter->InternationalRoaming = 0;
	   }
	   ScanOnChange = 1;
        }
/* check for changed SSID */ 
	if (Adapter->SSID_size != pDevConfig->SSIDlength || memcmp(Adapter->SelectedSSID, pDevConfig->SSID, Adapter->SSID_size)) {
		char buf[2 * ESSID_SIZE + 1] __attribute__ ((unused));

		assert(pDevConfig->SSIDlength <= ESSID_SIZE);

                if(Adapter->InternationalRoaming == TRUE){
                    Adapter->flags &= ~FIRST_SCAN_DONE;
                                for (i=0; i<14; i++){
                                        Adapter->ChannelVector[i] = 0;
                                }
                }
		Adapter->AssociationInfo.DesiredSSIDsize = pDevConfig->SSIDlength;
		memcpy(Adapter->AssociationInfo.DesiredSSID, pDevConfig->SSID, pDevConfig->SSIDlength);
                Adapter->SSID_size = Adapter->AssociationInfo.DesiredSSIDsize;
		memcpy(Adapter->SelectedSSID, Adapter->AssociationInfo.DesiredSSID, Adapter->SSID_size);

		dbgcond(DBG_CONFIG, "new SSID %s\n",
			ssid2str(buf, Adapter->AssociationInfo.DesiredSSID,
				 Adapter->AssociationInfo.DesiredSSIDsize));

		    ScanOnChange = 1;
	}

	/* checked for changed op mode (i.e. infrastructure or adhoc */
        if (Adapter->OperatingMode != (pDevConfig->OperatingMode + 1)) {
            Adapter->OperatingMode = pDevConfig->OperatingMode + 1;
            if(Adapter->OperatingMode == AD_HOC_MODE){
               Adapter->InternationalRoaming = 0;
            }

            ScanOnChange = 1;
        }


	if (Adapter->TxRate != pDevConfig->TxRate) {
		Adapter->TxRate = pDevConfig->TxRate;
		if (Adapter->TxRate == 4)
			Adapter->flags |= AutoRFChanged;
		/*
		   
		else
		        ScanOnChange = 1;

		*/
	}

	if (Adapter->AssociationInfo.PreambleType != pDevConfig->PreambleType) {
		Adapter->AssociationInfo.PreambleType = pDevConfig->PreambleType;
		Adapter->flags |= PreambleChanged;
	}

	if (Adapter->RadioIsOn != pDevConfig->RadioIsOn) {
		Adapter->RadioIsOn = pDevConfig->RadioIsOn;
                if ( Adapter->RadioIsOn == TRUE){
		    if (Adapter->InternationalRoaming){
			    Adapter->flags &= ~FIRST_SCAN_DONE;
                                for (i=0; i<14; i++){
                                        Adapter->ChannelVector[i] = 0;
                                }
		    }

                    Adapter->AssociationInfo.DesiredSSIDsize = pDevConfig->SSIDlength;
                    memcpy(Adapter->AssociationInfo.DesiredSSID, pDevConfig->SSID, pDevConfig->SSIDlength);

                    ChangeState(Adapter, STATION_STATE_SCANNING);
                    SetRadio(Adapter, TRUE);
                } else {
                        ChangeState(Adapter, STATION_STATE_RADIO_OFF);
                        SetRadio(Adapter, FALSE);
                        DropPendingTxPackets(Adapter);
                }
                return 0;
                        
	}

	if (((pDevConfig->FragmentationThreshold <= 2346) ||
	     (pDevConfig->FragmentationThreshold >= 256)) &&
	    (Adapter->FragmentationThreshold !=
	     pDevConfig->FragmentationThreshold)) {
		Adapter->flags |= FragChanged;
                if ((pDevConfig->FragmentationThreshold)%2){
                      pDevConfig->FragmentationThreshold -= 1;
                }
		Adapter->FragmentationThreshold = pDevConfig->FragmentationThreshold;
	}
	if (((pDevConfig->RtsCtsThreshold <= 2347) ||
	     (pDevConfig->RtsCtsThreshold >= 256)) &&
	    (Adapter->RtsThreshold != pDevConfig->RtsCtsThreshold)) {
		Adapter->flags |= RtsCtsChanged;
		Adapter->RtsThreshold = pDevConfig->RtsCtsThreshold;
	}

        if (Adapter->RadioIsOn){
		if(Adapter->OperatingMode == INFRASTRUCTURE_MODE){
			if(Adapter->PowerMgmtMode != pDevConfig->PowerMgmtMode){
				if(Adapter->PowerMgmtMode){
					UsbJoin(Adapter);

				}
					Adapter->PowerMgmtMode = pDevConfig->PowerMgmtMode;
                                        if((Adapter->flags & StationWasAssociated) && (Adapter->PowerMgmtMode)){

                                            SetAssocID(Adapter);
                                        }
			}
		}

            if(Adapter->CurrentTXpowerIndbm != pDevConfig->TxPowerLevel) {
              pDevConfig->TxPowerLevel = 200;
              Adapter->CurrentTXpowerIndbm = pDevConfig->TxPowerLevel; 
              SetTxPower(Adapter);
            }

            if (ScanOnChange) {

                    dbgcond(DBG_CONFIG, "rescanning\n");

#if 0
                    if (Adapter->flags & GetCommandPedding) {
                            VNetSchedule(Adapter, IOCTL_SLEEPING);

                            if (Adapter->flags & WAKE_UP_WITH_ERROR) {
                                    Adapter->flags &= ~IOCTL_SLEEPING;
                                    return -1;
                            }
                    }
#endif
                    dbgcond(DBG_CONFIG,
                            "state %02x status tx_urb %d ctrl_urb %d\n",
                            Adapter->StationState, Adapter->tx_urb->status,
                            Adapter->ctrl_urb->status);
                    DropPendingTxPackets(Adapter);
                    Adapter->flags &= ~StationWasAssociated;
                    ChangeState(Adapter, STATION_STATE_SCANNING);
                    SetIRoaming(Adapter);
                    dbgcond(DBG_CONFIG,
                            "state %02x status tx_urb %d ctrl_urb %d\n",
                            Adapter->StationState, Adapter->tx_urb->status,
                            Adapter->ctrl_urb->status);
            }

	/* if (ScanOnChange) */
            if (Adapter->flags & PreambleChanged) {
                    /* jal: this will start a chain of Set*,
                       see callbacks.c::GetCmd_callback()  case CMD_SetMib */
                    SetPreamble(Adapter);
                    Adapter->flags &= ~PreambleChanged;
                    return 0;
            }

            if (Adapter->flags & FragChanged) {
                    SetFragThreshold(Adapter);
                    Adapter->flags &= ~FragChanged;
                    return 0;
            }

            if (Adapter->flags & RtsCtsChanged) {
                    SetRTSThreshold(Adapter);
                    Adapter->flags &= ~RtsCtsChanged;
                    return 0;
            }

            if (Adapter->flags & AutoRFChanged) {
                    SetAutoRateFallback(Adapter);
                    Adapter->flags &= ~AutoRFChanged;
            }

        }
	return 0;
}

BOOLEAN ValidateChannel(PVNet_ADAPTER Adapter, UCHAR Channel)
{
	switch (Adapter->RegDomain) {
	case REG_DOMAIN_FCC:		//1-11
		if ((Channel < 1) || (Channel > 11))
			return FALSE;
		break;
	case REG_DOMAIN_DOC:		//11
		if (Channel != 11)
			return FALSE;
		break;
	case REG_DOMAIN_ETSI:		//1-13
		if ((Channel < 1) || (Channel > 13))
			return FALSE;
		break;
	case REG_DOMAIN_SPAIN:		//10-11
		if ((Channel < 10) || (Channel > 11))
			return FALSE;
		break;
	case REG_DOMAIN_FRANCE:		//10-13
		if ((Channel < 10) || (Channel > 13))
			return FALSE;
		break;
	case REG_DOMAIN_MKK:		//14
		if (Channel != 14)
			return FALSE;
		break;
	case REG_DOMAIN_MKK1:		//1-14
		if ((Channel < 1) || (Channel > 14))
			return FALSE;
		break;
	case REG_DOMAIN_ISRAEL:		//3-7
		if ((Channel < 3) || (Channel > 9))
			return FALSE;
		break;

	}
	return TRUE;
}


void TimerExpired(ULONG a)
{
	PVNet_ADAPTER Adapter = (PVNet_ADAPTER) a;
  UCHAR MaxCnt;
	ULONG flags;

  
	if ((!Adapter) || (!(Adapter->flags & VNET_RUNNING)
			   && (Adapter->StationState !=
			       STATION_STATE_FW_DOWNLOAD)))
		return;

  Adapter->lock = SPIN_LOCK_UNLOCKED;

	spin_lock_irqsave(&Adapter->lock, flags);
	Adapter->flags &= ~TimerReScheduled;

	switch (Adapter->StationState) {

	case STATION_STATE_FW_DOWNLOAD:
		// if usb_reset_device is called here the result will be :
		// Scheduling in interrupt,,,Kernel BUG at .....
		// In this case we wake the sleeping task (InitDevice in vnet.c)
		// and handle the reset there..........
		dbgcond(DBG_FW_DL, "going for reset\n");
		Adapter->flags |= USB_RESET_DEVICE;
		Adapter->cmdhandler = GetOpMode;
		wake_up_interruptible(Adapter->ctrl_wait);
		break;

	case STATION_STATE_AUTHENTICATING:
		dbgcond(DBG_AUTH,
			"auth retry: nr retries %d rx_urb->status %d\n",
			Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt,
			Adapter->rx_urb->status);
		Adapter->AuthenticationInfo.CurrentAuthentTransactionSeqNum = 0x0001;
		if (++Adapter->AuthenticationInfo.AuthenticationRequestRetryCnt >
		    MAX_AUTHENTICATION_RETRIES) {
			dbgcond(DBG_AUTH,
				"max auth retries (%d) reached - giving up\n",
				MAX_AUTHENTICATION_RETRIES);
			DropPendingTxPackets(Adapter);
			ChangeState(Adapter, STATION_STATE_SCANNING);
			UsbScan(Adapter);
		} else {
			SendAuthRequest(Adapter);
		}
		break;

	case STATION_STATE_ASSOCIATING:
		dbgcond(DBG_ASSOC, "assoc retry: nr retries %d\n",
			Adapter->AssociationInfo.AssociationRequestRetryCnt);
		if (++Adapter->AssociationInfo.AssociationRequestRetryCnt >
		    MAX_ASSOCIATION_RETRIES) {
			DropPendingTxPackets(Adapter);
			ChangeState(Adapter, STATION_STATE_SCANNING);
			UsbScan(Adapter);
		} else {
			SendAssocRequest(Adapter);
		}
		break;


	case STATION_STATE_REASSOCIATING:
		dbgcond(DBG_ASSOC, "assoc retry: nr retries %d\n",
			Adapter->AssociationInfo.ReAssociationRequestRetryCnt);
		if (++Adapter->AssociationInfo.ReAssociationRequestRetryCnt >
		    MAX_REASSOCIATION_RETRIES) {
			DropPendingTxPackets(Adapter);
			ChangeState(Adapter, STATION_STATE_SCANNING);
			UsbScan(Adapter);
		} else {
			SendReAssocRequest(Adapter);
		}
		break;

	case STATION_STATE_READY:
		// Out of range handler. Check every 2 seconds if HeardBeacons bit is set. If
		// not => OutOfRange
		if (Adapter->flags & HeardBeacons) {
       MaxCnt = PERIODIC_TIMER_OUT / Adapter->AssociationInfo.BeaconPeriod;

       if(Adapter->PowerMgmtMode == 1) {
                 MaxCnt = MaxCnt / Adapter->AssociationInfo.ListenInterval;
       }


			Adapter->flags &= ~HeardBeacons;
          if(Adapter->OperatingMode == INFRASTRUCTURE_MODE) {
              Adapter->LinkQuality = (UCHAR) ((Adapter->LinkQuality + ((Adapter->PeriodicBeaconsCnt * 100)/ MaxCnt) / 2));
              if(Adapter->LinkQuality > 100)
                Adapter->LinkQuality = 100;
          }
			VnetTimer(Adapter, 25);
			break;
		}
                Adapter->PeriodicBeaconsCnt = 0;
                printk("CAUTION : ==> OUT OF RANGE\n");
		DropPendingTxPackets(Adapter);
		ChangeState(Adapter, STATION_STATE_SCANNING);
		UsbScan(Adapter);
		break;

	case STATION_STATE_JOINING:
	        spin_unlock_irqrestore(&Adapter->lock, flags);
		dbgcond(DBG_JOIN, "Timer : ctrl_urb->status %d\n", Adapter->ctrl_urb->status);
		return;

	default:
		if (!(Adapter->flags & TimerReScheduled))
			VnetTimer(Adapter, 25);
		break;
	}
	spin_unlock_irqrestore(&Adapter->lock, flags);
}

void VnetReleaseResources(PVNet_ADAPTER Adapter)
{

	if (!Adapter)
		return;

	usb_free_urb(Adapter->rx_urb);
	Adapter->rx_urb = NULL;
	usb_free_urb(Adapter->tx_urb);
	Adapter->tx_urb = NULL;
	usb_free_urb(Adapter->ctrl_urb);
	Adapter->ctrl_urb = NULL;
	dbgcond(DBG_INIT, "Urb's freed\n");

	VNetFree(Adapter->CtrlTimer);
	Adapter->CtrlTimer = NULL;
	VNetFree(Adapter->ctrl_wait);
	Adapter->ctrl_wait = NULL;
	VNetFree(Adapter->remove_wait);
	Adapter->remove_wait = NULL;

	dbgcond(DBG_INIT, "Timer's OK freed\n");

	VNetFree(Adapter->netstats);
	Adapter->netstats = NULL;
	VNetFree(Adapter->w_stats);
	Adapter->w_stats = NULL;
	VNetFree(Adapter->devreq);
	Adapter->devreq = NULL;
	//VNetFree(&Adapter->lock);
	//Adapter->lock = SPIN_LOCK_UNLOCKED;
	VNetFree(Adapter->freeTx);
	Adapter->freeTx = NULL;

	VNetFree(Adapter);
}

void DropPendingTxPackets(PVNet_ADAPTER Adapter)
{
	struct list_head *p;
	PTX_LIST pTxList;

	dbgcond(DBG_TX, "freeing all pending tx packets\n");

	VNetIf_queue(Adapter->net, 0);

	/* free the TxList of pending Tx packets */
	for (p = Adapter->TxList->next; p != Adapter->TxList;) {
		pTxList = list_entry(p, TX_LIST, tx_list);
		list_del(p);
		p = p->next;	/* get the next pointer here, we de-alloc *p below */

		if (pTxList->status & UNDER_SUBMIT) {
			dbgcond(DBG_TX, "tx packet %p dropped\n", pTxList);
#if (LINUX_VERSION_CODE < 0x020607)
          wait_ms(10);
#else
          msleep(10);
#endif
			pTxList->status &= ~UNDER_SUBMIT;	/* jal: shall we inform USB somehow ??? */
		}

		Adapter->netstats->tx_dropped++;
		kfree(pTxList->tx_buff);
		kfree(pTxList);
	}
	Adapter->flags &= ~TX_SLEEP;
}


VOID init_CRCtable(PVNet_ADAPTER Adapter)
{
	ULONG crc;
	int i, j;

	for (i = 0; i < 256; i++) {
		crc = i;
		for (j = 8; j > 0; j--) {
			if (crc & 1)
				crc = (crc >> 1) ^ 0xedb88320;
			else
				crc >>= 1;
		}
		Adapter->CrcTable[i] = crc;
	}
}

ULONG Calculate_CRC32(PUCHAR p, ULONG FrameLength, PULONG crctab)
{
	ULONG crcv, t;
	ULONG nr = 0;

	crcv = 0xffffffff;
	while (nr < FrameLength) {
		t = (crcv ^ p[nr]) & 0xff;
		crcv = ((crcv >> 8) & 0x00ffffff) ^ crctab[t];
		nr++;
	}
	crcv = (crcv ^ 0xffffffff);

	return (ULONG) crcv;
}


USHORT Bytes2Ushort(PUCHAR pBuf)
{
    USHORT tmp;

      tmp = *(pBuf) << 8; //((USHORT)(*(pBuf))) << 8;
        tmp |= (*(pBuf+1));

          return tmp;
}

#define ROL32( A, n )  ( ((A) << (n)) | ( ((A)>>(32-(n)))  & ( (1UL << (n)) - 1 ) ) )
#define ROR32( A, n ) ROL32( (A), 32-(n) )


void WPA_MicCalculation(UCHAR *msg,UINT msgLength )
{
	UINT i;
	ULONG *M;
	
	M=(ULONG *) msg;
	for(i=0;i<msgLength;i++)
	{
		Lefty ^= M[i];
		Righty ^= ROL32( Lefty, 17 );
		Lefty += Righty;
		Righty ^= ((Lefty & 0xff00ff00) >> 8) | ((Lefty & 0x00ff00ff) << 8);
		Lefty += Righty;
		Righty ^= ROL32( Lefty, 3 );
		Lefty += Righty;
		Righty ^= ROR32( Lefty, 2 );
		Lefty += Righty;
	}
}

void setAddressesField(UCHAR *pPacket,UCHAR *addressesField)
{
	UCHAR tmpSA,tmpDA, i;

	tmpDA = (!(pPacket[1] & 0x01)) ? 4 : 16; //To DS
	tmpSA = (pPacket[1] & 0x02) ? 16 : 10;  //From DS

	for(i=0;i<6;i++)
	{
		addressesField[i]=pPacket[tmpDA+i];
		addressesField[6+i]=pPacket[tmpSA+i];
	}
	for(i=12;i<16;i++)
		addressesField[i]=0;

}

void WPA_michael(UCHAR *packet, UCHAR *micKey, UINT msgLength)
{
	UCHAR   numOfBytes2Add, i, addressesField[16];	

	Lefty=((ULONG *) micKey)[0];
	Righty=((ULONG *) micKey)[1];

	//Michael operation on address fields: DA and SA.
	setAddressesField(packet,addressesField);
	WPA_MicCalculation(addressesField,4);
	//Padding
	packet[msgLength]=0x5a;
	numOfBytes2Add=(msgLength+1)%4;
	if(numOfBytes2Add==0) numOfBytes2Add=4;
	else numOfBytes2Add=8-numOfBytes2Add;
	
	for(i=1;i<=numOfBytes2Add;i++) packet[msgLength+i]=0;

	//Calc Mic for raw data			
	WPA_MicCalculation(&packet[24],(UINT)((msgLength+1+numOfBytes2Add-24)>>2));
}

UCHAR WPA_validateMic(UCHAR *packet,UINT msgLength,UCHAR *micKey)
{

	UCHAR	rxMic[8];
	UCHAR	i, *mic;

	//Keep the MIC value of the received packet
	mic = packet + msgLength - 8;
	for(i=0;i<8;i++) rxMic[i]=mic[i];
	//Calculate MIC from the payload of the received packet
	WPA_michael(packet, micKey, (UINT)(msgLength-8) );
	//Append received mic value at the end of the packet
	for(i=0;i<8;i++) mic[i]=rxMic[i];
	
	//Check if mic value received and calculated match
	if( (Lefty!=(((ULONG *) rxMic)[0]))||(Righty!=(((ULONG *) (rxMic+4) )[0]) ) )	
	{
		return 0;
	}
	return 1;
}

void WPA_addMICToPacket(UCHAR *packet,UINT msgLength,UCHAR *micKey)
{
	UCHAR i, *mic;
			
	WPA_michael(packet, micKey, (UINT)(msgLength-8) );

	mic = packet + msgLength - 8;
	for(i=0;i< 4;i++)
	{
		mic[i]=(UCHAR) (Lefty>>(8*i) );
		mic[i+4]=(UCHAR)(Righty>>(8*i) );
	}
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	WPA_AddAPInBlackList()
//	Adds an AP to the MIC Black List
//
//	Parameters
//		MIC_BLACK_LIST*	Pointer to the black list
//		UCHAR			Size of the black list (entries, not bytes)
//		PUCHAR			Pointer to a parameter that keeps the entries that are in use
//		PUCHAR			Pointer to a 6-byte buffer that holds the AP's BSSID
//		PVOID			Pointer to a parameter that contains the timestamp (OS depended)
//		
//
//	returns:	TRUE  : if the AP was added in the list
//				FALSE : if the list was full
/////////////////////////////////////////////////////////////////////////////
BOOLEAN WPA_AddAPInBlackList(MIC_BLACK_LIST* MicBlackList, unsigned char ListEntries, unsigned char* pValidListEntries, unsigned char* pBSSID, void* pTime)
{
	UCHAR	i;
	UCHAR	ReservedBSSID[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

	// limit the search for empty entry to the ListEntries
	for(i=0; i<ListEntries; i++)
	{
		// if an entry is not used, use it
		if(!MicBlackList[i].InUse)
		{
			memcpy(MicBlackList[i].Bssid, pBSSID, 6);
			MicBlackList[i].TimeStamp = *((ULONG64*)pTime);
			// mark it as used
			MicBlackList[i].InUse = TRUE;

			// increase the valid list entries and return success
			(*pValidListEntries) ++;
			return TRUE;
		}
	}

	// if we are here it means that all the list members are in use.
	// we have to check if there are any entries that can be removed.
	// we use the InvalidBSSID to force the WPA_IsOkToConnect to invalidate
	// all unnecessary entries.
	WPA_IsOkToConnect(MicBlackList, ListEntries, pValidListEntries, ReservedBSSID, pTime);

	// we might have found an empty place, so try again to add the AP to the list
	for(i=0; i<ListEntries; i++)
	{
		if(!MicBlackList[i].InUse)
		{
			memcpy(MicBlackList[i].Bssid, pBSSID, 6);
			MicBlackList[i].TimeStamp = *((ULONG64*)pTime);
			MicBlackList[i].InUse = TRUE;

			(*pValidListEntries) ++;
			return TRUE;
		}
	}

	// could not add the AP in the list, return failure
	return FALSE;
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//		Compare_Array()  checks if x and y are identical 
//
//	returns:	TRUE  : x and y are identical
//				FALSE : x and y are not identical
/////////////////////////////////////////////////////////////////////////////
BOOLEAN Compare_Array(PUCHAR x,
                     PUCHAR y,
                     USHORT l)
{
    USHORT  i;

    for (i=0; i<l; i++)
    {
        if (x[i]!=y[i]) return FALSE;
    }

    return TRUE;
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
//	WPA_IsOkToConnect()
//	Checks if an AP is in the black list for more or less than 60 seconds
//	If it is more than 60 secondos, it invalidates that entry.
//
//	Parameters
//		MIC_BLACK_LIST*	Pointer to the black list
//		UCHAR			Size of the black list (entries, not bytes)
//		PUCHAR			Pointer to a parameter that keeps the entries that are in use
//		PUCHAR			Pointer to a 6-byte buffer that holds the AP's BSSID
//		PVOID			Pointer to a parameter that contains the timestamp (OS depended)
//		
//
//	returns:	TRUE  : if the AP was added in the list
//				FALSE : if the list was full
//
//	NOTE: The ReservedBSSID indicates that we should go throught the whole list
//		  until we find an AP that could be removed from the list, in order to
//		  create space for a new entry.
/////////////////////////////////////////////////////////////////////////////
BOOLEAN WPA_IsOkToConnect(MIC_BLACK_LIST* MicBlackList, unsigned char ListEntries, unsigned char* pValidListEntries, unsigned char* pBSSID, void* pCurrentTime)
{
	ULONG	TimeInterval;
	UCHAR	i;
	UCHAR	ValidEntriesInBlackListCounter = 0;
	UCHAR	ReservedBSSID[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

	// limit the search to the ListEntries
	for(i=0;i<ListEntries;i++)
	{
		// first check if this entry is valid
		if(MicBlackList[i].InUse)
		{
			ValidEntriesInBlackListCounter ++;

			// compare the requested BSSID with the list BSSID and the ReservedBSSID.
			if(Compare_Array(MicBlackList[i].Bssid, pBSSID, 6) || 
			   Compare_Array(ReservedBSSID, pBSSID, 6)
			  )
			{
				TimeInterval = (ULONG)((*(ULONG64*)pCurrentTime) - MicBlackList[i].TimeStamp);
				// check if the AP is more than 60 seconds
				if(TimeInterval > 600000000)
				{
					// invalidate the entry and reduce the valid list entries
					MicBlackList[i].InUse = FALSE;
					(*pValidListEntries) --;
					return TRUE;			
				}

				// if we are not going throught the whole list, indicate
				// that we have found an bad AP.
				if( !Compare_Array(ReservedBSSID, pBSSID, 6) )
					return FALSE;
			}

			if(ValidEntriesInBlackListCounter == *pValidListEntries)
				return TRUE;
		}
	}	
	return TRUE;
}
#define MAX_ENCRYPTION_KEYS   4

BOOLEAN RemoveKey(PVNet_ADAPTER Adapter)
{
	ULONG					i,j,k;



		printk("Remove ALL Keys\n");
					

		k=0;
		for(i=0;i<MAX_ENCRYPTION_KEYS;i++)
		{
			k=0;
			for(j=0;j<32;j++)
			{
				Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[i][j] = Adapter->AssociationInfo.CurrentAddress[k];
				k++;
				if(k==6)
					k=0;
			}
			for(j=32;j<40;j++)
			{
				Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[i][j] = 0;
			}

		}

		Adapter->KeyIndex = 5;	// to show that you will remove all keys...
    SetEncryptionStatus(Adapter, TRUE, TRUE, Adapter->KeyIndex);


	return TRUE;
}


BOOLEAN RemoveEncryptionSettings(PVNet_ADAPTER Adapter)
{
	RemoveKey(Adapter);

	if(Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption2Enabled || Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption2KeyAbsent)
		Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption2KeyAbsent;
#ifdef AES
	else if(Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption3Enabled || Adapter->SecurityInfo.EncryptionStatus == _802_11Encryption3KeyAbsent)
		Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption3KeyAbsent;
#endif
	else
		Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption1KeyAbsent; 


	Adapter->SecurityInfo.PairwiseKeyAvailable = FALSE;
	Adapter->SecurityInfo.GroupKeyAvailable = FALSE;

	Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = CIPHER_SUITE_NONE;
	Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite = CIPHER_SUITE_NONE;
	
	return TRUE;
}

