/* $Id: vnetusba.c,v 1.31 2004/09/16 16:26:18 tmpetsos Exp $ */
/***************************************************************************************
	Copyright 2000-2001 ATMEL Corporation.
	
	This file is part of atmel wireless lan drivers.

    Atmel wireless lan drivers is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Atmel wireless lan drivers is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmel wireless lan drivers; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************************/
//jal: there might be a makefile compiling all c source in one call ...
#ifndef MODULE
# define MODULE
#endif

#include "vnetusba.h"
#include <linux/kmod.h>		/* for call_usermodehelper only */
#include <linux/string.h>
#include <net/iw_handler.h>
#include "vnetioctl.h"
#include "config.h"

struct usb_device_id __devinitdata vnet_usb_id[] =
#if defined(RFMD)
    RFMD_DEVICES
#elif defined(R505)    
    R505_DEVICES
#elif defined(R505A)    
    R505A_DEVICES
#elif (defined(R505AMX))    
    R505AMX_DEVICES
#elif defined(R505_2958)    
    R505_2958_DEVICES
#endif
;

/* the number of entries in array above */
int const vnet_usb_id_len =
    sizeof(vnet_usb_id) / sizeof(struct usb_device_id);

static const struct {
	int reg_domain;
	int min, max;
} channel_table[] = { { REG_DOMAIN_FCC, 1, 11},
		      { REG_DOMAIN_DOC, 1, 11},
		      { REG_DOMAIN_ETSI, 1, 13},
		      { REG_DOMAIN_SPAIN, 10, 11},
		      { REG_DOMAIN_FRANCE, 10, 13},
		      { REG_DOMAIN_MKK, 14, 14},
		      { REG_DOMAIN_MKK1, 1, 14},
		      { REG_DOMAIN_ISRAEL, 9} };
/* forward for vnet_driver */
#if (LINUX_VERSION_CODE < 0x020500)
static void usb_vnet_disconnect(struct usb_device *dev, void *ptr);
static void *usb_vnet_probe(struct usb_device *dev, UINT interface,
			    const struct usb_device_id *id_table);
#else
static void usb_vnet_disconnect(struct usb_interface *interface);
int usb_vnet_probe(struct usb_interface *interface, 
			    const struct usb_device_id *id_table);
#endif

struct usb_driver vnet_driver = {
#if LINUX_VERSION_CODE < 0x020415
owner : THIS_MODULE,
#endif
#if defined(RFMD)
name : "usbvnetr",
#elif defined(R505)             
name : "usbvnet5",
#elif defined(R505A)             
name : "usbvnet5A",
#elif (defined(R505AMX))             
name : "usbvnet5AMX",
#elif defined(R505_2958)             
name : "usbvnet5_2958",
#endif
probe : usb_vnet_probe,
disconnect : usb_vnet_disconnect,
id_table : vnet_usb_id,
};

/* == PROC ssid2str ==
   converts a SSID into a zero terminated string in the given buffer */
char *ssid2str(char *buf, char *ssid, int len)
{
	assert(len <= ESSID_SIZE);
	if (len > ESSID_SIZE)
		len = ESSID_SIZE;
	memcpy(buf, ssid, len);
	buf[len] = '\0';

	return buf;
}

/* == PROC buf2str ==
   hexdump a memory area into a given buffer, zero terminated */
char *buf2str(char *str, void *buf, int len)
{
	unsigned char *ptr = buf;
	char *retval = str;
	while (len--) {
		*str++ = BIN2HEX(*ptr >> 4);
		*str++ = BIN2HEX(*ptr & 0xf);
		ptr++;
	}
	*str = '\0';
	return retval;
}


void mod_use(UCHAR a, PVNet_ADAPTER Adapter)
{
	if (a) {
		Adapter->CallbacksPending++;
#if (LINUX_VERSION_CODE < 0x020500)
		MOD_INC_USE_COUNT;
#endif
	} else {
		Adapter->CallbacksPending--;
#if (LINUX_VERSION_CODE < 0x020500)
		MOD_DEC_USE_COUNT;
#endif
	}
}

void ChangeState(PVNet_ADAPTER Adapter, UCHAR State)
{
		if ((Adapter->StationState == STATION_STATE_READY) 
        && (Adapter->StationState != State)){
			if (!(Adapter->flags & TX_STOPPED)) 
				netif_stop_queue(Adapter->net);
				Adapter->flags |= TX_STOPPED;
			} else if ((Adapter->StationState != State) 
		    && (State == STATION_STATE_READY)) {
			if (Adapter->flags & TX_STOPPED) 
				netif_wake_queue(Adapter->net);
				Adapter->flags &= ~TX_STOPPED;
		}
		dbgcond(DBG_STATE, "state change %d -> %d\n",
			Adapter->StationState, State);
	Adapter->StationState = State;
}

static int usb_vnet_open(struct net_device *dev)
{
	PVNet_ADAPTER Adapter = (PVNet_ADAPTER) dev->priv;

	if ((Adapter->StationState == STATION_STATE_INITIALIZING) ||
	    (Adapter->StationState == STATION_STATE_FW_DOWNLOAD))
		return -1;

	Adapter->flags |= VNET_RUNNING;

#if (LINUX_VERSION_CODE < 0x020500)
	MOD_INC_USE_COUNT;
#endif

	if (Adapter->flags & STOPPED_RX)
		UsbRxInit(Adapter);

  printk("-> vnet_open(%s)\n", dev->name);
  netif_start_queue (dev);
  if(!netif_running(dev)){
     if (netif_queue_stopped (dev)){
         printk("waiking up the net queue - (%s)\n", dev->name);
         netif_wake_queue (dev);
     }
  }

#if (LINUX_VERSION_CODE > 0x020500)
  SET_MODULE_OWNER (Adapter->net);
#endif

	if (Adapter->flags & TX_STOPPED) {
		netif_start_queue(Adapter->net);
		Adapter->flags &= ~TX_STOPPED;
	}
	return 0;
}

#if 0
static void
usb_vnet_tx_timeout(struct net_device *dev)
{
    PVNet_ADAPTER Adapter = (PVNet_ADAPTER ) dev->priv;

    if(Adapter->flags & TX_STOPPED){
    printk(KERN_INFO __FILE__ "Tx Timeout called\n");
        Adapter->flags &= ~TX_STOPPED;
        netif_wake_queue(dev);
    }

}
#endif

static int usb_vnet_close(struct net_device *dev)
{
	PVNet_ADAPTER Adapter = (PVNet_ADAPTER) dev->priv;

	Adapter->flags &= ~VNET_RUNNING;

#if (LINUX_VERSION_CODE < 0x020500)
#if defined (R505) || defined (RFMD) || defined (R505_2958) || defined (R505A) || defined (R505AMX)
    if (MOD_IN_USE > 2) {
        MOD_DEC_USE_COUNT;
    }
#else
      MOD_DEC_USE_COUNT;
#endif
#endif	// Linux v. < 2.5+

	printk("Device %s closed\n", dev->name);
	return 0;
}

static struct net_device_stats *usb_vnet_net_stats(struct net_device *dev)
{
	if (!dev->priv)
		return NULL;
	return ((PVNet_ADAPTER) dev->priv)->netstats;
}

static struct iw_statistics *wireless_stats(struct net_device *dev)
{
	PVNet_ADAPTER Adapter = (PVNet_ADAPTER) dev->priv;

  Adapter->w_stats->status = 0;
	Adapter->w_stats->qual.qual = (UCHAR) (Adapter->LinkQuality);
	Adapter->w_stats->qual.level = ((Adapter->Rssi > 40) ? 40 : Adapter->Rssi) * 10 / 4;
	Adapter->w_stats->qual.updated = 1;

	return (Adapter->w_stats);
}
//////////////////////////////////////////////////////////////////////////////////////////////////

/* IW-HANDLERS */
static const long frequency_list[] = { 2412, 2417, 2422, 2427, 2432, 2437, 2442,
				2447, 2452, 2457, 2462, 2467, 2472, 2484 };

static int vnet_set_encode(struct net_device *dev,
			    struct iw_request_info *info,
			    struct iw_point *wrq,
			    char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        int i __attribute__ ((unused));



		if ((wrq->flags & IW_ENCODE_DISABLED) || (wrq->flags & IW_ENCODE_NOKEY)) {
			Adapter->PrivacyInvoked = 0;
                        Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = 0;         
			Adapter->SecurityInfo.AuthenticationMode = C80211_MGMT_AAN_OPENSYSTEM;
			Adapter->SecurityInfo.EncryptionInfo.AuthenticationMode = Adapter->SecurityInfo.AuthenticationMode;
#if 0
#ifdef ATMEL_WLAN
      Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = CIPHER_SUITE_WEP_128;
#endif
#endif
                        // Disable Wep
		if (Adapter->flags & GetCommandPedding) {
			DECLARE_WAITQUEUE(wait, current);
			add_wait_queue(Adapter->ctrl_wait, &wait);
			set_current_state(TASK_INTERRUPTIBLE);
			Adapter->flags |= IOCTL_SLEEPING;
			schedule();
			set_current_state(TASK_RUNNING);
			remove_wait_queue(Adapter->ctrl_wait, &wait);
			if (Adapter->flags & WAKE_UP_WITH_ERROR) {
				Adapter->flags &= ~IOCTL_SLEEPING;
				return -1;
			}
		}
		DropPendingTxPackets(Adapter);
		dbgcond(DBG_IOCTL, "Tx Packets Dropped\n");
		SetWEPvalue(Adapter);
    //End of Disable - Wait for Callback
			return 0;
		} else {
    
    
      Adapter->PrivacyInvoked = 1;
		if (wrq->flags & IW_ENCODE_OPEN){
			Adapter->SecurityInfo.AuthenticationMode = C80211_MGMT_AAN_OPENSYSTEM;
			Adapter->SecurityInfo.EncryptionInfo.AuthenticationMode = Adapter->SecurityInfo.AuthenticationMode;
    } else if (wrq->flags & IW_ENCODE_RESTRICTED) {
			Adapter->SecurityInfo.AuthenticationMode = C80211_MGMT_AAN_SHAREDKEY;
			Adapter->SecurityInfo.EncryptionInfo.AuthenticationMode = Adapter->SecurityInfo.AuthenticationMode;
    }

		if ((wrq->length != 5) && (wrq->length != 13))
			return -EINVAL;

#ifdef ATMEL_WLAN
                  switch(wrq->length)
                  {
                  case 5 :
			                  Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = WEP_64BIT;
                        Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = CIPHER_SUITE_WEP_64;
                    break;
                  case 13 :
			                  Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = WEP_128BIT;
                        Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = CIPHER_SUITE_WEP_128;
                    break;
                  }
#else
		if (wrq->length == 5)
			Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = WEP_64BIT;
		else
			Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel = WEP_128BIT;
#endif

		Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse = (wrq->flags & IW_ENCODE_INDEX) - 1;
		dbgcond(DBG_IOCTL, "WepKeyToUse : %d\n", Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse);
		if (Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse > 3)
			return -EINVAL;
#ifdef ATMEL_WLAN
      memcpy(((PUCHAR) Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse]),
		      extra, wrq->length);
      Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse][39] = 1;
#else
		copy_from_user((PUCHAR) &Adapter->SecurityInfo.EncryptionInfo +
			       (5 + (13 * Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse)),
			       wrq->pointer, wrq->length);
#endif

		Adapter->flags &= ~WepIsSet;
		if (Adapter->flags & GetCommandPedding) {
			DECLARE_WAITQUEUE(wait, current);
			add_wait_queue(Adapter->ctrl_wait, &wait);
			set_current_state(TASK_INTERRUPTIBLE);
			Adapter->flags |= IOCTL_SLEEPING;
			schedule();
			set_current_state(TASK_RUNNING);
			remove_wait_queue(Adapter->ctrl_wait, &wait);
			if (Adapter->flags & WAKE_UP_WITH_ERROR) {
				Adapter->flags &= ~IOCTL_SLEEPING;
				return -1;
			}
		}
		dbgcond(DBG_IOCTL, "Setting Wep (%ld)\n",
			Adapter->Stats.RxDataPacketsOk);
		DropPendingTxPackets(Adapter);
		dbgcond(DBG_IOCTL, "Tx Packets Dropped\n");
		SetWEPvalue(Adapter);
		dbgcond(DBG_IOCTL, "out\n");
                }
        
	return 0;
}

static int vnet_get_encode(struct net_device *dev,
			    struct iw_request_info *info,
			    struct iw_point *wrq,
			    char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;


		if (Adapter->PrivacyInvoked == 0) {
			wrq->flags |= IW_ENCODE_DISABLED;
			wrq->length = 0;
		} else {
			wrq->flags &= ~IW_ENCODE_DISABLED;
			wrq->flags |= (Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse + 1);
			if (Adapter->SecurityInfo.EncryptionInfo.AuthenticationMode == C80211_MGMT_AAN_OPENSYSTEM)
				wrq->flags |= IW_ENCODE_OPEN;
      else
				wrq->flags |= IW_ENCODE_RESTRICTED;

			wrq->length = (Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == WEP_64BIT) ? 5 : 13;
		}
#ifdef ATMEL_WLAN
      memcpy(extra,
		    ((PUCHAR) Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse] +
		     ((Adapter->SecurityInfo.EncryptionInfo.WepKeyToUse * 13))),
		    wrq->length);
#else
		copy_to_user(wrq->pointer,
			     (PUCHAR) &Adapter->SecurityInfo.EncryptionInfo + (5 +
							    (Adapter->
							     SecurityInfo.EncryptionInfo.
							     WepKeyToUse *
							     13)),
			     wrq->length);
#endif
	return 0;
}

static int vnet_set_rts(struct net_device *dev,
			 struct iw_request_info *info,
			 struct iw_param *vwrq,
			 char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;

	UpdateAppInfo(Adapter, &DevConfig);
	if(vwrq->disabled)
		DevConfig.RtsCtsThreshold = 2347;
	if((DevConfig.RtsCtsThreshold < 255) || (DevConfig.RtsCtsThreshold > 2347)) {
		return -EINVAL;
	}
	DevConfig.RtsCtsThreshold = vwrq->value;
        ChangeConfiguration(Adapter, &DevConfig);
	return 0;		/* Call commit handler */
}
static int vnet_set_wap(struct net_device *dev,
			 struct iw_request_info *info,
			 struct sockaddr *awrq,
			 char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
       // DEVICE_CONFIGURATION DevConfig;
	int i;
        unsigned long flags __attribute__ ((unused));

	//static const u8 bcast[] = { 255, 255, 255, 255, 255, 255 };
	
	if (awrq->sa_family != 1)
		return -EINVAL;
	
	if (memcmp(Adapter->AssociationInfo.CurrentBSSID, awrq->sa_data, 6) != 0) {
                ChangeState(Adapter, STATION_STATE_SCANNING);
                 DropPendingTxPackets(Adapter);
                 UsbScan(Adapter);
		return 0;
	}
	for(i=0; i<Adapter->AssociationInfo.SiteS.BSSInList; i++) {
		if (memcmp(Adapter->AssociationInfo.SiteS.BssInfo[i].BSSID, awrq->sa_data, 6) == 0) {
#if 0
                    if (Adapter->AssociationInfo.SiteS.BssInfo[i].EncryptionIsOn == 1 && Adapter->PrivacyInvoked == 0){ 
                        Adapter->ErrorCode=0xa0;
                        ChangeState(Adapter, STATION_STATE_MGMT_ERROR);
                        printk("UsbJoin (1)\n");
                        return 0;
                        //continue;
                    } else if (Adapter->AssociationInfo.SiteS.BssInfo[i].EncryptionIsOn == 0 && Adapter->PrivacyInvoked == 1){ 
                        Adapter->ErrorCode=0xa1;
                        ChangeState(Adapter, STATION_STATE_MGMT_ERROR);
                        printk("UsbJoin (2)\n");
                        return 0;
                    }else {
#endif
                                UsbJoin(Adapter);
                                return 0;
		}
	}

	return 0;
}
static int vnet_get_wap(struct net_device *dev,
			 struct iw_request_info *info,
			 struct sockaddr *awrq,
			 char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        unsigned long flags __attribute__ ((unused));

	memcpy(awrq->sa_data, Adapter->AssociationInfo.CurrentBSSID, 6);
	awrq->sa_family = 1;

	return 0;
}

static int vnet_get_rts(struct net_device *dev,
			 struct iw_request_info *info,
			 struct iw_param *vwrq,
			 char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;


        UpdateAppInfo(Adapter, &DevConfig);
	
	vwrq->value = DevConfig.RtsCtsThreshold;
	vwrq->disabled = (vwrq->value >= 2347);
	vwrq->fixed = 1;

	return 0;
}
static int vnet_get_frag(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_param *vwrq,
			  char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;

        UpdateAppInfo(Adapter, &DevConfig);
	vwrq->value = DevConfig.FragmentationThreshold;
	vwrq->disabled = (vwrq->value >= 2346);
	vwrq->fixed = 1;

	return 0;
}
static int vnet_set_frag(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_param *vwrq,
			  char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;

        UpdateAppInfo(Adapter, &DevConfig);
	if(vwrq->disabled)
		DevConfig.FragmentationThreshold = 2346;
	if((DevConfig.FragmentationThreshold < 256) || (DevConfig.FragmentationThreshold > 2346)) {
		return -EINVAL;
	}
	DevConfig.FragmentationThreshold = vwrq->value;
  ChangeConfiguration(Adapter, &DevConfig);
	
	return 0;		/* Call commit handler */
}

static int vnet_set_rate(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_param *vwrq,
			  char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;
	
        UpdateAppInfo(Adapter, &DevConfig);
		/* Setting by frequency value */
			switch (vwrq->value) {
			case  1000000: DevConfig.TxRate = 0; break;
			case  2000000: DevConfig.TxRate = 1; break;
			case  5500000: DevConfig.TxRate = 2; break;
			case 11000000: DevConfig.TxRate = 3; break;
			default: return -EINVAL;
			}
        ChangeConfiguration(Adapter, &DevConfig);

	return 0;
}

static int vnet_set_freq(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_freq *fwrq,
			  char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;
	int rc = 0;		/* Call commit handler */
	
        UpdateAppInfo(Adapter, &DevConfig);
        
	
	/* If setting by frequency, convert to a channel */
	if((fwrq->e == 1) &&
	   (fwrq->m >= (int) 241200000) &&
	   (fwrq->m <= (int) 248700000)) {
		int f = fwrq->m / 100000;
		int c = 0;
		while((c < 14) && (f != frequency_list[c]))
			c++;
		/* Hack to fall through... */
		fwrq->e = 0;
		fwrq->m = c + 1;
	}
	/* Setting by channel number */
	if((fwrq->m > 1000) || (fwrq->e > 0))
		rc = -EOPNOTSUPP;
	else {
		int channel = fwrq->m;
			Adapter->Channel = (UCHAR) channel;
	}
        ChangeConfiguration(Adapter, &DevConfig);

	return rc;
}

static int vnet_get_freq(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_freq *fwrq,
			  char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;

        UpdateAppInfo(Adapter, &DevConfig);
	fwrq->m = frequency_list[(Adapter->Channel) -1] * 100000;
	fwrq->e = 1;
	return 0;
}
static int vnet_get_range(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_point *dwrq,
			   char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
	struct iw_range *range = (struct iw_range *) extra;
	int k,i,j;

	dwrq->length = sizeof(struct iw_range);
	memset(range, 0, sizeof(range));
	range->min_nwid = 0x0000;
	range->max_nwid = 0x0000;
	range->num_channels = 0;
	for (j = 0; j < sizeof(channel_table)/sizeof(channel_table[0]); j++)
		if (Adapter->RegDomain == channel_table[j].reg_domain) {
			range->num_channels = channel_table[j].max - channel_table[j].min + 1;
			break;
		}
	if (range->num_channels != 0) {
		for(k = 0, i = channel_table[j].min; i <= channel_table[j].max; i++) {
			range->freq[k].i = i; /* List index */
			range->freq[k].m = frequency_list[i-1] * 100000;
			range->freq[k++].e = 1;	/* Values in table in MHz -> * 10^5 * 10 */
		}
		range->num_frequency = k;
	}
	
	range->max_qual.qual = 100;
	range->max_qual.level = 100;
	range->max_qual.noise = 0;
	range->sensitivity = 0;

	range->bitrate[0] =  1000000;
	range->bitrate[1] =  2000000;
	range->bitrate[2] =  5500000;
	range->bitrate[3] = 11000000;
	range->num_bitrates = 4;

	range->min_rts = 0;
	range->max_rts = 2347;
	range->min_frag = 256;
	range->max_frag = 2346;

	range->encoding_size[0] = 5;
	range->encoding_size[1] = 13;
	range->num_encoding_sizes = 2;
	range->max_encoding_tokens = 4;
	
	range->pmp_flags = IW_POWER_ON;
	range->pmt_flags = IW_POWER_ON;
	range->pm_capa = 0;
	
	range->we_version_source = WIRELESS_EXT;
	range->we_version_compiled = WIRELESS_EXT;
	range->retry_capa = IW_RETRY_LIMIT ;
	range->retry_flags = IW_RETRY_LIMIT;
	range->r_time_flags = 0;
	range->min_retry = 1;
	range->max_retry = 65535;
	range->avg_qual.qual = 50;
	range->avg_qual.level = 50;
	range->avg_qual.noise = 0;

	return 0;
}
static int vnet_get_name(struct net_device *dev,
			  struct iw_request_info *info,
			  char *cwrq,
			  char *extra)
{
#ifdef RFMD
		strcpy(cwrq, "ATMEL RFMD503A");
#elif defined(R505)                
		strcpy(cwrq, "ATMEL RFMD505");
#elif defined(R505A)               
		strcpy(cwrq, "ATMEL RFMD505A");
#elif defined(R505AMX)                
		strcpy(cwrq, "ATMEL MX 505A");
#elif defined(R505_2958)                
		strcpy(cwrq, "ATMEL 2958505");
#endif
	return 0;
}
#if 0
static int vnet_config_commit(struct net_device *dev,
			       struct iw_request_info *info,	/* NULL */
			       void *zwrq,			/* NULL */
			       char *extra)			/* NULL */
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
//        Don't need reset anymore !!!
        printk(KERN_INFO "vnet_config_commit () !!!\n");        
//        ResetAdapter(Adapter, TRUE);
	return 0;
}
#endif

static int vnet_get_rate(struct net_device *dev,
			 struct iw_request_info *info,
			 struct iw_param *vwrq,
			 char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;

                UpdateAppInfo(Adapter, &DevConfig);

		vwrq->fixed = 1;
		switch(DevConfig.TxRate) {
		case 0: vwrq->value =  1000000; 
                        break;
		case 1: vwrq->value =  2000000; 
                        break;
		case 2: vwrq->value =  5500000; 
                        break;
		case 3: vwrq->value = 11000000; 
                        break;
                }
	return 0;
}
static int vnet_set_scan(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_param *vwrq,
			  char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        unsigned long flags __attribute__ ((unused));

        /* Timeout old surveys. */
        if (Adapter->flags & SITE_SURVEY_REQUEST)
          return -EBUSY;
              
        Adapter->lock = SPIN_LOCK_UNLOCKED;
    
spin_lock_irqsave(&Adapter->lock, flags);	

         Adapter->TempState = Adapter->StationState;
         ChangeState(Adapter, STATION_STATE_SCANNING);
         Adapter->flags |= SITE_SURVEY_REQUEST;
         DropPendingTxPackets(Adapter);
         UsbScan(Adapter);

spin_unlock_irqrestore(&Adapter->lock, flags);	

	return 0;
}

static int vnet_set_essid(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_point *dwrq,
			   char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;
        unsigned long flags __attribute__ ((unused));

        UpdateAppInfo(Adapter, &DevConfig);


        memcpy (DevConfig.SSID, extra, dwrq->length);
        DevConfig.SSID[dwrq->length] ='\0';
        DevConfig.SSIDlength = strlen(DevConfig.SSID);
        ChangeConfiguration (Adapter, &DevConfig);
	return 0;
}

static int vnet_get_essid(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_point *dwrq,
			   char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;
        unsigned long flags __attribute__ ((unused));


        UpdateAppInfo(Adapter, &DevConfig);

		dwrq->flags = 1;
		//if (Adapter->SSID_size == 0) {
		if (Adapter->AssociationInfo.DesiredSSIDsize == 0) {
			/* we don't camp on a beacon yet -> report the desired SSID */
			dwrq->length = Adapter->SSID_size;
	  memcpy (extra, Adapter->SelectedSSID, dwrq->length);
		} else {
			/* we have found a BSS -> report its SSID (in case we had "any") */
			dwrq->length = Adapter->AssociationInfo.DesiredSSIDsize;
	    memcpy (extra, Adapter->AssociationInfo.DesiredSSID, dwrq->length);
		}

	return 0;
}

static int vnet_set_mode(struct net_device *dev,
			  struct iw_request_info *info,
			  ULONG *uwrq,
			  char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;


	if (*uwrq != IW_MODE_ADHOC && *uwrq != IW_MODE_INFRA)
		return -EINVAL;

                UpdateAppInfo(Adapter, &DevConfig);
      switch (*uwrq)
	{
	case IW_MODE_ADHOC:
	  DevConfig.OperatingMode = 0;
	  break;
	case IW_MODE_INFRA:
	  DevConfig.OperatingMode = 1;
	  break;
	default:
	  return -EINVAL;
	}
      ChangeConfiguration (Adapter, &DevConfig);
	//return -EINPROGRESS;  
	return 0;  
}
static int vnet_get_mode(struct net_device *dev,
			  struct iw_request_info *info,
			  ULONG *uwrq,
			  char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
	
      switch (Adapter->OperatingMode)
	{
	case AD_HOC_MODE:
	  *uwrq = IW_MODE_ADHOC;
	  break;
	case INFRASTRUCTURE_MODE:
	  *uwrq = IW_MODE_INFRA;
	  break;
	  // For compatibility with wireless tools
	  // AdHoc in wt is 1 and for infrastructure return 0 (Auto in wt)
	}
	return 0;
}
#define MAX_WPA_IE_LEN 64

void VNetGetSystemTime(PVOID tCurrentTime) 
{
  u64* pTime = (u64*) tCurrentTime; 
  struct timeval timev;
  do_gettimeofday(&timev);
  (u64) (*pTime) = timev.tv_sec * 10000000;
};

int atmel_assoc_info_event (PVNet_ADAPTER Adapter)
{
	union iwreq_data wrqu;
    char buf[255];
    int j;
    UCHAR *p = buf;

    printk(KERN_INFO "Sending AssocInfo Event\n");

    p += sprintf(p, "ASSOCINFO(ReqIEs=");
    for (j = 0; j < Adapter->AssociationInfo.AssReassIEInfoBufferLength; j++) {
        
            p += sprintf(p, "%02x", Adapter->AssociationInfo.AssReassIEInfoBuffer[j]);
    }
    printk("AssocInfoBuffLen: %d\n", Adapter->AssociationInfo.AssReassIEInfoBufferLength);

    memcpy(wrqu.ap_addr.sa_data, Adapter->AssociationInfo.CurrentBSSID, 6);
    wrqu.ap_addr.sa_family = 1;
    wireless_send_event(Adapter->net, SIOCGIWAP, &wrqu, NULL);
    memset(&wrqu, 0, sizeof(wrqu));
    wrqu.data.length = strlen(buf);
    wireless_send_event(Adapter->net, IWEVCUSTOM, &wrqu, buf);

    return 0;
}

int atmel_scan_info_event (PVNet_ADAPTER Adapter)
{
	union iwreq_data wrqu;
    //char buf[255];
    //int j;
    //UCHAR *p = buf;


    wrqu.data.length = 0;
    wrqu.data.flags = 0;
    wireless_send_event(Adapter->net, SIOCGIWSCAN, &wrqu, NULL);
    return 0;
}

static int vnet_get_scan(struct net_device *dev,
			  struct iw_request_info *info,
			  struct iw_point *dwrq,
			  char *extra)
{
	int i, j;
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
	char *current_ev = extra;
	char buf[MAX_WPA_IE_LEN * 2 + 30];
	struct iw_event	iwe;
        unsigned long flags __attribute__ ((unused));
	

        
	for(i=0; i<Adapter->AssociationInfo.SiteS.BSSInList; i++) { 
		iwe.cmd = SIOCGIWAP;
		iwe.u.ap_addr.sa_family = 1;
		memcpy(iwe.u.ap_addr.sa_data, Adapter->AssociationInfo.SiteS.BssInfo[i].BSSID, 6);
		current_ev = iwe_stream_add_event(current_ev, extra + IW_SCAN_MAX_DATA, &iwe, IW_EV_ADDR_LEN);

		iwe.u.data.length =  Adapter->AssociationInfo.SiteS.BssInfo[i].SSIDsize;
		if (iwe.u.data.length > 32)
			iwe.u.data.length = 32;
		iwe.cmd = SIOCGIWESSID;
		iwe.u.data.flags = 1;
		current_ev = iwe_stream_add_point(current_ev, extra + IW_SCAN_MAX_DATA, &iwe, Adapter->AssociationInfo.SiteS.BssInfo[i].SSID);
		
		iwe.cmd = SIOCGIWMODE;
		iwe.u.mode = Adapter->AssociationInfo.SiteS.BssInfo[i].BSStype;
		current_ev = iwe_stream_add_event(current_ev, extra + IW_SCAN_MAX_DATA, &iwe, IW_EV_UINT_LEN);
	
		iwe.cmd = SIOCGIWFREQ;
		iwe.u.freq.m = Adapter->AssociationInfo.SiteS.BssInfo[i].Channel;
		iwe.u.freq.e = 0;
		current_ev = iwe_stream_add_event(current_ev, extra + IW_SCAN_MAX_DATA, &iwe, IW_EV_FREQ_LEN);
		
		iwe.cmd = SIOCGIWENCODE;
		if (Adapter->AssociationInfo.SiteS.BssInfo[i].EncryptionIsOn)
			iwe.u.data.flags = IW_ENCODE_ENABLED | IW_ENCODE_NOKEY;
		else
			iwe.u.data.flags = IW_ENCODE_DISABLED;
		iwe.u.data.length = 0;
		current_ev = iwe_stream_add_point(current_ev, extra + IW_SCAN_MAX_DATA, &iwe, NULL);

                if(Adapter->AssociationInfo.SiteS.BssInfo[i].WpaIELen > 0) 
                {
                    UCHAR *p = buf;
                    p += sprintf(p, "wpa_ie=");
                    for (j = 0; j < Adapter->AssociationInfo.SiteS.BssInfo[i].WpaIELen; j++) {
                            p += sprintf(p, "%02x", Adapter->AssociationInfo.SiteS.BssInfo[i].WpaIE[j]);
                    }

                    memset(&iwe, 0, sizeof(iwe));
                    iwe.cmd = IWEVCUSTOM;
                    iwe.u.data.length = strlen(buf);
                    current_ev = iwe_stream_add_point( current_ev, extra + IW_SCAN_MAX_DATA, &iwe, buf);

                
                }
	}

	/* Length of data */
	dwrq->length = (current_ev - extra);
	dwrq->flags = 0;	/* todo */

	
	return 0;
}


/*------------------------------------------------------------------*/
/*
 * Wireless Private Handler : get preamble
 */
static int vnet_ioctl_get_preamble(struct net_device *dev, struct iw_request_info *info,
                            union iwreq_data *wrq, char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;
        unsigned long flags __attribute__ ((unused));

        Adapter->lock = SPIN_LOCK_UNLOCKED;
	
        spin_lock_irqsave(&Adapter->lock, flags);
        UpdateAppInfo(Adapter, &DevConfig);
          if (wrq->data.pointer) {
              wrq->data.flags = 1;
              if (DevConfig.PreambleType)
                {
                  spin_unlock_irqrestore(&Adapter->lock, flags);
                  wrq->data.length = 6;
                  if (copy_to_user (wrq->data.pointer, "short", 6)) {
                  spin_unlock_irqrestore(&Adapter->lock, flags);
                  return -EFAULT;
                  }
                  spin_lock_irqsave(&Adapter->lock, flags);
                } else {
                  spin_unlock_irqrestore(&Adapter->lock, flags);
                  wrq->data.length = 5;
                  if (copy_to_user (wrq->data.pointer, "long", wrq->data.length)) {
                      spin_unlock_irqrestore(&Adapter->lock, flags);
                      return -EFAULT;
                  }
                  spin_lock_irqsave(&Adapter->lock, flags);
                }
          }
         spin_unlock_irqrestore(&Adapter->lock, flags);

  //return -EINPROGRESS;
  return 0;
}

static int vnet_set_txpower(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_param *vwrq,
			   char *extra)
{
   PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
   DEVICE_CONFIGURATION DevConfig;
   unsigned long flags __attribute__ ((unused));
  
  UpdateAppInfo(Adapter, &DevConfig);

  vwrq->flags = IW_TXPOW_DBM;
  DevConfig.TxPowerLevel = (10 * vwrq->value);

  if (vwrq->flags != IW_TXPOW_DBM) {
        printk("SIOCSIWTXPOW with mW is not supported; use dBm\n");
        return -EOPNOTSUPP;
  }

  ChangeConfiguration(Adapter, &DevConfig);
	return 0;
}

static int vnet_get_txpower(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_param *vwrq,
			   char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;
        unsigned long flags __attribute__ ((unused));
  
  UpdateAppInfo(Adapter, &DevConfig);
  
  if(Adapter->RadioIsOn) {
    vwrq->disabled = 0;
    vwrq->fixed = 1;
    vwrq->value = (Adapter->CurrentTXpowerIndbm)/(UCHAR)10;
  }
	return 0;
}


static int vnet_set_power(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_param *vwrq,
			   char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;
        unsigned long flags __attribute__ ((unused));

      UpdateAppInfo(Adapter, &DevConfig);


	DevConfig.PowerMgmtMode = vwrq->disabled ? 0 : 1;
  
  ChangeConfiguration (Adapter, &DevConfig);

	return 0;
}

static int vnet_get_power(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_param *vwrq,
			   char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;
        unsigned long flags __attribute__ ((unused));

      UpdateAppInfo(Adapter, &DevConfig);
	    vwrq->disabled = DevConfig.PowerMgmtMode ? 0 : 1;
	    vwrq->flags = IW_POWER_ON;

	return 0;
}

/*------------------------------------------------------------------*/
/*
 * Wireless Private Handler : set preamble
 */
static int vnet_ioctl_set_preamble(struct net_device *dev,
			    struct iw_request_info *info,
                            union iwreq_data *wrq,
			    char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        DEVICE_CONFIGURATION DevConfig;
        unsigned long flags __attribute__ ((unused));
	
        Adapter->lock = SPIN_LOCK_UNLOCKED;
        spin_lock_irqsave(&Adapter->lock, flags);
        UpdateAppInfo(Adapter, &DevConfig);
        DevConfig.PreambleType = *((int *) wrq->name) & 1;
	/* Enable interrupts and restore flags. */
        ChangeConfiguration (Adapter, &DevConfig);
        spin_unlock_irqrestore(&Adapter->lock, flags);
	return 0;
}



static int atmel_ioctl_get_wpa_param(struct net_device *dev,
					  struct iw_request_info *info, union iwreq_data *wrq, char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
        int ret = 0;
        unsigned long flags __attribute__ ((unused));
        int length = 0;
        char tmpbuff[9];

        Adapter->lock = SPIN_LOCK_UNLOCKED;
	spin_lock_irqsave(&Adapter->lock, flags);

	switch (wrq->data.flags) {
	case ATMEL_PARAM_WPA:
		break;
	case ATMEL_PARAM_PRIVACY_INVOKED:
		break;
	case ATMEL_PARAM_WPA_TYPE:
              switch(Adapter->SecurityInfo.EncryptionInfo.AuthenticationMode) {
              case 0:
                length = 5;
                memcpy(tmpbuff, "Open", 4);
                tmpbuff[5] = '\0';
                break;
              case 1:
                length = 7;
                memcpy(tmpbuff, "Shared", 6);
                tmpbuff[7] = '\0';
                break;
              case 3:
                length = 4;
                memcpy(tmpbuff, "WPA", 3);
                tmpbuff[4] = '\0';
                break;
              case 4:
                length = 8;
                memcpy(tmpbuff, "WPA PSK", 7);
                tmpbuff[8] = '\0';
                break;
              default :
                break;
              }
              if(wrq->data.pointer) {
               spin_unlock_irqrestore(&Adapter->lock, flags);
                  wrq->data.length = length;
                  if(copy_to_user(wrq->data.pointer, tmpbuff, wrq->data.length)) {
                    return -EFAULT; 
                  }
              spin_lock_irqsave(&Adapter->lock, flags);
              }
		break;
	default:
		ret = -EOPNOTSUPP;
		break;
	}
	spin_unlock_irqrestore(&Adapter->lock, flags);

  //return -EINPROGRESS;
  return 0;
}

static int atmel_ioctl_set_wpa_param(struct net_device *dev,
					  struct iw_request_info *info,
            union iwreq_data *wrq,
					  char *extra)
{
        PVNet_ADAPTER Adapter =  (PVNet_ADAPTER) dev->priv;
	int *i = (int *) extra;
	int param = *i;
	int value = *(i + 1);
	int ret = 0;
  unsigned long flags __attribute__ ((unused));
	//USHORT val;

    
        Adapter->lock = SPIN_LOCK_UNLOCKED;
	spin_lock_irqsave(&Adapter->lock, flags);
	switch (param) {
	case ATMEL_PARAM_WPA:
                Adapter->SecurityInfo.InWpaMode = 1;
		break;
	case ATMEL_PARAM_PRIVACY_INVOKED:
                Adapter->PrivacyInvoked = 1;
		break;
	case ATMEL_PARAM_WPA_TYPE:
                    Adapter->SecurityInfo.EncryptionInfo.AuthenticationMode = value;
                    if(value == 3){
                        printk("Entering WPA Mode - Waiting for the Completion of Key Exchange ...\n");
                    }else  if(value == 4){
                        printk("Entering WPA PSK Mode - Waiting for the Completion of Key Exchange ...\n");
                    } 
		break;
	default:
		printk(KERN_INFO "%s: atmel_param: unknown param %d\n",
		       dev->name, param);
		ret = -EOPNOTSUPP;
		break;
	}

	spin_unlock_irqrestore(&Adapter->lock, flags);
	return ret;
}

static const iw_handler		vnet_handler[] =
{
	(iw_handler) NULL,	                /* SIOCSIWCOMMIT */
	(iw_handler) vnet_get_name,    	/* SIOCGIWNAME */
	(iw_handler) NULL,			/* SIOCSIWNWID */
	(iw_handler) NULL,			/* SIOCGIWNWID */
	(iw_handler) vnet_set_freq,	        /* SIOCSIWFREQ */
	(iw_handler) vnet_get_freq,		/* SIOCGIWFREQ */
	(iw_handler) vnet_set_mode,		/* SIOCSIWMODE */
	(iw_handler) vnet_get_mode,		/* SIOCGIWMODE */
	(iw_handler) NULL,                  	/* SIOCSIWSENS */
	(iw_handler) NULL,           		/* SIOCGIWSENS */
	(iw_handler) NULL,			/* SIOCSIWRANGE */
	(iw_handler) vnet_get_range,           /* SIOCGIWRANGE */
	(iw_handler) NULL,			/* SIOCSIWPRIV */
	(iw_handler) NULL,			/* SIOCGIWPRIV */
	(iw_handler) NULL,			/* SIOCSIWSTATS */
	(iw_handler) NULL,			/* SIOCGIWSTATS */
	(iw_handler) NULL,			/* SIOCSIWSPY */
	(iw_handler) NULL,			/* SIOCGIWSPY */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) vnet_set_wap,		/* SIOCSIWAP */
	(iw_handler) vnet_get_wap,		/* SIOCGIWAP */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) NULL,       		/* SIOCGIWAPLIST */
	(iw_handler) vnet_set_scan,		/* SIOCSIWSCAN */
	(iw_handler) vnet_get_scan,		/* SIOCGIWSCAN */
	(iw_handler) vnet_set_essid,		/* SIOCSIWESSID */
	(iw_handler) vnet_get_essid,		/* SIOCGIWESSID */
	(iw_handler) NULL,         		/* SIOCSIWNICKN */
	(iw_handler) NULL,          		/* SIOCGIWNICKN */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) NULL,			/* -- hole -- */
	(iw_handler) vnet_set_rate,		/* SIOCSIWRATE */
	(iw_handler) vnet_get_rate,		/* SIOCGIWRATE */
	(iw_handler) vnet_set_rts,		/* SIOCSIWRTS */
	(iw_handler) vnet_get_rts,		/* SIOCGIWRTS */
	(iw_handler) vnet_set_frag,		/* SIOCSIWFRAG */
	(iw_handler) vnet_get_frag,		/* SIOCGIWFRAG */
	(iw_handler) vnet_set_txpower,         		/* SIOCSIWTXPOW */
	(iw_handler) vnet_get_txpower,          		/* SIOCGIWTXPOW */
	(iw_handler) NULL,		/* SIOCSIWRETRY */
	(iw_handler) NULL,		/* SIOCGIWRETRY */
	(iw_handler) vnet_set_encode,		/* SIOCSIWENCODE */
	(iw_handler) vnet_get_encode,		/* SIOCGIWENCODE */
	(iw_handler) vnet_set_power,		/* SIOCSIWPOWER */
	(iw_handler) vnet_get_power,		/* SIOCGIWPOWER */
};

static const iw_handler		vnet_private_handler[] =
{
	(iw_handler) vnet_ioctl_set_preamble,	/* +0 */
	(iw_handler) vnet_ioctl_get_preamble,	/* +1 */
  NULL,
	(iw_handler) atmel_ioctl_set_wpa_param,	/* +3 */
	(iw_handler) atmel_ioctl_get_wpa_param,	/* +4 */
};

static const struct iw_priv_args vnet_private_args[] = {
/*{ cmd,         set_args,                            get_args, name } */
  { SIOCIWPRIVSPREAM, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "set_preamble" },
  { SIOCIWPRIVGPREAM, 0, IW_PRIV_TYPE_CHAR | 6 | 1, "get_preamble" },
  { ATMEL_WPA_IOCTL_PARAM, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 2, 0, "atmlparam" },
  { ATMEL_WPA_IOCTL_GET_PARAM, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, "get_atmlparam" },
  { ATMEL_WPA_IOCTL_PARAM, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "" },
  { ATMEL_WPA_IOCTL_GET_PARAM, 0, IW_PRIV_TYPE_CHAR | 8 | 1, "" },
  { ATMEL_PARAM_WPA, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "wpa" },
  { ATMEL_PARAM_WPA, 0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, "getwpa" },
  { ATMEL_PARAM_PRIVACY_INVOKED, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "privacy_invoked" },
  { ATMEL_PARAM_PRIVACY_INVOKED, 0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, "getprivacy_invo" },
  { ATMEL_PARAM_WPA_TYPE, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "wpa_mode" },
  { ATMEL_PARAM_WPA_TYPE, 0, IW_PRIV_TYPE_CHAR | 8 | 1, "getwpa_mode" },
};

static const struct iw_handler_def	vnet_handler_def =
{
	.num_standard	= sizeof(vnet_handler)/sizeof(iw_handler),
	.num_private	= sizeof(vnet_private_handler)/sizeof(iw_handler), 
	.num_private_args = sizeof(vnet_private_args)/sizeof(struct iw_priv_args), 
	.standard	= (iw_handler *) vnet_handler,
	.private	= (iw_handler *) vnet_private_handler, 
	.private_args	= (struct iw_priv_args *) vnet_private_args
};

static int atmel_wpa_ioctl(PVNet_ADAPTER Adapter, struct iw_point *p)
{
	struct atmel_param *param;
        UCHAR mic_offset[8];
        UCHAR BroadcastBSSID[] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
        UCHAR EncryptionSuite;
        unsigned long flags __attribute__ ((unused));
        int i __attribute__ ((unused));
        UCHAR TmpKeyRsc[6];
        //int i;
	int ret = 0;
	int ap_ioctl = 0;

	if (p->length < sizeof(struct atmel_param) ||
	    p->length > 1024 || !p->pointer)
		return -EINVAL;

	param = (struct atmel_param *) kmalloc(p->length, GFP_KERNEL);
	if (param == NULL)
		return -ENOMEM;



	if (copy_from_user(param, p->pointer, p->length)) {
		ret = -EFAULT;
		goto out;
	}

	switch (param->cmd) {
	case SET_WPA_ENCRYPTION:

		if(!param->key_len){ 
        break;
    }

		if(!param->key_idx)
		{
			printk("Pairwise Key\n");
			Adapter->IsGroupKey = FALSE;
			Adapter->IsTxKey = TRUE;
		} else if((param->key_idx) && (param->set_tx & 0x80)) {
			printk("Pairwise Key for 1X (No-WPA)\n");
			Adapter->IsGroupKey = FALSE;
			Adapter->IsTxKey = TRUE;
		} else if((param->key_idx) && (param->set_tx)) {
			printk("Group Key / Tx / Rx\n");
			Adapter->IsGroupKey = TRUE;
			Adapter->IsTxKey = TRUE;
		} else if((param->key_idx) && (!param->set_tx)) {
			printk("Group Key Rx only\n");
			Adapter->IsGroupKey = TRUE;
			Adapter->IsTxKey = FALSE;
		} else {
      printk(KERN_INFO "Invalid Combination\n");
      break;
    }
    
		if((Adapter->OperatingMode == AD_HOC_MODE) && Adapter->IsGroupKey && memcmp(param->sta_addr, BroadcastBSSID, 6))
		{
			printk("*** ERROR: Invalid BSSID / Group Key combination in AdHoc Mode\n");
			break;
		}

		if(param->key_len == 5)
		{
			printk("Cipher Suite is WEP-64...\n");
			EncryptionSuite = CIPHER_SUITE_WEP_64;
		}
		else
		if(param->key_len == 13)
		{
			//IF_LOUD(DbgPrint("Cipher Suite is WEP-128...\n");)
			EncryptionSuite = CIPHER_SUITE_WEP_128;
		}
		else
#ifdef AES
		if(param->key_len == 16)
		{
			//IF_LOUD(DbgPrint("Cipher Suite is AES...\n");)
			EncryptionSuite = CIPHER_SUITE_AES;
		}
		else
#endif
		if(param->key_len == 32)
		{
			printk(KERN_INFO "Cipher Suite is TKIP...\n");
			EncryptionSuite = CIPHER_SUITE_TKIP;
		}
		else
		{
			//IF_LOUD(DbgPrint("Cipher Suite is NONE...\n");)
			EncryptionSuite = CIPHER_SUITE_NONE;
		}
                printk(KERN_INFO "#######################################################\n");
                printk(KERN_INFO "ENCRYPTION -SETTINGS : key_idx=%d set_tx=%d seq_len=%d "
                                              "key_len=%d\n", param->key_idx, param->set_tx, param->seq_len, param->key_len);
                printk(KERN_INFO "#######################################################\n");

		
                memcpy(TmpKeyRsc, param->seq, param->seq_len);
		  Adapter->SecurityInfo.EncryptionInfo.KeyRSC[param->key_idx][4] = TmpKeyRsc[0];
		  Adapter->SecurityInfo.EncryptionInfo.KeyRSC[param->key_idx][5] = TmpKeyRsc[1];
      for (i = 0; i < 4; i++) {
        Adapter->SecurityInfo.EncryptionInfo.KeyRSC[param->key_idx][i] = TmpKeyRsc[2+i];
      }

			if(!memcmp(param->sta_addr, BroadcastBSSID, 6) && (Adapter->OperatingMode == INFRASTRUCTURE_MODE )){
		     memcpy(Adapter->SecurityInfo.EncryptionInfo.BSSID, Adapter->AssociationInfo.CurrentBSSID, 6);
      }
      else {
		                memcpy(Adapter->SecurityInfo.EncryptionInfo.BSSID, param->sta_addr, 6);
      }

		memcpy(&mic_offset[0], &param->key[24], 8);
		memcpy(&Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[(param->key_idx)][0], param->key, param->key_len);
		memcpy(&Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[(param->key_idx)][24], &param->key[16], 8);
		memcpy(&Adapter->SecurityInfo.EncryptionInfo.KeyMaterial[(param->key_idx)][16], &mic_offset[0], 8);
		
		if(Adapter->IsGroupKey)
		{
      Adapter->IsPairwiseKey = FALSE;
			if(!((param->key_len == 5 || param->key_len == 13) && !Adapter->IsTxKey))
				Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex = param->key_idx;
        Adapter->TmpKeyIndex = Adapter->SecurityInfo.EncryptionInfo.GroupKeyIndex;

			Adapter->SecurityInfo.EncryptionInfo.GroupKeyLength = param->key_len;
			Adapter->SecurityInfo.EncryptionInfo.GroupCipherSuite = EncryptionSuite;

			if(!memcmp(param->sta_addr, BroadcastBSSID, 6) || 
				(Adapter->OperatingMode == INFRASTRUCTURE_MODE && !memcmp(param->sta_addr, Adapter->AssociationInfo.CurrentBSSID, 6)))
			{ //configure immediately
				printk(">>>> Configure Immediately (Group key with Broadcast BSSID or our BSSID in Infrastructure mode)\n");

				//Adapter->SecurityInfo.PairwiseKeyAvailable = TRUE;
			//	if(IsTxKey)
				{
					Adapter->SecurityInfo.GroupKeyAvailable = TRUE;
//					Adapter->LastGroupKeyID = (UCHAR)(pNdisKey->KeyIndex & 0xff);
				}
				if(SetEncryptionStatus(Adapter, TRUE, FALSE, param->key_idx) != 0)
				{
					printk("Encryption configured with SUCCESS...(NOT)\n");
          return -EFAULT;
				}
          if (Adapter->flags & GetCommandPedding) {
			DECLARE_WAITQUEUE(wait, current);
			add_wait_queue(Adapter->ctrl_wait, &wait);
			set_current_state(TASK_INTERRUPTIBLE);
			Adapter->flags |= IOCTL_SLEEPING;
			schedule();
			set_current_state(TASK_RUNNING);
			remove_wait_queue(Adapter->ctrl_wait, &wait);
            if (Adapter->flags & WAKE_UP_WITH_ERROR) {
              Adapter->flags &= ~IOCTL_SLEEPING;
              return -1;	/* jal: real error code ??? */
            }
          }
			}
		}
		else //Pairwise Key
		{
			Adapter->SecurityInfo.PairwiseKeyAvailable = TRUE;
      Adapter->IsPairwiseKey = TRUE;
			if(!((param->key_len == 5 || param->key_len == 13) && !Adapter->IsTxKey))
				Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex = param->key_idx;
        Adapter->TmpKeyIndex = Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex;
        printk("(PAIRWISE) TMPKEYINDEX : %d\n",Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyIndex);

			Adapter->SecurityInfo.EncryptionInfo.PairwiseKeyLength = param->key_len;
			Adapter->SecurityInfo.EncryptionInfo.PairwiseCipherSuite = EncryptionSuite;

                        printk(KERN_INFO "#### PAIRWISE index = %d###\n", param->key_idx);
	//		if(IsTxKey)
			{
		//		Adapter->SecurityInfo.PairwiseKeyAvailable = TRUE;
//				Adapter->LastPairwiseKeyID = (UCHAR)(pNdisKey->KeyIndex & 0xff);
			}
			if(SetEncryptionStatus(Adapter, TRUE, FALSE, param->key_idx))
			{
					printk("Encryption configured with SUCCESS...(NOT)\n");
          return -EFAULT;
			}
          if (Adapter->flags & GetCommandPedding) {
			DECLARE_WAITQUEUE(wait, current);
			add_wait_queue(Adapter->ctrl_wait, &wait);
			set_current_state(TASK_INTERRUPTIBLE);
			Adapter->flags |= IOCTL_SLEEPING;
			schedule();
			set_current_state(TASK_RUNNING);
			remove_wait_queue(Adapter->ctrl_wait, &wait);
            if (Adapter->flags & WAKE_UP_WITH_ERROR) {
              Adapter->flags &= ~IOCTL_SLEEPING;
              return -1;	/* jal: real error code ??? */
            }
          }
		}
		if(EncryptionSuite == CIPHER_SUITE_WEP_64 || EncryptionSuite == CIPHER_SUITE_WEP_128)
		{
			Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption1Enabled;
		}
		else
		if(EncryptionSuite == CIPHER_SUITE_TKIP)
		{
			if(Adapter->OperatingMode == INFRASTRUCTURE_MODE)
			{
				if(Adapter->IsTxKey)
				{
					if(Adapter->SecurityInfo.PairwiseKeyAvailable)
						Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption2Enabled;
					// Maybe this case is wrong, since if we get a group key, without already
					// having a pairwise, with the Tx flag on, we will disable the encryption.
					else
						Adapter->SecurityInfo.EncryptionStatus = _802_11WEPDisabled;
				}
			}
			else
			{
				if(Adapter->IsTxKey)
					Adapter->SecurityInfo.EncryptionStatus = _802_11Encryption2Enabled;
			}
		}

		break;
	case SET_CIPHER_SUITES:
                printk(KERN_INFO "SET_CIPHER_SUITES\n");
		break;
	case MLME_STA_DEAUTH:
            printk(KERN_INFO "MLME_STA_DEAUTH\n");
		break;
	case MLME_STA_DISASSOC:
            printk(KERN_INFO "MLME_STA_DISASSOC\n");
            RemoveEncryptionSettings(Adapter);
		break;
	default:
            printk(KERN_INFO "DEFAULT IOCTL COMMAND for Supplicant\n");
		ap_ioctl = 1;
		break;
	}

	if (ret == 1 || !ap_ioctl) {
		if (copy_to_user(p->pointer, param, p->length)) {
			ret = -EFAULT;
			goto out;
		} else if (ap_ioctl)
			ret = 0;
	}

 out:
	if (param != NULL) {
		kfree(param);
        }

	return ret;
}




////////////////////////////////////////////////////////////////////////////////////////////////////
static int usb_vnet_ioctl(struct net_device *dev, struct ifreq *ifr,
			  int cmd)
{
	PVNet_ADAPTER Adapter = (PVNet_ADAPTER) dev->priv;
	struct iwreq *wrq = (struct iwreq *) ifr;
	UCHAR old_EncL = Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel;
	DEVICE_CONFIGURATION DevConfig;
	VERSION_INFO VersionInfo;
  unsigned long flags __attribute__ ((unused));
  int wpa_result;
	int AP_Selected, i=0;

	dbgcond(DBG_IOCTL, "cmd x%x StationState %d\n",
	cmd, Adapter->StationState);
      if (!capable (CAP_NET_ADMIN))
        {
          return -EPERM;
        }

	if ((Adapter->StationState == STATION_STATE_INITIALIZING) ||
	    (Adapter->StationState == STATION_STATE_FW_DOWNLOAD)) {
		printk("Device not ready yet\n");
		return -EIO;
	}

	if (Adapter->flags & IOCTL_SLEEPING) {
		printk(KERN_INFO "ioctl already pending\n");
		return -EINPROGRESS;
	}
	UpdateAppInfo(Adapter, &DevConfig);

        Adapter->lock = SPIN_LOCK_UNLOCKED;
	spin_lock_irqsave(&Adapter->lock, flags);

	switch (cmd) {

	case SIOCDEVPRIVATE:
		switch (wrq->u.data.flags) {
		case GET_SUPPORT_INFO:
			// Give our configuration to the user
			wrq->u.data.length = sizeof(DEVICE_CONFIGURATION);
#if (LINUX_VERSION_CODE > 0x020500)
			spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
			if(copy_to_user(wrq->u.data.pointer, &DevConfig, wrq->u.data.length))
                           return -EFAULT;
#if (LINUX_VERSION_CODE > 0x020500)
	                spin_lock_irqsave(&Adapter->lock, flags);
#endif
			break;
		case SET_SUPPORT_INFO:
			// Set config we got from the user
#if (LINUX_VERSION_CODE > 0x020500)
			spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
			if(copy_from_user(&DevConfig, wrq->u.data.pointer, wrq->u.data.length))
                        return -EFAULT;

#if (LINUX_VERSION_CODE > 0x020500)
			spin_lock_irqsave(&Adapter->lock, flags);
#endif
			if (!ValidateChannel(Adapter, DevConfig.Channel)) {
				dbgcond(DBG_IOCTL,
					"Trying to set invalid channel %d",
					DevConfig.Channel);
				return -EINVAL;
			}
			ChangeConfiguration(Adapter, &DevConfig);
			break;
		case GET_ENCRYPTION_INFO:
			{
				ENCRYPTION_INFO TmpEncryption;
				wrq->u.data.length = sizeof(ENCRYPTION_INFO);
				memcpy(&TmpEncryption, &Adapter->SecurityInfo.EncryptionInfo,
				       wrq->u.data.length);
				if (Adapter->PrivacyInvoked == 0)
					TmpEncryption.EncryptionLevel = 0;
#if (LINUX_VERSION_CODE > 0x020500)
			spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
				if(copy_to_user(wrq->u.data.pointer, &TmpEncryption, wrq->u.data.length))
                                    return -EFAULT;

#if (LINUX_VERSION_CODE > 0x020500)
			spin_lock_irqsave(&Adapter->lock, flags);
#endif
			}
			break;
		case SET_ENCRYPTION_INFO:
#if (LINUX_VERSION_CODE > 0x020500)
			spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
			if(copy_from_user(&Adapter->SecurityInfo.EncryptionInfo, wrq->u.data.pointer, wrq->u.data.length))
                            return -EFAULT;
#if (LINUX_VERSION_CODE > 0x020500)
			spin_lock_irqsave(&Adapter->lock, flags);
#endif
			if (Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel == 0) {
				Adapter->PrivacyInvoked = 0;
				Adapter->SecurityInfo.EncryptionInfo.EncryptionLevel =
				    old_EncL;
			} else
				Adapter->PrivacyInvoked = 1;

#if 0
			dbgcond(DBG_IOCTL, "AuthType %d WepKey1[0]=%02x\n",
				Adapter->SecurityInfo.EncryptionInfo.AuthenticationMode,
				Adapter->SecurityInfo.EncryptionInfo.WepKey1[0]);
#endif

			Adapter->flags &= ~WepIsSet;

			if (Adapter->StationState ==
			    STATION_STATE_SCANNING)
				break;

			spin_unlock_irqrestore(&Adapter->lock, flags);
			if (Adapter->flags & GetCommandPedding) {
			DECLARE_WAITQUEUE(wait, current);
			add_wait_queue(Adapter->ctrl_wait, &wait);
			set_current_state(TASK_INTERRUPTIBLE);
			Adapter->flags |= IOCTL_SLEEPING;
			schedule();
			set_current_state(TASK_RUNNING);
			remove_wait_queue(Adapter->ctrl_wait, &wait);
				if (Adapter->flags & WAKE_UP_WITH_ERROR) {
					Adapter->flags &= ~IOCTL_SLEEPING;
					return -1;	/* jal: real error code ??? */
				}
			}
			dbgcond(DBG_IOCTL, "Setting Wep (%lu)\n",
				Adapter->Stats.RxDataPacketsOk);
			DropPendingTxPackets(Adapter);
			dbgcond(DBG_IOCTL, "Tx Packets Dropped\n");
			SetWEPvalue(Adapter);
			dbgcond(DBG_IOCTL, "out\n");
			return 0;
		case GET_VERSION_INFO:
			VersionInfo.DriverMajorVersion =
			    USB_VNET_MAJOR_VERSION;
			VersionInfo.DriverMinorVersion =
			    USB_VNET_MINOR_VERSION;
			VersionInfo.DriverSubVersion =
			    USB_VNET_SUB_VERSION;
			VersionInfo.DriverBuild = USB_VNET_BUILD;
			VersionInfo.FwMajorVersion =
			    (USHORT) Adapter->FwVersion[0];
			VersionInfo.FwMinorVersion =
			    (USHORT) Adapter->FwVersion[1];
			VersionInfo.FwSubVersion = (USHORT) Adapter->FwVersion[2];
			VersionInfo.FwBuild =
			    (USHORT) Adapter->FwVersion[3];
			wrq->u.data.length = sizeof(VERSION_INFO);
#if (LINUX_VERSION_CODE > 0x020500)
			spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
			if (copy_to_user(wrq->u.data.pointer, &VersionInfo, wrq->u.data.length))
                            return -EFAULT;
#if (LINUX_VERSION_CODE > 0x020500)
	spin_lock_irqsave(&Adapter->lock, flags);
#endif
			dbgcond(DBG_IOCTL, "net queue %s\n",
				netif_queue_stopped(Adapter->
						    net) ? "stopped" :
				"running");
			dbgcond(DBG_IOCTL,
				"status: tx_urb %d, rx_urb %d\n",
				Adapter->tx_urb->status,
				Adapter->rx_urb->status);
			break;
		case CLEAR_PACKETS:
			memset((PUCHAR) & Adapter->Stats, 0,
			       sizeof(STATISTICS));
			break;
		case GET_STATISTICS:
			wrq->u.data.length = sizeof(STATISTICS);
#if (LINUX_VERSION_CODE > 0x020500)
			spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
			if (copy_to_user(wrq->u.data.pointer, &Adapter->Stats, wrq->u.data.length))
                            return -EFAULT;

#if (LINUX_VERSION_CODE > 0x020500)
	spin_lock_irqsave(&Adapter->lock, flags);
#endif
			break;
		case GET_MAC_ADDRESS:
			wrq->u.data.length = 6;
#if (LINUX_VERSION_CODE > 0x020500)
			spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
			if (copy_to_user(wrq->u.data.pointer, Adapter->StationAddress, wrq->u.data.length))
                            return -EFAULT;

#if (LINUX_VERSION_CODE > 0x020500)
	spin_lock_irqsave(&Adapter->lock, flags);
#endif
			break;
		case SITE_SURVEY:
			//spin_unlock_irqrestore(&Adapter->lock, flags);
                        if (Adapter->RadioIsOn){
                            if (Adapter->flags & GetCommandPedding) {
			DECLARE_WAITQUEUE(wait, current);
			add_wait_queue(Adapter->ctrl_wait, &wait);
			set_current_state(TASK_INTERRUPTIBLE);
			Adapter->flags |= IOCTL_SLEEPING;
			schedule();
			set_current_state(TASK_RUNNING);
			remove_wait_queue(Adapter->ctrl_wait, &wait);
                                    if (Adapter->flags & WAKE_UP_WITH_ERROR) {
                                            Adapter->flags &= ~IOCTL_SLEEPING;
#if (LINUX_VERSION_CODE > 0x020500)
			                    spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
                                            return -1;
                                    }
                            }
                            Adapter->TempState = Adapter->StationState;
                            ChangeState(Adapter, STATION_STATE_SCANNING);
                            Adapter->flags |= SITE_SURVEY_REQUEST;
                            DropPendingTxPackets(Adapter);
                            UsbScan(Adapter);
                        }
                        break;
		case SITE_SURVEY_STATE:
			if (Adapter->StationState != STATION_STATE_SCANNING){
				*(PUCHAR)wrq->u.data.pointer = 2;	//SITE_SURVEY_COMPLETED
                        }
			break;
		case GET_AP_INFO:
                        if(!Adapter->RadioIsOn){
                            Adapter->AssociationInfo.SiteS.BSSInList = 0;
                        }
			if (Adapter->AssociationInfo.SiteS.BSSInList == 0)
				return 0;
			//wrq->u.data.length = Adapter->AssociationInfo.SiteS.BSSInList * sizeof(BSS_INFO);
			wrq->u.data.length = Adapter->AssociationInfo.SiteS.BSSInList * sizeof(BSS_INFO_EX);
#if (LINUX_VERSION_CODE > 0x020500)
			spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
			if (copy_to_user(wrq->u.data.pointer, Adapter->AssociationInfo.SiteS.BssInfo, wrq->u.data.length))
                            return -EFAULT;

#if (LINUX_VERSION_CODE > 0x020500)
	spin_lock_irqsave(&Adapter->lock, flags);
#endif
			break;
		case STOP_SITE_SURVEY:
			    spin_unlock_irqrestore(&Adapter->lock, flags);
                            if(Adapter->RadioIsOn){
                                //Let's see What about Ad-hoc
                                if(Adapter->OperatingMode !=AD_HOC_MODE){
                                    if(Adapter->InternationalRoaming){
                                        for(i=0; i<14; i++){
                                            if ( Adapter->ChannelVector[i] != 0 )
                                                break;
                                        }
                                        if(i<14){
                                            if(Adapter->ErrorCode == 0xa0){
                                                    Adapter->StationState = STATION_STATE_MGMT_ERROR;
                                                    //HERE
#if (LINUX_VERSION_CODE > 0x020500)
			    spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
                                                return -1;
                                            } else if(Adapter->ErrorCode == 0xa1){
                                                    //HERE
#if (LINUX_VERSION_CODE > 0x020500)
			    spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
                                                return -1;
                                            }else{
                                                    Adapter->StationState = STATION_STATE_MGMT_ERROR;
                                                    UsbJoin(Adapter);
                                            }
                                        }else {
                                                Adapter->StationState = STATION_STATE_SCANNING;   
                                                UsbScan(Adapter);
                                        }
                                    }else {
                                            if(Adapter->ErrorCode == 0xa0){
                                                    Adapter->StationState = STATION_STATE_MGMT_ERROR;
                                                    //HERE
#if (LINUX_VERSION_CODE > 0x020500)
			    spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
                                                return -1;
                                            } else if(Adapter->ErrorCode == 0xa1){
                                                    //HERE
#if (LINUX_VERSION_CODE > 0x020500)
			    spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
                                                return -1;
                                            }else{
                                                    Adapter->StationState = STATION_STATE_MGMT_ERROR;
                                                    UsbJoin(Adapter);
                                            }
                                    }
                                } else {
                                           Adapter->StationState = STATION_STATE_READY;
                               }

                            }
			break;
		case SELECT_AP_BY_INDEX:
                            if(Adapter->RadioIsOn){
                                AP_Selected = *(PUCHAR)wrq->u.data.pointer - 1;
                                //spin_unlock_irqrestore(&Adapter->lock, flags);
                                if (Adapter->flags & GetCommandPedding) {
			DECLARE_WAITQUEUE(wait, current);
			add_wait_queue(Adapter->ctrl_wait, &wait);
			set_current_state(TASK_INTERRUPTIBLE);
			Adapter->flags |= IOCTL_SLEEPING;
			schedule();
			set_current_state(TASK_RUNNING);
			remove_wait_queue(Adapter->ctrl_wait, &wait);
                                        if (Adapter->flags & WAKE_UP_WITH_ERROR) {
                                                Adapter->flags &= ~IOCTL_SLEEPING;
                                                    //HERE
#if (LINUX_VERSION_CODE > 0x020500)
			                spin_unlock_irqrestore(&Adapter->lock, flags);
#endif
                                                return -1;
                                        }
                                }

                                Adapter->SSID_size =
                                    Adapter->AssociationInfo.SiteS.BssInfo[AP_Selected].SSIDsize;
                                Adapter->AssociationInfo.DesiredSSIDsize = Adapter->SSID_size;
                                memcpy(Adapter->SelectedSSID,
                                       Adapter->AssociationInfo.SiteS.BssInfo[AP_Selected].SSID,
                                       Adapter->SSID_size);
                                memcpy(Adapter->AssociationInfo.DesiredSSID,
                                       Adapter->AssociationInfo.SiteS.BssInfo[AP_Selected].SSID,
                                       Adapter->SSID_size);
                                memcpy((PUCHAR) Adapter->AssociationInfo.CurrentBSSID,
                                       (PUCHAR) Adapter->AssociationInfo.SiteS.
                                       BssInfo[AP_Selected].BSSID, 6);
                                //Copy the Channel of the corresponding Channel
                                Adapter->Channel = Adapter->AssociationInfo.SiteS.BssInfo[AP_Selected].Channel;
                                Adapter->OperatingMode = Adapter->AssociationInfo.SiteS.BssInfo[AP_Selected].BSStype;
			        spin_unlock_irqrestore(&Adapter->lock, flags);
                                Adapter->StationState = STATION_STATE_JOINING;
                                UsbJoin(Adapter);
                                return 0;
                                                    //HERE
                            }
                }
		break;
        case ATMEL_WPA_IOCTL : 
		            spin_unlock_irqrestore(&Adapter->lock, flags);
                wpa_result = atmel_wpa_ioctl(Adapter, &wrq->u.data); 
                spin_lock_irqsave(&Adapter->lock, flags);
          break;
	default:
		spin_unlock_irqrestore(&Adapter->lock, flags);
		dbgcond(DBG_IOCTL, "unknown ioctl x%x\n", cmd);
		return -EOPNOTSUPP;
	}
	spin_unlock_irqrestore(&Adapter->lock, flags);
	return 0;
}

static int usb_vnet_set_mac(struct net_device *dev, void *addr)
{
	/*
	   PVNet_ADAPTER Adapter = (PVNet_ADAPTER)dev->priv;
	   struct sockaddr *sa = addr;        
	   UCHAR tmpMAC[6];
	   memcpy(Adapter->tmpMAC, sa->sa_data);      
	   SetMacAddress(Adapter);    */
	return 0;
}

static int usb_vnet_tx(struct sk_buff *skb, struct net_device *dev)
{
	PVNet_ADAPTER Adapter = (PVNet_ADAPTER) dev->priv;
	USHORT len;
	ULONG flags;
	PTX_LIST NewTxL;

	if (!Adapter) {
		Adapter->Stats.TxDataPacketsError++;
		Adapter->netstats->tx_dropped++;
		netif_stop_queue(dev);
		Adapter->flags |= TX_STOPPED;

		return NET_XMIT_DROP;
	}

	if (!netif_device_present(Adapter->net)
	    || !(Adapter->flags & VNET_RUNNING)) {
		Adapter->Stats.TxDataPacketsError++;
		Adapter->netstats->tx_dropped++;
		netif_stop_queue(dev);
		Adapter->flags |= TX_STOPPED;
		return NET_XMIT_DROP;
	}

	if (Adapter->StationState != STATION_STATE_READY)
		return -EAGAIN;

	len = (ETH_ZLEN < skb->len) ? skb->len : ETH_ZLEN;

  Adapter->lock = SPIN_LOCK_UNLOCKED;
	spin_lock_irqsave(&Adapter->lock, flags);

	if (!(NewTxL = AllocPacket(Adapter, len))) {
    printk("Failure in Packet Allocation\n");
		spin_unlock_irqrestore(&Adapter->lock, flags);
		return NET_XMIT_DROP;
	}

  if(Adapter->OperatingMode == INFRASTRUCTURE_MODE){ //In ESS the final DA is always the AP
            Adapter->IsBroadcast = FALSE;
  }else {
    if ( (*((PUCHAR)(skb->data))) & 0x01 ){
      Adapter->IsBroadcast = TRUE;
    }else {
      Adapter->IsBroadcast = FALSE;
    }
  }
	NewTxL->skb = skb;

	// add the packet in the buffer (the first 12 bytes are going to be used in the
	// wireless header.
	memcpy(&NewTxL->tx_buff->WirelessPacket[18], (PUCHAR) skb->data,
	       len);

	// add the wireless header.....
	Ethernet2Wireless(Adapter, NewTxL);

	/* we are inside the Adapter->lock spinlock ... */
	DeQueueTxPacket(Adapter, NewTxL);

	// remember time of transmit
	Adapter->net->trans_start = jiffies;
	spin_unlock_irqrestore(&Adapter->lock, flags);

	// packet used. Wait until callback to free sk_buff....
	return NET_XMIT_SUCCESS;
}

#if (LINUX_VERSION_CODE < 0x020500)
#define RETERR NULL
static void *usb_vnet_probe(struct usb_device *dev, UINT interface,
			    const struct usb_device_id *id_table)
{
	PVNet_ADAPTER Adapter;
        struct net_device *netdev;
	int i;
#else
#define RETERR -ENODEV
int usb_vnet_probe(struct usb_interface *interface,
			    const struct usb_device_id *id_table)
{
	PVNet_ADAPTER Adapter;
        struct net_device *netdev;
	int i;
	struct usb_device *dev = interface_to_usbdev(interface);
#endif

	/* jal: do we really need to probe as we pass an id_table
	   in vnet_driver struct ??? */
	    printk(KERN_INFO "usb vendor/product=%x/%x\n", dev->descriptor.idVendor, dev->descriptor.idProduct);

	for (i = 0; i < vnet_usb_id_len; i++)
#if (LINUX_VERSION_CODE < 0x020500)
		if ((dev->descriptor.idVendor == vnet_usb_id[i].idVendor &&
		    dev->descriptor.idProduct == vnet_usb_id[i].idProduct) && ( interface == 0))
#else
		if ((dev->descriptor.idVendor == vnet_usb_id[i].idVendor &&
		    dev->descriptor.idProduct == vnet_usb_id[i].idProduct) && ( interface->altsetting->desc.bInterfaceClass == 0xfe || interface->altsetting->desc.bInterfaceClass == 0xff))
#endif
			break;

	if (i == vnet_usb_id_len) {
		err("Device Descriptor not matching");
		return RETERR;
	}

	Adapter = (PVNet_ADAPTER) kmalloc(sizeof(VNet_ADAPTER), GFP_KERNEL);

	if (!Adapter) {
		err("Not enough memory");
		return RETERR;
	}

#if (LINUX_VERSION_CODE < 0x020500)
	MOD_INC_USE_COUNT;
#endif
	memset((PUCHAR) Adapter, 0, sizeof(VNet_ADAPTER));

	Adapter->dbg_mask = 
#ifdef USBDBG
		0xffffffff & ~(DBG_BEACON | DBG_RXDATA | DBG_TX | DBG_CMDS | DBG_TXDATA | 
				DBG_DFU_CB | DBG_TX_CB | DBG_IOCTL | DBG_STATE);	/* switch on all debugs but ... */
#else
	0;
#endif

	ChangeState(Adapter, STATION_STATE_INITIALIZING);
#ifdef RFMD
	Adapter->BoardType = RFMD_Board;
#elif defined(R505)        
	Adapter->BoardType = R505_Board;
#elif defined(R505A)        
	Adapter->BoardType = R505A_Board;
#elif defined(R505AX)        
	Adapter->BoardType = R505AMX_Board;
#elif defined(R505_2958)        
	Adapter->BoardType = R505_2958_Board;
#endif
	Adapter->config = dev->config;
	Adapter->usb = dev;
#if (LINUX_VERSION_CODE > 0x020500)
        usb_get_dev(Adapter->usb);
#endif
	if (InitDevice(Adapter) != 0) {
		err("device initialization failed");
		ChangeState(Adapter, STATION_STATE_EXITING);
#if (LINUX_VERSION_CODE < 0x020500)
		MOD_DEC_USE_COUNT;
#endif
		del_timer_sync(Adapter->CtrlTimer);
		kfree(Adapter);
		return RETERR;
	}
#if (LINUX_VERSION_CODE > 0x020500)
        usb_set_intfdata(interface, Adapter);

#endif

#if (LINUX_VERSION_CODE < 0x020500)
	usb_inc_dev_use(dev);
#endif

//    Start Polling In Bulk In Endpoint
#if (LINUX_VERSION_CODE < 0x020500)
	Adapter->BulkInAddr =
	    	(Adapter->usb->actconfig->interface->altsetting->endpoint->
	     		bEndpointAddress) & ~0x80;
#elif (LINUX_VERSION_CODE < 0x020600)
	Adapter->BulkInAddr = 
		(Adapter->usb->actconfig->interface[0].altsetting->endpoint->desc.bEndpointAddress) & ~0x80; 
#else
	Adapter->BulkInAddr =  (Adapter->usb->actconfig->interface[0]->altsetting->endpoint->desc.bEndpointAddress) & ~0x80; 
#endif

	ChangeState(Adapter, STATION_STATE_SCANNING);
  SetRadio(Adapter, TRUE);

	if (UsbRxInit(Adapter) != 0) {
#if (LINUX_VERSION_CODE < 0x020500)
		MOD_DEC_USE_COUNT;
#endif
		kfree(Adapter);
		usb_dec_dev_use(dev);
		return RETERR;
	}
	dbgcond(DBG_INIT, "RxInit OK\n");
	Adapter->flags |= VNET_RUNNING;


#if (LINUX_VERSION_CODE < 0x020414)
	if (!(Adapter->net = init_etherdev(NULL, 0))) {
		MOD_DEC_USE_COUNT;
		kfree(Adapter);
		usb_dec_dev_use(dev);
		return RETERR;
	}
	ether_setup(Adapter->net);
#else
	netdev = alloc_etherdev(sizeof(VNet_ADAPTER)); 
        if(!netdev) {
		kfree(Adapter);
		usb_dec_dev_use(dev);
		return RETERR;
	}
        Adapter->net = netdev;
#endif

	Adapter->net->open = usb_vnet_open;
	Adapter->net->stop = usb_vnet_close;
	Adapter->net->hard_start_xmit = usb_vnet_tx;
#if 0
  Adapter->net->tx_timeout = usb_vnet_tx_timeout;
  Adapter->net->watchdog_timeo = TX_TIMEOUT_JIFFIES;
#endif
  Adapter->net->wireless_handlers = (struct iw_handler_def *)&vnet_handler_def;
	Adapter->net->do_ioctl = usb_vnet_ioctl;
	Adapter->net->get_wireless_stats = wireless_stats;
	Adapter->net->set_mac_address = usb_vnet_set_mac;
	Adapter->net->priv = Adapter;
	Adapter->net->get_stats = usb_vnet_net_stats;
	Adapter->net->hard_header_len = 14;
	Adapter->net->mtu = 1500;
	Adapter->net->addr_len = 6;
        strcpy (Adapter->net->name, "atml%d");
	memcpy(Adapter->net->dev_addr, Adapter->StationAddress,
	       Adapter->net->addr_len);

#if 0
	if (Adapter->net->flags & IFF_MULTICAST)
		printk("Multicast is enabled\n");
	if (Adapter->net->flags & IFF_PROMISC)
		printk("Promiscuous mode is enabled\n");
#endif

#if 0
	ChangeState(Adapter, STATION_STATE_SCANNING);
        SetRadio(Adapter, TRUE);
#endif

/*	if (UsbScan(Adapter) != 0) {
		err("Scan Command failed");
		MOD_DEC_USE_COUNT;
		usb_dec_dev_use(dev);
		kfree(Adapter);
		return NULL;
	}
  */
#if (LINUX_VERSION_CODE < 0x020413)
	if (!register_netdev(Adapter->net)) {
		err("register_netdev failed");
#else
	if (register_netdev(Adapter->net)) {
		err("register_netdev failed");
        Adapter->net->flags |= IFF_UP;
#endif
#if (LINUX_VERSION_CODE < 0x020500)
		MOD_DEC_USE_COUNT;
#else
        SET_NETDEV_DEV(Adapter->net, &interface->dev);
#endif
		kfree(Adapter);
		usb_dec_dev_use(dev);
		return RETERR;
	}
  netif_device_attach (Adapter->net);
	dbgcond(DBG_INIT, "usb %s initialized and registered\n", Adapter->net->name);
#if (LINUX_VERSION_CODE < 0x020500)
	return Adapter;
#else
	return 0;
#endif
}

#undef RETERR

#if (LINUX_VERSION_CODE < 0x020500)
static void usb_vnet_disconnect(struct usb_device *dev, void *ptr)
{
	PVNet_ADAPTER Adapter = (PVNet_ADAPTER) ptr;
#else
static void usb_vnet_disconnect(struct usb_interface *interface)
{
	PVNet_ADAPTER Adapter = usb_get_intfdata(interface);
        struct usb_device *dev;
        
#endif	// Linux v. < 2.5+
  unsigned long flags __attribute__ ((unused));
	char *argv[4];

#if (LINUX_VERSION_CODE > 0x020500)
        usb_set_intfdata(interface, NULL);
#endif	// Linux v. < 2.5+
	if (!Adapter)
		return;
#if (LINUX_VERSION_CODE > 0x020500)
        dev = interface_to_usbdev(interface);
#endif	// Linux v. < 2.5+

	netif_stop_queue(Adapter->net);


#if (LINUX_VERSION_CODE < 0x020500)
	argv[0] = "/sbin/modprobe";
#else
	argv[0] = "/usr/local/sbin/modprobe";
#endif
	argv[1] = "-r";
	argv[2] = (char *) vnet_driver.name;
	argv[3] = 0;

	dbgcond(DBG_INIT, " device %s module %s\n", Adapter->net->name,
		argv[2]);


	Adapter->flags &= ~VNET_RUNNING;
	Adapter->StationState = STATION_STATE_EXITING;

	if (waitqueue_active(Adapter->ctrl_wait)) {
		Adapter->flags |= WAKE_UP_WITH_ERROR;
		wake_up_interruptible(Adapter->ctrl_wait);
	}

  Adapter->lock = SPIN_LOCK_UNLOCKED;
	spin_lock_irqsave(&Adapter->lock, flags);

	DropPendingTxPackets(Adapter);

	if (timer_pending(Adapter->CtrlTimer)) {
		del_timer_sync(Adapter->CtrlTimer);
	}

#if (LINUX_VERSION_CODE < 0x020500)
	Adapter->rx_urb->transfer_flags &= ~USB_ASYNC_UNLINK;
	Adapter->tx_urb->transfer_flags &= ~USB_ASYNC_UNLINK;
	Adapter->ctrl_urb->transfer_flags &= ~USB_ASYNC_UNLINK;
	usb_unlink_urb(Adapter->rx_urb);
	usb_unlink_urb(Adapter->tx_urb);
	usb_unlink_urb(Adapter->ctrl_urb);
#else
	Adapter->rx_urb->transfer_flags &= ~URB_ASYNC_UNLINK;
	Adapter->tx_urb->transfer_flags &= ~URB_ASYNC_UNLINK;
	Adapter->ctrl_urb->transfer_flags &= ~URB_ASYNC_UNLINK;
	usb_kill_urb(Adapter->rx_urb);
	usb_kill_urb(Adapter->tx_urb);
	usb_kill_urb(Adapter->ctrl_urb);
#endif	// Linux v. < 2.5+

	spin_unlock_irqrestore(&Adapter->lock, flags);
	// Since it isn't clear if the Callback function is called whenever 
	// usb_unlink_urb is used.....(patenta).

	Adapter->flags |= TX_STOPPED;

#if (LINUX_VERSION_CODE < 0x020500)
	while (MOD_IN_USE > 1) {
		MOD_DEC_USE_COUNT;
	}

#endif
	dbgcond(DBG_INIT, "trying to unregister net device\n");
        if(Adapter->net != NULL)
	unregister_netdev(Adapter->net);
	dbgcond(DBG_INIT, "unregistered\n");

	/* jal: TODO: remove this and look at mod_use() calls, right now
	   we are unable to have more than one USB adapter handled by this driver !!! */

#if (LINUX_VERSION_CODE < 0x020500)
	while (MOD_IN_USE > 0) {
		MOD_DEC_USE_COUNT;
	}

	usb_dec_dev_use(dev);
#endif
	VnetReleaseResources(Adapter);

#if (LINUX_VERSION_CODE > 0x020500)
        usb_put_dev(dev);
#endif
	/* jal: Do we really need this ? IMHO the kmod will remove the module,
	   _if_ we handle the MOD_IN_USE counter correctly (see above) !!! */
#if (LINUX_VERSION_CODE < 0x020500)
	call_usermodehelper(argv[0], argv, NULL);
#endif
	printk("device disconnected\n");
}

static int __init usb_vnet_init(void)
{
	return usb_register(&vnet_driver);
}

static void __exit usb_vnet_exit(void)
{
#if (LINUX_VERSION_CODE < 0x020500)
	wait_ms(250);		//wait for call_usermodehelper to finish.
#endif
	//printk("AtmelWlanDriver Is deregistered!\n");
	usb_deregister(&vnet_driver);
}

/**************************************************************/
MODULE_DESCRIPTION("ATMEL 503A usb Card");
MODULE_AUTHOR("ATMEL Drivers Group");
MODULE_DEVICE_TABLE(usb, vnet_usb_id);
MODULE_LICENSE("GPL");



module_init(usb_vnet_init);	// initialization function
module_exit(usb_vnet_exit);	// exit function
/**************************************************************/
